#!/usr/bin/env node

import {
  confirm,
  intro,
  outro,
  spinner,
  note,
  log,
  text,
  isCancel,
} from "@clack/prompts";
import tiged from "tiged";
import { DegitError } from "tiged/src/utils";
import logo from "./logo";
import sade from "sade";
import { bold, underline, magenta } from "picocolors";
import { init } from "@paralleldrive/cuid2";
import { getEnvPkgManager, pnpmRc } from "./util";
import { readFile, writeFile } from "fs/promises";
import { join, resolve } from "path";
import { exec as execCb } from "node:child_process";
import { promisify } from "node:util";
import { execSync } from "child_process";
const exec = promisify(execCb);

const prog = sade("create-nitropage [dest]", true)
  .version("alpha")
  .action(async function (dest) {
    console.log(magenta(logo) + "\n");
    intro(`create-nitropage`);

    if (dest) {
      dest = "./" + dest;
    }

    const destDir =
      dest ||
      (await text({
        message: "Where should we create your project?",
        defaultValue: "./my-project",
        placeholder: "./my-project",
      }));

    if (isCancel(destDir)) {
      process.exit();
    }

    const destDirAbs = resolve(process.cwd(), destDir);

    const s = spinner();
    s.start("Cloning the starter kit");
    // TODO: Add git.lufrai.org as a redundant sourcecode host
    const emitter = tiged(
      "https://codeberg.org/nitropage/nitropage/packages/starter",
    );
    emitter.on("info", (info: any) => {
      s.stop(info.message);
    });

    try {
      await emitter.clone(destDir, {
        disableCache: true,
      });
    } catch (err) {
      if (!(err instanceof Error)) {
        log.error("Something went wrong");
        process.exit(1);
      }

      if (err instanceof DegitError) {
        log.error(
          err.code === "DEST_NOT_EMPTY"
            ? "Destination directory is not empty, aborting."
            : err.message,
        );
        process.exit(1);
      }

      log.error(err.message);
      process.exit(1);
    }

    s.start("Generating random salts.");
    const envFilePath = join(destDirAbs, ".env");
    let envContent = await readFile(envFilePath, {
      encoding: "utf8",
    });

    envContent = envContent.replace("change_this_1", init({ length: 20 })());
    envContent = envContent.replace("change_this_2", init({ length: 20 })());
    envContent = envContent.replace(
      "some_not_random_password_that_is_at_least_32_characters",
      init({ length: 32 })(),
    );
    await writeFile(envFilePath, envContent);
    s.stop("Salts generated.");

    const pkgManager = getEnvPkgManager();

    // TODO: This probably can be removed <3
    if (pkgManager === "pnpm") {
      await writeFile(join(destDirAbs, ".npmrc"), pnpmRc());
    }

    const doInstall = await confirm({
      message: `Install dependencies with ${bold(pkgManager)}?`,
    });
    if (isCancel(doInstall)) {
      process.exit();
    }
    if (doInstall) {
      s.start("Installing dependencies");

      const { stdout, stderr } = await exec(`${pkgManager} install`, {
        cwd: destDirAbs,
      });
      if (stderr.length) {
        log.error(stderr);
        process.exit(1);
      }
    }

    s.stop("Dependencies installed.");

    // TODO: Add options for postgres
    const prismaCommand = "npx prisma migrate dev";
    const setupDatabase = await confirm({
      message: "Initialize the database with SQLite?",
    });
    if (isCancel(setupDatabase)) {
      process.exit();
    }
    if (setupDatabase) {
      s.start("Initializing the database");
      try {
        execSync(prismaCommand, {
          cwd: destDirAbs,
          stdio: "inherit",
        });
      } catch (err) {
        console.error(err);
      }

      s.stop("Database initialized.");
    } else {
      note(
        `Please edit ${bold(destDir + "/prisma/schema.prisma")}.
Further help: ${underline(
          "https://www.prisma.io/docs/reference/api-reference/prisma-schema-reference#fields",
        )}

Afterwards run:
${prismaCommand}`,
        "Database configuration.",
      );
    }

    note(
      `cd ${destDir}
${pkgManager} run dev

Then open: ${underline("http://localhost:3000/admin")}`,
      "Next steps.",
    );

    outro(`Nitropage is ready 🚀`);
  })
  .parse(process.argv);
