import { BlueprintData, EditorUiState, Element, State } from "../types";
import { createId } from "@paralleldrive/cuid2";
import { klona } from "klona";
import {
  Accessor,
  createEffect,
  createMemo,
  onCleanup,
  onMount,
  useContext,
} from "solid-js";
import { capitalize } from "froebel";
import { EditorUiContext, SettingsContext } from "../components/context/admin";
import { StateContext } from "../components/context/page";
import type { Settings } from "../server/settings";
import { produce } from "solid-js/store";

export const duplicateElement = function (
  element: Element,
  draft: State,
  parentId?: string,
  siblings?: string[],
) {
  const clonedElement = klona(element);
  clonedElement.id = createId();
  clonedElement.elementId = "";
  clonedElement.slots = {};

  if (parentId != null) {
    clonedElement.parentId = parentId;
  }

  draft.elements[clonedElement.id] = clonedElement;

  if (siblings) {
    siblings.push(clonedElement.id);
  } else {
    const siblingElements =
      element.parentId != null
        ? draft.elements[element.parentId].slots[element.parentSlot].elements
        : draft.root[element.parentSlot as keyof typeof draft.root];

    const origIndex = siblingElements.indexOf(element.id);
    siblingElements.splice(origIndex + 1, 0, clonedElement.id);
  }

  for (const [slotKey, slot] of Object.entries(element.slots)) {
    const elements: string[] = [];
    clonedElement.slots[slotKey] = {
      elements,
    };

    for (const childElementId of slot.elements) {
      duplicateElement(
        draft.elements[childElementId],
        draft,
        clonedElement.id,
        elements,
      );
    }
  }
};

export const useSelectedElement = function () {
  const [editorUi] = useContext(EditorUiContext)!;
  const [state] = useContext(StateContext)!;

  return createMemo(function () {
    if (!editorUi.selectedElementId) {
      return;
    }

    return state.elements[editorUi.selectedElementId];
  });
};

export const useSelectedElementPreset = function (
  element = useSelectedElement(),
) {
  const [state] = useContext(StateContext)!;

  return createMemo(function () {
    const presetId = element()?.presetId;
    if (!presetId) {
      return;
    }

    return state.presets[presetId];
  });
};

export const blueprintDataTitle = function <T extends BlueprintData>(
  data: T,
  fallback: string,
) {
  const titleFn = data.title;
  return titleFn ? titleFn() : capitalize(fallback);
};

export const useDarkMode = function (editorUi: EditorUiState) {
  createEffect(function () {
    if (editorUi.darkMode) {
      document.documentElement.classList.add("dark");
    } else {
      document.documentElement.classList.remove("dark");
    }
  });
};

export const useSetting = function <T extends keyof Settings>(
  key: T,
  from: Accessor<"pages" | "projects">,
) {
  const [settings, setSettings] = useContext(SettingsContext)!;
  const getter = (): Settings[T]["value"] => settings[from()][key]?.value;
  const setter = function (
    fn: (draft: Settings[T]["value"]) => Settings[T]["value"],
  ) {
    setSettings(
      produce(function (draft) {
        const from_ = from();
        const result = fn(draft[from_][key]?.value);
        draft[from_][key]!.value = result;
        draft[from_][key]!.from = result != null ? from_ : "default";
      }),
    );
  };
  return [getter, setter] as [typeof getter, typeof setter];
};

export const isElementInViewport = (el: globalThis.Element) => {
  const screenBorder = 150;
  const elRect = el.getBoundingClientRect();
  const elCord = {
    top: elRect.y,
    bottom: elRect.y + elRect.height,
    // TODO: Also check the horizontal axis
    // left: elRect.x,
    // right: elRect.x + elRect.width,
  };
  const vCord = {
    top: window.screenTop + screenBorder,
    bottom: window.screenTop + window.innerHeight - screenBorder,
    // left: window.screenLeft,
    // right: window.screenLeft + window.innerWidth,
  };

  // The viewport is inside of the element (for large elements)
  if (elCord.top <= vCord.top && elCord.bottom >= vCord.bottom) {
    return true;
  }

  // The top of the element is in the viewport
  if (elCord.top >= vCord.top && elCord.top <= vCord.bottom) {
    return true;
  }

  // The bottom of the element is in the viewport
  if (elCord.bottom >= vCord.top && elCord.bottom <= vCord.bottom) {
    return true;
  }

  return false;
};

export const useScrollRestoration = function (
  editorUi: EditorUiState,
  focusElementById: (id: string) => void,
) {
  // Store information about the document, before it changed to a new toolbarStatus
  // The old information is used to calculate a proper scroll position in %
  // This is needed because the restoration effect below might run too late, after the document already has started to change
  let restorationInfo = { scrollHeight: 0, scrollTop: 0 };
  let lastToolbarStatus = editorUi.toolbarStatus;
  onMount(function () {
    let interval = setInterval(function () {
      if (lastToolbarStatus != editorUi.toolbarStatus) return;
      restorationInfo.scrollHeight = document.documentElement.scrollHeight;
      restorationInfo.scrollTop = document.documentElement.scrollTop;
    }, 1000);
    onCleanup(() => clearInterval(interval));
  });

  // Restore the scroll position when switching between different editor modes
  let skip = true;
  createEffect(function () {
    lastToolbarStatus = editorUi.toolbarStatus;

    if (skip) {
      skip = false;
      return;
    }

    const document = window.document.documentElement;
    const outerHeight = document.clientHeight;
    const oldScroll = restorationInfo.scrollTop;
    const oldHeight = restorationInfo.scrollHeight;
    const precision = 100;
    const scrollPercent =
      Math.round((precision / (oldHeight - outerHeight)) * oldScroll) /
      precision;

    let frames = 20;

    const restoreScroll = () => {
      if (editorUi.creatingElementId) {
        focusElementById(editorUi.creatingElementId);
      } else {
        const nextHeight = document.scrollHeight;
        const newScroll = (nextHeight - outerHeight) * scrollPercent;
        document.scrollTo({ behavior: "instant", top: newScroll });
      }

      frames -= 1;
      if (frames == 0) {
        return;
      }

      requestAnimationFrame(restoreScroll);
    };

    restoreScroll();
  });
};

export type EditorIframeEvent =
  | { type: "savePage" }
  | { type: "publishPage" }
  | { type: "setState"; data: State }
  | { type: "setEditorUi"; data: EditorUiState }
  | { type: "focusElement"; data: string };
