import { Accessor, useContext } from "solid-js";
import { produce } from "solid-js/store";
import { MoveEvent, SortableEvent, SortableOptions } from "sortablejs";
import { EditorUiContext } from "../components/context/admin";
import { StateContext, ConfigContext } from "../components/context/page";
import { Config, Element, State } from "../types";

export const blueprintAllowed = function (
  declaration: Config,
  elementBlueprint?: string,
  targetBlueprint?: string,
  targetSlot?: string,
) {
  if (elementBlueprint == null) {
    return true;
  }

  const elBlueprint = declaration.blueprints[elementBlueprint];
  const parentIsRoot = targetBlueprint == null || targetBlueprint === "root";

  if (elBlueprint.acceptParentType === "root" && !parentIsRoot) {
    return false;
  }
  if (elBlueprint.acceptParentType === "element" && parentIsRoot) {
    return false;
  }

  if (
    elBlueprint.acceptParent &&
    !elBlueprint.acceptParent(targetBlueprint || "root")
  ) {
    return false;
  }

  if (targetBlueprint === "root") {
    return true;
  }

  if (targetBlueprint == null) {
    return true;
  }

  if (targetSlot == null) {
    return true;
  }

  const blueprint = declaration.blueprints[targetBlueprint];
  if (!blueprint.slots) {
    return true;
  }

  if (!blueprint.slots[targetSlot]) {
    return true;
  }

  const acceptBlueprint = blueprint.slots[targetSlot].acceptBlueprint;

  if (acceptBlueprint) {
    return acceptBlueprint(elementBlueprint);
  }

  return true;
};
export const useSortableElementMove = function (elementId: Accessor<string>) {
  const [state, setState] = useContext(StateContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const [declaration] = useContext(ConfigContext)!;

  return function (childProduce: any, evt: SortableEvent) {
    const sameList = evt.from === evt.to;
    const toElementId = sameList ? elementId() : evt.to.dataset.elementId;
    if (!toElementId) {
      return false;
    }
    const toRoot = toElementId === "root";
    const fromSlot = evt.from.dataset.slot || "default";
    const toSlot = (sameList ? fromSlot : evt.to.dataset.slot) || "default";

    setEditorUi("movingElementId", false);
    if (sameList) {
      if (toRoot) {
        setState("root", toSlot as "default", produce(childProduce));
      } else {
        setState(
          "elements",
          toElementId,
          "slots",
          toSlot,
          "elements",
          produce(childProduce),
        );
      }

      return true;
    }

    const builderInfo = getBuilderInfoFromSortableEvent(evt, state);
    if (!canMoveToOtherSlot(declaration, builderInfo)) {
      return false;
    }

    const itemElement = evt.item.dataset.element;
    if (!itemElement) {
      return false;
    }

    const fromRoot = elementId() === "root";
    const removeInList = produce<string[]>(function (list) {
      list.splice(evt.oldIndex!, 1);
    });

    // Remove element from old list
    if (fromRoot) {
      setState("root", fromSlot as "default", removeInList);
    } else {
      setState(
        "elements",
        elementId(),
        "slots",
        fromSlot,
        "elements",
        removeInList,
      );
    }

    // Add element to new list
    if (toRoot) {
      setState(
        "root",
        toSlot as "default",
        produce((l) => l.splice(evt.newIndex!, 0, itemElement)),
      );
    } else {
      setState(
        "elements",
        toElementId,
        "slots",
        produce((s) => {
          if (!s[toSlot]) {
            s[toSlot] = createSlot();
          }

          s[toSlot].elements.splice(evt.newIndex!, 0, itemElement);
        }),
      );
    }
    setState(
      "elements",
      itemElement,
      produce(function (element) {
        element.parentSlot = toSlot;
        element.parentId = toRoot ? undefined : toElementId;
      }),
    );
  };
};
export const getBuilderInfoFromSortableEvent = function (
  evt: MoveEvent | SortableEvent,
  state: State,
) {
  const domElement = (evt as SortableEvent).item || (evt as MoveEvent).dragged;
  const elementId = domElement.dataset.element;
  const element = elementId ? state.elements[elementId] : undefined;
  const fromElementId = evt.from.dataset.elementId;
  const fromElement = fromElementId ? state.elements[fromElementId] : undefined;
  const targetElementId = evt.to.dataset.elementId;
  const targetElement = targetElementId
    ? state.elements[targetElementId]
    : undefined;

  return {
    element,
    elementId,
    elementBlueprint: element ? element.blueprint : "root",
    fromElementId,
    fromElement,
    fromBlueprint: fromElement ? fromElement.blueprint : "root",
    fromSlot: evt.from.dataset.slot || "default",
    targetElement,
    targetElementId,
    targetBlueprint: targetElement ? targetElement.blueprint : "root",
    targetSlot: evt.to.dataset.slot || "default",
  };
};
export const canMoveToOtherSlot = function (
  declaration: Config,
  builderInfo: ReturnType<typeof getBuilderInfoFromSortableEvent>,
) {
  if (
    !slotHasSpace(
      declaration,
      builderInfo.targetElement,
      builderInfo.targetSlot,
    )
  ) {
    return false;
  }
  if (
    !blueprintAllowed(
      declaration,
      builderInfo.element?.blueprint,
      builderInfo.targetElement?.blueprint,
      builderInfo.targetSlot,
    )
  ) {
    return false;
  }

  return true;
};
export const pageViewSortableConfig: Partial<SortableOptions> = {
  forceFallback: true,
  fallbackOnBody: true,
  swapThreshold: 0.3,
  invertSwap: true,
};
export const slotHasSpace = function (
  declaration: Config,
  targetElement?: Element,
  targetSlot?: string,
) {
  if (!targetElement) {
    return true;
  }
  if (targetElement.id === "root") {
    return true;
  }

  if (!targetSlot) {
    return true;
  }

  const blueprint = declaration.blueprints[targetElement.blueprint];
  if (!blueprint) {
    return true;
  }

  const slot = blueprint.slots?.[targetSlot];
  if (!slot) {
    return false;
  }

  if (slot.max != null) {
    const targetElementSlot = targetElement.slots[targetSlot];
    if (
      targetElementSlot?.elements &&
      targetElementSlot?.elements.length >= slot.max
    ) {
      return false;
    }
  }

  return true;
};
export const createSlot = () => ({ elements: [] });
