import { onCleanup, onMount } from "solid-js";
import { isServer } from "solid-js/web";
import { EditorUiState, State } from "../types";

type IMQ = { _$IMQ?: Record<string, number> };

const INIT_TYPE = "_$init";

export const createIframeMessageQueue = function <T extends { type: string }>(
  name: string,
  receive: (event: T) => void,
  options: {
    connected?: () => void;
    child?: () => HTMLIFrameElement;
  } = {},
) {
  const { child } = options;
  const isChild = !child;
  const storage = (!isServer ? window : {}) as Window & IMQ;
  const imq = (storage._$IMQ = storage._$IMQ || {});
  imq[name] = 0;

  let connected = !child;

  const target = (): (Window & IMQ) | null => {
    if (child) {
      return child?.()?.contentWindow;
    }

    return window.parent;
  };

  const getState = () => imq[name];
  const getTargetState = () => {
    return target()?._$IMQ?.[name] || -1;
  };

  let batch: Function[] = [];

  const schedule = <T extends (...args: any) => void>(
    fn: T,
    connected = () => true,
  ) =>
    ((...args: any[]) => {
      if (connected() && synced()) {
        fn(...args);
        return;
      }

      batch.push(() => fn(...args));
    }) as T;

  const send = schedule(
    function (event: T) {
      const targetWindow = target();
      if (!targetWindow) {
        return;
      }

      imq[name]++;
      targetWindow.postMessage({
        id: imq[name],
        ...event,
      });
    },
    () => connected,
  );

  const synced = function () {
    return getState() >= getTargetState();
  };

  const onMessage = (_e: MessageEvent<T & { id?: number }>) => {
    const event = _e.data;
    if (event.id == null) {
      return;
    }
    if (connected) {
      receive(event);
    }
    if (event.type == INIT_TYPE) {
      connected = true;
      options.connected?.();
    }

    imq[name] = event.id;

    if (batch.length && synced()) {
      batch.map((f) => f());
      batch = [];
    }
  };

  onMount(() => {
    window.addEventListener("message", onMessage);
    if (isChild) {
      send({
        type: INIT_TYPE,
      } as any);
      options.connected?.();
    }
    onCleanup(() => {
      delete imq[name];
      window.removeEventListener("message", onMessage);
    });
  });

  return [send, schedule, synced] as [
    typeof send,
    typeof schedule,
    typeof synced,
  ];
};

export type EditorIframeEvent =
  | { type: "savePage" }
  | { type: "publishPage" }
  | { type: "setState"; data: State }
  | { type: "setEditorUi"; data: EditorUiState }
  | { type: "focusElement"; data: string };
