#!/usr/bin/env node
"use strict";

import "dotenv/config";
import { hashPassword } from "../server/auth/util";
import c from "picocolors";
import sade from "sade";
import { userCommands } from "./user";
import { confirm, group, intro, outro, spinner } from "@clack/prompts";
import { name, version } from "../../package.json";
const { spawnSync } = require("child_process");
const path = require("path");
const { readFile, writeFile, access } = require("fs/promises");
const { getConfig: getPrismaConfig } = require("@prisma/internals");
const { capitalize } = require("froebel");
const { createId } = require("@paralleldrive/cuid2");

const prog = sade(name).version(version);

prog
  .command("import-prisma")
  .describe("Import the Nitropage specific prisma models into your schema")
  .option(
    "--migrate, -m",
    `Run ${c.magenta("prisma migrate dev")} after import`,
  )
  .action(async (opts: { migrate: boolean }) => {
    const cwd = process.cwd();
    const depPath = __dirname;

    // TODO: find the prisma schema location based on prisma conventions
    // https://www.prisma.io/docs/concepts/components/prisma-schema#prisma-schema-file-location
    const schemaTargetPath = path.resolve(cwd, "prisma", "schema.prisma");
    let targetSchema = await readFile(schemaTargetPath, { encoding: "utf8" });

    const schemaConfig = await getPrismaConfig({
      datamodel: targetSchema,
      ignoreEnvVarErrors: true,
    });

    let provider = schemaConfig.datasources[0]?.provider || "sqlite";
    const supportedProviders = ["sqlite"];
    if (!supportedProviders.includes(provider)) {
      const fallbackProvider = "sqlite";
      console.warn(
        c.magenta(
          `Detected datasource provider ${c.bold(
            provider,
          )} is not officially supported. Trying ${c.bold(
            fallbackProvider,
          )} instead.`,
        ),
      );
      provider = fallbackProvider;
    }

    console.log(`Importing schema for ${c.bold(provider)} datasource provider`);

    const schemaSourcePath = path.resolve(
      depPath,
      "prisma",
      `${provider}.prisma`,
    );
    const sourceSchema = await readFile(schemaSourcePath, { encoding: "utf8" });

    const sourceRegex = /(model[\S\s]*)/;
    const sourceModels = sourceSchema.match(sourceRegex)[1];

    const targetRegex = /(\/\/ NITROPAGE>\n)([\S\s]*?)(\/\/ <NITROPAGE)/;
    if (!targetSchema.match(targetRegex)) {
      targetSchema += `
// NITROPAGE>
// <NITROPAGE
`;
    }

    const updatedSchema = targetSchema.replace(
      targetRegex,
      `$1${sourceModels}
$3`,
    );
    await writeFile(schemaTargetPath, updatedSchema);

    console.log(`Updated prisma schema: ${c.green(schemaTargetPath)}`);

    if (!opts.migrate) {
      return;
    }

    spawnSync("pnpm", "prisma migrate dev".split(" "), {
      stdio: "inherit",
    });
  });

prog
  .command("blueprint <name>")
  .describe("Generate a new blueprint")
  .example("blueprint button")
  .example("blueprint menu")
  .option("--out-dir, -d", "Provide path to elements", "src/blueprints")
  .action(async (blueprintName: string, opts: { d: string }) => {
    const componentsPath = path.join(process.cwd(), opts.d);

    await access(componentsPath);

    const findAvailableName = async function (
      number?: number,
    ): Promise<{ filename: string; filepath: string; number?: number }> {
      if (number != null && number > 9) {
        throw new Error(
          `Blueprint name ${c.bold(
            blueprintName,
          )} is already used nine times, what are you doing?`,
        );
      }

      const suffix = number != null ? number : "";
      const filename = `${blueprintName}${suffix}.np.tsx`;
      const filepath = path.join(componentsPath, filename);
      const exists = await access(filepath)
        .then(() => true)
        .catch(() => false);
      if (!exists) {
        return { filename, filepath, number };
      }

      number = number || 1;
      number++;

      return findAvailableName(number);
    };

    const { filepath, number } = await findAvailableName();
    const titleSuffix = number != null ? " " + number : "";
    const title = `${capitalize(blueprintName)}${titleSuffix}`;
    const titleJson =
      number != null ? "\n" + `    title: () => "${title}",` : "";

    const template = `import { createBlueprint, BlueprintComponent } from "nitropage";
import { css } from "nitropage/css";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  return (
    <div 
      class={css({
        padding: "3rem"
      })}
      {...props.elementProps}
    >
      {props.handles}
      <props.Slot name="default" />
    </div>
  );
};

export const blueprint = createBlueprint(() => {
  return {${titleJson}
    slots: { default: {} },
    data: {},
  };
}, import.meta.hot);

export default MyComponent;
export const id = "${createId()}";
`;

    await writeFile(filepath, template);

    console.log(
      c.green(
        `Scaffolded new blueprint "${c.bold(title)}" in ${c.bold(filepath)}`,
      ),
    );
  });

userCommands(prog);

prog
  .command("demo")
  .describe(
    "Cleans up the database, media and fonts and seeds the database with example data",
  )
  .option(
    "--force, -f",
    "Don't show a confirmation dialog before deleting everything",
  )
  .action(async function ({ force }) {
    intro("Demo mode preparation");
    if (!force) {
      const result = await group(
        {
          ask1: () =>
            confirm({
              message: `Do you really wanna ${c.bold(
                "delete all data",
              )} and replace it with example data?`,
              initialValue: false,
            }),
          ask2: () =>
            confirm({
              message: `Are you 100% sure?`,
              initialValue: false,
            }),
        },
        {
          onCancel(opts) {
            outro("That was close!");
            process.exit();
          },
        },
      );

      if (!result.ask1 || !result.ask2) {
        outro("That was close!");
        process.exit();
      }
    }

    const s = spinner();
    s.start("Cleaning up all existing data");

    const { PrismaClient } = require("@prisma/client");
    const client = new PrismaClient();

    await client.nitroElementSlot.deleteMany();
    await client.nitroElement.deleteMany();
    await client.nitroElementHistory.deleteMany();
    await client.nitroFontFace.deleteMany();
    await client.nitroFont.deleteMany();
    await client.nitroMedia.deleteMany();
    await client.nitroPageRevision.deleteMany();
    await client.nitroSetting.deleteMany();
    await client.nitroPage.deleteMany();
    await client.nitroProject.deleteMany();
    await client.nitroUser.deleteMany();

    s.stop("Cleanup done.");

    s.start("Seeding example data");

    const hashedPassword = await hashPassword(`1234`);
    const user = await client.nitroUser.create({
      data: {
        email: "demo@example.com",
        username: "admin",
        password: hashedPassword,
        admin: true,
      },
    });
    const project = await client.nitroProject.create({
      data: {
        title: "Demo",
        domain: "",
        publicId: createId(),
      },
    });

    await client.nitroSetting.create({
      data: {
        name: "colors",
        value: JSON.stringify({
          ajwmhd2ql2shvwue35fqz0bv: { light: { r: 194, g: 88, b: 88 } },
          fgytns2iopaiwk0w63lze1fw: { light: { r: 182, g: 224, b: 198 } },
          ozkonvsv3eeb839w3quu70sn: { light: { r: 67, g: 25, b: 117 } },
          rp33x4o7czcumt74fu529o2l: { light: { r: 255, g: 255, b: 255 } },
        }),
        project: {
          connect: {
            id: project.id,
          },
        },
      },
    });

    const HEADER_BLUEPRINT = "pynj0gs5qlwztg60j596bkgm";
    const SECTION_BLUEPRINT = "usedfsqfwa4g4aedh2wjkqkp";
    const INTRO_BLUEPRINT = "ae8f3r696oky0tpbdlu1accp";
    const CONTAINER_BLUEPRINT = "oyf5ut4svjmudc6kl6hhm69h";
    const FEATURE_BLUEPRINT = "vt5a4mjfbbbzn5d6y08pho35";
    const BUTTON_BLUEPRINT = "m2vqbubuu61s8cu8l9z3cllu";

    const font = await client.nitroFont.create({
      data: {
        publicId: createId(),
        family: "Baskerville",
        fallback: "Baskerville Old Face, Garamond, Times New Roman, serif",
        project: {
          connect: {
            id: project.id,
          },
        },
      },
    });

    await client.nitroFont.create({
      data: {
        publicId: createId(),
        family: "Arial",
        fallback: "Helvetica Neue, Helvetica, sans-serif",
        project: {
          connect: {
            id: project.id,
          },
        },
      },
    });

    await client.nitroSetting.create({
      data: {
        name: "blueprintDefaults",
        value: JSON.stringify({
          [HEADER_BLUEPRINT]: {
            titleColor: {
              value: {
                id: "rp33x4o7czcumt74fu529o2l",
                rgb: { r: 255, g: 255, b: 255 },
                css: "255, 255, 255",
              },
            },
            linkColor: {
              value: {
                rgb: { r: 255, g: 255, b: 255 },
                css: "255, 255, 255",
                id: "rp33x4o7czcumt74fu529o2l",
              },
            },
            links: {
              value: {
                byId: {
                  i329892045: {
                    url: { content: "/" },
                    text: { content: '<span class="block">Home</span>' },
                  },
                },
                ids: ["i329892045"],
              },
            },
          },
          [INTRO_BLUEPRINT]: {
            titleFont: {
              value: {
                id: font.id,
                publicId: font.publicId,
              },
            },
          },
        }),
        project: {
          connect: {
            id: project.id,
          },
        },
      },
    });

    const page = await client.nitroPage.create({
      data: {
        project: {
          connect: {
            id: project.id,
          },
        },
      },
    });

    const revision = await client.nitroPageRevision.create({
      data: {
        title: "Nitropage Demo",
        urlPath: "/",
        publishedPage: {
          connect: {
            id: page.id,
          },
        },
        page: {
          connect: {
            id: page.id,
          },
        },
      },
    });

    const createSlot = async (name: string, revision?: any) => {
      return await client.nitroElementSlot.create({
        data: {
          name,
          parentPageRevision: revision
            ? {
                connect: {
                  id: revision.id,
                },
              }
            : undefined,
        },
      });
    };

    const createElement = async function (
      blueprint: string,
      data: any,
      parentSlot: any,
      position: number,
      slot?: any,
    ) {
      return await client.nitroElement.create({
        data: {
          blueprint,
          position,
          data: JSON.stringify(data),
          page: {
            connect: {
              id: revision.id,
            },
          },
          parentSlot: {
            connect: {
              id: parentSlot.id,
            },
          },
          slots: slot
            ? {
                connect: {
                  id: slot.id,
                },
              }
            : undefined,
          history: {
            create: {
              page: {
                connect: {
                  id: page.id,
                },
              },
            },
          },
        },
      });
    };

    const headerSlot = await createSlot("header", revision);
    const mainSlot = await createSlot("default", revision);
    const introSectionSlot = await createSlot("default");
    const introSlot = await createSlot("default");
    const featureSectionSlot = await createSlot("default");
    const featureContainerSlot = await createSlot("default");

    await createElement(HEADER_BLUEPRINT, {}, headerSlot, 0);

    await createElement(
      SECTION_BLUEPRINT,
      {
        backgroundColor: {
          value: {
            id: "ozkonvsv3eeb839w3quu70sn",
            rgb: { r: 150, g: 109, b: 179 },
          },
          override: true,
        },
      },
      mainSlot,
      0,
      introSectionSlot,
    );
    await createElement(
      INTRO_BLUEPRINT,
      {
        title: {
          value: {
            content:
              '<span class="block">NITROPAGE</span><span class="block">PLAYGROUND</span>',
          },
          override: true,
        },
        titleColor: {
          value: {
            css: "255, 255, 255",
            rgb: { r: 255, g: 255, b: 255 },
            id: "rp33x4o7czcumt74fu529o2l",
          },
          override: true,
        },
        descriptionColor: {
          value: { css: "247, 167, 255", rgb: { r: 247, g: 167, b: 255 } },
          override: true,
        },
        description: {
          value: {
            content:
              '<span class="block">Welcome to the playground! Have fun and enjoy the trip :)</span>',
          },
          override: true,
        },
      },
      introSectionSlot,
      0,
      introSlot,
    );
    await createElement(
      BUTTON_BLUEPRINT,
      {
        link: { override: true, value: { content: "/admin" } },
        content: {
          override: true,
          value: { content: '<span class="block">ADMIN LOGIN</span>' },
        },
        size: { override: true, value: { selected: "lg" } },
        textColor: {
          override: true,
          value: { id: "ozkonvsv3eeb839w3quu70sn" },
        },
        bgColor: {
          override: true,
          value: {
            css: "255, 255, 255",
            rgb: { r: 255, g: 255, b: 255 },
            id: "rp33x4o7czcumt74fu529o2l",
          },
        },
      },
      introSlot,
      0,
    );

    await createElement(
      SECTION_BLUEPRINT,
      {
        backgroundColor: {
          value: {
            id: "fgytns2iopaiwk0w63lze1fw",
            rgb: { r: 217, g: 219, b: 178 },
          },
          override: true,
        },
      },
      mainSlot,
      1,
      featureSectionSlot,
    );
    await createElement(
      CONTAINER_BLUEPRINT,
      {
        topPadding: { value: { selected: "lg" }, override: true },
        bottomPadding: { value: { selected: "lg" }, override: true },
      },
      featureSectionSlot,
      0,
      featureContainerSlot,
    );
    await createElement(
      FEATURE_BLUEPRINT,
      {
        textColor: {
          override: true,
          value: {
            css: "186, 24, 102",
            rgb: { r: 176, g: 40, b: 40 },
            id: "ozkonvsv3eeb839w3quu70sn",
          },
        },
        iconColor: {
          value: {
            css: "189, 78, 30",
            rgb: { r: 104, g: 33, b: 156 },
            id: "ozkonvsv3eeb839w3quu70sn",
          },
          override: true,
        },
        iconBackgroundColor: {
          value: {
            css: "255, 255, 255",
            rgb: { r: 255, g: 255, b: 255 },
            id: "rp33x4o7czcumt74fu529o2l",
          },
          override: true,
        },
      },
      featureContainerSlot,
      0,
    );

    s.stop("Seeding finished.");
    outro("Everything done.");
  });

prog.parse(process.argv);
