import {
  Component,
  FlowComponent,
  Resource,
  createContext,
  splitProps,
} from "solid-js";
import { css } from "nitropage/css";
import logoIcon from "./../../../assets/icon.svg";
import { NitroProject } from "@prisma/client";
import { ValidateField } from "@modular-forms/solid";
import { getDomainAvailabilityRpc } from "../../server/rpc/project";
import { useAuth } from "../../hooks/auth";
import { isServer } from "solid-js/web";
import { getRequest } from "../../server/util";

export const ProjectContext =
  createContext<Resource<NitroProject | undefined>>();

export const StartContainer: FlowComponent = function (p) {
  return (
    <>
      <div
        class={
          "np-h-screen " +
          css({
            width: "100%",
          })
        }
      >
        <div
          class={css({
            width: "32rem",
            height: "100%",
            margin: "0 auto",
            display: "flex",
            color: "rgb(15 23 42)",
            justifyContent: "center",
            flexDirection: "column",
          })}
        >
          {p.children}
        </div>
      </div>
    </>
  );
};

export const useVisitLink = function () {
  const port = isServer ? new URL(getRequest().url).port : location.port;
  const portUrl = port === "80" || port === "443" ? "" : `:${port}`;

  return function (domain = "", path = "") {
    if (domain === "") {
      return path;
    }

    return `//${domain}${portUrl}${path}`;
  };
};

export const useDomainValidator = function (
  currentValue?: () => string | undefined | null,
): ValidateField<string> {
  const error = "The provided domain is not available.";

  return async (value?: string) => {
    if (currentValue && value === currentValue()) {
      return "";
    }

    const sanitizedValue = value || "";

    if (!sanitizedValue.match(/^[a-z0-9\.-]*$/)) {
      return "Please specify a valid domain.";
    }

    const result = await getDomainAvailabilityRpc({
      domain: sanitizedValue,
    });

    return result.available ? "" : error;
  };
};

export const FooterButton: FlowComponent = function (p) {
  return (
    <div
      class={css({
        display: "flex",
        margin: "3rem 11rem 0 11rem",
        textAlign: "center",
      })}
    >
      {p.children}
    </div>
  );
};

export const Logo: Component = function (p) {
  return (
    <span
      class={css({
        display: "inline-flex",
        gap: "0.25em",
        alignItems: "center",
        fontFamily: "Jost",
        color: "var(--np-accent-900)",
        letterSpacing: "-0.025em",
        "& img": {
          width: "1.25em",
        },
      })}
    >
      <img src={logoIcon} alt="nitropage logo" />
      <span>Nitropage</span>
    </span>
  );
};

export const FormLogo: Component = function () {
  return (
    <div
      class={css({
        marginBottom: "3rem",
        fontSize: "3rem",
        textAlign: "center",
      })}
    >
      <Logo />
    </div>
  );
};

export const FormLayout: FlowComponent = function (p) {
  const [local, rest] = splitProps(p, ["children"]);
  return (
    <div
      class={css({
        display: "flex",
        flexDirection: "column",
        gap: "0.75rem",
      })}
      {...rest}
    >
      {local.children}
    </div>
  );
};

export const StyleContainer: FlowComponent = function (props) {
  return <div class="np-admin-colors np-admin-font">{props.children}</div>;
};
