import {
  Component,
  JSX,
  Show,
  children,
  createMemo,
  createSignal,
  splitProps,
} from "solid-js";
import { css } from "../../css";
import { Select as KSelect, TextField } from "@kobalte/core";
import { AsteriskIcon, CheckIcon, SelectorIcon } from "./icons";
import clsx from "clsx";

export const Input: Component<
  {
    theme?: "lg";
    error?: string;
    label?: JSX.Element;
  } & JSX.InputHTMLAttributes<HTMLInputElement>
> = function (props) {
  const [local, rootProps, inputProps] = splitProps(
    props,
    ["theme", "label"],
    ["name", "required", "disabled"],
  );

  const label = children(() => props.label);

  return (
    <TextField.Root
      {...rootProps}
      validationState={props.error ? "invalid" : "valid"}
    >
      <Show when={label()}>
        <TextField.Label
          class={css({
            display: "flex",
            alignItems: "center",
            gap: "0.5rem",
            color: "#374151",
            fontSize: "0.875rem",
            paddingBottom: "0.25rem",
          })}
        >
          {label()}
          <Show when={rootProps.required}>
            <AsteriskIcon width="0.8em" height="0.8em" />
          </Show>
        </TextField.Label>
      </Show>
      <TextField.Input
        value={props.value != null ? props.value : ""}
        class={css({
          fontSize: local.theme !== "lg" ? "0.875rem" : undefined,
          lineHeight: 1,
          color: "#374151",
          width: "100%",
          backgroundColor: "#f9fafb",
          borderBottomColor: "#9ca3afaa",
          borderRadius: "0.125rem",
          borderBottomWidth: "1px",
          boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.1)",
          transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
          padding:
            local.theme !== "lg"
              ? "0.25rem 0.5rem"
              : "0.75rem 1rem calc(0.75rem - 1px) 1rem",
          ":hover": {
            boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-alt-300)",
            outline: "none",
          },
          ":focus": {
            borderBottomColor: "transparent",
            boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-400)",
            outline: "none",
          },
          "[data-invalid]": {
            borderBottomColor: "#fb7185",
          },
        })}
        type={props.type || "text"}
        {...inputProps}
      />
      <TextField.ErrorMessage
        class={css({
          color: "#e11d48",
          paddingTop: "0.125rem",
          fontSize: "0.875rem",
        })}
      >
        {props.error}
      </TextField.ErrorMessage>
    </TextField.Root>
  );
};

export const FileInput: Component<JSX.InputHTMLAttributes<HTMLInputElement>> =
  function (props) {
    const [local, rest] = splitProps(props, ["class"]);
    return (
      <input
        type="file"
        class={clsx(
          local.class,
          css({
            width: "100%",
            color: "#374151",
            fontSize: "0.75rem",
            marginRight: "0.5rem",
            "::file-selector-button": {
              cursor: "pointer",
              fontSize: "0.9rem",
              padding: "0.25rem 0.75rem",
              lineHeight: 1,
              color: "#374151",
              borderRadius: "0.25rem",
              border: "1px solid #9ca3afaa",
              backgroundColor: "#e5e7eb",
              ":hover": {
                backgroundColor: "#d1d5db",
              },
            },
          }),
          "file:rounded-full file:border-0",
        )}
        {...rest}
      />
    );
  };

export const NumberInput: Component<
  {
    setValue: (v: any, change: boolean) => void;
  } & JSX.InputHTMLAttributes<HTMLInputElement>
> = function (props) {
  const [localProps, restProps] = splitProps(props, ["setValue"]);
  const [focused, setFocused] = createSignal(false);
  let rangeInput: HTMLInputElement = undefined!;
  let textInput: HTMLInputElement = undefined!;
  const onBlur = function () {
    requestAnimationFrame(function () {
      if (rangeInput === document.activeElement) {
        return;
      }

      if (textInput === document.activeElement) {
        return;
      }

      const value = Number.parseFloat(textInput.value);
      const numberValue = Number.isNaN(value) ? null : value;
      if (numberValue != props.value) {
        props.setValue(numberValue, true);
      }

      setFocused(false);
    });
  };

  const setValue = function (value: any, change = false) {
    const v = value === "" ? undefined : Number.parseFloat(value);
    if (Number.isNaN(v)) {
      return;
    }
    props.setValue(v, change);
  };

  return (
    <div
      class={css({
        display: "flex",
        gap: "0.2rem",
        position: "relative",
        overflow: "visible",
        width: "100%",
      })}
    >
      <input
        ref={rangeInput}
        class={css({
          display: focused() ? "inline" : "none",
          position: "absolute",
          zIndex: 1,
          bottom: "-1.2rem",
          width: "100%",
          minWidth: "100px",
          marginLeft: "50%",
          transform: "translateX(-50%)",
        })}
        max={props.max}
        min={props.min}
        step={props.step}
        type="range"
        onBlur={onBlur}
        value={props.value}
        onInput={(e) => props.setValue(e.currentTarget.valueAsNumber, true)}
      />
      <Input
        ref={textInput}
        onFocusIn={() => setFocused(true)}
        onBlur={onBlur}
        onKeyUp={(e) => {
          if (e.key !== "Enter") {
            return;
          }
          setValue(textInput.value, true);
        }}
        onInput={(e) => {
          const v = e.currentTarget.value;
          // Only set the value if it ends with a number
          if (v != "" && Number.isNaN(parseInt(v.slice(-1)))) {
            return;
          }
          setValue(v, false);
        }}
        {...restProps}
      />
    </div>
  );
};

export const Select: Component<{
  state: any;
  setState: any;
  options: { value: string; label: string }[];
}> = function (props) {
  const selectedItem = createMemo(function () {
    const item = props.options.find((o) => o.value === props.state);
    if (item) {
      return item;
    }

    return props.options.find((o) => o.value === "");
  });

  return (
    <KSelect.Root
      class={css({ width: "100%", minWidth: 0 })}
      optionValue="value"
      optionTextValue="label"
      value={{ label: "", value: props.state } as any}
      defaultValue=""
      gutter={5}
      onChange={(v) => {
        // In certain cases Kobalte triggers an onChange even though the value hasn't changed
        if (v?.value === props.state) {
          return;
        }
        props.setState(v?.value);
      }}
      options={props.options}
      itemComponent={(props) => {
        return (
          <KSelect.Item
            class={css({
              width: "100%",
              userSelect: "none",
              cursor: "pointer",
              padding: "0.15rem 0.5rem",
              borderRadius: "0.25rem",
              display: "flex",
              gap: "0.5rem",
              alignItems: "center",
              ":hover": {
                backgroundColor: "#e5e7eb",
              },
            })}
            item={props.item}
          >
            <KSelect.ItemLabel>{props.item.rawValue.label}</KSelect.ItemLabel>
            <KSelect.ItemIndicator>
              <CheckIcon />
            </KSelect.ItemIndicator>
          </KSelect.Item>
        );
      }}
    >
      <KSelect.Trigger
        title={selectedItem()?.label}
        class={css({
          display: "flex",
          alignItems: "center",
          textAlign: "left",
          color: "#374151",
          width: "100%",
          backgroundColor: "#f9fafb",
          borderColor: "#9ca3afaa",
          borderRadius: "0.125rem",
          borderBottomWidth: "1px",
          boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.15)",
          transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
          padding: "0.15rem 0.5rem",
          "> div": {
            flex: 1,
            whiteSpace: "nowrap",
            overflow: "hidden",
            textOverflow: "ellipsis",
          },
          ":hover": {
            boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-400)",
            outline: "none",
          },
          "[data-expanded]": {
            borderBottomColor: "transparent",
            boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-400)",
            outline: "none",
          },
        })}
      >
        <div>{selectedItem()?.label}</div>
        <SelectorIcon />
      </KSelect.Trigger>
      <KSelect.Portal>
        <KSelect.Content
          class={
            "np-admin-font " +
            css({
              zIndex: 60,
              padding: "0.5rem",
              borderRadius: "0.75rem",
              backgroundColor: "white",
              boxShadow:
                "0 20px 25px -5px rgb(0 0 0 / 0.1), 0 8px 10px -6px rgb(0 0 0 / 0.1)",
              transformOrigin: "var(--kb-select-content-transform-origin)",
              animation: "npSelectHide 250ms ease-in forwards",
              "&[data-expanded]": {
                animation: "npSelectShow 250ms ease-out",
              },
            })
          }
        >
          <KSelect.Listbox />
        </KSelect.Content>
      </KSelect.Portal>
    </KSelect.Root>
  );
};

export const TextArea: Component<
  JSX.TextareaHTMLAttributes<HTMLTextAreaElement>
> = function (props) {
  const [local, rest] = splitProps(props, ["class"]);
  return (
    <TextField.Root>
      <TextField.TextArea
        class={
          css({
            fontSize: "0.875rem",
            color: "#374151",
            width: "100%",
            backgroundColor: "#f9fafb",
            borderColor: "#9ca3afaa",
            borderRadius: "0.125rem",
            borderBottomWidth: "1px",
            boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.15)",
            transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
            padding: "0.15rem 0.5rem",
            ":hover": {
              boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-alt-300)",
              outline: "none",
            },
            ":focus": {
              borderBottomColor: "transparent",
              boxShadow: "0 0 0 1px white, 0 0 0 3px var(--np-accent-400)",
              outline: "none",
            },
          }) +
          " " +
          (local.class || "")
        }
        {...rest}
      />
    </TextField.Root>
  );
};
