import {
  Component,
  createMemo,
  createSignal,
  For,
  Show,
  useContext,
} from "solid-js";
import { A, createRouteAction, createRouteData, useParams } from "solid-start";
import { useAuth } from "nitropage";
import { css } from "nitropage/css";
import server$, { createServerAction$ } from "solid-start/server";
import { prisma } from "../../../prisma";
import {
  ButtonSpinner,
  ConfirmButton,
  SmallButton,
} from "../../common/buttons";
import {
  CopyIcon,
  ExternalLinkIcon,
  PlusIcon,
  TrashIcon,
} from "../../common/icons";
import {
  deletePageRpc,
  duplicatePage,
  getSettings,
  newPageAllowed,
} from "../../../server/rpc";
import { format, formatDistanceToNow } from "date-fns";
import { authData } from "../../../server/auth";
import { Input } from "../../common/inputs";
import slugify from "@sindresorhus/slugify";
import { slugifyUrl } from "../../../admin/slugify";
import Container from "../container";
import { createForm, required, reset, setValue } from "@modular-forms/solid";
import { NitroPage, NitroPageRevision } from "@prisma/client";
import { ConfirmDialog } from "../../common/dialog";
import { ProjectContext, useVisitLink } from "../../admin/util";
import { catchAuthErrors } from "../../../admin/auth";

type PageForm = {
  title: string;
  urlPath: string;
};

const CreatePageForm: Component<{ projectId: string }> = function (props) {
  const user = useAuth();
  const [createPageState, createPageAction] = createServerAction$(
    async (args: {
      csrf: string;
      title: string;
      urlPath: string;
      projectId: string;
    }) => {
      await authData(server$.request, {
        csrf: args.csrf || "",
        error: "throw",
      });
      await newPageAllowed();

      const newPage = await prisma.nitroPage.create({
        include: {
          revisions: true,
        },
        data: {
          project: {
            connect: {
              id: args.projectId,
            },
          },
          revisions: {
            create: {
              title: args.title,
              urlPath: args.urlPath,
            },
          },
        },
      });
    },
  );
  const [createPageForm, { Form, Field }] = createForm<PageForm>();

  return (
    <Form
      onSubmit={async (data) => {
        if (createPageState.pending) {
          return;
        }

        let urlPath = data.urlPath;
        if (!urlPath.startsWith("/")) {
          urlPath = "/" + urlPath;
        }

        await createPageAction({
          title: data.title,
          urlPath,
          projectId: props.projectId,
          csrf: user()?.csrf!,
        });

        reset(createPageForm);
      }}
    >
      <div
        class={css({
          display: "flex",
          alignItems: "flex-start",
        })}
      >
        <div class={css({ width: "30%", marginRight: "1.5rem" })}>
          <Field
            name="title"
            validate={[required("Please enter a page title.")]}
          >
            {(field, props) => (
              <Input
                {...props}
                error={field.error}
                value={field.value}
                theme="lg"
                placeholder="Page Title"
                onInput={(e) => {
                  const value = e.currentTarget.value;
                  props.onInput(e);
                  setValue(createPageForm, "urlPath", slugify(value));
                }}
                autofocus
              />
            )}
          </Field>
        </div>
        <div
          class={css({
            display: "flex",
            alignItems: "center",
            height: "1.7em",
            color: "#475569",
            marginRight: "0.5rem",
            fontSize: "1.5rem",
            lineHeight: 1,
            fontWeight: "500",
          })}
        >
          /
        </div>
        <div class={css({ width: "30%", marginRight: "1.5rem" })}>
          <Field name="urlPath">
            {(field, props) => (
              <Input
                {...props}
                error={field.error}
                value={field.value}
                theme="lg"
                placeholder="URL Path"
                onChange={(e) => {
                  setValue(
                    createPageForm,
                    "urlPath",
                    slugifyUrl(e.currentTarget.value),
                  );
                }}
              />
            )}
          </Field>
        </div>

        <ConfirmButton type="submit">
          <ButtonSpinner show={createPageState.pending}>
            <PlusIcon />
            <div
              class={css({
                lineHeight: 1,
              })}
            >
              Create Page
            </div>
          </ButtonSpinner>
        </ConfirmButton>
      </div>
    </Form>
  );
};

const getPages = server$(async ({ projectId }: { projectId: string }) => {
  await authData(server$.request, { error: "throw" });

  return await prisma.nitroPage.findMany({
    include: {
      revisions: {
        orderBy: { savedAt: "desc" },
        take: 1,
      },
      publishedRevision: true,
    },
    where: {
      projectId,
    },
    orderBy: {
      publishedRevision: {
        urlPath: "asc",
      },
    },
  });
});

type NitroPageWithRevision = NitroPage & {
  revisions: NitroPageRevision[];
  publishedRevision?: NitroPageRevision | null;
};

const getPageRevision = function (page: NitroPageWithRevision) {
  return page.publishedRevision ? page.publishedRevision : page.revisions[0];
};

const Pages: Component = function (props) {
  const user = useAuth();
  const params = useParams();
  const project = useContext(ProjectContext)!;
  const pages = createRouteData(
    async (key) => {
      return await catchAuthErrors(() => getPages({ projectId: key[1] }));
    },
    {
      key: () => ["pages", params.id],
    },
  );
  const settings = createRouteData(async () => {
    return await catchAuthErrors(() =>
      getSettings({
        projectId: params.id,
      }),
    );
  });
  const [deletePageState, deletePageAction] = createRouteAction(
    async (params: Parameters<typeof deletePageRpc>[0]) => {
      return await deletePageRpc(params);
    },
  );
  const [duplicatePageState, duplicatePageAction] = createRouteAction(
    async (params: Parameters<typeof duplicatePage>[0]) => {
      return await duplicatePage(params);
    },
  );
  const [deletePage, setDeletePage] = createSignal<NitroPageWithRevision>();
  const visitLink = useVisitLink();

  return (
    <Container>
      <CreatePageForm projectId={params.id} />

      <ConfirmDialog
        icon={TrashIcon}
        open={!!deletePage()}
        onConfirm={() => {
          deletePageAction({
            csrf: user()?.csrf!,
            id: deletePage()!.id,
          });
          setDeletePage(undefined);
        }}
        onCancel={() => setDeletePage(undefined)}
      >
        Do you really want to delete the page{" "}
        <strong>{getPageRevision(deletePage()!).title}</strong> with the url "
        {getPageRevision(deletePage()!).urlPath}"?
      </ConfirmDialog>

      <div
        class={css({
          marginTop: "2rem",
          display: "grid",
          gridTemplateColumns: "auto auto auto 1fr",
          //justifyContent:"center",
          alignItems: "center",
          flex: "1",
          rowGap: "0.5rem",
          columnGap: "2rem",
        })}
      >
        <For each={pages()}>
          {(page) => {
            const revision = createMemo(function () {
              return getPageRevision(page);
            });
            const url = createMemo(function () {
              return visitLink(project()?.domain!, revision().urlPath);
            });
            return (
              <Show when={revision()}>
                <div
                  class={css({
                    display: "flex",
                    alignItems: "center",
                    gap: "0.5rem",
                  })}
                >
                  <A
                    href={`/admin/page/${page.id}`}
                    class={css({
                      color: "#4f46e5",
                      fontWeight: "500",
                      ":hover": {
                        textDecoration: "underline",
                      },
                    })}
                  >
                    {revision().title}
                  </A>
                  <Show
                    when={settings()?.notFoundPage.value === revision().pageId}
                  >
                    <div
                      class={css({
                        fontWeight: 600,
                        borderRadius: "9999px",
                        backgroundColor: "#fed7aa",
                        padding: "0.375rem 0.375rem",
                        color: "#334155",
                        fontSize: "0.75rem",
                        lineHeight: 1,
                      })}
                    >
                      404
                    </div>
                  </Show>
                </div>
                <div>
                  <Show
                    when={page.publishedRevisionId === revision().id}
                    fallback={revision().urlPath}
                  >
                    <A
                      class={css({
                        display: "flex",
                        alignItems: "center",
                        gap: "0.5rem",
                        color: "#4f46e5",
                        ":hover": {
                          textDecoration: "underline",
                        },
                      })}
                      href={url()}
                      target={url()}
                      title="Visit"
                    >
                      {revision().urlPath}{" "}
                      <ExternalLinkIcon width="0.9em" height="0.9em" />
                    </A>
                  </Show>
                </div>
                <div
                  title={
                    revision().savedAt &&
                    format(new Date(revision().savedAt), "dd.MM.yy - kk:mm")
                  }
                >
                  {formatDistanceToNow(new Date(revision().savedAt))}
                </div>
                <div
                  class={css({
                    display: "flex",
                    justifyContent: "flex-end",
                    gap: "0.75rem",
                  })}
                >
                  <SmallButton
                    title="Duplicate"
                    onClick={() => {
                      duplicatePageAction({
                        csrf: user()?.csrf!,
                        id: page.id,
                      });
                    }}
                  >
                    <ButtonSpinner
                      show={
                        duplicatePageState.pending &&
                        duplicatePageState.input?.id === page.id
                      }
                    >
                      <CopyIcon />
                    </ButtonSpinner>
                  </SmallButton>
                  <SmallButton
                    onClick={() => {
                      setDeletePage(page);
                    }}
                  >
                    <ButtonSpinner
                      show={
                        deletePageState.pending &&
                        deletePageState.input?.id === page.id
                      }
                    >
                      <TrashIcon />
                    </ButtonSpinner>
                  </SmallButton>
                </div>
              </Show>
            );
          }}
        </For>
      </div>
    </Container>
  );
};

export default Pages;
