import { Component, For, Show, createSignal } from "solid-js";
import {
  A,
  createRouteAction,
  createRouteData,
  Title,
  useNavigate,
} from "solid-start";
import { useAuth } from "nitropage";
import { css } from "nitropage/css";
import server$, { createServerAction$ } from "solid-start/server";
import { prisma } from "../../prisma";
import { ButtonSpinner, ConfirmButton, SmallButton } from "../common/buttons";
import { authData, throwOnDemo } from "nitropage/server";
import { PlusIcon, SettingsIcon, TrashIcon } from "../common/icons";
import { format, formatDistanceToNow } from "date-fns";
import { Input } from "../common/inputs";
import Container from "./container";
import { createId } from "@paralleldrive/cuid2";
import { createForm, required, reset } from "@modular-forms/solid";
import { deleteProjectRpc } from "../../server/rpc/project";
import { useDomainValidator } from "../admin/util";
import { ConfirmDialog } from "../common/dialog";
import { NitroProject } from "@prisma/client";
import { demoEnabled } from "../../util";
import { catchAuthErrors } from "../../admin/auth";

const getProjects = server$(async () => {
  await authData(server$.request, { error: "throw" });

  return await prisma.nitroProject.findMany({
    orderBy: {
      title: "asc",
    },
  });
});

type ProjectForm = {
  title: string;
  domain: string;
};

export const CreateProjectForm: Component = function () {
  const user = useAuth();
  const navigate = useNavigate();
  const [projectForm, { Form, Field }] = createForm<ProjectForm>();
  const [createProjectState, createProjectAction] = createServerAction$(
    async ({
      csrf,
      title,
      domain,
    }: {
      csrf: string;
      title: string;
      domain: string;
    }) => {
      await authData(server$.request, { csrf: csrf || "", error: "throw" });
      throwOnDemo();

      const newProject = await prisma.nitroProject.create({
        data: {
          title,
          domain,
          publicId: createId(),
        },
      });

      return newProject;
    },
  );
  const domainValidator = useDomainValidator();

  return (
    <Form
      onSubmit={async (data) => {
        if (createProjectState.pending) {
          return;
        }

        const project = await createProjectAction({
          title: data.title,
          domain: data.domain,
          csrf: user()?.csrf!,
        });

        if (!project) {
          // TODO: Add this error to the frontent
          console.error("Project could not be created");
          reset(projectForm);
          return;
        }

        navigate(`/admin/project/${project?.id}`);
      }}
    >
      <div
        class={css({
          display: "flex",
          alignItems: "flex-start",
          gap: "1rem",
        })}
      >
        <div class={css({ flex: 1 })}>
          <Field
            name="title"
            validate={[required("Please specify your project title.")]}
          >
            {(field, props) => (
              <Input
                {...props}
                theme="lg"
                placeholder="Project Title"
                error={field.error}
                required
                autofocus
              />
            )}
          </Field>
        </div>

        <div class={css({ flex: 1 })}>
          <Field name="domain" validate={[domainValidator]}>
            {(field, props) => (
              <Input
                {...props}
                theme="lg"
                placeholder="Domain"
                error={field.error}
                required
              />
            )}
          </Field>
        </div>

        <ConfirmButton type="submit">
          <ButtonSpinner show={createProjectState.pending}>
            <PlusIcon />
            Create Project
          </ButtonSpinner>
        </ConfirmButton>
      </div>
    </Form>
  );
};

const Projects: Component = function () {
  const user = useAuth();
  const data = createRouteData(async () => {
    return await catchAuthErrors(getProjects);
  });
  const [deleteProjectState, deleteProjectAction] = createRouteAction(
    async (params: Parameters<typeof deleteProjectRpc>[0]) => {
      return await deleteProjectRpc(params);
    },
  );
  const [deleteProject, setDeleteProject] = createSignal<NitroProject>();

  return (
    <>
      <Title>Nitropage - Dashboard</Title>
      <Container>
        <Show when={!demoEnabled()}>
          <CreateProjectForm />
        </Show>
        <ConfirmDialog
          icon={TrashIcon}
          open={!!deleteProject()}
          onConfirm={() => {
            deleteProjectAction({
              csrf: user()?.csrf!,
              id: deleteProject()!.id,
            });
            setDeleteProject(undefined);
          }}
          onCancel={() => setDeleteProject(undefined)}
        >
          Do you really want to delete the project{" "}
          <strong>{deleteProject()!.title}</strong>?
        </ConfirmDialog>
        <div
          class={css({
            marginTop: "1.15rem",
            marginBottom: "0.7rem",
            fontSize: "1.25rem",
            fontWeight: "600",
          })}
        >
          Projects
        </div>
        <div
          class={css({
            display: "grid",
            gridTemplateColumns: "auto auto auto 1fr",
            //justifyContent:"center",
            alignItems: "center",
            flex: "1",
            rowGap: "0.5rem",
            columnGap: "2rem",
          })}
        >
          <For each={data()}>
            {(project) => {
              return (
                <>
                  <div
                    class={css({
                      display: "flex",
                      alignItems: "center",
                      gap: "0.5rem",
                    })}
                  >
                    <div class={css({ fontWeight: "500", color: "#4f46e5" })}>
                      <A href={`/admin/project/${project.id}`}>
                        {project.title || project.id}
                      </A>
                    </div>
                  </div>
                  <div>{project.domain}</div>
                  <div
                    title={
                      project.updatedAt &&
                      format(new Date(project.updatedAt), "dd.MM.yy - kk:mm")
                    }
                  >
                    {formatDistanceToNow(new Date(project.updatedAt))}
                  </div>
                  <div
                    class={css({
                      display: "flex",
                      justifyContent: "flex-end",
                      gap: "0.75rem",
                    })}
                  >
                    <SmallButton
                      component={A}
                      href={`/admin/project/${project.id}/settings`}
                      theme="primary"
                    >
                      <SettingsIcon />
                    </SmallButton>
                    <Show when={!demoEnabled()}>
                      <SmallButton onClick={() => setDeleteProject(project)}>
                        <ButtonSpinner
                          show={
                            deleteProjectState.pending &&
                            deleteProjectState.input?.id === project.id
                          }
                        >
                          <TrashIcon />
                        </ButtonSpinner>
                      </SmallButton>
                    </Show>
                  </div>
                </>
              );
            }}
          </For>
        </div>
      </Container>
    </>
  );
};

export default Projects;
