import { produce } from "solid-js/store";
import { BlueprintDataInput, RgbColor } from "../../../types";
import {
  Component,
  createEffect,
  createMemo,
  createSignal,
  For,
  Show,
  splitProps,
  startTransition,
  useContext,
} from "solid-js";
import type { FlowComponent, JSX } from "solid-js";
import type { HslColorPicker } from "vanilla-colorful/hsl-color-picker";
import { createId } from "@paralleldrive/cuid2";
import { Value, cssFromRgbColor } from ".";
import {
  BlueprintDataInputLayout,
  BlueprintDataTitle,
} from "../../editor/sidebar/blueprintData";
import classes from "./color.module.css";
import { css } from "../../../css/fela";
import {
  LinkOffIcon,
  LinkIcon,
  MoonIcon,
  PlusIcon,
  SettingsIcon,
  SunIcon,
  SwitchHorizontalIcon,
  TrashIcon,
} from "../../common/icons";
import { rgbVal } from "../../page/colors";
import { Input, NumberInput } from "../../common/inputs";
import clsx from "clsx";
import tinycolor from "tinycolor2";
import { createAutoAnimateDirective } from "@formkit/auto-animate/solid";
import { ConfirmDialog } from "../../common/dialog";
import { StateContext } from "../../context/page";

declare module "solid-js" {
  namespace JSX {
    interface IntrinsicElements {
      "hsl-color-picker": JSX.HTMLAttributes<HslColorPicker>;
    }
  }
}

const ColorPlaceholder: Component<{
  onlyLeft?: boolean;
  showLabel?: boolean;
  left?: RgbColor;
  right?: RgbColor;
  opacity?: number;
}> = (p) => {
  const leftCss = createMemo(() => cssFromRgbColor(p.left, p.opacity));
  const rightCss = createMemo(() => cssFromRgbColor(p.right, p.opacity));

  return (
    <>
      <div class={css({ width: "100%", height: "100%" }) + " np-bg-iso"} />
      <div class={classes["color-placeholder"]}>
        <div>
          <div
            class={css({
              height: "100%",
              width: p.onlyLeft ? "100%" : "50%",
              display: "flex",
              alignItems: "center",
              backgroundColor: leftCss(),
            })}
          >
            <Show when={p.showLabel && p.left}>
              <div class={classes["color-placeholder__item"]}>
                {tinycolor(p.left).toHexString()}
              </div>
            </Show>
          </div>
          <Show when={!p.onlyLeft && p.right}>
            <div
              class={css({
                height: "100%",
                width: "50%",
                display: "flex",
                alignItems: "center",
                justifyContent: "flex-end",
                backgroundColor: rightCss(),
              })}
            >
              <Show when={p.showLabel}>
                <div class={classes["color-placeholder__item"]}>
                  {tinycolor(p.right).toHexString()}
                </div>
              </Show>
            </div>
          </Show>
        </div>
      </div>
    </>
  );
};

const ColorButtonIcon: FlowComponent = function (props) {
  return (
    <div
      class={css({
        fontSize: "1rem",
        display: "flex",
        alignItems: "center",
        justifyContent: "center",
        "& svg": {
          strokeWidth: 1.5,
        },
      })}
    >
      {props.children}
    </div>
  );
};

const [visibleValue, setVisibleValue] = createSignal();

const colorDataInput: BlueprintDataInput<Value, { alpha?: boolean }> = (p) => {
  const [state, setState] = useContext(StateContext)!;
  const pickerIdentity = {};
  let bodyRef: HTMLDivElement = undefined!;

  const pickerVisible = createMemo(function () {
    return visibleValue() === pickerIdentity;
  });
  const [editPreset, setEditPreset] = createSignal(false);

  const cssVar = createMemo(function () {
    if (!p.value?.id) {
      return;
    }

    return state.colors?.[p.value.id];
  });

  const hasDark = createMemo(function () {
    return cssVar()?.dark != null;
  });

  if (typeof window !== "undefined") {
    import("vanilla-colorful/hsl-color-picker.js");
  }

  // https://github.com/web-padawan/vanilla-colorful

  const ColorButton: Component<
    {
      id?: string;
      active?: boolean;
      nav?: boolean;
    } & JSX.HTMLAttributes<HTMLButtonElement>
  > = function (colorP) {
    const [_, restProps] = splitProps(colorP, ["active", "id"]);
    return (
      <button
        title={colorP.id}
        class={clsx(
          classes["preset"],
          colorP.active && classes["preset--active"],
          colorP.nav && classes["preset--nav"],
        )}
        {...restProps}
      >
        <Show when={colorP.id}>
          <ColorPlaceholder
            onlyLeft={state.colors[colorP.id!].dark == null}
            left={state.colors[colorP.id!].light}
            right={state.colors[colorP.id!].dark}
          />
        </Show>
        {colorP.children}
      </button>
    );
  };

  const emptyValue = () => ({ r: 255, g: 255, b: 255 });

  const ColorPicker: Component<{ scheme?: "light" | "dark" }> = function (
    pickerProps,
  ) {
    const [state, setState] = useContext(StateContext)!;
    const color = createMemo(function () {
      if (!p.value?.id || pickerProps.scheme == null) {
        return p.value?.rgb;
      }

      const rgbColor = cssVar()?.[pickerProps.scheme];

      return rgbColor;
    });

    const update = (color?: RgbColor) => {
      const id = p.value?.id || createId();
      const scheme = pickerProps.scheme;
      if (scheme != null) {
        color = color || emptyValue();
        setState(
          produce(function (draft) {
            draft.colors[id] = draft.colors[id] || [];
            draft.colors[id][scheme] = color!;

            return draft;
          }),
        );
        p.setValue((draft) => {
          draft = draft || {};
          draft.id = id;
          return draft;
        });
      } else {
        p.setValue((draft) => {
          draft = draft || {};
          draft.css = rgbVal(color);
          return draft;
        });
      }
      p.setValue((draft) => {
        draft = draft || {};
        draft.rgb = color;
        return draft;
      });
    };

    type HSL = { h: number; s: number; l: number; a: number };

    const tiny = createMemo(() => tinycolor(color() || emptyValue()));
    const [hsl, setHsl] = createSignal<HSL | undefined>();
    let skipHslUpdate = false;
    createEffect(function () {
      const value = tiny().toHsl();
      if (skipHslUpdate) return;
      setHsl(value);
    });
    const setHslValue = function (value: Partial<HSL>) {
      try {
        const newHsl = {
          ...hsl()!,
          ...value,
        };
        const color = tinycolor(newHsl as any);
        const { r, g, b } = color.toRgb();
        skipHslUpdate = true;
        update({
          r,
          g,
          b,
        });
        skipHslUpdate = false;
        setHsl(newHsl);
      } catch (err) {
        console.log(err);
        // color.js throws when the input is not parsable
      }
    };

    return (
      <div class={classes["picker-wrap"]}>
        <hsl-color-picker
          class={classes["picker"]}
          ref={(el) => {
            el.addEventListener("color-changed", (event) => {
              const v = event.detail.value;
              setHslValue({
                h: v.h,
                s: v.s / 100,
                l: v.l / 100,
              });
            });
            createEffect(function () {
              const v = hsl();
              if (!v) return;

              el.color = {
                h: v.h,
                s: v.s * 100,
                l: v.l * 100,
              };
            });
          }}
        ></hsl-color-picker>
        <div
          class={css({
            display: "flex",
            gap: "0.25rem",
            "& > div": {
              flex: 1,
            },
            "& > .f2": {
              flex: 2,
            },
          })}
        >
          <div class="f2">
            <Input
              placeholder="Color"
              value={tiny() ? tiny()!.toHexString() : ""}
              onChange={(e) => {
                try {
                  if (!e.currentTarget.value) return update();
                  const color = tinycolor(e.currentTarget.value);
                  const { r, g, b } = color.toRgb();
                  update({
                    r,
                    g,
                    b,
                  });
                } catch (err) {
                  // color.js throws when the input is not parsable
                }
              }}
            />
          </div>
          <NumberInput
            placeholder="Hue"
            title="Hue"
            min={0}
            max={359}
            step={1}
            value={hsl() ? Math.round(hsl()!.h) : ""}
            setValue={(h, change) => {
              if (!change) return;
              h = h || 0;
              setHslValue({ h });
            }}
          />
          <NumberInput
            placeholder="Saturation"
            title="Saturation"
            min={0}
            max={100}
            step={1}
            value={hsl() ? Math.round(hsl()!.s * 100) : ""}
            setValue={(s, change) => {
              if (!change) return;
              s = s || 0;
              setHslValue({ s: s / 100 });
            }}
          />
          <NumberInput
            placeholder="Lightness"
            title="Lightness"
            min={0}
            max={100}
            step={1}
            value={hsl() ? Math.round(hsl()!.l * 100) : ""}
            setValue={(l, change) => {
              if (!change) return;
              l = l || 0;
              setHslValue({ l: l / 100 });
            }}
          />
        </div>
      </div>
    );
  };

  const ActivePlaceholder: Component<{ showLabel?: boolean }> = function (
    props,
  ) {
    return (
      <>
        <Show when={cssVar()}>
          <ColorPlaceholder
            showLabel={props.showLabel}
            onlyLeft={!hasDark()}
            left={cssVar()?.[!p.value?.invert ? "light" : "dark"]}
            right={cssVar()?.[!p.value?.invert ? "dark" : "light"]}
            opacity={
              p.data.custom?.alpha !== false ? p.value?.alpha : undefined
            }
          />
        </Show>
        <Show when={!cssVar()}>
          <ColorPlaceholder
            showLabel={props.showLabel}
            onlyLeft={true}
            left={p.value?.rgb}
            opacity={
              p.data.custom?.alpha !== false ? p.value?.alpha : undefined
            }
          />
        </Show>
      </>
    );
  };

  const Header = function () {
    return (
      <BlueprintDataTitle {...p}>
        <button
          onClick={async () => {
            await startTransition(function () {
              if (pickerVisible()) {
                setEditPreset(false);
              }
              setVisibleValue(!pickerVisible() ? pickerIdentity : undefined);
            });
            setTimeout(function () {
              if (pickerVisible() && bodyRef) {
                bodyRef.scrollIntoView({ behavior: "smooth", block: "center" });
              }
            }, 200);
          }}
          class={classes["spoiler-button"]}
        >
          <ActivePlaceholder />
        </button>
      </BlueprintDataTitle>
    );
  };

  const presets = createMemo(() => Object.entries(state.colors));
  const autoAnimate = createAutoAnimateDirective();
  autoAnimate;

  const [deletePresetId, setDeletePresetId] = createSignal<
    undefined | string
  >();

  return (
    <BlueprintDataInputLayout header={<Header />}>
      <ConfirmDialog
        open={!!deletePresetId()}
        icon={TrashIcon}
        onConfirm={() => {
          const v = deletePresetId()!;
          setDeletePresetId();
          p.setValue((draft) => {
            draft = draft || {};
            draft.invert = undefined;
            draft.id = undefined;
            return draft;
          });
          setState(
            produce(function (draft) {
              delete draft.colors[v];
              return draft;
            }),
          );
          setEditPreset(false);
        }}
        onCancel={() => setDeletePresetId()}
      >
        Do you really wanna delete the selected preset?
      </ConfirmDialog>
      <div
        use:autoAnimate={{ duration: 200 }}
        class={clsx(
          classes["layout"],
          pickerVisible() && classes["layout--open"],
        )}
      >
        <Show when={pickerVisible()}>
          <div ref={bodyRef} class={classes["spoiler-body"]}>
            <div
              class={css({
                display: "flex",
                alignItems: "center",
                gap: "0.75rem",
                marginBottom: "0.75rem",
                "& .side": {
                  width: "3rem",
                },
              })}
            >
              <div
                class={css({
                  height: "2.5rem",
                  flex: 1,
                  position: "relative",
                  boxShadow:
                    "0 4px 6px -1px rgb(0 0 0 / 0.1), 0 2px 4px -2px rgb(0 0 0 / 0.1)",
                })}
              >
                <ActivePlaceholder showLabel />
                <Show when={hasDark()}>
                  <button
                    title={
                      p.value?.invert
                        ? "Don't invert light/dark variants"
                        : "Invert light/dark variants"
                    }
                    class={css({
                      padding: "0.25rem",
                      borderRadius: "0.375rem",
                      backgroundColor: p.value?.invert ? "#f8fafc" : "#9ca3af",
                      color: p.value?.invert
                        ? "var(--np-accent-500)"
                        : "#f8fafc",
                      boxShadow:
                        "0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1)",
                      border: `2px solid ${
                        p.value?.invert ? "#6b7280" : "#f8fafc"
                      }`,
                      position: "absolute",
                      top: "50%",
                      left: "50%",
                      transform:
                        "translate(-50%,-50%)" +
                        (p.value?.invert ? " scale(-1, 1)" : ""),
                      transition:
                        "transform 150ms cubic-bezier(0.4, 0, 0.2, 1)",
                    })}
                    onClick={() => {
                      p.setValue((d) => {
                        d!.invert = !d!.invert;
                        return d;
                      });
                    }}
                  >
                    <SwitchHorizontalIcon />
                  </button>
                </Show>
              </div>

              <Show when={p.data.custom?.alpha !== false}>
                <div class="side">
                  <NumberInput
                    placeholder="Alpha"
                    title="Alpha"
                    min={0}
                    max={100}
                    step={1}
                    value={
                      p.value?.alpha != null
                        ? Math.round(p.value?.alpha * 100)
                        : undefined
                    }
                    setValue={(v, change) => {
                      if (!change) return;
                      p.setValue((d) => {
                        d = d || {};
                        d.alpha = v != null ? v / 100 : undefined;
                        return d;
                      });
                    }}
                  />
                </div>
              </Show>
            </div>

            <div
              class={css({
                position: "relative",
                overflow: !cssVar() || editPreset() ? "visible" : "hidden",
                margin:
                  !cssVar() || editPreset() ? "0.25rem 0 1rem 0" : undefined,
              })}
              use:autoAnimate={{ duration: 200 }}
            >
              <Show when={cssVar() && editPreset()}>
                <div
                  class={css({
                    border: "1px solid #e2e8f0",
                    boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                    borderRadius: "0.5rem",
                    padding: "0.625rem",
                    position: "relative",
                  })}
                >
                  <div
                    class={css({
                      color: "#64748b",
                      fontSize: "0.85rem",
                      position: "absolute",
                      top: "-10px",
                      padding: "0 0.5rem",
                      backgroundColor: "#f8fafc",
                      left: "50%",
                      transform: "translate(-50%)",
                    })}
                  >
                    Edit preset
                  </div>
                  <div
                    class={css({
                      display: "flex",
                      alignItems: "center",
                      justifyContent: "center",
                      marginTop: "0.5rem",
                      marginBottom: "0.365rem",
                      "& .title": {
                        border: "1px solid #475569",
                        backgroundColor: "white",
                        borderRadius: "1rem",
                        padding: "0 1rem",
                        display: "flex",
                        gap: "0.365rem",
                        fontWeight: 500,
                        alignItems: "center",
                        textAlign: "center",
                        fontSize: "0.75rem",
                        lineHeight: "1rem",
                      },
                      "& .gap": {
                        height: "1px",
                        width: "3rem",
                        transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
                        backgroundColor: hasDark() ? "transparent" : "#64748b",
                      },
                      "& .title--dark": {
                        color: "white",
                        backgroundColor: "#475569",
                      },
                      "& button": {
                        display: "flex",
                        gap: "0.25rem",
                        alignItems: "center",
                        borderRadius: "0.25rem",
                        padding: "0 1rem",
                        border: "1px solid #b5b9c2",
                        boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                        transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
                        color: hasDark() ? "var(--np-accent-500)" : "#475569",
                        "&:hover": {
                          backgroundColor: "#e5e7eb",
                        },
                      },
                    })}
                  >
                    <div class="title">
                      <SunIcon /> Light
                    </div>
                    <div class="gap" />
                    <button
                      title={`${
                        hasDark() ? "Disable" : "Enable"
                      } color mode variants`}
                      onClick={() => {
                        const v = !hasDark();
                        if (!v) {
                          p.setValue((d) => {
                            d!.invert = undefined;
                            return d;
                          });
                        }
                        setState(
                          produce(function (draft) {
                            draft.colors[p.value!.id!].dark = !v
                              ? undefined
                              : cssVar()?.light;
                            return draft;
                          }),
                        );
                      }}
                    >
                      <Show when={hasDark()} fallback={<LinkIcon />}>
                        <LinkOffIcon />
                      </Show>
                    </button>
                    <div class="gap" />
                    <div class="title title--dark">
                      <MoonIcon /> Dark
                    </div>
                  </div>

                  <div
                    class={css({
                      display: "flex",
                      justifyContent: "space-between",
                      gap: "0.75rem",
                      "& > *": {
                        flex: 1,
                      },
                    })}
                  >
                    <div>
                      <ColorPicker scheme="light" />
                    </div>
                    <Show when={hasDark()}>
                      <div>
                        <ColorPicker scheme="dark" />
                      </div>
                    </Show>
                  </div>
                </div>
              </Show>

              <Show when={!cssVar()}>
                <ColorPicker />
              </Show>
            </div>

            <div class={classes["presets-layout"]}>
              <div
                class={clsx(
                  classes["presets-layout__nav"],
                  presets().length && classes["presets-layout__nav--items"],
                )}
              >
                <div class={classes["presets-layout__nav-title"]}>Presets</div>
                <ColorButton
                  title={"New"}
                  nav
                  onClick={() => {
                    const newId = createId();
                    setState(
                      produce(function (draft) {
                        const newColor = p.value?.rgb || {
                          r: 255,
                          g: 255,
                          b: 255,
                        };
                        const val = cssVar();
                        draft.colors[newId] = val
                          ? { ...val }
                          : { light: newColor, dark: undefined };

                        return draft;
                      }),
                    );
                    p.setValue((draft) => {
                      draft = draft || {};
                      draft.id = newId;
                      return draft;
                    });
                    setEditPreset(true);
                  }}
                >
                  <ColorButtonIcon>
                    <PlusIcon />
                  </ColorButtonIcon>
                </ColorButton>
                <Show when={cssVar()}>
                  <ColorButton
                    title={"Edit selected preset"}
                    nav
                    active={editPreset()}
                    onClick={() => setEditPreset(!editPreset())}
                  >
                    <ColorButtonIcon>
                      <SettingsIcon />
                    </ColorButtonIcon>
                  </ColorButton>
                  <ColorButton
                    title="Delete selected preset"
                    nav
                    onClick={() => {
                      setDeletePresetId(p.value!.id!);
                    }}
                  >
                    <ColorButtonIcon>
                      <TrashIcon />
                    </ColorButtonIcon>
                  </ColorButton>
                </Show>
              </div>

              <div class={classes["presets-layout__presets"]}>
                <For each={presets()}>
                  {([id, value]) => (
                    <ColorButton
                      active={id === p.value?.id}
                      onClick={() => {
                        if (p.context !== "project") {
                          p.setOverride(() => true);
                        }
                        const set = id !== p.value?.id;
                        if (!set) {
                          setEditPreset(false);
                        }
                        p.setValue((draft) => {
                          draft = draft || {};
                          draft.id = set ? id : undefined;
                          draft.invert = undefined;
                          draft.rgb = value.light;
                          draft.css = rgbVal(value.light);
                          return draft;
                        });
                      }}
                      id={id}
                    />
                  )}
                </For>
              </div>
            </div>
          </div>
        </Show>
      </div>
    </BlueprintDataInputLayout>
  );
};

export default colorDataInput;
