import { lazy } from "solid-js";
import { clamp, css } from "../../../css";
import { BlueprintData } from "../../../types";

const id = "@npgjs/font";

export type Value = {
  id?: string;
  publicId?: string;
  weight?: number;
  italic?: boolean;
  size?: number;
  sizeMax?: number;
  lineHeight?: number;
  letterSpacing?: number;
  letterSpacingMax?: number;
};

export type Font = {
  family?: string;
  weight?: number;
  italic?: boolean;
  size?: string;
  sizeMin?: string;
  sizeMax?: string;
  lineHeight?: number;
  letterSpacing?: string;
  letterSpacingMin?: string;
  letterSpacingMax?: string;
};

export const fontCss = function (font?: Font) {
  const result: {
    fontFamily?: string;
    fontWeight?: number;
    fontStyle?: string;
    fontSize?: string;
    lineHeight?: number;
    letterSpacing?: string;
  } = {};

  if (!font) return result;

  if (font.family) result.fontFamily = font.family;
  if (font.weight) result.fontWeight = font.weight;
  if (font.italic) result.fontStyle = "italic";
  if (font.size) result.fontSize = font.size;
  if (font.lineHeight) result.lineHeight = font.lineHeight;
  if (font.letterSpacing) result.letterSpacing = font.letterSpacing;

  return result;
};

export const fontClass = function (font?: Font) {
  return font && css(fontCss(font));
};

const fontData = ({
  responsive = true,
  family = true,
  weight = false,
  italic = false,
  size = false,
  lineHeight = false,
  letterSpacing = false,
  title,
}: {
  responsive?: boolean;
  family?: boolean;
  weight?: boolean;
  italic?: boolean;
  size?: boolean;
  lineHeight?: boolean;
  letterSpacing?: boolean;
  title?: BlueprintData["title"];
} = {}) => {
  const getter = function (v?: Value): Font {
    if (!v) {
      return {};
    }

    const size = v.size != null ? v.size + "rem" : undefined;
    const sizeMax = v.sizeMax != null ? v.sizeMax + "rem" : undefined;

    let sizeClamp: string | undefined = size || sizeMax;
    if (v.size != null && v.sizeMax != null) {
      sizeClamp = clamp(v.size, v.sizeMax);
    }

    const letterSpacing =
      v.letterSpacing != null ? v.letterSpacing + "em" : undefined;
    const letterSpacingMax =
      v.letterSpacingMax != null ? v.letterSpacingMax + "em" : undefined;

    let letterSpacingClamp: string | undefined =
      letterSpacing || letterSpacingMax;
    if (v.letterSpacing != null && v.letterSpacingMax != null) {
      letterSpacingClamp = clamp(v.letterSpacing, v.letterSpacingMax);
    }

    return {
      family: v.publicId ? `var(--font-${v.publicId})` : undefined,
      weight: v.weight,
      italic: v.italic,
      size: responsive ? sizeClamp : size,
      sizeMin: size,
      sizeMax,
      lineHeight: v.lineHeight,
      letterSpacing: responsive ? letterSpacingClamp : letterSpacing,
      letterSpacingMin: letterSpacing,
      letterSpacingMax,
    };
  };

  return {
    id: id,
    getter,
    title,
    defaultValue: () =>
      ({
        id: undefined,
        weight: undefined,
        italic: undefined,
        size: undefined,
        sizeMax: undefined,
        lineHeight: undefined,
        letterSpacing: undefined,
      }) as Value,
    input: lazy(async () => import("./input")),
    custom: {
      responsive,
      family,
      weight,
      italic,
      size,
      lineHeight,
      letterSpacing,
    },
  } satisfies BlueprintData<Value>;
};

export default fontData;
