import { BlueprintDataInput } from "../../../types";
import {
  BlueprintDataInputLayout,
  BlueprintDataTitle,
} from "../../editor/sidebar/blueprintData";
import { Value } from ".";
import {
  createMemo,
  FlowComponent,
  JSX,
  ParentComponent,
  Show,
  splitProps,
  useContext,
} from "solid-js";
import { css } from "nitropage/css";
import Checkbox from "../../common/checkbox";
import {
  FontFamilyIcon,
  LetterSpacingIcon,
  LineHeightIcon,
  TextBoldIcon,
  TextItalicIcon,
  TextSizeIcon,
} from "../../common/icons";
import { NumberInput, Select } from "../../common/inputs";
import { StateContext } from "../../context/page";

const TextInput: BlueprintDataInput<Value> = function (props) {
  const bottomColumns = createMemo(function () {
    let columns = 0;
    if (props.data.custom.size) {
      columns++;
    }
    if (props.data.custom.lineHeight) {
      columns++;
    }
    if (props.data.custom.letterSpacing) {
      columns++;
    }

    return columns;
  });
  const columns = createMemo(function () {
    let topRow = 0;

    if (props.data.custom.family) {
      topRow++;
    }
    if (props.data.custom.weight) {
      topRow++;
    }
    if (props.data.custom.italic) {
      topRow++;
    }

    const bottomRow = bottomColumns();

    return topRow > bottomRow ? topRow : bottomRow;
  });

  const SizeSuffix: FlowComponent = function (props) {
    return (
      <span class={css({ fontSize: "0.7rem", color: "#475569" })}>
        {props.children}
      </span>
    );
  };

  const SizeInput: ParentComponent<
    {
      key:
        | "size"
        | "sizeMax"
        | "weight"
        | "letterSpacing"
        | "letterSpacingMax"
        | "lineHeight";
    } & JSX.InputHTMLAttributes<HTMLInputElement>
  > = function (props_) {
    const [p, inputProps] = splitProps(props_, ["value", "children", "key"]);

    return (
      <div
        class={
          "input " +
          css({
            width: "100%",
            display: "flex",
            gap: "0.2rem",
            alignItems: "center",
          })
        }
      >
        <NumberInput
          value={props.value?.[p.key]}
          setValue={(v) => {
            props.setValue(function (draft) {
              draft[p.key] = v;
              return draft;
            });
          }}
          {...inputProps}
        />

        {p.children}
      </div>
    );
  };

  type Options = { label: string; value: string }[];
  const [state] = useContext(StateContext)!;
  const familyOptions = createMemo(function () {
    const optionsSorted: Options = [];
    const options: Options = [];

    for (const font of state.fonts) {
      (font.family ? optionsSorted : options).push({
        label: font.family || font.id!,
        value: font.id!,
      });
    }

    optionsSorted.sort((a, b) => a.label.localeCompare(b.label));
    options.unshift({ label: "Default", value: "" }, ...optionsSorted);

    return options;
  });
  return (
    <BlueprintDataInputLayout header={<BlueprintDataTitle {...props} />}>
      <div
        class={css({
          width: "100%",
          display: "grid",
          alignItems: "center",
          gridTemplateColumns: `${
            props.data.custom.responsive ? "auto " : ""
          }repeat(${columns()}, auto minmax(0, 1fr))`,
          gridTemplateRows: "1fr",
          whiteSpace: "nowrap",
          fontSize: "0.9rem",
          columnGap: "0.5rem",
          rowGap: "0.5rem",
          marginRight: "-20px",
          padding: "0.5rem 0 0.5rem 1rem",
          "& .row-1": {
            gridRowStart: "1",
          },
          "& .row-2": {
            gridRowStart: "2",
          },
          "& .row-end-4": {
            gridRowEnd: props.data.custom.responsive ? "4" : undefined,
          },
          "& .row-3": {
            gridRowStart: "3",
          },
          "& .input": {
            width: "100%",
            paddingRight: "1rem",
          },
        })}
      >
        <Show when={props.data.custom.family}>
          <Show when={props.data.custom.responsive}>
            <div class="row-1"></div>
          </Show>
          <label class="row-1" title="Family">
            <FontFamilyIcon />
          </label>
          <div class="input row-1">
            <Select
              options={familyOptions()}
              state={props.value?.id || ""}
              setState={(v: string) =>
                props.setValue(function (draft) {
                  draft.id = v;
                  draft.publicId = state.fonts.find(
                    (f) => f.id === v,
                  )?.publicId;
                  return draft;
                })
              }
            />
          </div>
        </Show>
        <Show when={props.data.custom.weight}>
          <label class="row-1" title="Weight">
            <TextBoldIcon />
          </label>
          <div class="row-1">
            <SizeInput
              key="weight"
              placeholder="Weight"
              min={100}
              step={50}
              max={950}
            />
          </div>
        </Show>

        <Show when={props.data.custom.italic}>
          <label class="row-1" title="Italic">
            <TextItalicIcon />
          </label>
          <div class="row-1">
            <Checkbox
              state={!!props.value?.italic}
              setState={(v) =>
                props.setValue(function (draft) {
                  draft.italic = v;
                  return draft;
                })
              }
            />
          </div>
        </Show>

        <Show when={bottomColumns()}>
          <Show when={props.data.custom.responsive}>
            <div class="row-2">Min</div>
          </Show>

          <Show when={props.data.custom.size}>
            <label class="row-2" title="Size">
              <TextSizeIcon />
            </label>
            <div class="row-2">
              <SizeInput
                key="size"
                placeholder="Size"
                min={0.75}
                step={0.125}
                max={8}
              >
                <SizeSuffix>rem</SizeSuffix>
              </SizeInput>
            </div>
          </Show>

          <Show when={props.data.custom.letterSpacing}>
            <label class="row-2" title="Tracking (Letter-Spacing)">
              <LetterSpacingIcon />
            </label>
            <div class="row-2">
              <SizeInput
                key="letterSpacing"
                placeholder="Tracking"
                min={-0.05}
                step={0.025}
                max={0.1}
              >
                <SizeSuffix>em</SizeSuffix>
              </SizeInput>
            </div>
          </Show>

          <Show when={props.data.custom.lineHeight}>
            <label class="row-2 row-end-4" title="Leading (Line-Height)">
              <LineHeightIcon />
            </label>
            <div class="row-2 row-end-4">
              <SizeInput
                key="lineHeight"
                placeholder="Leading"
                min={0.75}
                step={0.125}
                max={2.5}
              />
            </div>
          </Show>

          <Show when={props.data.custom.responsive}>
            <div class="input row-3">Max</div>

            <Show when={props.data.custom.size}>
              <label class="row-3" title="Size Max">
                <TextSizeIcon />
              </label>
              <div class="row-3">
                <SizeInput
                  key="sizeMax"
                  placeholder="Size Max"
                  min={0.75}
                  step={0.125}
                  max={8}
                >
                  <SizeSuffix>rem</SizeSuffix>
                </SizeInput>
              </div>
            </Show>

            <Show when={props.data.custom.letterSpacing}>
              <label class="row-3" title="Tracking (Letter-Spacing)">
                <LetterSpacingIcon />
              </label>
              <div class="row-3">
                <SizeInput
                  key="letterSpacingMax"
                  placeholder="Tracking Max"
                  min={-0.05}
                  step={0.025}
                  max={0.1}
                >
                  <SizeSuffix>em</SizeSuffix>
                </SizeInput>
              </div>
            </Show>
          </Show>
        </Show>
      </div>
    </BlueprintDataInputLayout>
  );
};

export default TextInput;
