import { lazy, onCleanup } from "solid-js";
import { isServer } from "solid-js/web";
import {
  BlueprintData,
  BlueprintDataInput,
  BlueprintDataPropDefault,
  BlueprintDataRecord,
  BlueprintDataRecordPropGetter,
  State,
} from "../../../types";

type Value<T extends BlueprintDataRecord = BlueprintDataRecord> = {
  ids: string[];
  byId: Record<string, BlueprintDataRecordPropGetter<T>>;
};

type Custom<T extends BlueprintDataRecord = BlueprintDataRecord> = {
  title?: BlueprintData["title"];
  defaultItems?: number;
  itemData: T;
  itemTitle?: (
    item: Partial<BlueprintDataPropDefault<T>>,
  ) => string | undefined;
};

const input = lazy(async () => import("./input")) as BlueprintDataInput<
  Value,
  Custom
>;

export const sortItems = function <T extends Value>(data?: T) {
  if (!data?.ids) {
    return [];
  }

  return data.ids.map(function (id) {
    return [data.byId[id], id];
  }) as [T["byId"]["0"], string][];
};

const proxyContexts = new Map<
  any,
  { itemsProxy?: any; itemProxies: Map<any, any> }
>();

const listData = <T extends BlueprintDataRecord = BlueprintDataRecord>(
  customOpts: Custom<T>,
) => {
  const itemPropertyGetter = function (
    item: any,
    property: string,
    state: State,
  ) {
    const data = customOpts.itemData[property];
    if (!data) {
      return;
    }

    const value = item[property];
    return data.getter ? data.getter(value, state) : value;
  };

  const dataKeys = Object.keys(customOpts.itemData);

  return {
    id: "@npgjs/list",
    title: customOpts.title,
    getter(v: Value, state): Value<T> | undefined {
      if (!v) {
        return;
      }

      if (isServer) {
        return {
          ...v,
          byId: Object.keys(v.byId).reduce(
            function (acc, itemId) {
              const item = v.byId[itemId];
              if (!item) {
                return acc;
              }

              acc[itemId] = dataKeys.reduce(function (itemAcc, dataKey) {
                itemAcc[dataKey] = itemPropertyGetter(item, dataKey, state);
                return itemAcc;
              }, {} as any);

              return acc;
            },
            {} as Value<T>["byId"],
          ),
        };
      }

      let context = proxyContexts.get(v);
      if (!context) {
        context = {
          itemProxies: new Map(),
        };
        // console.info("new proxy", v);
        onCleanup(function () {
          proxyContexts.delete(v);
          // console.info("cleaned up proxy", v);
        });

        context.itemsProxy = new Proxy(
          {},
          {
            ownKeys() {
              return Object.keys(v.byId);
            },
            getOwnPropertyDescriptor(target, prop) {
              return {
                enumerable: true,
                configurable: true,
              };
            },
            get(_, itemId: string) {
              const item = v.byId[itemId];
              if (!item) {
                return;
              }

              let itemProxy = context!.itemProxies.get(item);
              if (itemProxy) {
                return itemProxy;
              }

              itemProxy = new Proxy(
                {},
                {
                  ownKeys() {
                    return dataKeys;
                  },
                  getOwnPropertyDescriptor(target, prop) {
                    return {
                      enumerable: true,
                      configurable: true,
                    };
                  },
                  get(_, property: string) {
                    return itemPropertyGetter(item, property, state);
                  },
                },
              );

              context!.itemProxies.set(item, itemProxy);

              return itemProxy;
            },
          },
        );

        proxyContexts.set(v, context);
      }

      return {
        ids: v.ids,
        byId: context.itemsProxy,
      };
    },
    defaultValue: (): Value<T> => {
      const result: Value<T> = {
        byId: {},
        ids: [],
      };
      if (!customOpts.defaultItems) {
        return result;
      }
      for (let i = 0; i < customOpts.defaultItems; i++) {
        const itemValue = Object.entries(customOpts.itemData).reduce(function (
          acc,
          [key, data],
        ) {
          acc[key as keyof T] = data.defaultValue();
          return acc;
        }, {} as BlueprintDataRecordPropGetter<T>);
        const id = "i" + ("" + Math.random()).slice(2, 11);
        result.byId[id] = itemValue;
        result.ids.push(id);
      }

      return result;
    },
    input,
    custom: customOpts,
  } satisfies BlueprintData;
};

export default listData;
