import { createMemo, createResource, lazy, Signal, useContext } from "solid-js";
import { BlueprintData, BlueprintDataInput, ResizeOffer } from "../../../types";
import { getPublicMedia, optimizeImage } from "../../../server/rpc";
import { NitroMedia } from "@prisma/client";
import server$ from "solid-start/server";
import { createResize } from "../../../server/media";
import { createStore, reconcile, unwrap } from "solid-js/store";
import { isServer } from "solid-js/web";
import { useIsRouting } from "solid-start";
import { calculateAspectRatio } from "../../../util/media";
import { StateContext } from "../../context/page";

const id = "@npgjs/media";

export type Value = { mediaIds: string[] };

const getter = function (v?: Value) {
  return v?.mediaIds;
};

type MediaType = "image" | "video" | "audio";
export type MediaTypeObject = {
  image?: boolean;
  video?: boolean;
  audio?: boolean;
};
type Custom = {
  type?: MediaType | MediaTypeObject;
  max?: number;
  title?: BlueprintData["title"];
};

const input = lazy(async () => import("./input")) as BlueprintDataInput<
  Value,
  Custom
>;

const mediaData = ({ type, max, title }: Custom = {}) => {
  let typeNormalized: MediaTypeObject | undefined;
  if (type) {
    const typeIsObject = typeof type === "object";
    if (typeIsObject) {
      typeNormalized = type;
    } else {
      typeNormalized = {};
      typeNormalized[type] = true;
    }
  }

  return {
    id: id,
    title,
    getter,
    defaultValue: () => {
      return { mediaIds: [] } as Value;
    },
    input,
    custom: {
      max,
      type: typeNormalized,
    },
  } satisfies BlueprintData<Value, Custom>;
};

export default mediaData;

const createDeepSignal = function createDeepSignal<T>(value: T): Signal<T> {
  value = value || ([] as any);
  const [store, setStore] = createStore({
    value,
  });

  return [
    () => store.value,

    (v: T) => {
      if (typeof v === "function") {
        const unwrapped = unwrap(store.value);
        v = v(unwrapped);
      }

      setStore("value", reconcile(v));

      return store.value;
    },
  ] as Signal<T>;
};

/*
 * Unsplash URL: https://unsplash.com/photos/75715CVEJhI
 * Artist: https://unsplash.com/@sadmax
 */
export const PlaceholderImage = {
  rawUrl: "/np-placeholder.webp",
  url: "/np-placeholder.webp",
  width: 2400,
  height: 2400,
  color: "185,131,60",
  blurhash: "UIE.U:.8A^Ouysx^ozS$K+xbr=s,%#WBWAV@",
} as NitroMedia;

export type Media = Awaited<ReturnType<typeof getPublicMedia>>[0];

export const useMedia = function (
  key: () => string[] | undefined,
  options: {
    dataUrl?: boolean;
    admin?: boolean;
    noRouting?: boolean;
  } = {},
) {
  const [state] = useContext(StateContext)!;
  const isRouting = options.noRouting ? () => false : useIsRouting();

  const res = createResource(
    createMemo(function () {
      const ids = key() || [];
      return { key: [...ids], cacheKey: state.cacheKey };
    }),
    async function ({ key }) {
      const media =
        key.length > 0
          ? await getPublicMedia({
              ids: key,
              includeDataUrl: options.dataUrl,
              admin: options.admin,
            })
          : [];
      const byId = media.reduce(
        function (acc, media) {
          acc[media.id] = media;
          return acc;
        },
        {} as Record<string, Media>,
      );
      const byOrder = key.reduce((acc, id) => {
        const media = byId[id];
        if (!media) {
          return acc;
        }

        acc.push(media);

        return acc;
      }, [] as Media[]);

      return { byId, byOrder };
    },
    {
      initialValue: { byId: {}, byOrder: [] },
      storage: createDeepSignal,
    },
  );

  res[0]();

  const result = function () {
    if (state.admin) {
      return res[0].latest;
    }

    if (import.meta.hot && !isRouting()) {
      return res[0].latest;
    }

    return res[0]();
  };
  const actions = res[1];

  return [result, actions, res[0]] as [
    typeof result,
    typeof actions,
    (typeof res)[0],
  ];
};

export const thumResize = server$(
  createResize({
    sizes: { thumb: { width: 100 } },
  }),
);

const resizeResultsCache = new Map();

export const useResize = function <
  T extends () => Promise<{ key: string; Offer: ResizeOffer }>,
>(fn: T) {
  const [data, actions] = createResource<Awaited<ReturnType<T>>>(
    function () {
      const cachedResult = resizeResultsCache.get(fn);
      if (cachedResult) {
        return cachedResult;
      }

      const fnPromise = fn();
      resizeResultsCache.set(fn, fnPromise);

      return fnPromise.then((v) => {
        resizeResultsCache.set(fn, v);
        return v;
      });
    },
    { initialValue: {} as any },
  );

  const serverData = data();

  // Warm up the resize results cache client side
  if (serverData && !isServer && !resizeResultsCache.get(fn)) {
    resizeResultsCache.set(fn, serverData);
  }

  type ResizeOffer_ = Awaited<ReturnType<T>>["Offer"];
  type ResizeSize = keyof ResizeOffer_["sizes"];
  type ResizeAspectRatio =
    ResizeOffer_["aspectRatios"] extends readonly string[]
      ? ResizeOffer_["aspectRatios"][number]
      : undefined;

  type Source = {
    size: ResizeSize;
    aspectRatio?: ResizeAspectRatio;
    screen?: number;
  };

  const mediaSource = function (media: Media, sources: Source[]) {
    return sources.map(function (source) {
      const { size, screen } = source;
      const baseOptions = {
        aspectRatio: source.aspectRatio,
        key: media.optimizeKey || undefined,
        position: (media.optimizeKey && media.resizePosition) || undefined,
      };

      const isSvg = media.mime === "image/svg+xml";
      const url = isSvg
        ? media.rawUrl
        : optimizeUrl(media.rawUrl, size, baseOptions);
      const url2x = isSvg
        ? undefined
        : optimizeUrl(media.rawUrl, size, { ...baseOptions, pixelRatio: 2 });

      const sizes = data.latest?.Offer?.sizes;
      const sizeData = sizes?.[size as any];

      let width = media.width || 1;
      let height = media.height || 1;

      if (!isSvg && sizeData) {
        const aspectRatio = calculateAspectRatio(source.aspectRatio);

        if (sizeData.width) {
          if (!sizeData.height) {
            height = aspectRatio
              ? sizeData.width * aspectRatio
              : (sizeData.width / width) * height;
          }
          width = sizeData.width;
        }
        if (sizeData.height) {
          if (!sizeData.width) {
            width = aspectRatio
              ? sizeData.height / aspectRatio
              : (sizeData.height / height) * width;
          }
          height = sizeData.height;
        }
      }

      width = Math.round(width);
      height = Math.round(height);

      return {
        url,
        url2x,
        width,
        height,
        screen,
      };
    });
  };

  const optimizeUrl = function (
    imageUrl: string,
    size: ResizeSize,
    options: {
      position?: string;
      key?: string;
      aspectRatio?: ResizeAspectRatio;
      pixelRatio?: 1 | 2;
    } = {},
  ) {
    const sizes = data.latest?.Offer?.sizes;
    if (!sizes) {
      return imageUrl;
    }

    if (!sizes[size as any]) {
      return imageUrl;
    }

    const key = data.latest.key;

    const params = new URLSearchParams();

    params.append("image", imageUrl);
    params.append("key", key);
    params.append("size", size as string);

    const { aspectRatio, pixelRatio } = options;

    if (aspectRatio) {
      params.append("aspectRatio", aspectRatio + "");
    }

    if (pixelRatio) {
      params.append("pixelRatio", pixelRatio + "");
    }

    if (options?.key) {
      params.append("optionsKey", options.key);
      params.append(
        "options",
        JSON.stringify({
          position: options.position,
        }),
      );
    }

    const url = optimizeImage.url + `?${params}`;

    return url;
  };

  return [mediaSource, optimizeUrl, { Source: null as any }] as [
    typeof mediaSource,
    typeof optimizeUrl,
    { Source: Source },
  ];
};
