import { InputFromBlueprintData } from "../../../types";
import {
  BlueprintDataInputLayout,
  BlueprintDataTitle,
} from "../../editor/sidebar/blueprintData";
import { MediaTypeObject, thumResize, useMedia, useResize } from ".";
import { useAuth } from "nitropage";
import { css } from "nitropage/css";
import { createRouteAction, createRouteData, useRouteData } from "solid-start";
import { ButtonSpinner, SmallButton } from "../../common/buttons";
import { uploadFilesRpc } from "../../../server/rpc";
import {
  Component,
  createMemo,
  createSignal,
  For,
  Show,
  useContext,
} from "solid-js";
import Image, { ImageContainer } from "../../image";
import { sortable, createOnEnd } from "../../../directives/sortable";
import {
  deleteMediaRpc,
  getAllMedia,
  setMediaResizePosition,
} from "../../../server/rpc/media";
import { produce } from "solid-js/store";
import { CancelIcon, TrashIcon } from "../../common/icons";
import config from ".";
import { demoEnabled } from "../../../util";
import { adminRouteData } from "../../admin/routeData";
import clsx from "clsx";
import {
  audioTypes,
  fileType,
  imageTypes,
  mediaType,
  videoTypes,
} from "../../../util/media";
import { NitroMedia } from "@prisma/client";
import { ConfirmDialog } from "../../common/dialog";
import { FileInput } from "../../common/inputs";
import { StateContext } from "../../context/page";

const MediaInput: InputFromBlueprintData<ReturnType<typeof config>> = (
  props,
) => {
  sortable;
  const [_, setState] = useContext(StateContext)!;
  const [optimizeMedia] = useResize(thumResize);
  const data = useRouteData<typeof adminRouteData>();
  const auth = useAuth();
  const [__, deleteMediaAction] = createRouteAction(
    async (args: { id: string }) => {
      return await deleteMediaRpc({ id: args.id, csrf: auth()?.csrf! });
    },
    { invalidate: "mediaLibrary" },
  );
  const newFilesAllowed = function () {
    if (props.data.custom?.max == null) {
      return true;
    }

    if (props.value?.mediaIds.length == null) {
      return true;
    }

    return props.data.custom.max > props.value?.mediaIds.length;
  };
  const [deleteMedia, setDeleteMedia] = createSignal<NitroMedia | undefined>();

  let fileInput: HTMLInputElement = undefined!;
  const [uploadState, { Form }] = createRouteAction(
    async (form: FormData) => {
      fileInput.value = null as any;
      form.set("csrf", auth()?.csrf!);
      form.set("projectId", data.editorPage()?.projectId!);
      const files = await uploadFilesRpc(form);
      const replaceFiles = files.length === props.value?.mediaIds.length;
      props.setValue(function (draft) {
        if (replaceFiles) {
          draft.mediaIds = [];
        }
        for (const file of files) {
          if (!newFilesAllowed()) {
            break;
          }
          if (!replaceFiles && props.value?.mediaIds.includes(file.id)) {
            continue;
          }
          draft.mediaIds.push(file.id);
        }
        return draft;
      });
    },
    { invalidate: "mediaLibrary" },
  );

  const [files, filesActions] = useMedia(() => props.value?.mediaIds || [], {
    admin: true,
  });

  const [showFileLibrary, setShowFileLibrary] = createSignal(false);

  const availableFiles = createRouteData(
    function () {
      return getAllMedia({
        projectId: data.editorPage()?.projectId!,
      });
    },
    { key: "mediaLibrary" },
  );

  const acceptedTypes = createMemo(function () {
    const types: string[] = [];

    const type = props.data.custom?.type as MediaTypeObject | undefined;
    if (!type) {
      return types;
    }

    if (type.image) {
      types.push(...imageTypes);
    }
    if (type.video) {
      types.push(...videoTypes);
    }
    if (type.audio) {
      types.push(...audioTypes);
    }

    return types;
  });

  const filteredFiles = createMemo(function () {
    const files = availableFiles();
    const types = acceptedTypes();

    if (!files) {
      return [];
    }

    if (!types.length) {
      return files;
    }

    return files.filter((file) => types.includes(file.mime));
  });

  const MediaThumbnail: Component<{ media: NitroMedia }> = function (props) {
    const isImage = createMemo(function () {
      return fileType(props.media.mime) === "image";
    });

    return (
      <Show
        when={isImage()}
        fallback={
          <div
            class={css({
              fontSize: "0.6rem",
              padding: "0.25rem",
              wordBreak: "break-all",
            })}
          >
            {props.media.name}.{props.media.extension}
          </div>
        }
      >
        <Image
          fadeDuration={1000}
          sources={optimizeMedia(props.media, [{ size: "thumb" }])}
        />
      </Show>
    );
  };

  return (
    <BlueprintDataInputLayout
      header={
        <BlueprintDataTitle {...props}>
          <Show when={props.data.custom?.max != null}>
            {files().byOrder.length} / {props.data.custom?.max}
          </Show>
        </BlueprintDataTitle>
      }
    >
      <ConfirmDialog
        icon={TrashIcon}
        open={!!deleteMedia()}
        onConfirm={async () => {
          const id = deleteMedia()!.id;
          await deleteMediaAction({ id });
          const idx = props.value?.mediaIds.indexOf(id);
          if (idx != null && idx !== -1) {
            props.setValue(function (draft) {
              draft.mediaIds.splice(idx, 1);
              return draft;
            });
          }

          setState(
            produce(function (draft) {
              draft.cacheKey++;
            }),
          );
          setDeleteMedia(undefined);
        }}
        onCancel={() => setDeleteMedia(undefined)}
      >
        Do you really want to delete the file{" "}
        <strong>
          {deleteMedia()!.name}.{deleteMedia()!.extension}
        </strong>
        ?
      </ConfirmDialog>
      <div
        class={css({
          display: "flex",
          flexDirection: "column",
          gap: "1rem",
          padding: "0.5rem 1rem",
        })}
      >
        <Show when={files().byOrder.length}>
          <div
            use:sortable={{
              handle: ".handle",
              animation: 150,
              forceFallback: true,
              onEnd: createOnEnd(function (childProduce) {
                props.setValue(function (draft) {
                  draft.mediaIds = childProduce(draft.mediaIds);
                  return draft;
                });
              }),
            }}
            class={css({
              width: "100%",
              display: "grid",
              gap: "0.5rem",
              gridTemplateColumns: "repeat(5, 1fr)",
            })}
          >
            <For each={props.value?.mediaIds}>
              {(mediaId, idx) => {
                const media = createMemo(() => files().byId[mediaId]!);
                const isImage = createMemo(
                  () => fileType(media()?.mime) === "image",
                );
                return (
                  <div
                    class={css({
                      position: "relative",
                      display: "flex",
                      flexDirection: "column",
                      justifyContent: "flex-end",
                    })}
                  >
                    <Show
                      when={media()}
                      fallback={<div class={css({ height: "60px" })} />}
                    >
                      <ImageContainer
                        height="60px"
                        fit="contain"
                        title={`${media().name}.${media().extension}`}
                        class={clsx(
                          "handle",
                          css({
                            ":hover": {
                              cursor: "grab",
                            },
                            backgroundColor: "#f3f4f6",
                            boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                            overflow: "hidden",
                            borderRadius: "1rem",
                            border: "1px solid rgb(156, 163, 175)",
                          }),
                          isImage() && "np-bg-iso",
                        )}
                      >
                        <MediaThumbnail media={media()} />
                      </ImageContainer>
                    </Show>
                    <SmallButton
                      title="Unselect file"
                      onClick={() => {
                        props.setValue(function (draft) {
                          draft.mediaIds.splice(idx(), 1);
                          return draft;
                        });
                      }}
                      style={{
                        "z-index": 1,
                        "background-color": "#e5e7eb55",
                        "backdrop-filter": "blur(8px)",
                        padding: "0.25rem",
                        top: "-0.25rem",
                        "font-size": "0.75rem",
                        right: "-0.25rem",
                        position: "absolute",
                        "border-radius": "1rem",
                      }}
                    >
                      <CancelIcon />
                    </SmallButton>
                  </div>
                );
              }}
            </For>
          </div>
        </Show>
        <div class={css({ display: "flex", gap: "1.5rem" })}>
          <Show when={!demoEnabled()}>
            <Form
              method="post"
              enctype="multipart/form-data"
              class={css({
                flex: "1",
                padding: "0.25rem 0.75rem",
                boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                border: "1px solid #e2e8f0",
                borderRadius: "0.25rem",
                display: "flex",
                alignItems: "center",
                fontSize: "0.9rem",
              })}
            >
              <FileInput
                ref={fileInput}
                name="files"
                multiple
                required
                accept={acceptedTypes().join(",")}
              />
              <SmallButton type="submit" theme="primary">
                <ButtonSpinner show={uploadState.pending}>Upload</ButtonSpinner>
              </SmallButton>
            </Form>
          </Show>
          <Show when={filteredFiles()?.length}>
            <SmallButton
              theme="secondary"
              onClick={() => setShowFileLibrary(!showFileLibrary())}
            >
              Library
            </SmallButton>
          </Show>
        </div>

        <Show when={showFileLibrary()}>
          <div
            class={css({
              width: "100%",
              display: "grid",
              gap: "0.5rem",
              gridTemplateColumns: "repeat(7, 1fr)",
              padding: "0 0.5rem",
              "& .item": {
                minWidth: 0,
                height: "6rem",
                display: "flex",
                flexDirection: "column",
              },
              "& select": {
                padding: "0.125rem",
                width: "100%",
                textAlign: "center",
                fontSize: "0.75rem",
              },
            })}
          >
            <For each={filteredFiles()}>
              {(file) => {
                const fileType = createMemo(() => mediaType(file));
                const isImage = createMemo(() => fileType() === "image");
                return (
                  <div class="item">
                    <ImageContainer
                      onClick={() => {
                        if (props.value?.mediaIds.includes(file.id)) {
                          return;
                        }
                        if (newFilesAllowed()) {
                          props.setValue(function (draft) {
                            draft.mediaIds.push(file.id);
                            return draft;
                          });
                        } else if (
                          props.value?.mediaIds.length === 1 &&
                          props.data.custom?.max === 1
                        ) {
                          props.setValue(function (draft) {
                            draft.mediaIds.splice(0, 1);
                            draft.mediaIds.push(file.id);
                            return draft;
                          });
                        }
                      }}
                      title={file.name + "." + file.extension}
                      fit="contain"
                      class={clsx(
                        css({
                          ":hover": {
                            cursor: "pointer",
                          },
                          flex: 1,
                          overflow: "hidden",
                          backgroundColor: "#f3f4f6",
                          borderTop: "1px solid rgb(156, 163, 175)",
                          borderLeft: "1px solid rgb(156, 163, 175)",
                          borderRight: "1px solid rgb(156, 163, 175)",
                          borderTopLeftRadius: "1rem",
                          borderTopRightRadius: "1rem",
                          "& img": {
                            transition: "all 0.25s ease-out",
                          },
                          ":hover img": {
                            transform: "scale(1.5)",
                            transition: "all 0.75s ease-out",
                          },
                        }),
                        isImage() && "np-bg-iso",
                      )}
                    >
                      <MediaThumbnail media={file} />
                    </ImageContainer>
                    <Show when={isImage()}>
                      <select
                        class={css({
                          borderTop: "1px solid rgb(156, 163, 175)",
                          borderLeft: "1px solid rgb(156, 163, 175)",
                          borderRight: "1px solid rgb(156, 163, 175)",
                        })}
                        onChange={async (e) => {
                          const value = e.currentTarget.value;
                          await setMediaResizePosition({
                            id: file.id,
                            resizePosition: value == "" ? undefined : value,
                            csrf: auth()?.csrf!,
                          });
                          await filesActions.refetch();
                          setState(
                            produce(function (draft) {
                              draft.cacheKey++;
                            }),
                          );
                        }}
                      >
                        <For
                          each={[
                            "none",
                            "attention",
                            "entropy",
                            "center",
                            "north",
                            "northeast",
                            "northwest",
                            "south",
                            "southeast",
                            "southwest",
                            "east",
                            "west",
                          ]}
                        >
                          {(item) => {
                            return (
                              <option
                                value={item === "none" ? "" : item}
                                selected={item === file.resizePosition}
                              >
                                {item}
                              </option>
                            );
                          }}
                        </For>
                      </select>
                    </Show>
                    <SmallButton
                      title="Delete file"
                      onClick={async () => {
                        setDeleteMedia(file);
                      }}
                      style={{
                        "z-index": 1,
                        "font-size": "0.75rem",
                        width: "100%",
                        "border-top-left-radius": "0rem",
                        "border-top-right-radius": "0rem",
                        "border-bottom-left-radius": "1rem",
                        "border-bottom-right-radius": "1rem",
                      }}
                    >
                      <TrashIcon />
                    </SmallButton>
                  </div>
                );
              }}
            </For>
          </div>
        </Show>
      </div>
    </BlueprintDataInputLayout>
  );
};

export default MediaInput;
