import {
  Component,
  createEffect,
  createMemo,
  For,
  onMount,
  ParentComponent,
  Show,
  splitProps,
  Suspense,
  useContext,
} from "solid-js";
import { produce } from "solid-js/store";
import { Dynamic } from "solid-js/web";
import {
  Blueprint,
  DragSlotFunction,
  Element,
  ElementData,
  RteComponent,
  SlotComponent,
  State,
} from "../../../types";
import RteEditor from "../../rteEditor";
import { createOnEnd, sortable } from "../../../directives/sortable";
import {
  pageViewSortableConfig,
  getBuilderInfoFromSortableEvent,
  useSortableElementMove,
  slotHasSpace,
  blueprintAllowed,
} from "../../../admin/element";
import { css } from "../../../css/fela";
import classes from "./element.module.css";
import { toolbarDefaults, Value } from "../../data/rte";
import { setLookupRteToolbar, useElementData } from "../../../util/elementData";
import { ErrorBoundary } from "solid-start";
import { SmallButton } from "../../common/buttons";
import { Popover } from "@kobalte/core";
import { klona } from "klona";
import { Richtext } from "../../page";
import { blueprintDataTitle } from "../../../admin/editor";
import { capitalize } from "froebel";
import clsx from "clsx";
import gridImage from "../../../../assets/editor-hover-grid.png";
import { ConfigContext, StateContext } from "../../context/page";
import { EditorUiContext } from "../../context/admin";

const FrameElement: ParentComponent<{
  element: Element;
  blueprint: Blueprint;
  last?: boolean;
  outline?: "vertical" | "horizontal";
}> = (p) => {
  const [state, setState] = useContext(StateContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;

  const preset = createMemo(function () {
    return p.element.presetId ? state.presets[p.element.presetId] : undefined;
  });

  const [declaration] = useContext(ConfigContext)!;
  const Rte: RteComponent = (ep) => {
    const toolbar = createMemo(function () {
      if (ep.path) {
        setLookupRteToolbar(true);
        try {
          const toolbar = ep.path((d) => data[d]);
          setLookupRteToolbar(false);
          return toolbar;
        } catch (err) {
          console.error(err);
        }

        setLookupRteToolbar(false);
      }

      if (ep.data) {
        return p.blueprint.data[ep.data].rte?.toolbar;
      }
    });

    let dataKey = "";

    const value = createMemo(() => {
      if (ep.path) {
        return ep.path((d) => {
          dataKey = d;
          return data[d];
        });
      }

      if (ep.data) {
        dataKey = ep.data;
        return data[ep.data];
      }
    });

    // Figures out where to store incoming data, depending on overrides and ui state
    const getWriteMode = createMemo<[string, boolean]>(function () {
      const elementOverride =
        !!state.elements[p.element.id].data[dataKey]?.override;
      if (editorUi.sidebarElementsTab === "element" || elementOverride) {
        return ["element", elementOverride];
      }

      const preset_ = preset();
      if (preset_) {
        const presetOverride = !!preset_.data[dataKey]?.override;
        if (
          editorUi.sidebarElementsTab === "preset" ||
          editorUi.sidebarElementsTab === "presets" ||
          presetOverride
        ) {
          return ["preset", presetOverride];
        }
      }

      return ["project", true];
    });

    const overridePending = createMemo(function () {
      if (!ep.path) {
        return false;
      }

      return !getWriteMode()[1];
    });

    const showOverride = createMemo(
      () => overridePending() && editorUi.toolbarStatus === "write",
    );

    return (
      <Dynamic
        component={ep.component || "span"}
        title={
          showOverride()
            ? `Override "${blueprintDataTitle(
                p.blueprint.data[dataKey],
                dataKey,
              )}" in ${capitalize(getWriteMode()[0])}?`
            : undefined
        }
        class={clsx(ep.class, showOverride() && "ql-override")}
        onClick={() => {
          if (!showOverride()) {
            return;
          }
          setEditorUi(
            produce(function (draft) {
              draft.selectedElementId = p.element.id;
            }),
          );

          const [writeMode] = getWriteMode();

          // Clones the data from the proper parent
          setState(
            produce((draft) => {
              let initData: undefined | ElementData;
              let elementDataDraft: undefined | ElementData;

              const projectData =
                state.blueprintDefaults.project[p.blueprint.id][dataKey];

              if (writeMode === "preset") {
                elementDataDraft =
                  draft.presets[p.element.presetId!].data[dataKey];
                initData = projectData;
              }
              if (writeMode === "element") {
                elementDataDraft = draft.elements[p.element.id].data[dataKey];
                const presetData = preset()?.data[dataKey];
                initData = presetData?.override ? presetData : projectData;
              }

              if (!elementDataDraft) {
                return;
              }

              elementDataDraft.override = true;
              elementDataDraft.value = klona(initData?.value);
            }),
          );
        }}
      >
        <Show
          when={!overridePending()}
          fallback={<Richtext>{value()}</Richtext>}
        >
          <RteEditor
            inline={true}
            toolbar={toolbar() || toolbarDefaults}
            disable={editorUi.toolbarStatus != "write"}
            value={value()}
            onFocus={function () {
              setEditorUi(
                produce(function (draft) {
                  draft.selectedElementId = p.element.id;
                }),
              );
            }}
            setValue={(v) => {
              setState(
                produce<State>(function (draft) {
                  const getDataRecord = (data: string) => {
                    const [writeMode] = getWriteMode();

                    let rootDataDraft =
                      draft.blueprintDefaults.project[p.blueprint.id];

                    if (writeMode === "preset") {
                      rootDataDraft = draft.presets[p.element.presetId!].data;
                      rootDataDraft[data].override = true;
                    }

                    if (writeMode === "element") {
                      rootDataDraft = draft.elements[p.element.id].data;
                      rootDataDraft[data].override = true;
                    }

                    return rootDataDraft[data];
                  };

                  const dataDraft: Value = ep.path
                    ? ep.path((data) => getDataRecord(data).value)
                    : getDataRecord(ep.data!).value;
                  dataDraft.content = v;
                }),
              );
            }}
          />
        </Show>
      </Dynamic>
    );
  };

  sortable;

  const data = useElementData(state, p.element.id);

  const handleElementMove = useSortableElementMove(() => p.element.id);

  const useDragSlot: DragSlotFunction = function (el, config) {
    const sortableInstance = sortable(el, () => ({
      ...pageViewSortableConfig,
      handle: ".handle-" + p.element.id,
      group: {
        name: "shared",
        pull: true,
      },
      direction: config().direction || "vertical",
      dragClass: "np-drag-card",
      ghostClass: "np-ghost-card",
      filter: ".drag-disabled",
      onStart: (evt) => {
        const builderInfo = getBuilderInfoFromSortableEvent(evt, state);
        const elementId = builderInfo.element?.id;
        if (!elementId) return false;
        setEditorUi(
          produce(function (draft) {
            draft.movingElementId = elementId;
            draft.createMoveTarget.elementId = p.element.parentId || "root";
            draft.createMoveTarget.slot = p.element.parentSlot;
          }),
        );
      },
      onMove: (evt) => {
        const builderInfo = getBuilderInfoFromSortableEvent(evt, state);

        if (evt.from === evt.to) {
          setEditorUi(
            produce(function (draft) {
              draft.createMoveTarget.elementId = builderInfo.targetElementId;
              draft.createMoveTarget.slot = builderInfo.targetSlot;
            }),
          );

          return true;
        }

        if (
          builderInfo.targetElementId === "root" &&
          builderInfo.targetSlot !== "default"
        ) {
          return false;
        }

        if (!evt.to.dataset.slotAccepting) {
          return false;
        }

        setEditorUi(
          produce(function (draft) {
            draft.createMoveTarget.elementId = builderInfo.targetElementId;
            draft.createMoveTarget.slot = builderInfo.targetSlot;
          }),
        );
      },
      onEnd: createOnEnd((childProduce, evt) => {
        setEditorUi(
          produce(function (draft) {
            draft.createMoveTarget = {};
          }),
        );

        return handleElementMove(childProduce, evt);
      }),
    }));
    if (sortableInstance) {
      createEffect(function () {
        sortableInstance.option("direction", config().direction || "vertical");
      });
    }
  };

  const Handles = () => {
    const moveHandleClass = createMemo(function () {
      return "handle-" + (state.elements[p.element.id]?.parentId || "root");
    });
    const outline = () => {
      if (editorUi.toolbarStatus === "creating") {
        return true;
      }
      if (editorUi.toolbarStatus === "moving" && editorUi.movingElementId) {
        return true;
      }
      return false;
    };
    ["creating", "moving"].includes(editorUi.toolbarStatus);
    const showing = createMemo(() => {
      return (
        ["deleting", "selecting", "moving"].includes(editorUi.toolbarStatus) ||
        editorUi.hoveredElementId === p.element.id
      );
    });

    const hoverStyles = {
      opacity: 1,
      backgroundImage: `url(${gridImage})`,
    };
    let handleRef: HTMLDivElement = undefined!;

    const style = createMemo(function () {
      return css({
        display: !showing() || editorUi.movingElementId ? "none" : undefined,
        cursor: (
          {
            selecting: "pointer",
            deleting: "no-drop",
            moving: "move",
          } as Record<string, string>
        )[editorUi.toolbarStatus],
        position: "absolute",
        zIndex: 9999,
        left: "0px",
        right: "0px",
        top: "0px",
        bottom: "0px",
        color: "white",
        fontSize: "3rem",
        lineHeight: 1,
        userSelect: "none",
        transitionProperty: "opacity",
        transitionTimingFunction: "cubic-bezier(0, 0, 0.2, 1)",
        transitionDuration: "150ms",
        opacity: 0,
        "&.selectable:active": {
          backgroundColor: "rgb(255 255 255 / 0.3)",
        },
        "&.hover": hoverStyles,
      });
    });

    const hovered = () =>
      (editorUi.hoveredElementId === p.element.id &&
        !editorUi.movingElementId) ||
      p.element.id === editorUi.deletingElementId;

    return (
      <>
        <Show when={outline() && p.outline}>
          <div
            class={clsx(
              classes["outline"],
              classes["bg-rainbow"],
              classes[
                p.outline === "horizontal" ? "outline--left" : "outline--top"
              ],
            )}
          />
          <Show when={p.last}>
            <div
              class={clsx(
                classes["outline"],
                classes["bg-rainbow"],
                classes[
                  p.outline === "horizontal"
                    ? "outline--right"
                    : "outline--bottom"
                ],
              )}
            />
          </Show>
        </Show>
        <Show when={state.elements[p.element.id]?.blueprint !== "root"}>
          <div
            ref={handleRef}
            onClick={(evt) => {
              if (editorUi.toolbarStatus === "deleting") {
                evt.stopPropagation();
                setEditorUi(
                  produce((d) => (d.deletingElementId = p.element.id)),
                );
              }
              if (editorUi.toolbarStatus === "selecting") {
                evt.stopPropagation();
                setEditorUi("selectedElementId", p.element.id);
                setEditorUi("sidebarIdleTab", "elements");
              }
            }}
            data-np-handle-id={p.element.id}
            class={clsx(
              style(),
              editorUi.selectedElementId !== p.element.id && "selectable",
              hovered() && "hover",
              editorUi.toolbarStatus === "moving" && moveHandleClass(),
            )}
          >
            <Popover.Root
              flip={false}
              anchorRef={() => handleRef}
              open={hovered()}
              slide={false}
              arrowPadding={0}
              overlap={false}
              modal={false}
              placement="top-start"
              gutter={0}
            >
              <Popover.Portal>
                <Popover.Content
                  class={css({
                    left: 0,
                    pointerEvents: "none !important",
                    "> .title": {
                      position: "relative",
                      display: "flex",
                      gap: "0.5rem",
                      justifyContent: "center",
                      alignItems: "center",
                      zIndex: 1,
                      top: "-30px",
                      fontSize: "1.5rem",
                      fontWeight: "700",
                      textAlign: "center",
                      "--color": "#FFFF",
                      textShadow:
                        "1px 1px var(--color), -1px -1px var(--color), -1px 0px var(--color), 0px -1px var(--color), 1px 0px var(--color), 0px 1px var(--color)",
                      "& span": {
                        fontSize: "0.75rem",
                      },
                    },
                  })}
                  style={{
                    position: "absolute",
                    outline: "none",
                    "z-index": 10000,
                    width: handleRef.clientWidth + "px",
                    height: handleRef.clientHeight + "px",
                  }}
                >
                  <div class="title">
                    {p.blueprint.title()}
                    <Show when={data.id?.length}>
                      <span>#{data.id}</span>
                    </Show>
                  </div>
                  <SelectionBorder />
                </Popover.Content>
              </Popover.Portal>
            </Popover.Root>
          </div>
        </Show>
      </>
    );
  };

  const lax = createMemo(function () {
    return !["creating", "moving", "deleting", "selecting"].includes(
      editorUi.toolbarStatus,
    );
  });

  const Slot: SlotComponent = function (ps) {
    const [localProps, restProps] = splitProps(ps, [
      "editorPadding",
      "direction",
      "name",
      "ref",
      "optional",
      "component",
    ]);
    const elements = createMemo(function () {
      if (p.element.blueprint === "root") {
        return state.root[ps.name as "default"];
      }

      return p.element.slots[ps.name]?.elements;
    });

    const acceptsElements = createMemo(() => {
      const slotName = ps.name || "default";
      if (!slotHasSpace(declaration, p.element, slotName)) {
        return false;
      }
      const element =
        editorUi.movingElementId && state.elements[editorUi.movingElementId];

      if (
        element &&
        !blueprintAllowed(
          declaration,
          element.blueprint,
          p.element.blueprint,
          slotName,
        )
      ) {
        return false;
      }

      return true;
    });

    const ElementsView = () => {
      return (
        <For each={elements()}>
          {(childElementId, idx) => {
            const blueprint = createMemo(function () {
              return declaration.blueprints[
                state.elements[childElementId].blueprint
              ];
            });
            return (
              <Show
                when={blueprint()}
                fallback={
                  <div
                    class={css({
                      padding: "0.75rem",
                      "--tw-ring-offset-shadow":
                        "var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color)",
                      "--tw-ring-shadow":
                        "var(--tw-ring-inset) 0 0 0 calc(2px + var(--tw-ring-offset-width)) var(--tw-ring-color)",
                      boxShadow:
                        "var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000)",
                      "--tw-ring-opacity": 1,
                      "--tw-ring-color":
                        "rgb(254 202 202 / var(--tw-ring-opacity))",
                    })}
                    {...elementProps()}
                  >
                    Blueprint not found
                  </div>
                }
              >
                <FrameElement
                  outline={
                    acceptsElements()
                      ? localProps.direction || "vertical"
                      : undefined
                  }
                  blueprint={blueprint()}
                  last={elements().length === idx() + 1}
                  element={state.elements[childElementId]}
                />
              </Show>
            );
          }}
        </For>
      );
    };

    const SlotInner = function () {
      let slotEl: HTMLElement = undefined!;

      onMount(function () {
        const config = createMemo(function () {
          return {
            slot: ps.name,
            direction: ps.direction,
          };
        });

        useDragSlot(slotEl, config);
      });

      return (
        <Dynamic
          component={ps.component || "div"}
          data-element-id={p.element.id}
          data-slot={ps.name || "default"}
          data-slot-accepting={acceptsElements() ? true : undefined}
          data-element-preview={
            p.element.id === editorUi.creatingElementId ? true : undefined
          }
          data-editor-padding={ps.editorPadding !== false ? true : undefined}
          ref={(el: any) => {
            slotEl = el;
            (ps.ref as any)?.(el);
          }}
          {...restProps}
        >
          <Show when={elements()?.length}>
            <ElementsView />
          </Show>
        </Dynamic>
      );
    };

    return (
      <Show
        when={
          !localProps.optional ||
          elements()?.length ||
          (p.element.id !== "root" &&
            ["selecting", "creating", "moving"].indexOf(
              editorUi.toolbarStatus,
            ) != -1) ||
          editorUi.creatingElementId != null
        }
      >
        <SlotInner />
      </Show>
    );
  };

  const SelectionBorder: Component = (p) => {
    return (
      <div class={classes["selection-border"]}>
        <div
          class={clsx(
            classes["selection-border"],
            classes["selection-border--inner"],
          )}
        ></div>
      </div>
    );
  };

  const elementProps = createMemo(function () {
    return {
      id: data.id?.length ? data.id : undefined,
      "data-element": p.element.id,
    };
  });

  const slotElements = function (name: string) {
    return p.element.slots?.[name]?.elements || [];
  };

  const Element = () => {
    return (
      <>
        <Show when={p.element?.blueprint !== "none"}>
          <Suspense>
            <Show when={p.blueprint?.component} keyed>
              <ErrorBoundary
                fallback={(err, reset) => (
                  <div
                    class={
                      "np-admin-font " +
                      css({
                        fontSize: "1.1rem",
                        display: "flex",
                        flexDirection: "column",
                        background: "white",
                        color: "#DC2626",
                        padding: "1rem",
                        borderRadius: "1rem",
                        gap: "0.5rem",
                      })
                    }
                    {...elementProps()}
                  >
                    <div>
                      Blueprint <strong>{p.blueprint.title()}</strong> failed to
                      render.
                    </div>
                    <div>{err.toString()}</div>
                    <div>
                      <SmallButton onClick={() => reset()}>Retry</SmallButton>
                    </div>
                    <Handles />
                  </div>
                )}
              >
                <Dynamic
                  admin={true}
                  blueprint={p.blueprint}
                  preview={p.element.id === editorUi.creatingElementId}
                  element={{ id: p.element.id }}
                  elementProps={elementProps()}
                  handles={<Handles />}
                  Slot={Slot}
                  Rte={Rte}
                  component={p?.blueprint?.component as any}
                  slotElements={slotElements}
                  data={data}
                  lax={lax()}
                />
              </ErrorBoundary>
            </Show>
          </Suspense>
        </Show>
        <Show when={p.element?.blueprint === "none"}>
          <div
            class={clsx(
              "np-admin-font",
              classes["bg-rainbow"],
              classes["no-blueprint-element-placeholder"],
            )}
            {...elementProps()}
          >
            Choose a blueprint in the sidepanel
            <Handles />
          </div>
        </Show>
      </>
    );
  };

  return <Element />;
};

export default FrameElement;
