import {
  Component,
  createEffect,
  createMemo,
  createSignal,
  lazy,
  onCleanup,
  onMount,
  Show,
  untrack,
  useContext,
} from "solid-js";
import { createStore, produce, reconcile, unwrap } from "solid-js/store";
import {
  Blueprint,
  CustomBlueprintComponent,
  State,
  Element,
} from "../../../types";
import { sortable } from "../../../directives/sortable";
import { onClientCleanup } from "../../../util";
import {
  createSlot,
  slotHasSpace,
  pageViewSortableConfig,
  getBuilderInfoFromSortableEvent,
} from "../../../admin/element";
import { createId } from "@paralleldrive/cuid2";
import { useMousePosition } from "@solid-primitives/mouse";
import { debounce } from "froebel";
import Sortable from "sortablejs";
import { createElementData } from "../sidebar/blueprintData";
import { PlusIcon } from "../../common/icons";
import { css } from "../../../css/fela";
import "../../../css/editor.css";
import Styles from "../../page/styles";
import clsx from "clsx";
import {
  EditorIframeEvent,
  isElementInViewport,
  useDarkMode,
  useScrollRestoration,
} from "../../../admin/editor";
import { ConfigContext, StateContext } from "../../context/page";
import { createAppStoreState } from "../../../stores/state";
import { createEditorUiStore } from "../../../stores/admin";
import { EditorUiContext, ElementFocusContext } from "../../context/admin";
import { createIframeMessageQueue } from "../../../admin/iframeMessageQueue";

const FrameElement = lazy(async () => import("./element"));

const Frame: Component = function (props) {
  const [config] = useContext(ConfigContext)!;
  const [state, setState_] = createStore<State>(
    createAppStoreState({ admin: true }),
  );
  const [editorUi, setEditorUi_] = createEditorUiStore({
    pageId: "",
    revisionId: "",
  });

  const focusElementById = function (id: string) {
    const el = document.querySelectorAll(`[data-element="${id}"]`)[0];
    if (!el) {
      return;
    }

    const isCreatingElement = editorUi.creatingElementId == id;
    if (!isCreatingElement && isElementInViewport(el)) {
      return;
    }

    el.scrollIntoView({
      behavior: isCreatingElement ? "instant" : "smooth",
      block: "center",
    });
  };

  const [connected, setConnected] = createSignal(false);
  const [send, schedule] = createIframeMessageQueue<EditorIframeEvent>(
    "editor-frame",
    function (event) {
      if (event.type === "setState") {
        setState_(reconcile(event.data, { key: null, merge: true }));
        if (!connected()) {
          setConnected(true);
        }
        return;
      }

      if (event.type === "setEditorUi") {
        setEditorUi_(reconcile(event.data, { key: null, merge: true }));
        if (!connected()) {
          setConnected(true);
        }
        return;
      }

      if (event.type === "focusElement") {
        focusElementById(event.data);
        return;
      }
    },
  );

  const sendState = debounce(
    () =>
      send({
        type: "setState",
        data: unwrap(state),
      }),
    50,
  );

  const savePageShortcutHandler = (e: KeyboardEvent) => {
    if (!e.ctrlKey) {
      return;
    }

    if (e.key === "s") {
      e.preventDefault();
      send({ type: "savePage" });
      return;
    }

    if (e.key === "p") {
      e.preventDefault();
      send({ type: "publishPage" });
      return;
    }
  };

  onMount(function () {
    document.addEventListener("keydown", savePageShortcutHandler);
  });

  onClientCleanup(function () {
    document.removeEventListener("keydown", savePageShortcutHandler);
  });

  const setState: typeof setState_ = schedule(function (...args: any) {
    (setState_ as any)(...args);
    sendState();
  });
  const setEditorUi: typeof setEditorUi_ = schedule(function (...args: any) {
    (setEditorUi_ as any)(...args);
    send({
      type: "setEditorUi",
      data: unwrap(editorUi),
    });
  });

  useDarkMode(editorUi);

  // TODO: Implement zoom
  // This doesn't fully work, e.g. it breaks the mouse-following icon during element creation
  // Since transform scale on the body breaks ALL nested mouse tracking,
  // nitropage would need to provide a custom createMousePosition primitive
  /*
  createEffect(function() {
    const zoomLevel = 0.5;
    document.body.style['transform'] = `scale(${zoomLevel})`;
    document.body.style['width'] = `calc(1/${zoomLevel}*100%)`;
    document.body.style['transformOrigin'] = `0 0`;
  })
  */

  const Root: CustomBlueprintComponent = (p) => {
    const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
    const style = createMemo(function () {
      return css({
        paddingBottom: p.lax ? undefined : "1.25rem",
        paddingTop: p.lax ? undefined : "1.25rem",
        height: state.root.default.length === 0 ? "100%" : undefined,
      });
    });

    return (
      <div
        class={clsx(
          "np-root",
          style(),
          !p.lax && "np-slot-expand np-elements-interact",
        )}
        onMouseOver={(evt) => {
          const el = evt.target as HTMLDivElement;
          const elementId = el?.dataset.npHandleId;
          if (!elementId) {
            return;
          }
          setEditorUi("hoveredElementId", elementId);
        }}
        onMouseOut={(evt) => {
          const el = evt.target as HTMLDivElement;
          const elementId = el?.dataset.npHandleId;
          if (editorUi.hoveredElementId !== elementId) {
            return;
          }
          setEditorUi("hoveredElementId", undefined);
        }}
      >
        <p.Slot name="header" component="header" optional />
        <p.Slot name="default" component="main" />
        <p.Slot name="footer" component="footer" optional />
      </div>
    );
  };

  const rootBlueprint: Blueprint = {
    id: "root",
    title: () => "Root",
    data: {},
    slots: {},
    component: Root,
  };

  const pos = useMousePosition();

  const [createHandleEl, setCreateHandleEl] = createSignal<HTMLDivElement>();
  const dispatchMouseEvent = (type: string, EventType = MouseEvent) => {
    const handleEl = createHandleEl();
    if (!handleEl) return;
    const el = handleEl.children[0];
    if (!el) {
      return;
    }
    const mouseY = untrack(() => pos.y);
    const scrollY = document.documentElement.scrollTop;
    const mousedown = new EventType(type, {
      bubbles: true,
      cancelable: true,
      view: window,
      detail: 1,
      screenX: 1,
      screenY: mouseY,
      clientX: 0,
      clientY: -scrollY,
      ctrlKey: false,
      altKey: false,
      shiftKey: false,
      metaKey: false,
      button: 0,
      relatedTarget: el,
    });

    el.dispatchEvent(mousedown);
  };

  let inserting = false;
  const startCreating = () => {
    inserting = true;
    // Its very likely that the document height changes when the user starts inserting an element.
    // We give the browser some time to update the document height, before triggering the insert mode.
    // This way the cursor position of the insert icon will be properly tracked.
    setTimeout(function () {
      dispatchMouseEvent("pointerdown");
    }, 100);
  };
  onMount(function () {
    createEffect(function () {
      if (editorUi.toolbarStatus !== "creating") {
        inserting = false;
        dispatchMouseEvent("mouseup");
        // createAndDispatch("dragend");
        return;
      }

      startCreating();

      //createAndDispatch("MouseEvent", "mousedown");
      // createAndDispatch("MouseEvent", "pointerup");
      // createAndDispatch("dragstart");
    });
  });

  sortable;

  let draggedEl: HTMLElement | undefined;
  onCleanup(function () {
    if (!draggedEl) {
      return;
    }
    draggedEl.remove();
  });

  const CreateElementSortable = () => {
    return (
      <div
        class={css({
          position: "absolute",
          "& > .np-create-card": {
            visibility: "hidden",
          },
        })}
        ref={(el) => setCreateHandleEl(el)}
        use:sortable={{
          ...pageViewSortableConfig,
          animation: 150,
          group: {
            name: "shared",
            pull: "clone",
            put: false,
          },
          direction: "vertical",
          dragClass: "np-drag-card",
          ghostClass: "np-ghost-card",
          onStart: (evt) => {
            draggedEl = Sortable.ghost!;
            setEditorUi("toolbarStatus", "creating");
          },
          onMove: (evt) => {
            const builderInfo = getBuilderInfoFromSortableEvent(evt, state);
            if (
              builderInfo.targetElementId === "root" &&
              builderInfo.targetSlot !== "default"
            ) {
              return false;
            }
            const result = slotHasSpace(
              config,
              builderInfo.targetElement,
              builderInfo.targetSlot,
            );
            if (!result) {
              return false;
            }

            setEditorUi(
              produce(function (draft) {
                draft.createMoveTarget.elementId = builderInfo.targetElementId;
                draft.createMoveTarget.slot = builderInfo.targetSlot;
              }),
            );
          },
          onEnd: (evt) => {
            setEditorUi(
              produce(function (draft) {
                draft.createMoveTarget = {};
              }),
            );

            if (!inserting) {
              return false;
            }

            const newIndex = evt.newIndex;

            // Element was placed in the own list, just restart the creating in this case
            if (evt.from === evt.to || newIndex == null) {
              requestAnimationFrame(function () {
                startCreating();
              });

              return;
            }

            const builderInfo = getBuilderInfoFromSortableEvent(evt, state);

            const targetSlot = builderInfo.targetSlot;

            const newElement: Element = {
              id: createId(),
              elementId: "",
              parentSlot: targetSlot!,
              blueprint: "none",
              data: {},
              ssrData: {},
              slots: {},
            };

            const targetElementId = builderInfo.targetElementId;
            if (targetElementId != "root") {
              newElement.parentId = targetElementId;
              if (targetSlot && targetElementId) {
                const parentElement = state.elements[targetElementId];
                const parentBlueprint =
                  config.blueprints[parentElement.blueprint];
                const slotBlueprint =
                  parentBlueprint.slots &&
                  parentBlueprint.slots[targetSlot]?.blueprint;
                if (slotBlueprint) {
                  newElement.blueprint = slotBlueprint;
                  newElement.data = createElementData(config, slotBlueprint);
                }
              }
            }

            setState("elements", newElement.id, newElement);

            if (targetElementId && targetSlot && targetElementId != "root") {
              setState(
                "elements",
                targetElementId,
                produce(function (e) {
                  if (!e.slots[targetSlot]) {
                    e.slots[targetSlot] = createSlot();
                  }
                  e.slots[targetSlot].elements.splice(
                    newIndex,
                    0,
                    newElement.id,
                  );
                }),
              );
            } else {
              setState(
                "root",
                targetSlot as "default",
                produce(function (e) {
                  e.splice(newIndex, 0, newElement.id);
                }),
              );
            }

            // If the new element has a blueprint we can skip the manual blueprint selection and directly go back to creating further elements
            if (newElement.blueprint !== "none") {
              requestAnimationFrame(function () {
                startCreating();
              });

              return;
            }

            setEditorUi("creatingElementId", newElement.id);
            setEditorUi("sidebarStatus", "chooseType");
            setEditorUi("toolbarStatus", "locked");
            // onDragEnd(evt)
          },
        }}
      >
        <div class="np-drag-indicator np-create-card">
          <div class="np-drag-indicator__circle">
            <PlusIcon />
          </div>
        </div>
      </div>
    );
  };

  useScrollRestoration(editorUi, focusElementById);

  return (
    <EditorUiContext.Provider value={[editorUi, setEditorUi]}>
      <StateContext.Provider value={[state, setState]}>
        <ElementFocusContext.Provider value={focusElementById}>
          <Styles />
          <CreateElementSortable />
          <Show when={connected() && config.blueprints["none"]}>
            <FrameElement
              blueprint={rootBlueprint}
              element={{
                blueprint: "root",
                elementId: "root",
                parentSlot: "default",
                id: "root",
                slots: {},
                data: {},
                ssrData: {},
              }}
            />
          </Show>
        </ElementFocusContext.Provider>
      </StateContext.Provider>
    </EditorUiContext.Provider>
  );
};

export default Frame;
