import {
  batch,
  Component,
  createEffect,
  createMemo,
  createSignal,
  Show,
  untrack,
  useContext,
} from "solid-js";
import { produce, reconcile, unwrap } from "solid-js/store";
import { throttle } from "froebel";
import Toolbar from "./toolbar";
import Sidebar from "./sidebar";
import "../../css/editor.css";
import classes from "./index.module.css";
import { Title, useParams, useRouteData } from "solid-start";
import { adminRouteData } from "../admin/routeData";
import {
  normalizePage,
  normalizeSettings,
  normalizePresets,
  normalizeFonts,
} from "../../util/normalization";
import { useAuth } from "nitropage";
import { css } from "../../css";
import { LoaderIcon } from "../common/icons";
import Styles from "../page/styles";
import { StyleContainer } from "../admin/util";
import { EditorIframeEvent, useDarkMode } from "../../admin/editor";
import type { Settings } from "../../server/settings";
import { State } from "../../types";
import { createPageStore } from "../../stores/state";
import { createEditorUiStore, createSettingsStore } from "../../stores/admin";
import { ConfigContext, StateContext } from "../context/page";
import {
  EditorUiContext,
  ElementFocusContext,
  SettingsContext,
} from "../context/admin";
import {
  createFontData,
  createPresetsData,
  createRevisionData,
  createSettingsPageData,
  createSettingsProjectData,
} from "./routeData";
import DeleteElementDialog from "./dialogs/deleteElementDialog";
import { createIframeMessageQueue } from "../../admin/iframeMessageQueue";

const Frame: Component<{ ref: any }> = function (p) {
  let el: HTMLIFrameElement;
  return (
    <div
      class={classes.frame}
      onWheel={(evt) => {
        el.contentWindow?.scrollTo({
          top: el.contentWindow.scrollY + evt.deltaY / 2,
        });
      }}
    >
      <div class={classes["frame__scroll-area"]}>
        <iframe
          ref={(el_) => {
            el = el_;
            p.ref(el_);
          }}
          src="/admin/frame"
        />
      </div>
    </div>
  );
};

const Editor: Component = () => {
  const params = useParams();
  const data = useRouteData<typeof adminRouteData>();
  const user = useAuth();
  const [state, setState_, mounted] = createPageStore(params.id, {
    admin: true,
    pendingDeletes: {
      presets: [],
    },
  });
  const [config] = useContext(ConfigContext)!;
  const [editorUi, setEditorUi_] = createEditorUiStore({
    pageId: params.id,
    revisionId: data.editorPageRevision()?.id as string,
  });

  const revision = createRevisionData(editorUi);
  const settingsProjectData = createSettingsProjectData();
  const settingsPageData = createSettingsPageData();
  const fontData = createFontData();
  const presetsData = createPresetsData();

  const [settings, setSettings_] = createSettingsStore();
  const [settingsVersion, setSettingsVersion] = createSignal(0);
  const setSettings = function (...args: [any]) {
    const result = setSettings_(...args);
    untrack(function () {
      setSettingsVersion(settingsVersion() + 1);
    });
    return result;
  } as typeof setSettings_;

  createEffect(function () {
    if (settingsProjectData.loading) return;
    const data = settingsProjectData();
    if (!data) return;
    setSettings("projects", reconcile(data));
  });
  createEffect(function () {
    if (settingsPageData.loading) return;
    const data = settingsPageData();
    if (!data) return;
    setSettings("pages", reconcile(data));
  });

  // This proxy emulates the settings inheritance from the server
  // Meaning: Use project setting across pages and optionally override it per page
  const mergedSettings = new Proxy(
    {},
    {
      ownKeys() {
        return Object.keys(settingsProjectData() || {});
      },
      getOwnPropertyDescriptor(target, prop) {
        return {
          enumerable: true,
          configurable: true,
        };
      },
      get(_, property: string) {
        const page = settings.pages;
        const project = settings.projects;
        if (page && page[property as keyof Settings]?.from === "pages") {
          return page[property as keyof Settings];
        }
        if (project) {
          return project[property as keyof Settings];
        }
      },
      set(_, property, value) {
        return false;
      },
    },
  ) as Settings;

  const mergeState = function (newState: Partial<State>) {
    for (const key of Object.keys(newState)) {
      setState(
        key as keyof State,
        reconcile(newState[key as keyof State], { key: null, merge: true }),
      );
    }
  };

  const loading = createMemo(() =>
    [
      revision.loading,
      presetsData.loading,
      fontData.loading,
      settingsPageData.loading,
      settingsProjectData.loading,
    ].includes(true),
  );

  // Normalize fonts and merge with state
  createEffect(function () {
    if (
      fontData.loading ||
      settingsPageData.loading ||
      settingsProjectData.loading
    )
      return;
    const fonts = fontData() || [];
    const newState = normalizeFonts({
      fonts,
      settings: mergedSettings,
    });

    untrack(() => mergeState(newState));
  });

  let lastRevisionId: string | undefined;

  // Normalize settings and merge with state
  createEffect(function () {
    settingsVersion();
    const pageRevision = revision();
    if (loading() || !pageRevision) return;
    let newState: Partial<State> = normalizeSettings({
      pageRevision,
      settings: mergedSettings,
    });

    // On first run and when the selected revision changes, the whole state has to be merged
    // In other cases we only have to sync data that is primarely stored in the new settings store
    if (lastRevisionId === pageRevision.id) {
      newState = {
        backgroundColor: newState.backgroundColor,
        metaImage: newState.metaImage,
        favicon: newState.favicon,
      };
    }

    /*
    console.info(
      `merge ${
        lastRevisionId === pageRevision.id ? "partial" : "full"
      } settings`,
    );
    */

    lastRevisionId = pageRevision.id;

    requestAnimationFrame(function () {
      untrack(function () {
        mergeState(newState);
      });
    });
  });

  const [stateVersion, setStateVersion] = createSignal(0);

  // Normalize page & presets and merge with state
  // Presets are part of this effect, because the page elements depend on the presets
  createEffect(async () => {
    const presets = presetsData();
    const pageRevision = revision();

    if (
      presetsData.loading ||
      revision.loading ||
      !presets ||
      !pageRevision ||
      !mounted()
    )
      return;

    // Run if one of these changes (new revision or revision saved)
    pageRevision.id;
    pageRevision.savedAt;

    untrack(function () {
      if (import.meta.env.DEV) {
        console.info(
          "Normalizing server data and merging with client state after page save",
        );
      }

      const newState = {
        presets: normalizePresets({ presets }),
        ...normalizePage({
          pageRevision,
          page: data.editorPage(),
        }),
      };

      batch(function () {
        setState("admin", true);
        mergeState(newState);
        setStateVersion(stateVersion() + 1);
      });
    });
  });

  // This sets up blueprint data default values
  const blueprintVersions: Record<string, number> = {};
  let lastStateVersion: number | undefined;
  createEffect(function () {
    if (!connected()) {
      return;
    }

    const newState = lastStateVersion !== stateVersion();
    if (stateVersion() === 0) {
      return;
    }

    lastStateVersion = stateVersion();

    for (const blueprint of Object.keys(config.blueprints)) {
      const version = config.blueprints[blueprint].version;
      const newVersion = blueprintVersions[blueprint] !== version;
      if (!newState && !newVersion) {
        continue;
      }

      if (version) {
        blueprintVersions[blueprint] = version;
      }

      untrack(function () {
        const dataItems = Object.entries(config.blueprints[blueprint].data);
        const presetIds = Object.keys(state.presets);
        const elementIds = Object.keys(state.elements);

        setState(
          produce(function (draft) {
            if (!state.blueprintDefaults.project[blueprint]) {
              draft.blueprintDefaults.project[blueprint] = {};
            }

            // TODO: Remove page blueprint defaults
            /*
            if (!state.blueprintDefaults.page[blueprint]) {
              draft.blueprintDefaults.page[blueprint] = {};
            }
            */

            for (const [key, data] of dataItems) {
              if (!state.blueprintDefaults.project[blueprint][key]) {
                draft.blueprintDefaults.project[blueprint][key] = {
                  value: data.defaultValue(),
                };
              }

              // TODO: Remove page blueprint defaults
              /*
              if (!state.blueprintDefaults.page[blueprint][key]) {
                draft.blueprintDefaults.page[blueprint][key] = {
                  value: data.defaultValue(),
                  override: false,
                };
              }
              */

              for (const elementId of elementIds) {
                const element = state.elements[elementId];
                if (element.blueprint !== blueprint) {
                  continue;
                }

                if (element.data[key]) {
                  continue;
                }

                draft.elements[elementId].data[key] = {
                  value: data.defaultValue(),
                  override: false,
                };
              }

              for (const presetId of presetIds) {
                const preset = state.presets[presetId];
                if (preset.blueprint !== blueprint) {
                  continue;
                }

                if (preset.data[key]) {
                  continue;
                }

                draft.presets[presetId].data[key] = {
                  value: data.defaultValue(),
                  override: false,
                };
              }
            }
          }),
        );
      });
    }
  });

  const [connected, setConnected] = createSignal(false);
  let pageViewIframe: HTMLIFrameElement = undefined!;
  const [send, schedule] = createIframeMessageQueue<EditorIframeEvent>(
    "editor-frame",
    function (event) {
      if (event.type === "setState") {
        setState_(reconcile(event.data, { key: null, merge: true }));
      }
      if (event.type === "setEditorUi") {
        setEditorUi_(reconcile(event.data, { key: null, merge: true }));
      }

      if (event.type === "savePage") {
        document.dispatchEvent(
          new KeyboardEvent("keydown", { key: "s", ctrlKey: true }),
        );
      }
      if (event.type === "publishPage") {
        document.dispatchEvent(
          new KeyboardEvent("keydown", { key: "p", ctrlKey: true }),
        );
      }
    },
    {
      child: () => pageViewIframe,
      connected() {
        if (!Object.keys(state.elements).length) {
          setEditorUi(produce((draft) => (draft.toolbarStatus = "creating")));
        }
        sendStateToIframe();
        sendEditorUiToIframe();
        setConnected(true);
      },
    },
  );

  const sendStateToIframe = schedule(
    throttle(() => send({ type: "setState", data: unwrap(state) }), 100),
  );
  const sendEditorUiToIframe = schedule(() =>
    send({ type: "setEditorUi", data: unwrap(editorUi) }),
  );

  const setState: typeof setState_ = function (...args: any) {
    (setState_ as any)(...args);
    connected() && sendStateToIframe();
  };

  const setEditorUi: typeof setEditorUi_ = schedule(function (...args: any) {
    (setEditorUi_ as any)(...args);
    connected() && sendEditorUiToIframe();
  });

  const focusElement = (id: string) => {
    send({
      type: "focusElement",
      data: id,
    });
  };

  useDarkMode(editorUi);

  return (
    <EditorUiContext.Provider value={[editorUi, setEditorUi]}>
      <SettingsContext.Provider value={[settings, setSettings]}>
        <StateContext.Provider value={[state, setState]}>
          <ElementFocusContext.Provider value={focusElement}>
            <Show when={config.blueprints["none"] && mounted()}>
              <Title>Editor - {state.title}</Title>
              <Styles />
              <StyleContainer>
                <DeleteElementDialog />
                <div class={classes.layout}>
                  <div class={classes["layout__frame"]}>
                    <Toolbar visible={connected()} />
                    <Frame ref={pageViewIframe} />
                    <Show when={!connected()}>
                      <div
                        class={css({
                          display: "flex",
                          justifyContent: "center",
                          alignItems: "center",
                          color: "#334155",
                          position: "absolute",
                          zIndex: 1,
                          width: "100%",
                          height: "100%",
                          fontSize: "5rem",
                        })}
                      >
                        <LoaderIcon
                          class={css({
                            animation: "np-spin 0.9s linear infinite",
                          })}
                        />
                      </div>
                    </Show>
                  </div>
                  <div class={classes["layout__sidebar"]}>
                    <Sidebar />
                  </div>
                </div>
              </StyleContainer>
            </Show>
          </ElementFocusContext.Provider>
        </StateContext.Provider>
      </SettingsContext.Provider>
    </EditorUiContext.Provider>
  );
};

export default Editor;
