import {
  Accessor,
  Component,
  createMemo,
  FlowComponent,
  For,
  JSXElement,
  ParentComponent,
  Show,
  Suspense,
  useContext,
} from "solid-js";
import { produce } from "solid-js/store";
import { Dynamic } from "solid-js/web";
import {
  Blueprint,
  BlueprintDataInput,
  BlueprintData,
  Element,
  Config,
  ElementDataRecord,
  BlueprintDataInputProps,
  SidebarElementsTab,
} from "../../../types";
import { blueprintEntries } from "../../../admin/blueprint";
import { css } from "../../../css/fela";
import { SidebarCard, TabSubtitle } from "./util";
import Checkbox from "../../common/checkbox";
import { klona } from "klona";
import { SmallButton } from "../../common/buttons";
import { RefreshIcon, SelectorIcon } from "../../common/icons";
import {
  blueprintDataTitle,
  useSelectedElement,
  useSelectedElementPreset,
} from "../../../admin/editor";
import ElementPresets from "./elementPresets";
import clsx from "clsx";
import { ConfigContext, StateContext } from "../../context/page";
import { EditorUiContext } from "../../context/admin";

type Value = { value: any; override?: boolean };
type ChildProduce = (draft: Value) => Value;

type BlueprintSettingsComponent = Component<{
  element?: Element;
  blueprint: Blueprint;
  values: Record<string, Value>;
  hideResetButton?: boolean;
  context: "element" | "preset" | "project";
  setter: (
    produce: (value: Record<string, Value>) => Record<string, Value>,
  ) => void;
}>;

type BlueprintSettingComponent = Component<{
  blueprint?: Blueprint;
  element?: Element;
  data: BlueprintData;
  value?: Value;
  key: string;
  context: "element" | "preset" | "project";
  setter: (produce: (value: Value) => Value) => void;
}>;

export const BlueprintSetting: BlueprintSettingComponent = (p) => {
  const valueSetter = function (childProduce: ChildProduce) {
    p.setter(function (draft: Value) {
      const newDraft = !draft;
      if (newDraft) {
        draft = {
          value: p.data.defaultValue(),
          override: false,
        };
      }

      if (p.context !== "project" && !draft.override) {
        const currentValue = value();
        if (currentValue != null) {
          draft.value = klona(currentValue);
        } else if (!newDraft) {
          draft.value = p.data.defaultValue();
        }

        draft.override = true;
      }

      draft = childProduce(draft);
      return draft;
    });
  };

  const [state] = useContext(StateContext)!;

  const value = createMemo(function () {
    if (p.context === "project" || !p.blueprint || p.value?.override) {
      return p.value?.value != null ? p.value?.value : p.data.defaultValue();
    }

    if (p.context !== "preset" && p.element?.presetId) {
      const presetData = state.presets[p.element.presetId].data[p.key];
      if (presetData?.override) {
        return presetData.value;
      }
    }

    // TODO: Remove page blueprint defaults
    /*
    if (
      p.context !== "page" &&
      state.blueprintDefaults.page?.[p.blueprint.id]?.[p.key]?.override
    ) {
      return state.blueprintDefaults.page[p.blueprint.id][p.key].value;
    }
    */

    const projectValue =
      state.blueprintDefaults.project?.[p.blueprint.id]?.[p.key]?.value;
    return projectValue != null ? projectValue : p.data.defaultValue();
  });

  return (
    <Dynamic<BlueprintDataInput>
      key={p.key}
      data={p.data}
      context={p.context}
      setValue={
        ((childProduce: Function) => {
          return valueSetter(function (draft) {
            draft["value"] = childProduce(draft["value"]);
            return draft;
          });
        }) as any
      }
      setOverride={
        ((childProduce: Function) => {
          return valueSetter(function (draft) {
            draft["override"] = childProduce(draft["override"]);
            return draft;
          });
        }) as any
      }
      element={p.element}
      component={p.data.input}
      value={value()}
      override={!!p.value?.override}
    />
  );
};

export const BlueprintSettingsContainer: FlowComponent = function (p) {
  return (
    <div
      class={css({ display: "flex", flexDirection: "column", gap: "0.25rem" })}
    >
      {p.children}
    </div>
  );
};

export const DefaultBlueprintSettings: BlueprintSettingsComponent = (p) => {
  const [state, setState] = useContext(StateContext)!;
  const [config] = useContext(ConfigContext)!;

  const blueprint = createMemo(function () {
    return p.blueprint;
  });

  return (
    <BlueprintSettingsContainer>
      <Show when={p.context === "project" && p.hideResetButton !== true}>
        <SmallButton
          theme="primary"
          onClick={() => {
            setState(
              produce(function (draft) {
                for (const [key, data] of Object.entries(
                  config.blueprints[blueprint().id].data,
                )) {
                  draft.blueprintDefaults.project[blueprint().id][key] = {
                    value: data.defaultValue(),
                  };
                }
              }),
            );
          }}
        >
          <RefreshIcon /> Reset all
        </SmallButton>
      </Show>
      <For
        each={Object.entries({
          id: {} as BlueprintData,
          ...((blueprint().data || {}) as Record<string, BlueprintData>),
        })}
      >
        {([k, v]) => {
          const valueSetter = function (childProduce: ChildProduce) {
            p.setter(function (draft) {
              draft[k] = childProduce(draft[k]);
              return draft;
            });
          };

          return (
            <BlueprintSetting
              {...p}
              setter={valueSetter}
              value={p.values[k]}
              key={k}
              data={v}
            />
          );
        }}
      </For>
    </BlueprintSettingsContainer>
  );
};

const BlueprintSettingsListItem: FlowComponent<{ blueprint: Blueprint }> =
  function (p) {
    const [editorUi, setEditorUi] = useContext(EditorUiContext)!;

    return (
      <div
        class={css({
          display: "flex",
          flexDirection: "column",
          borderWidth: "1px",
          "--tw-border-opacity": 1,
          borderColor: "rgb(203 213 225 / var(--tw-border-opacity))",
          "&:not(:last-child)": {
            borderBottomWidth: "0",
          },
          ":first-child": {
            borderTopLeftRadius: "0.25rem",
            borderTopRightRadius: "0.25rem",
            "> button": {
              borderTopLeftRadius: "0.25rem",
              borderTopRightRadius: "0.25rem",
            },
          },
          ":last-child": {
            borderBottomLeftRadius: "0.25rem",
            borderBottomRightRadius: "0.25rem",
            "> button": {
              borderBottomLeftRadius: "0.25rem",
              borderBottomRightRadius: "0.25rem",
            },
          },
          "--tw-bg-opacity": 1,
          backgroundColor: "white",
          "--tw-shadow": "0 1px 2px 0 rgb(0 0 0 / 0.05)",
          "--tw-shadow-colored": "0 1px 2px 0 var(--tw-shadow-color)",
          boxShadow:
            "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
          "> button": {
            padding: ".375rem .75rem",
            fontSize: "0.875rem",
            "--tw-ring-opacity": 1,
            "--tw-ring-color": "var(--np-accent-400)",
            ":hover": {
              "--tw-ring-offset-shadow":
                "var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color)",
              "--tw-ring-shadow":
                "var(--tw-ring-inset) 0 0 0 calc(2px + var(--tw-ring-offset-width)) var(--tw-ring-color)",
              boxShadow:
                "var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000)",
              zIndex: 10,
            },
            "&.active": {
              fontWeight: "600",
              borderBottomColor: "transparent",
              borderBottomLeftRadius: "0",
              borderBottomRightRadius: "0",
            },
          },
        })}
      >
        <button
          class={clsx(
            editorUi.selectedBlueprint === p.blueprint.id && "active",
          )}
          onClick={() => {
            setEditorUi(
              "selectedBlueprint",
              editorUi.selectedBlueprint === p.blueprint.id
                ? undefined
                : p.blueprint.id,
            );
          }}
        >
          {p.blueprint.title()}
        </button>
        <Show when={editorUi.selectedBlueprint === p.blueprint.id}>
          <div
            class={css({
              padding: "0.75rem",
              backgroundColor: "#f1f5f9",
              borderRadius: "0.25rem",
              borderTopWidth: "1px",
              "--tw-border-opacity": 1,
              borderColor: "rgb(226 232 240 / var(--tw-border-opacity))",
            })}
          >
            {p.children}
          </div>
        </Show>
      </div>
    );
  };

// TODO: Remove page blueprint defaults
/*
export const PageBlueprintSettings = function () {
  const [declaration] = useContext(ConfigContext)!;
  const [state, setState] = useContext(StateContext)!;

  return (
    <div>
      <div
        class={css({
          display: "flex",
          flexDirection: "column",
        })}
      >
        <For each={blueprintEntries(declaration.blueprints)}>
          {([type, v]) => {
            return (
              <BlueprintSettingsListItem blueprint={v}>
                <Suspense>
                  <Dynamic
                    component={DefaultBlueprintSettings}
                    context="page"
                    blueprint={v}
                    values={state.blueprintDefaults.page[type] || {}}
                    setter={(childProduce: Function) => {
                      setState(
                        produce(function (draft) {
                          if (!draft.blueprintDefaults.page[type]) {
                            draft.blueprintDefaults.page[type] = {};
                          }
                          draft.blueprintDefaults.page[type] = childProduce(
                            draft.blueprintDefaults.page[type],
                          );
                        }),
                      );
                    }}
                  />
                </Suspense>
              </BlueprintSettingsListItem>
            );
          }}
        </For>
      </div>
    </div>
  );
};
*/

export const createElementData = function (
  declaration: Config,
  blueprintId: string,
) {
  const data: ElementDataRecord = {};
  const blueprint = declaration.blueprints[blueprintId];
  if (!blueprint?.data) {
    return data;
  }

  for (const [k, v] of Object.entries(blueprint.data)) {
    data[k] = { override: false, value: v.defaultValue() };
  }

  return data;
};

export const ProjectBlueprintSettings = function () {
  const [declaration] = useContext(ConfigContext)!;

  return (
    <div>
      <div
        class={css({
          display: "flex",
          flexDirection: "column",
        })}
      >
        <For each={blueprintEntries(declaration.blueprints)}>
          {([type, v]) => {
            const props = useBlueprintSettingsProps(
              () => "project",
              () => v,
            );
            return (
              <BlueprintSettingsListItem blueprint={v}>
                <Suspense>
                  <DefaultBlueprintSettings {...props()} />
                </Suspense>
              </BlueprintSettingsListItem>
            );
          }}
        </For>
      </div>
    </div>
  );
};

const BlueprintSettingsRightColumn = function () {
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;

  const Button: FlowComponent<{ tab: SidebarElementsTab }> = function (props) {
    return (
      <button
        class={clsx(editorUi.sidebarElementsTab === props.tab && "active")}
        onClick={() =>
          setEditorUi(
            produce((draft) => (draft.sidebarElementsTab = props.tab)),
          )
        }
      >
        {props.children}
      </button>
    );
  };

  const preset = useSelectedElementPreset();

  const presetButtonStyle = createMemo(function () {
    return css({
      "--border": preset() ? "#9ca3af" : (undefined as any),
      minWidth: 0,
      width: "9rem",
      justifyContent: "center",
    });
  });

  return (
    <div
      class={css({
        minWidth: 0,
        display: "flex",
        gap: "1rem",
        alignItems: "flex-end",
        "--border": "#d1d5db",
        "--active-border": "var(--np-accent-500)",
        "& button": {
          transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
          padding: "0 1rem 0.25rem 1rem",
          display: "flex",
          alignItems: "center",
          textOverflow: "ellipsis",
          gap: "0.5rem",
          fontSize: "0.875rem",
          fontWeight: 500,
          color: "#64748b",
          borderBottom: "2px solid var(--border)",
          "&:first-child": {
            borderBottomLeftRadius: "0.375rem",
          },
          "&:last-child": {
            borderBottomRightRadius: "0.375rem",
          },
          "&:hover:not(.active)": {
            color: "#1e293b",
            borderColor: "var(--np-accent-alt-400)",
          },
          "&.active": {
            color: "#1e293b",
            borderColor: "var(--active-border)",
          },
        },
      })}
    >
      <div
        class={css({
          display: "flex",
          minWidth: 0,
        })}
      >
        <button
          title={preset() ? `Preset: ${preset()!.name}` : "Presets"}
          onClick={() =>
            setEditorUi(
              produce((draft) => (draft.sidebarElementsTab = "presets")),
            )
          }
          class={clsx(
            presetButtonStyle(),
            editorUi.sidebarElementsTab === "presets" && "active",
          )}
        >
          <span
            class={css({
              fontWeight: 400,
              fontSize: "0.75rem",
              whiteSpace: "nowrap",
              overflow: "hidden",
              textOverflow: "ellipsis",
            })}
          >
            {preset() ? preset()?.name : "Presets"}
          </span>{" "}
          <SelectorIcon width="1rem" height="1rem" />
        </button>
      </div>
      <div
        class={css({
          display: "flex",
        })}
      >
        <Button tab="element">Element</Button>
        <Button tab="preset">Preset</Button>
        <Button tab="project">Project</Button>
      </div>
    </div>
  );
};

const useBlueprintSettingsProps = function (
  getContext: Accessor<"element" | "preset" | "project">,
  getBlueprint: Accessor<Blueprint>,
) {
  const [state, setState] = useContext(StateContext)!;
  const [editorUi] = useContext(EditorUiContext)!;

  return createMemo(function () {
    const context = getContext();
    const blueprint = getBlueprint();

    const element = editorUi.selectedElementId
      ? state.elements[editorUi.selectedElementId!]
      : undefined;

    if (context === "element" && element) {
      return {
        context,
        blueprint,
        element,
        values: element.data,
        setter: (childProduce: Function) => {
          setState(
            produce(function (draft) {
              draft.elements[element.id]["data"] = childProduce(
                draft.elements[element.id]["data"],
              );
            }),
          );
        },
      };
    }

    if (context === "preset" && element) {
      return {
        context,
        blueprint,
        values: element.presetId ? state.presets[element.presetId].data : {},
        setter: (childProduce: Function) => {
          const presetId = element.presetId;
          if (!presetId) {
            return;
          }

          setState(
            produce(function (draft) {
              draft.presets[presetId]["data"] = childProduce(
                draft.presets[presetId]["data"],
              );
            }),
          );
        },
      };
    }

    return {
      context,
      blueprint,
      values: state.blueprintDefaults.project[blueprint.id] || {},
      setter: (childProduce: Function) => {
        setState(
          produce(function (draft) {
            if (!draft.blueprintDefaults.project[blueprint.id]) {
              draft.blueprintDefaults.project[blueprint.id] = {};
            }
            draft.blueprintDefaults.project[blueprint.id] = childProduce(
              draft.blueprintDefaults.project[blueprint.id],
            );
          }),
        );
      },
    };
  });
};

export const ElementBlueprintSettings = function () {
  const [declaration] = useContext(ConfigContext)!;
  const [editorUi] = useContext(EditorUiContext)!;

  const element = useSelectedElement();

  const blueprint = createMemo(function () {
    if (!element()) {
      return;
    }

    return declaration.blueprints[element()!.blueprint];
  });

  const settingsProps = useBlueprintSettingsProps(
    () =>
      editorUi.sidebarElementsTab === "presets"
        ? "preset"
        : editorUi.sidebarElementsTab,
    () => blueprint()!,
  );

  const showPresets = createMemo(function () {
    if (editorUi.sidebarElementsTab === "presets") {
      return true;
    }

    if (editorUi.sidebarElementsTab === "preset" && !element()?.presetId) {
      return true;
    }

    return false;
  });

  return (
    <div>
      <TabSubtitle
        class={css({
          position: "sticky",
          top: "33px",
          zIndex: 39,
          paddingBottom: "0.5rem",
          paddingTop: "0.5rem",
          marginTop: "-0.5rem",
          marginBottom: 0,
          background:
            "linear-gradient(0deg, rgba(234,236,244,0) 0%, rgba(234,236,244,1) 20%)",
        })}
        rightColumn={<BlueprintSettingsRightColumn />}
      >
        Settings
      </TabSubtitle>
      <div>
        <Show when={!showPresets()} fallback={<ElementPresets />}>
          <Suspense>
            <DefaultBlueprintSettings hideResetButton {...settingsProps()} />
          </Suspense>
        </Show>
      </div>
    </div>
  );
};

export const BlueprintDataTitle: ParentComponent<
  {} & BlueprintDataInputProps<any, any>
> = function (p) {
  return (
    <div
      class={css({
        display: "flex",
        fontSize: "0.75rem",
        gap: "1rem",
        alignItems: "center",
        justifyContent: "space-between",
      })}
    >
      <div
        class={css({
          fontWeight: "500",
        })}
      >
        {blueprintDataTitle(p.data, p.key)}
      </div>
      <div
        class={css({
          width: "35%",
          display: "flex",
          alignItems: "center",
          gap: "0.5rem",
          justifyContent: "flex-end",
        })}
      >
        {p.children}
        <SmallButton
          theme="transparent"
          onClick={() => {
            p.setValue(function () {
              return p.data.defaultValue();
            });
          }}
          title="Reset"
        >
          <RefreshIcon />
        </SmallButton>
        <Show when={p.context !== "project"}>
          <Checkbox
            title={"Override"}
            state={p.override}
            setState={(v) => {
              p.setOverride(() => v);
            }}
          />
        </Show>
      </div>
    </div>
  );
};

export const BlueprintDataInputLayout: ParentComponent<{ header: JSXElement }> =
  function (p) {
    return (
      <SidebarCard>
        <div
          class={css({
            borderRadius: "0.35rem",
            padding: "0.25rem 0.5rem",
            backgroundColor: "white",
          })}
        >
          {p.header}
        </div>
        {p.children}
      </SidebarCard>
    );
  };
