import type { IStyle } from "fela";
import { Component, createMemo, For, Show, useContext } from "solid-js";
import { cssMemo } from "../../../css/fela";
import { useElementData } from "../../../util/elementData";
import { TreeIcon } from "../../common/icons";
import clsx from "clsx";
import { ConfigContext, StateContext } from "../../context/page";
import { EditorUiContext, ElementFocusContext } from "../../context/admin";

const Breadcrumbs: Component = function () {
  const [state] = useContext(StateContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const [declaration] = useContext(ConfigContext)!;

  const elementBreadCrumbs = createMemo(function () {
    const ids = ["root"];

    if (editorUi.selectedElementId == null) {
      return ids;
    }

    const addElementToCrumb = function (id: string) {
      if (state.elements[id]) {
        const parentId = state.elements[id].parentId;
        if (parentId) {
          addElementToCrumb(parentId);
        }
      }
      ids.push(id);
    };

    addElementToCrumb(editorUi.selectedElementId);

    return ids;
  });

  const css = cssMemo(() => {
    const activeClass: IStyle = {
      "--tw-ring-offset-shadow":
        "var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color)",
      "--tw-ring-shadow":
        "var(--tw-ring-inset) 0 0 0 calc(2px + var(--tw-ring-offset-width)) var(--tw-ring-color)",
      boxShadow:
        "var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000)",
      zIndex: 1,
      color: "#1e293b",
      maxWidth: "inherit",
    };

    return {
      display: "flex",
      marginBottom: "0.75rem",
      flexWrap: "wrap",
      "& .spacer": {
        color: "#475569",
        height: "1.5rem",
        position: "relative",
        fontSize: "0.875rem",
        width: "3rem",
        "--shadow-x": "-4px",
        "&:first-child": {
          width: "2.4rem",
          "& .item": {
            fontSize: "1.125rem",
            width: "100%",
            justifyContent: "center",
            position: "relative",
            height: "100%",
            "--shadow-x": "0",
            padding: "0",
            borderTopRightRadius: "0",
            borderBottomRightRadius: "0",
          },
        },
        "&:nth-child(2) .item": {
          borderTopLeftRadius: "0",
          borderBottomLeftRadius: "0",
          "--shadow-x": "0",
        },
        "&:last-child": {
          color: "#1e293b",
          fontWeight: 600,
          "& .item": {
            maxWidth: "inherit",
          },
        },
      },
      "& .item": {
        top: 0,
        bottom: 0,
        position: "absolute",
        display: "flex",
        alignItems: "center",
        gap: "0.25rem",
        maxWidth: "5rem",
        overflow: "hidden",
        paddingLeft: "0.75rem",
        paddingRight: "1rem",
        "--tw-shadow": "var(--shadow-x) 1px 2px 0 rgb(0 0 0 / 0.05)",
        "--tw-shadow-colored": "0 1px 2px 0 var(--tw-shadow-color)",
        boxShadow:
          "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
        "--tw-bg-opacity": 1,
        backgroundColor: "rgb(248 250 252 / var(--tw-bg-opacity))",
        borderWidth: "1px",
        "--tw-border-opacity": 1,
        borderColor: "rgb(226 232 240 / var(--tw-border-opacity))",
        "--tw-ring-offset-width": "2px",
        "--tw-ring-opacity": 1,
        "--tw-ring-color": "var(--np-accent-400)",
        borderTopLeftRadius: "0.375rem",
        borderBottomLeftRadius: "0.375rem",
        borderTopRightRadius: "1.5rem",
        borderBottomRightRadius: "1.5rem",
        whiteSpace: "nowrap",
      },
      "& .item:active": {
        backgroundColor: "rgb(226 232 240 / var(--tw-bg-opacity))",
      },
      "& .item.active": activeClass,
      "& .item:hover": activeClass,
    };
  });
  const focusElement = useContext(ElementFocusContext)!;

  return (
    <div class={css()}>
      <For each={elementBreadCrumbs()}>
        {(id) => {
          const data = useElementData(state, id);
          return (
            <div class="spacer" title={id === "root" ? "Overview" : undefined}>
              <button
                title={data.id?.length ? `#${data.id}` : undefined}
                class={clsx(
                  "item",
                  (editorUi.hoveredElementId === id ||
                    editorUi.deletingElementId === id) &&
                    "active",
                )}
                onClick={() => {
                  setEditorUi(
                    "selectedElementId",
                    id === "root" ? undefined : id,
                  );
                  focusElement(id);
                }}
                onMouseEnter={() => {
                  if (id === "root") {
                    return;
                  }
                  setEditorUi("hoveredElementId", id);
                }}
                onMouseLeave={() => {
                  if (editorUi.hoveredElementId !== id) {
                    return;
                  }
                  setEditorUi("hoveredElementId", undefined);
                }}
              >
                <Show when={id !== "root"} fallback={<TreeIcon />}>
                  {declaration.blueprints[state.elements[id].blueprint]
                    ? declaration.blueprints[
                        state.elements[id].blueprint
                      ].title()
                    : "Invalid"}
                </Show>
              </button>
            </div>
          );
        }}
      </For>
    </div>
  );
};

export default Breadcrumbs;
