import { createMemo, createSignal, For, Show, useContext } from "solid-js";
import { produce } from "solid-js/store";
import { css } from "../../../css/fela";
import { SmallButton } from "../../common/buttons";
import {
  CheckIcon,
  CopyIcon,
  CursorTextIcon,
  PlusIcon,
  TrashIcon,
} from "../../common/icons";
import {
  useSelectedElement,
  useSelectedElementPreset,
} from "../../../admin/editor";
import { Input } from "../../common/inputs";
import {
  createForm,
  focus,
  required,
  reset,
  setValue,
} from "@modular-forms/solid";
import { createId } from "@paralleldrive/cuid2";
import { createElementData } from "./blueprintData";
import { klona } from "klona";
import { ConfirmDialog } from "../../common/dialog";
import clsx from "clsx";
import { ConfigContext, StateContext } from "../../context/page";

type PresetForm = {
  name: string;
};

const ElementPresets = function () {
  const [config] = useContext(ConfigContext)!;
  const [state, setState] = useContext(StateContext)!;
  const element = useSelectedElement();
  const preset = useSelectedElementPreset(element);
  const presets = createMemo(function () {
    if (!element()) {
      return [];
    }

    return Object.values(state.presets)
      .filter((preset) => preset.blueprint === element()?.blueprint)
      .sort((a, b) => a.name.localeCompare(b.name));
  });

  const [createPresetForm, { Form, Field, FieldArray }] =
    createForm<PresetForm>();

  const [editingPresetId, setEditingPresetId] = createSignal<
    undefined | string
  >();

  const [deletePresetId, setDeletePresetId] = createSignal<
    undefined | string
  >();

  return (
    <div>
      <ConfirmDialog
        open={!!deletePresetId()}
        icon={TrashIcon}
        onConfirm={() => {
          const id = deletePresetId()!;
          setDeletePresetId();
          setState(
            produce((draft) => {
              for (const element of Object.values(draft.elements)) {
                if (element.presetId !== id) {
                  continue;
                }

                element.presetId = undefined;
              }
              if (!state.presets[id].new) {
                draft.pendingDeletes?.presets.push(id);
              }
              delete draft.presets[id];
            }),
          );
        }}
        onCancel={() => setDeletePresetId()}
      >
        Do you really wanna delete the preset "
        <strong>{state.presets[deletePresetId()!].name}</strong>"?
      </ConfirmDialog>
      <Form
        class={css({
          display: "flex",
          alignItems: "flex-start",
          gap: "0.5rem",
          marginBottom: "1rem",
        })}
        onSubmit={(values) => {
          const id = createId();
          setState(
            produce((draft) => {
              if (!element()?.presetId) {
                draft.elements[element()!.id].presetId = id;
              }

              const blueprint = element()!.blueprint;

              draft.presets[id] = {
                id,
                new: true,
                blueprint,
                name: values.name,
                data: createElementData(config, blueprint),
              };
            }),
          );
          reset(createPresetForm);
        }}
      >
        <div class={css({ flex: 1 })}>
          <Field name="name" validate={[required("Please enter a name.")]}>
            {(field, props) => (
              <Input
                {...props}
                value={field.value}
                error={field.error}
                placeholder="New preset name"
              />
            )}
          </Field>
        </div>
        <SmallButton type="submit" theme="primary">
          <PlusIcon width="1em" height="1em" /> Create preset
        </SmallButton>
      </Form>
      <Show
        when={presets()}
        fallback={
          <div
            class={css({
              color: "#64748b",
            })}
          >
            No presets...
          </div>
        }
      >
        <div
          class={css({
            display: "flex",
            gap: "0.25rem",
            flexDirection: "column",
            "& .preset": {
              display: "flex",
              gap: "1rem",
              alignItems: "center",
              "&.preset--selected .preset__name": {
                color: "#1e293b",
                "& svg": {
                  color: "#1e293b",
                },
              },
              ":last-child .preset__name": {
                borderColor: "transparent",
              },
            },
            "& .preset__name": {
              color: "#64748b",
              minHeight: "27px",
              flex: 1,
              display: "flex",
              gap: "0.25rem",
              alignItems: "center",
              justifyContent: "center",
              borderBottom: "1px solid #cbd5e1",

              ":hover": {
                color: "#1e293b",
              },
              "& svg": {
                color: "transparent",
                width: "0.8rem",
                height: "0.8rem",
              },
            },
          })}
        >
          <For each={presets()}>
            {(presetItem) => {
              const selected = createMemo(() => preset()?.id === presetItem.id);
              const [renamePresetForm, { Form, Field, FieldArray }] =
                createForm<PresetForm>();
              return (
                <div class={clsx("preset", selected() && "preset--selected")}>
                  <Show
                    when={editingPresetId() !== presetItem.id}
                    fallback={
                      <Form
                        class="preset__name"
                        onSubmit={(values) => {
                          setState(
                            produce((draft) => {
                              draft.presets[presetItem.id].name = values.name;
                            }),
                          );
                          setEditingPresetId();
                        }}
                        onKeyDown={(e) => {
                          if (e.key !== "Escape") {
                            return;
                          }
                          setEditingPresetId();
                        }}
                      >
                        <Field
                          name="name"
                          validate={[required("Please enter a name.")]}
                        >
                          {(field, props) => (
                            <Input
                              {...props}
                              value={field.value}
                              error={field.error}
                            />
                          )}
                        </Field>
                        <CheckIcon />
                      </Form>
                    }
                  >
                    <button
                      class="preset__name"
                      onClick={() =>
                        setState(
                          produce((draft) => {
                            draft.elements[element()!.id].presetId = !selected()
                              ? presetItem.id
                              : undefined;
                          }),
                        )
                      }
                    >
                      {presetItem.name}
                      <CheckIcon />
                    </button>
                  </Show>
                  <div
                    class={css({
                      display: "flex",
                      gap: "0.5rem",
                      "& svg": {
                        width: "1em",
                        height: "1em",
                      },
                    })}
                  >
                    <SmallButton
                      title="Rename"
                      onClick={() => {
                        if (editingPresetId() === presetItem.id) {
                          renamePresetForm.element?.requestSubmit();
                          return;
                        }
                        setValue(renamePresetForm, "name", presetItem.name);
                        setEditingPresetId(presetItem.id);
                        focus(renamePresetForm, "name");
                      }}
                    >
                      <CursorTextIcon />
                    </SmallButton>
                    <SmallButton
                      title="Duplicate"
                      onClick={() => {
                        const newId = createId();
                        setState(
                          produce((draft) => {
                            draft.presets[newId] = {
                              id: newId,
                              blueprint: presetItem.blueprint,
                              new: true,
                              name: `${presetItem.name} Copy`,
                              data: klona(presetItem.data),
                            };
                          }),
                        );
                      }}
                    >
                      <CopyIcon />
                    </SmallButton>
                    <SmallButton
                      title="Delete"
                      onClick={() => setDeletePresetId(presetItem.id)}
                    >
                      <TrashIcon />
                    </SmallButton>
                  </div>
                </div>
              );
            }}
          </For>
        </div>
      </Show>
    </div>
  );
};

export default ElementPresets;
