import {
  Accessor,
  Component,
  createMemo,
  createSignal,
  For,
  Show,
  useContext,
} from "solid-js";
import { produce } from "solid-js/store";
import { ElementBlueprintSettings } from "./blueprintData";
import {
  canMoveToOtherSlot,
  getBuilderInfoFromSortableEvent,
  slotHasSpace,
  useSortableElementMove,
} from "../../../admin/element";
import Breadcrumbs from "./breadcrumbs";
import { createOnEnd, sortable } from "../../../directives/sortable";
import { cssMemo, css } from "../../../css/fela";
import { TabSubtitle } from "./util";
import {
  ArrowsDiagonalIcon,
  ArrowsDiagonalMinimizeIcon,
  CopyIcon,
  ArrowsMoveIcon,
  TrashIcon,
  TreeIcon,
} from "../../common/icons";
import { SmallButton } from "../../common/buttons";
import { Slots } from "../../../types";
import { useElementData } from "../../../util/elementData";
import { useDeleteElement } from "../../../hooks/element";
import { duplicateElement } from "../../../admin/editor";
import clsx from "clsx";
import { ConfigContext, StateContext } from "../../context/page";
import { EditorUiContext, ElementFocusContext } from "../../context/admin";

const [draggingInSidebar, setDraggingInSidebar] = createSignal(false);

const SidebarElementSlot: Component<{
  slot: string;
  recursive?: boolean;
  deep?: boolean;
  showLabelPadding?: boolean;
  elementId: string;
  elementIds: string[];
}> = function (p) {
  const [state, setState] = useContext(StateContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const handleElementMove = useSortableElementMove(() => p.elementId);
  const [declaration] = useContext(ConfigContext)!;
  const deleteElement = useDeleteElement();

  sortable;
  const activeItemStyle = {
    zIndex: "1",
    "--tw-ring-offset-shadow":
      "var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color)",
    "--tw-ring-shadow":
      "var(--tw-ring-inset) 0 0 0 calc(2px + var(--tw-ring-offset-width)) var(--tw-ring-color)",
    boxShadow:
      "var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000)",
  };
  const style = cssMemo(() => {
    return {
      display: "flex",
      flexDirection: "column",
      gap: "0.25rem",
      "--tw-bg-opacity": 1,
      backgroundColor: p.deep ? undefined : "#F1F5F9",
      borderWidth: p.deep ? 0 : "1px",
      borderTopWidth: 0,
      padding: p.deep
        ? `${p.showLabelPadding ? 1.25 : 0.25}rem 0 0.5rem 0`
        : "0.5rem 0.75rem",
      transition: "all 0.5s",
      borderRadius: "0.25rem",
      borderTopLeftRadius: "0",
      borderTopRightRadius: "0",
      borderColor: p.deep
        ? "transparent"
        : editorUi.createMoveTarget.elementId !== p.elementId ||
            editorUi.createMoveTarget.slot !== p.slot
          ? "#F1F5F9"
          : "var(--np-accent-alt-400)",
      "--tw-shadow": "0 1px 2px 0 rgb(0 0 0 / 0.05)",
      "--tw-shadow-colored": "0 1px 2px 0 var(--tw-shadow-color)",
      boxShadow: p.deep
        ? undefined
        : "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
    };
  });
  const focusElement = useContext(ElementFocusContext)!;

  return (
    <div
      class={style()}
      data-slot={p.slot}
      data-element-id={p.elementId}
      use:sortable={{
        handle: ".handle",
        animation: 150,
        invertSwap: true,
        direction: "vertical",
        swapThreshold: 0.65,
        emptyInsertThreshold: 5,
        group: "sidebarElements",
        filter: ".drag-disabled",
        forceFallback: true,
        fallbackOnBody: true,
        onStart: () => {
          setDraggingInSidebar(true);
        },
        onMove: (evt) => {
          if (evt.from === evt.to) {
            return true;
          }

          const builderInfo = getBuilderInfoFromSortableEvent(evt, state);
          return canMoveToOtherSlot(declaration, builderInfo);
        },
        onEnd: createOnEnd((childProduce, evt) => {
          setDraggingInSidebar(false);
          return handleElementMove(childProduce, evt);
        }),
      }}
    >
      <For each={p.elementIds}>
        {(id) => {
          const element = createMemo(function () {
            return state.elements[id];
          });
          const blueprint = createMemo(function () {
            return declaration.blueprints[element().blueprint];
          });
          const slots = createMemo(function () {
            if (!blueprint()?.slots) {
              return [];
            }

            const blueprintSlots = Object.keys(blueprint().slots!);
            const result = blueprintSlots.reduce(
              function (acc, id) {
                const slot = blueprint().slots[id];
                const elements = () => element().slots?.[id]?.elements || [];
                if (!editorUi.expandElementsTree && !elements().length) {
                  return acc;
                }

                acc.push({
                  id,
                  elements,
                  slot,
                });
                return acc;
              },
              [] as {
                id: string;
                elements: Accessor<string[]>;
                slot: Slots[""] & { _ghost?: boolean };
              }[],
            );

            // If a slot was renamed or removed in the blueprint, but still has elements in the database,
            // then the slot will appear here
            for (const [id, data] of Object.entries(element().slots)) {
              if (!data.elements.length || blueprintSlots.includes(id)) {
                continue;
              }

              result.push({
                id,
                elements: () => data.elements,
                slot: {
                  max: 0,
                  _ghost: true,
                } as any,
              });
            }

            return result;
          });
          const targeted = createMemo(function () {
            return editorUi.createMoveTarget.elementId === id;
          });
          const hovered = createMemo(function () {
            return (
              editorUi.hoveredElementId === id ||
              editorUi.deletingElementId === id ||
              targeted()
            );
          });
          const data = useElementData(state, id);
          return (
            <div
              data-element={id}
              class={css({
                display: "flex",
                flexDirection: "column",
                "&.sortable-drag": {
                  opacity: "0.2 !important",
                },
                "&.sortable-drag .item__slots": {
                  display: "none",
                },
                "&.sortable-ghost .item__slots": {
                  display: "none",
                },
              })}
            >
              <div
                class={css({
                  "--tw-ring-opacity": 1,
                  "--tw-ring-color": "var(--np-accent-400)",
                  "--tw-ring-offset-width": "1px",
                  display: "flex",
                  borderWidth: "1px",
                  backgroundColor: blueprint() ? "white" : "#fda4af",
                  color: blueprint() ? undefined : "#4c0519",
                  fontWeight: blueprint() ? undefined : 500,
                  fontSize: "0.75rem",
                  lineHeight: "1.25rem",
                  borderRadius: "0.375rem",
                  ":hover":
                    draggingInSidebar() || !blueprint()
                      ? undefined
                      : activeItemStyle,
                  ...(hovered() ? activeItemStyle : {}),
                  "& .item__button": {
                    overflow: "hidden",
                    textOverflow: "ellipsis",
                    padding: "0.125rem 0.75rem",
                    flex: "1 1 0%",
                    cursor: blueprint() ? "pointer" : undefined,
                  },
                  "& .item__id": {
                    marginLeft: "0.25rem",
                    fontWeight: 600,
                    color: "#64748b",
                  },
                  "& .item__invalid_blueprint": {
                    fontWeight: 700,
                  },
                  "& button": {
                    fontSize: "1.05rem",
                    display: "flex",
                    alignItems: "center",
                    justifyContent: "center",
                    padding: "0 0.5rem",
                  },
                  "& .item__handle--move": {
                    color: "rgb(51 65 85)",
                    cursor: "grab",
                  },
                  "& .item__handle--remove": {
                    cursor: "pointer",
                  },
                })}
                onMouseEnter={() => {
                  if (!blueprint()) {
                    return;
                  }
                  if (draggingInSidebar()) {
                    return;
                  }
                  setEditorUi("hoveredElementId", id);
                }}
                onMouseLeave={() => {
                  if (!blueprint()) {
                    return;
                  }
                  if (editorUi.hoveredElementId === id) {
                    setEditorUi("hoveredElementId", undefined);
                  }
                }}
              >
                <div
                  class={clsx(
                    "item__button",
                    !blueprint() && "item__button--error",
                  )}
                  onClick={() => {
                    if (!blueprint()) {
                      return;
                    }
                    setEditorUi("selectedElementId", id);
                    if (editorUi.hoveredElementId === id) {
                      setEditorUi("hoveredElementId", undefined);
                    }
                    focusElement(id);
                  }}
                >
                  <Show when={blueprint()}>
                    {blueprint().title()}
                    <Show when={data.id?.length}>
                      <span class="item__id">#{data.id}</span>
                    </Show>
                  </Show>
                  <Show when={!blueprint()}>
                    {"Invalid blueprint: "}
                    <span class="item__invalid_blueprint">
                      {element().blueprint}
                    </span>
                  </Show>
                </div>
                <Show
                  when={
                    blueprint() &&
                    (!element().parentId ||
                      slotHasSpace(
                        declaration,
                        state.elements[element().parentId!],
                        element().parentSlot,
                      ))
                  }
                >
                  <button
                    title="Duplicate"
                    onClick={function () {
                      setState(
                        produce(function (draft) {
                          duplicateElement(element(), draft);
                        }),
                      );
                    }}
                  >
                    <CopyIcon />
                  </button>
                </Show>
                <button
                  title="Drag and drop"
                  onClick={function () {
                    if (blueprint()) {
                      return;
                    }

                    deleteElement(id);
                  }}
                  class={clsx(
                    "item__handle",
                    blueprint()
                      ? "item__handle--move handle"
                      : "item__handle--remove",
                  )}
                >
                  <Show when={blueprint()} fallback={<TrashIcon />}>
                    <ArrowsMoveIcon />
                  </Show>
                </button>
              </div>
              <Show
                when={
                  element()?.blueprint != "none" &&
                  p.recursive &&
                  slots().length
                }
              >
                <div
                  class={
                    "item__slots " +
                    css({
                      background: targeted()
                        ? "linear-gradient(to right, var(--np-accent-alt-400), var(--np-accent-alt-300) 3px, transparent 3rem)"
                        : "linear-gradient(to right, #cbd5e1cc, #e2e8f0cc 3px, transparent 3rem)",
                      backgroundRepeat: "no-repeat",
                      marginLeft: "0.5rem",
                      borderBottomLeftRadius: "0.5rem",
                      paddingTop: slots().length > 1 ? "0.25rem" : undefined,
                      paddingLeft: "1rem",
                    })
                  }
                >
                  <For each={slots()}>
                    {({ id, slot, elements }) => {
                      const show = createMemo(function () {
                        return editorUi.expandElementsTree || elements().length;
                      });
                      return (
                        <div
                          class={css({
                            height: show() ? undefined : "0px",
                            visibility: show() ? undefined : "hidden",
                            backgroundColor: slot._ghost
                              ? "#f8717133"
                              : undefined,
                            border: slot._ghost
                              ? "1px solid #f87171"
                              : undefined,
                            borderRadius: slot._ghost ? "0.375rem" : undefined,
                            padding: slot._ghost ? "0 0.25rem" : undefined,
                            margin: slot._ghost ? "0.125rem 0" : undefined,
                          })}
                          title={
                            slot._ghost
                              ? `Slot "${id}" doesn't exist, but has elements!`
                              : undefined
                          }
                        >
                          <Show when={slots().length > 1}>
                            <div
                              class={css({
                                position: "absolute",
                                cursor: "default",
                                pointerEvents: "none",
                                paddingLeft: "0.5rem",
                                fontSize: "0.8rem",
                                color:
                                  targeted() &&
                                  id === editorUi.createMoveTarget.slot
                                    ? "var(--np-accent-900)"
                                    : slot._ghost
                                      ? "#b91c1c"
                                      : "#64748b",
                              })}
                            >
                              {id}
                            </div>
                          </Show>
                          <SidebarElementSlot
                            elementId={element().id}
                            elementIds={elements()}
                            slot={id}
                            recursive
                            deep
                            showLabelPadding={slots().length > 1}
                          />
                        </div>
                      );
                    }}
                  </For>
                </div>
              </Show>
            </div>
          );
        }}
      </For>
    </div>
  );
};

const SidebarSlots: Component = function () {
  const [state] = useContext(StateContext)!;
  const [editorUi] = useContext(EditorUiContext)!;
  const [declaration] = useContext(ConfigContext)!;
  const style = cssMemo(() => {
    return {
      display: "flex",
      flexDirection: "column",
      gap: "0.5rem",
      "& .item": {
        overflow: "hidden",
        borderTopLeftRadius: "0.25rem",
        borderTopRightRadius: "1rem",
        borderBottomLeftRadius: "0.25rem",
        borderBottomRightRadius: "0.25rem",
        boxShadow: "rgba(0, 0, 0, 0.05) 0px 1px 2px 0px",
      },
      "& .item__title": {
        fontSize: "0.875rem",
        lineHeight: "1.25rem",
        letterSpacing: "-0.025em",
        color: "#475569",
        backgroundColor: "#cbd5e1",
        padding: "1px 1.25rem 1px 1rem",
        fontWeight: 600,
      },
      "& .item__title--active": {
        backgroundColor: "var(--np-accent-alt-300)",
        color: "var(--np-accent-900)",
      },
    };
  });

  const ItemTitle: Component<{
    slotLabel?: string;
    slot: string;
    elementId?: string;
  }> = function (props) {
    return (
      <div
        class={clsx(
          "item__title",
          props.elementId === editorUi.createMoveTarget.elementId &&
            props.slot === editorUi.createMoveTarget.slot &&
            "item__title--active",
        )}
      >
        {props.slotLabel || props.slot}
      </div>
    );
  };

  return (
    <div class={style()}>
      <Show
        when={
          editorUi.selectedElementId &&
          state.elements[editorUi.selectedElementId] &&
          declaration.blueprints["none"]
        }
      >
        <For
          each={Object.keys(
            declaration.blueprints[
              state.elements[editorUi.selectedElementId!].blueprint
            ]?.slots,
          )}
        >
          {(v) => {
            const slot = createMemo(
              () => state.elements[editorUi.selectedElementId!].slots[v] || {},
            );

            return (
              <div class="item">
                <ItemTitle elementId={editorUi.selectedElementId!} slot={v} />
                <SidebarElementSlot
                  elementId={editorUi.selectedElementId!}
                  elementIds={slot().elements}
                  slot={v}
                />
              </div>
            );
          }}
        </For>
      </Show>
      <Show when={!editorUi.selectedElementId}>
        <Show
          when={state.root.header.length > 0 || editorUi.expandElementsTree}
        >
          <div class="item">
            <ItemTitle slot="header" elementId="root" />
            <SidebarElementSlot
              elementId="root"
              elementIds={state.root.header}
              recursive
              slot="header"
            />
          </div>
        </Show>

        <div class="item">
          <ItemTitle slotLabel="main" slot="default" elementId="root" />
          <SidebarElementSlot
            elementId="root"
            elementIds={state.root.default}
            recursive
            slot="default"
          />
        </div>

        <Show
          when={state.root.footer.length > 0 || editorUi.expandElementsTree}
        >
          <div class="item">
            <ItemTitle slot="footer" elementId="root" />
            <SidebarElementSlot
              elementId="root"
              elementIds={state.root.footer}
              recursive
              slot="footer"
            />
          </div>
        </Show>
      </Show>
    </div>
  );
};

const ElementsTab = function () {
  const [state] = useContext(StateContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const [declaration] = useContext(ConfigContext)!;
  const showElements = createMemo(function () {
    if (!editorUi.selectedElementId) {
      return true;
    }

    const element = state.elements[editorUi.selectedElementId];
    if (!element) {
      return true;
    }

    const blueprint = declaration.blueprints[element.blueprint];
    const blueprintSlots = Object.keys(blueprint.slots);
    if (blueprintSlots.length) {
      return true;
    }

    return false;
  });

  return (
    <>
      <Show
        when={
          editorUi.selectedElementId &&
          state.elements[editorUi.selectedElementId]
        }
      >
        <Breadcrumbs />
      </Show>
      <div
        class={css({
          display: "flex",
          flexDirection: "column",
          gap: "1rem",
        })}
      >
        <Show when={showElements()}>
          <div>
            <TabSubtitle
              rightColumn={
                <Show when={!editorUi.selectedElementId}>
                  <SmallButton
                    title="Show/hide empty slots"
                    theme="primary"
                    onClick={() =>
                      setEditorUi(
                        produce(function (d) {
                          d.expandElementsTree = !d.expandElementsTree;
                        }),
                      )
                    }
                  >
                    <Show
                      when={!editorUi.expandElementsTree}
                      fallback={
                        <>
                          <ArrowsDiagonalMinimizeIcon /> Shrink
                        </>
                      }
                    >
                      <ArrowsDiagonalIcon /> Expand
                    </Show>
                  </SmallButton>
                </Show>
              }
            >
              <Show when={!editorUi.selectedElementId} fallback={"Slots"}>
                <TreeIcon width="1.35rem" height="1.35rem" /> Overview
              </Show>
            </TabSubtitle>

            <SidebarSlots />
          </div>
        </Show>
        <Show
          when={
            editorUi.selectedElementId &&
            state.elements[editorUi.selectedElementId]
          }
        >
          <Show
            when={
              declaration.blueprints[
                state.elements[editorUi.selectedElementId!].blueprint
              ]
            }
          >
            <ElementBlueprintSettings />
          </Show>
        </Show>
      </div>
    </>
  );
};

export default ElementsTab;
