import { Component, createSignal, For, Show, useContext } from "solid-js";
import { produce } from "solid-js/store";
import { createRouteAction, useRouteData } from "solid-start";
import { css } from "../../../css/fela";
import { useAuth } from "../../../hooks/auth";
import {
  createFontRpc,
  createFontFace,
  deleteFontRpc,
  deleteFontFace,
  updateFontsRpc,
  uploadFontsRpc,
} from "../../../server/rpc";
import { SmallButton } from "../../common/buttons";
import Checkbox from "../../common/checkbox";
import { PlusIcon, TrashIcon } from "../../common/icons";
import { sortable, createOnEnd } from "../../../directives/sortable";
import { defaultFallbacks } from "../../page/fonts";
import { SidebarCard } from "./util";
import { demoEnabled } from "../../../util";
import { FileInput, Input } from "../../common/inputs";
import { adminRouteData } from "../../admin/routeData";
import { StateContext } from "../../context/page";

const textAbc = "abcdefghijklm\nnopqrstuvwxyz";

export const Fonts: Component = function (props) {
  sortable;
  const auth = useAuth();
  const data = useRouteData<typeof adminRouteData>();
  const [state, setState] = useContext(StateContext)!;
  const commonUpdateFonts = () =>
    updateFontsRpc({ csrf: auth()?.csrf!, fonts: state.fonts });
  const [createFontState, createFontAction] = createRouteAction(
    async () => {
      await commonUpdateFonts();
      return await createFontRpc({
        projectId: data.editorPage()?.project.id!,
        csrf: auth()?.csrf!,
      });
    },
    { invalidate: "fonts" },
  );
  const [createFontFaceState, createFontFaceAction] = createRouteAction(
    async ({ id }: { id: string }) => {
      await commonUpdateFonts();
      await createFontFace({ csrf: auth()?.csrf!, id });
    },
    { invalidate: "fonts" },
  );
  const [deleteFontFaceState, deleteFontFaceAction] = createRouteAction(
    async ({ id }: { id: string }) => {
      await commonUpdateFonts();
      await deleteFontFace({ csrf: auth()?.csrf!, id });
    },
    { invalidate: "fonts" },
  );
  const [deleteFontState, deleteFontAction] = createRouteAction(
    async ({ id }: { id: string }) => {
      await commonUpdateFonts();
      return await deleteFontRpc({ csrf: auth()?.csrf!, id });
    },
    { invalidate: "fonts" },
  );
  const [uploadFontsState, uploadFontAction] = createRouteAction(
    async (form: FormData) => {
      await commonUpdateFonts();
      form.set("csrf", auth()?.csrf!);
      const id = form.get("id") as string;
      const files = await uploadFontsRpc(form);
      setShowFontFaces(id);
    },
    { invalidate: "fonts" },
  );
  const { Form: UploadForm } = uploadFontAction;
  const [showFontFaces, setShowFontFaces] = createSignal<string>();

  return (
    <div>
      <SmallButton theme="primary" onClick={() => createFontAction()}>
        Add font <PlusIcon />
      </SmallButton>
      <Show when={state.fonts.length}>
        <div
          class={css({
            marginTop: "0.5rem",
            display: "flex",
            gap: "1rem",
            flexDirection: "column",
          })}
        >
          <For each={state.fonts}>
            {(font, idx) => {
              const [testText, setTestText] = createSignal(
                textAbc + "\n" + textAbc.toUpperCase() + "\n0123456789",
              );
              return (
                <SidebarCard>
                  <div
                    class={css({
                      fontSize: "0.9rem",
                      padding: "0.5rem",
                    })}
                  >
                    <div class={css({ display: "flex", gap: "0.5rem" })}>
                      <div
                        class={css({
                          display: "grid",
                          gridTemplateColumns: "repeat(2, auto)",
                          alignItems: "center",
                          gap: "0.5rem",
                        })}
                      >
                        <div>Family</div>
                        <Input
                          value={font.family}
                          onInput={(e) =>
                            setState(
                              produce(
                                (draft) =>
                                  (draft.fonts[idx()].family = e.target.value),
                              ),
                            )
                          }
                        />
                        <div>Default</div>
                        <div
                          class={css({
                            display: "flex",
                            justifyContent: "flex-end",
                          })}
                        >
                          <Checkbox
                            state={state.defaultFont === font.publicId}
                            setState={() =>
                              setState(
                                produce(function (draft) {
                                  draft.defaultFont =
                                    draft.defaultFont === font.publicId
                                      ? undefined
                                      : font.publicId;
                                }),
                              )
                            }
                          />
                        </div>
                        <div>Fallbacks</div>
                        <Input
                          placeholder={defaultFallbacks.join(", ")}
                          value={font.fallbacks.join(",")}
                          onChange={(e) =>
                            setState(
                              produce(function (draft) {
                                draft.fonts[idx()].fallbacks =
                                  e.target.value.split(",");
                              }),
                            )
                          }
                        />
                        <div>Preload all</div>
                        <div
                          class={css({
                            display: "flex",
                            justifyContent: "flex-end",
                          })}
                        >
                          <Checkbox
                            state={!!font.preload}
                            setState={(v) =>
                              setState(
                                produce(
                                  (draft) => (draft.fonts[idx()].preload = v),
                                ),
                              )
                            }
                          />
                        </div>
                        <Show when={!demoEnabled()}>
                          <UploadForm
                            method="post"
                            enctype="multipart/form-data"
                            class={css({
                              display: "flex",
                              justifyContent: "space-between",
                              gridColumnStart: "1",
                              gridColumnEnd: "3",
                            })}
                          >
                            <input name="id" type="hidden" value={font.id} />
                            <FileInput
                              name="files"
                              multiple
                              style={{ width: "200px" }}
                            />
                            <SmallButton type="submit">Upload</SmallButton>
                          </UploadForm>
                        </Show>
                        <div>
                          <Show when={font.faces.length}>
                            <SmallButton
                              onClick={() => {
                                setShowFontFaces(
                                  showFontFaces() === font.id
                                    ? undefined
                                    : font.id,
                                );
                              }}
                            >
                              {font.faces.length} Faces
                            </SmallButton>
                          </Show>
                        </div>
                        <div
                          class={css({
                            display: "flex",
                            justifyContent: "flex-end",
                          })}
                        >
                          <SmallButton
                            onClick={() => {
                              createFontFaceAction({ id: font.id! });
                              setShowFontFaces(font.id);
                            }}
                          >
                            Add face <PlusIcon />
                          </SmallButton>
                        </div>
                      </div>
                      <div
                        class={css({
                          width: "50%",
                          display: "flex",
                          gap: "0.5rem",
                          minWidth: 0,
                        })}
                      >
                        <div
                          title="Testarea"
                          spellcheck={false}
                          contenteditable
                          class={css({
                            display: "flex",
                            flex: 1,
                            minWidth: 0,
                            flexDirection: "column",
                            justifyContent: "center",
                            resize: "none",
                            whiteSpace: "pre",
                            overflowWrap: "normal",
                            overflowX: "auto",
                            height: "100%",
                            lineHeight: "1",
                            backgroundColor: "#f9fafb",
                            borderColor: "#9ca3af",
                            borderRadius: "0.125rem",
                            borderWidth: "1px",
                            boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                            transition:
                              "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
                            padding: "0.15rem 0.5rem",
                            fontSize: "1.25rem !important",
                            fontFamily: `var(--font-${font.publicId})`,
                            ":hover": {
                              boxShadow:
                                "0 0 0 1px white, 0 0 0 3px var(--np-accent-alt-300)",
                              outline: "none",
                            },
                            ":focus": {
                              boxShadow:
                                "0 0 0 1px white, 0 0 0 3px var(--np-accent-400)",
                              outline: "none",
                            },
                          })}
                          textContent={testText()}
                        />

                        <div
                          class={css({
                            display: "flex",
                            justifyContent: "flex-end",
                          })}
                        >
                          <SmallButton
                            theme="primary"
                            onClick={() => deleteFontAction({ id: font.id! })}
                          >
                            <TrashIcon />
                          </SmallButton>
                        </div>
                      </div>
                    </div>

                    <Show
                      when={showFontFaces() === font.id && font.faces.length}
                    >
                      <div
                        class={css({
                          display: "grid",
                          gridTemplateColumns: "repeat(2, minmax(0, 1fr))",
                          gap: "0.25rem",
                          marginTop: "0.5rem",
                          fontSize: "0.8rem",
                          padding: "0 0.5rem",
                        })}
                      >
                        <For each={font.faces}>
                          {(face, faceIdx) => {
                            return (
                              <div
                                class={css({
                                  display: "flex",
                                  flexDirection: "column",
                                  gap: "0.5rem",
                                  justifyContent: "space-between",
                                  padding: "0.375rem 0.5rem",
                                  backgroundColor: "#f8fafc",
                                  borderRadius: "0.5rem",
                                  borderColor: "#e2e8f0",
                                  borderWidth: "1px",
                                  boxShadow: "0 1px 2px 0 rgb(0 0 0 / 0.05)",
                                })}
                              >
                                <div
                                  class={css({
                                    display: "grid",
                                    gridTemplateColumns: "repeat(2, auto)",
                                    alignItems: "center",
                                    columnGap: "0.5rem",
                                    rowGap: "0.25rem",
                                  })}
                                >
                                  <div>Weight</div>
                                  <Input
                                    value={face.weight || ""}
                                    onInput={(e) => {
                                      setState(
                                        produce(
                                          (draft) =>
                                            (draft.fonts[idx()].faces[
                                              faceIdx()
                                            ].weight = parseInt(
                                              e.target.value,
                                            )),
                                        ),
                                      );
                                    }}
                                  />
                                  <div>Style</div>
                                  <Input
                                    value={face.style || ""}
                                    onInput={(e) => {
                                      setState(
                                        produce(
                                          (draft) =>
                                            (draft.fonts[idx()].faces[
                                              faceIdx()
                                            ].style = e.target.value),
                                        ),
                                      );
                                    }}
                                  />
                                  <div>Preload</div>
                                  <Checkbox
                                    state={!!face.preload}
                                    setState={(v) => {
                                      setState(
                                        produce(
                                          (draft) =>
                                            (draft.fonts[idx()].faces[
                                              faceIdx()
                                            ].preload = v),
                                        ),
                                      );
                                    }}
                                  />
                                  Files
                                  <div
                                    data-face={faceIdx()}
                                    use:sortable={{
                                      forceFallback: true,
                                      group: {
                                        name: "fonts",
                                        pull: true,
                                      },
                                      onMove: function (evt) {
                                        const sameList = evt.from === evt.to;
                                        if (sameList) {
                                          return false;
                                        }
                                      },
                                      onEnd: createOnEnd(
                                        function (childProduce, evt) {
                                          const sameList = evt.from === evt.to;
                                          const { oldIndex, newIndex } = evt;
                                          if (
                                            sameList ||
                                            oldIndex == null ||
                                            newIndex == null
                                          ) {
                                            return;
                                          }

                                          const targetFace =
                                            evt.to.dataset.face;
                                          if (!targetFace) {
                                            return;
                                          }

                                          const targetFaceIdx =
                                            parseInt(targetFace);

                                          setState(
                                            produce(function (draft) {
                                              const file =
                                                state.fonts[idx()].faces[
                                                  faceIdx()
                                                ].files[oldIndex];

                                              draft.fonts[idx()].faces[
                                                faceIdx()
                                              ].files.splice(oldIndex, 1);
                                              draft.fonts[idx()].faces[
                                                targetFaceIdx
                                              ].files.splice(newIndex, 0, file);
                                            }),
                                          );
                                        },
                                      ),
                                    }}
                                    class={css({
                                      padding: "0 0.25rem",
                                      minHeight: "1rem",
                                      display: "flex",
                                      flexWrap: "wrap",
                                      gap: "0.2rem",
                                      fontSize: "0.7rem",
                                      borderRadius: "0.15rem",
                                      backgroundColor: "#f1f5f9",
                                      overflow: "hidden",
                                    })}
                                  >
                                    <For each={face.files}>
                                      {(file) => {
                                        return (
                                          <div
                                            title={file.name}
                                            class={css({
                                              whiteSpace: "nowrap",
                                              overflow: "hidden",
                                              textOverflow: "ellipsis",
                                              cursor: "grab",
                                            })}
                                          >
                                            {file.name}
                                          </div>
                                        );
                                      }}
                                    </For>
                                  </div>
                                </div>
                                <SmallButton
                                  onClick={() =>
                                    deleteFontFaceAction({ id: face.id })
                                  }
                                >
                                  <TrashIcon />
                                </SmallButton>
                              </div>
                            );
                          }}
                        </For>
                      </div>
                    </Show>
                  </div>
                </SidebarCard>
              );
            }}
          </For>
        </div>
      </Show>
    </div>
  );
};
