import {
  batch,
  Component,
  createMemo,
  createSignal,
  FlowComponent,
  onCleanup,
  onMount,
  Show,
  startTransition,
  useContext,
} from "solid-js";
import ChooseTypeTab from "./chooseTypeTab";
import ElementsTab from "./elementsTab";
import { css, cssMemo } from "../../../css/fela";
import { IStyle } from "fela";
import {
  A,
  createRouteAction,
  refetchRouteData,
  useParams,
  useRouteData,
} from "solid-start";
import {
  DashboardIcon,
  ExternalLinkIcon,
  LoaderIcon,
  PublishIcon,
  SaveIcon,
} from "../../common/icons";
import PageTab from "./pageTab";
import { savePage, updateFontsRpc } from "../../../server/rpc";
import { useAuth } from "nitropage";
import { adminRouteData } from "../../admin/routeData";
import ProjectTab from "./projectTab";
import { produce } from "solid-js/store";
import { useVisitLink } from "../../admin/util";
import { onClientCleanup } from "../../../util";
import { toast } from "../../common/toast";
import clsx from "clsx";
import { EditorUiContext, SettingsContext } from "../../context/admin";
import { StateContext } from "../../context/page";
import { TabBody } from "./util";

const SidebarIdleTab: FlowComponent<{ tab: "elements" | "page" | "project" }> =
  function (p) {
    const [editorUi, setEditorUi] = useContext(EditorUiContext)!;

    const style = cssMemo(() => {
      const activeStyle: IStyle = {
        backgroundColor: "white",
        color: "#0f172a",
        "--tw-shadow": "0 1px 2px 0 rgb(0 0 0 / 0.05)",
        "--tw-shadow-colored": "0 1px 2px 0 var(--tw-shadow-color)",
        boxShadow:
          "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
        "--tw-border-opacity": "1 !important",
        borderColor: "var(--np-accent-500) !important",
      };

      return {
        backgroundColor: "rgb(241 245 249)",
        color: "#64748b",
        padding: "0.5rem 0.75rem",
        marginBottom: "-1px",
        borderBottomWidth: "2px",
        borderColor: "transparent",
        transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
        flex: "1 1 0%",
        fontSize: "1.125rem",
        lineHeight: "1.75rem",
        fontWeight: 500,
        ":last-child": {
          borderTopRightRadius: "9999px",
          borderBottomRightRadius: "9999px",
        },
        ":hover": {
          zIndex: 10,
          color: "#1e293b",
          "--tw-border-opacity": 1,
          borderColor: "var(--np-accent-alt-400)",
        },
        "&.active": activeStyle,
      };
    });

    return (
      <button
        class={clsx(style(), editorUi.sidebarIdleTab === p.tab && "active")}
        onClick={() =>
          setEditorUi(
            produce(function (draft) {
              draft.sidebarIdleTab = p.tab;
            }),
          )
        }
      >
        {p.children}
      </button>
    );
  };

const Sidebar: Component = function () {
  const [state] = useContext(StateContext)!;
  const [settings] = useContext(SettingsContext)!;
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const data = useRouteData<typeof adminRouteData>();
  const visitLink = useVisitLink();

  const externalUrl = createMemo(function () {
    return visitLink(data.editorPage()?.project.domain!, state.urlPath);
  });

  const [saveState, saveAction] = createRouteAction(
    async (params: Parameters<typeof savePage>[0]) => {
      try {
        await updateFontsRpc({ csrf: params.csrf, fonts: params.state.fonts });
        const result = await savePage(params);
        toast(`Successfully ${params.publish ? "published" : "saved"}!`, {
          status: "success",
        });

        return result;
      } catch (err) {
        toast(
          `Something went wrong, ${
            params.publish ? "publish" : "save"
          } failed!`,
          { status: "error" },
        );

        throw err;
      }
    },
    { invalidate: ["nothing"] },
  );

  const choosingElementType = () =>
    editorUi.sidebarStatus === "chooseType" &&
    state.elements[editorUi.creatingElementId!];

  const onStickyMount = function (el: HTMLElement) {
    const observer = new IntersectionObserver(
      ([e]) => {
        e.target.classList.toggle("is-pinned", e.intersectionRatio < 1);
      },
      { threshold: [1] },
    );

    observer.observe(el);

    onCleanup(function () {
      observer.disconnect();
    });
  };

  const style = cssMemo(() => {
    return {
      width: "100%",
      height: "100%",
      position: "relative",
      overflowY: choosingElementType() ? undefined : "scroll",
      overflowX: "hidden",
      borderRadius: "0.25rem",
      borderTopLeftRadius: "1rem",
      borderBottomLeftRadius: "1rem",
      backgroundColor: "#eaecf4",
      "--tw-text-opacity": 1,
      color: "rgb(15 23 42 / var(--tw-text-opacity))",
      padding: "0.75rem 0",
      "--tw-shadow-color": "rgb(30 41 59 / 0.15)",
      boxShadow:
        "0 10px 15px -3px var(--tw-shadow-color), 0 4px 4px -4px rgb(30 41 59 / 0.2)",
      transitionProperty: "all",
      transitionTimingFunction: "cubic-bezier(0.4, 0, 0.2, 1)",
      transitionDuration: "150ms",
      "& .save-button": {
        color: "#475569",
        display: "flex",
        alignItems: "center",
        fontSize: "0.9rem",
        height: "100%",
        padding: "0 1rem",
        transition: "all 150ms cubic-bezier(0.4, 0, 0.2, 1)",
        "--tw-shadow":
          "0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1)",
        "--tw-shadow-colored":
          "0 1px 3px 0 var(--tw-shadow-color), 0 1px 2px -1px var(--tw-shadow-color)",
        boxShadow:
          "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
        "--tw-bg-opacity": 1,
        backgroundColor: "rgb(248 250 252 / var(--tw-bg-opacity))",
        "--tw-ring-opacity": 1,
        "--tw-ring-color": "var(--np-accent-400)",
        "--tw-ring-offset-width": "2px",
        ":hover": {
          color: "#0f172a",
          zIndex: "1",
          borderColor: "transparent",
          "--tw-ring-offset-shadow":
            "var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color)",
          "--tw-ring-shadow":
            "var(--tw-ring-inset) 0 0 0 calc(2px + var(--tw-ring-offset-width)) var(--tw-ring-color)",
          boxShadow:
            "var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000)",
        },
        ":active": {
          color: "#0f172a",
          "--tw-bg-opacity": 1,
          backgroundColor: "rgb(226 232 240 / var(--tw-bg-opacity))",
        },
      },
      "& .save-button:first-child": {
        borderTopLeftRadius: "0.25rem",
        borderBottomLeftRadius: "0.25rem",
        borderRightWidth: "1px",
        borderColor: "rgb(241 245 249)",
      },
      "& .save-button:last-child": {
        borderTopRightRadius: "0.25rem",
        borderBottomRightRadius: "0.25rem",
      },
      "& .save-button--exit:last-child": {
        borderTopLeftRadius: "0.25rem",
        borderBottomLeftRadius: "0.25rem",
        borderTopRightRadius: "0",
        borderBottomRightRadius: "0",
      },
      "& .tab-container": {
        display: "flex",
        width: "100%",
        "--tw-shadow":
          "0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1)",
        "--tw-shadow-colored":
          "0 1px 3px 0 var(--tw-shadow-color), 0 1px 2px -1px var(--tw-shadow-color)",
        "box-shadow":
          "var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)",
        "--tw-bg-opacity": 1,
        borderTopRightRadius: "9999px",
        borderBottomRightRadius: "9999px",
        borderBottomWidth: "0px",
        "--tw-border-opacity": 1,
        borderColor: "rgb(203 213 225 / var(--tw-border-opacity))",
      },
    };
  });

  const params = useParams();
  const user = useAuth();

  const handleSavePage = async (publish = false) => {
    await saveAction({
      csrf: user()?.csrf!,
      publish,
      pageId: params.id,
      pageRevisionId: editorUi.selectedRevisionId,
      state,
      settings,
    });

    await startTransition(function () {
      batch(function () {
        if (saveState.result?.newRevisionId) {
          console.info("new revision id", saveState.result?.newRevisionId);
          setEditorUi("selectedRevisionId", saveState.result?.newRevisionId);
        }

        if (editorUi.selectedElementId) {
          const newId =
            saveState.result?.elementIdMap?.[editorUi.selectedElementId];
          if (newId) {
            setEditorUi("selectedElementId", newId);
          }
        }
      });
    });

    await refetchRouteData();
  };

  const savePageShortcutHandler = (e: KeyboardEvent) => {
    if (!e.ctrlKey) {
      return;
    }

    if (e.key !== "s" && e.key !== "p") {
      return;
    }

    e.preventDefault();
    handleSavePage(e.key === "p");
  };

  onMount(function () {
    document.addEventListener("keydown", savePageShortcutHandler);
  });
  onClientCleanup(function () {
    document.removeEventListener("keydown", savePageShortcutHandler);
  });

  return (
    <div class={style()}>
      <Show when={choosingElementType()}>
        <div
          class={css({
            height: "100%",
          })}
        >
          <ChooseTypeTab />
        </div>
      </Show>
      <Show when={editorUi.sidebarStatus === "idle"}>
        <div
          ref={onStickyMount}
          class={css({
            position: "sticky",
            // -0.75rem would put the header exactly to the top when its sticky, but:
            // This is -0.76 instead of -0.75 to make the IntersectionObserver working!
            // https://css-tricks.com/how-to-detect-when-a-sticky-element-gets-pinned/
            top: "-0.76rem",
            zIndex: 50,
            width: "100%",
            display: "flex",
            alignItems: "stretch",
            marginBottom: "1rem",
            gap: "2rem",
            boxShadow: "0 -5px 10px -10px #0000",
            transition: "box-shadow 0.3s ease-in-out",
            "&.is-pinned": {
              boxShadow: "0 0.5rem 1rem #0002",
              backgroundColor: "#eaecf4",
            },
          })}
        >
          <div class="tab-container">
            <SidebarIdleTab tab="elements">Elements</SidebarIdleTab>
            <SidebarIdleTab tab="page">Page</SidebarIdleTab>
            <SidebarIdleTab tab="project">Project</SidebarIdleTab>
          </div>
          <div
            class={css({
              margin: "0.45rem 0",
              display: "flex",
              gap: "0.5rem",
              alignItems: "center",
            })}
          >
            <div
              class={css({
                height: "100%",
                display: "flex",
                justifyItems: "stretch",
              })}
            >
              <button
                class="save-button"
                title="Save (Ctrl+S)"
                onClick={() => handleSavePage()}
              >
                <Show
                  when={saveState.input?.publish || !saveState.pending}
                  fallback={
                    <LoaderIcon
                      class={css({
                        animation: "np-spin 1s linear infinite",
                      })}
                    />
                  }
                >
                  <SaveIcon />
                </Show>
              </button>
              <button
                class="save-button save-button--right"
                title="Save & Publish (Ctrl+P)"
                onClick={() => handleSavePage(true)}
              >
                <Show
                  when={!saveState.input?.publish || !saveState.pending}
                  fallback={
                    <LoaderIcon
                      class={css({
                        animation: "np-spin 1s linear infinite",
                      })}
                    />
                  }
                >
                  <PublishIcon />
                </Show>
              </button>
              <Show when={data.editorPage()?.publishedRevisionId}>
                <A
                  href={externalUrl()}
                  target={externalUrl()}
                  class="save-button save-button--right"
                  title="Visit"
                >
                  <ExternalLinkIcon />
                </A>
              </Show>
            </div>
            <A
              href={"/admin/project/" + data.editorPage()?.projectId}
              title="Leave editor"
              class={"save-button save-button--exit"}
            >
              <DashboardIcon />
            </A>
          </div>
        </div>
        <TabBody>
          <Show when={editorUi.sidebarIdleTab === "elements"}>
            <ElementsTab />
          </Show>
          <Show when={editorUi.sidebarIdleTab === "page"}>
            <PageTab />
          </Show>
          <Show when={editorUi.sidebarIdleTab === "project"}>
            <ProjectTab />
          </Show>
        </TabBody>
      </Show>
    </div>
  );
};

export default Sidebar;
