import { Component, For, Show, Suspense, useContext } from "solid-js";
import { css } from "../../../css/fela";
import { ButtonSpinner, SmallButton } from "../../common/buttons";
import { createRouteAction, useRouteData } from "solid-start";
import { adminRouteData } from "../../admin/routeData";
import { EyeOffIcon, SelectIcon, TrashIcon } from "../../common/icons";
import {
  SettingsAccordionItem,
  SettingsAccordion,
  SidebarCard,
  MetaImageSetting,
} from "./util";
import { deleteRevisionRpc, unpublishPageRpc } from "../../../server/rpc";
import { useAuth } from "nitropage";
import { format } from "date-fns";
import { Input, TextArea } from "../../common/inputs";
import { produce } from "solid-js/store";
import { slugifyUrl } from "../../../admin/slugify";
import Checkbox from "../../common/checkbox";
import clsx from "clsx";
import { BlueprintSettingsContainer } from "./blueprintData";
import { StateContext } from "../../context/page";
import { EditorUiContext } from "../../context/admin";

/*
<SettingsAccordionItem
  value={"blueprintSettings"}
  title={"Blueprint Settings"}
>
  <PageBlueprintSettings />
</SettingsAccordionItem>
*/

const PageTab: Component = function () {
  const [state, setState] = useContext(StateContext)!;
  const data = useRouteData<typeof adminRouteData>();
  const [editorUi, setEditorUi] = useContext(EditorUiContext)!;
  const user = useAuth();

  const [deleteRevisionState, deleteRevisionAction] = createRouteAction(
    async (params: Parameters<typeof deleteRevisionRpc>[0]) =>
      await deleteRevisionRpc(params),
    {
      invalidate: ["editorPage"],
    },
  );

  const [unpublishPageState, unpublishPageAction] = createRouteAction(
    async (params: Parameters<typeof unpublishPageRpc>[0]) =>
      await unpublishPageRpc(params),
  );

  return (
    <>
      <div
        class={css({
          marginBottom: "1rem",
        })}
      >
        <BlueprintSettingsContainer>
          <Suspense>
            <SidebarCard>
              <div
                class={css({
                  display: "grid",
                  padding: "0.25rem 0.5rem",
                  gridTemplateColumns: "1fr auto 2fr",
                  alignItems: "center",
                  gap: "0.25rem",
                  fontSize: "0.875rem",
                  "& .label": {
                    fontWeight: "500",
                    fontSize: "0.75rem",
                  },
                })}
              >
                <div class="label">Title</div>
                <div />
                <Input
                  value={state.title}
                  onInput={(e) =>
                    setState(produce((d) => (d.title = e.target.value)))
                  }
                />
                <div class="label">URL Path</div>
                <div
                  class={css({
                    fontSize: "1.25rem",
                    fontWeight: 500,
                    marginRight: "0.125rem",
                    lineHeight: 1,
                    color: "#475569",
                  })}
                >
                  /
                </div>
                <Input
                  value={state.urlPath.slice(1)}
                  onInput={(e) =>
                    setState(produce((d) => (d.urlPath = "/" + e.target.value)))
                  }
                  onChange={() =>
                    setState(
                      produce((d) => (d.urlPath = slugifyUrl(d.urlPath))),
                    )
                  }
                />
                <div class="label">Include In Feed</div>
                <div />
                <Checkbox
                  state={!!state.includeInFeed}
                  setState={(v) =>
                    setState(
                      produce(function (draft) {
                        draft.includeInFeed = v;
                      }),
                    )
                  }
                />
                <div class="label">Is 404 Page</div>
                <div />
                <Checkbox
                  state={!!state.notFound}
                  setState={(v) =>
                    setState(
                      produce(function (draft) {
                        draft.notFound = v;
                      }),
                    )
                  }
                />
                <Show when={data.editorPage.latest?.publishedRevision != null}>
                  <div class="label">Published At</div>
                  <div></div>
                  <div
                    class={css({
                      display: "flex",
                      alignItems: "center",
                      justifyContent: "space-between",
                    })}
                  >
                    <div>
                      {format(
                        new Date(
                          data.editorPage.latest!.publishedRevision!.savedAt,
                        ),
                        "dd.MM.yy - kk:mm",
                      )}
                    </div>
                    <div>
                      <SmallButton
                        theme="primary"
                        onClick={() =>
                          unpublishPageAction({
                            csrf: user()!.csrf,
                            id: data.editorPage.latest!.id,
                          })
                        }
                      >
                        <ButtonSpinner show={unpublishPageState.pending}>
                          <EyeOffIcon /> Unpublish
                        </ButtonSpinner>
                      </SmallButton>
                    </div>
                  </div>
                </Show>
                <div class="label">Meta Description</div>
                <div></div>
                <TextArea
                  onInput={(e) =>
                    setState(
                      produce(
                        (draft) =>
                          (draft.metaDescription = e.currentTarget.value),
                      ),
                    )
                  }
                >
                  {state.metaDescription}
                </TextArea>
              </div>
            </SidebarCard>
            <MetaImageSetting from="pages" />
          </Suspense>
        </BlueprintSettingsContainer>
      </div>
      <SettingsAccordion defaultValue={"blueprintSettings"}>
        <SettingsAccordionItem
          value={"revisions"}
          title={`Revisions (${data.editorPage.latest?.revisions.length})`}
        >
          <div
            class={css({
              display: "grid",
              rowGap: "0.5rem",
              gridTemplateColumns: "1fr auto auto",
              "& .col": {
                padding: "0.2rem 1rem",
              },
              "& .publish-label": {
                display: "flex",
                alignItems: "center",
                fontSize: "0.7rem",
              },
              "& .btn": {
                marginLeft: "0.5rem",
                gap: "0.5rem",
                display: "flex",
                alignItems: "center",
              },
              "& .selected": {
                backgroundColor: "#0001",
              },
            })}
          >
            <For each={data.editorPage.latest?.revisions}>
              {(rev, idx) => {
                return (
                  <>
                    <div
                      class={clsx(
                        "col",
                        editorUi.selectedRevisionId === rev.id && "selected",
                      )}
                    >
                      {format(new Date(rev.savedAt), "dd.MM.yy - kk:mm")}
                    </div>
                    <div
                      class={clsx(
                        "col publish-label",
                        editorUi.selectedRevisionId === rev.id && "selected",
                      )}
                    >
                      {data.editorPage.latest?.publishedRevisionId === rev.id
                        ? "Published"
                        : ""}
                    </div>
                    <div class="btn">
                      <SmallButton
                        onClick={() => {
                          setEditorUi("selectedRevisionId", rev.id);
                        }}
                        disabled={editorUi.selectedRevisionId === rev.id}
                      >
                        <SelectIcon />
                      </SmallButton>
                      <Show
                        when={
                          idx() !=
                          (data.editorPage.latest?.revisions?.length || 0) - 1
                        }
                      >
                        <SmallButton
                          data-test={idx()}
                          onClick={() => {
                            deleteRevisionAction({
                              id: rev.id,
                              csrf: user()?.csrf!,
                            });
                          }}
                          disabled={
                            data.editorPage.latest?.publishedRevisionId ===
                              rev.id || editorUi.selectedRevisionId === rev.id
                          }
                        >
                          <TrashIcon />
                        </SmallButton>
                      </Show>
                    </div>
                  </>
                );
              }}
            </For>
          </div>
        </SettingsAccordionItem>
      </SettingsAccordion>
    </>
  );
};

export default PageTab;
