import { NitroMedia } from "@prisma/client";
import { decodeBlurHash } from "fast-blurhash";
import {
  Component,
  createEffect,
  createMemo,
  JSX,
  Show,
  splitProps,
} from "solid-js";

export const MediaBlur: Component<
  {
    media: NitroMedia;
    detail?: number;
  } & JSX.HTMLAttributes<HTMLCanvasElement>
> = function (props) {
  const [_, restProps] = splitProps(props, ["media", "detail"]);
  return (
    <Show when={props.media.blurhash}>
      <Blurhash
        detail={props.detail}
        width={props.media.width}
        height={props.media.height}
        hash={props.media.blurhash!}
        {...restProps}
      />
    </Show>
  );
};

const Blurhash: Component<
  {
    hash: string;
    width?: number | null;
    height?: number | null;
    detail?: number;
  } & JSX.HTMLAttributes<HTMLCanvasElement>
> = function (props) {
  let canvasRef: HTMLCanvasElement | undefined = undefined;
  const [_, restProps] = splitProps(props, [
    "hash",
    "width",
    "height",
    "detail",
  ]);

  const dimensions = createMemo(function () {
    const detail = props.detail || 5;
    const width = detail;
    const height =
      props.width && props.height
        ? Math.round((detail / props.width) * props.height)
        : detail;

    return { width, height };
  });
  createEffect(function () {
    if (!props.hash) return;
    if (!canvasRef) return;
    const ctx = canvasRef.getContext("2d");
    if (!ctx) {
      return;
    }

    const d = dimensions();
    const pixels = decodeBlurHash(props.hash, d.width, d.height);
    const imageData = ctx.createImageData(d.width, d.height);
    imageData.data.set(pixels);
    ctx.putImageData(imageData, 0, 0);
  });

  return (
    <Show when={props.hash}>
      <canvas
        width={dimensions().width}
        height={dimensions().height}
        {...restProps}
        ref={canvasRef}
      ></canvas>
    </Show>
  );
};

export default Blurhash;
