import {
  children,
  Component,
  createMemo,
  createSignal,
  FlowComponent,
  For,
  mergeProps,
  onMount,
  Show,
  splitProps,
} from "solid-js";
import type { JSX } from "solid-js";
import { css } from "../../css/fela";
import { Link } from "solid-start";
import { NitroMedia } from "@prisma/client";
import { rgb } from "../../css";
import Blurhash from "./blurhash";
import clsx from "clsx";
import { useVisibilityObserver } from "../common/visibilityObserver";

declare module "solid-js" {
  namespace JSX {
    interface SourceHTMLAttributes<T> extends HTMLAttributes<T> {
      width?: number | null;
      height?: number | null;
    }
  }
}

const positionMap: Record<string, string> = {
  north: "center top",
  northeast: "right top",
  northwest: "left top",
  south: "center bottom",
  southeast: "right bottom",
  southwest: "left bottom",
  east: "right center",
  west: "left center",
};

export const MediaContainer: Component<
  {
    media: Partial<NitroMedia>;
    imageClass?: string;
    sources?: ImageSources;
    lazy?: boolean;
    fadeDuration?: number;
    alt?: string;
    preload?: boolean;
    visibilityTimeout?: number;
    width?: string | false;
    height?: string | false;
    placeholder?: boolean;
    fit?: "fill" | "contain" | "cover" | "scale-down";
    position?: string;
    blurhash?: boolean;
  } & JSX.HTMLAttributes<HTMLDivElement>
> = function (props) {
  const defaults = mergeProps({ placeholder: true }, props);
  const [local, rest] = splitProps(defaults, [
    "alt",
    "blurhash",
    "children",
    "lazy",
    "fadeDuration",
    "media",
    "position",
    "imageClass",
    "placeholder",
    "preload",
    "sources",
  ]);

  return (
    <ImageContainer
      color={
        local.placeholder
          ? rgb({ value: local.media.color || undefined })
          : undefined
      }
      position={local.position || positionMap[local.media.resizePosition || ""]}
      background={
        local.placeholder && local.blurhash && local.media.blurhash ? (
          <Blurhash
            hash={local.media.blurhash}
            width={local.media.width}
            height={local.media.height}
          />
        ) : local.placeholder && local.media.dataUrl ? (
          <img
            src={local.media.dataUrl}
            width={local.media.width!}
            height={local.media.height!}
            alt={local.media.description!}
          />
        ) : undefined
      }
      {...rest}
    >
      <MediaImage
        class={local.imageClass}
        media={local.media}
        sources={local.sources}
        lazy={local.lazy}
        fadeDuration={local.fadeDuration}
        alt={local.alt}
        preload={local.preload}
      />
    </ImageContainer>
  );
};
export const MediaImage: Component<
  {
    media: Partial<NitroMedia>;
    alt?: string;
    preload?: boolean;
    sources?: ImageSources;
    lazy?: boolean;
    fadeDuration?: number;
  } & JSX.HTMLAttributes<HTMLImageElement>
> = function (props) {
  const [local, rest] = splitProps(props, [
    "alt",
    "children",
    "media",
    "sources",
  ]);

  return (
    <Image
      alt={props.media.description || props.alt}
      sources={
        props.sources || [
          {
            url: props.media.url!,
            url2x: props.media.url2x!,
            width: props.media.width,
            height: props.media.height,
          },
        ]
      }
      {...rest}
    />
  );
};

export const ImageContainer: FlowComponent<
  {
    width?: string | false;
    height?: string | false;
    fit?: "fill" | "contain" | "cover" | "scale-down";
    position?: string;
    color?: string | null;
    background?: JSX.Element;
  } & JSX.HTMLAttributes<HTMLDivElement>
> = function (p) {
  const props = mergeProps(
    {
      class: "",
      width: "100%",
      height: "100%",
      fit: "cover",
      position: "center center",
    },
    p,
  );
  const backgroundElement = children(() => props.background);
  const [_, restProps] = splitProps(props, [
    "class",
    "width",
    "height",
    "children",
    "fit",
    "background",
    "color",
    "position",
    "class",
  ]);

  const style = createMemo(function () {
    return css({
      position: "relative",
      zIndex: 0,
      width: props.width !== false ? props.width : undefined,
      height: props.height !== false ? props.height : undefined,
      backgroundColor: props.color,
      "& > .background": {
        zIndex: -20,
        position: "absolute",
      },
      "& > .background > *": {
        objectFit: props.fit,
        objectPosition: props.position,
        width: "100%",
        height: "100%",
      },
      "& > *": {
        width: "100%",
        height: "100%",
      },
      "& img": {
        zIndex: -10,
        position: "relative",
        objectFit: props.fit,
        objectPosition: props.position,
        width: "100%",
        height: "100%",
      },
    });
  });

  return (
    <div class={clsx(style(), props.class)} {...restProps}>
      <Show when={backgroundElement()}>
        <div class="background">{backgroundElement()}</div>
      </Show>

      {props.children}
    </div>
  );
};

export type ImageSources = {
  url: string;
  url2x?: string | null;
  width?: number | null;
  height?: number | null;
  screen?: number;
}[];

const Image: Component<
  {
    lazy?: boolean;
    fadeDuration?: number;
    alt?: string | null;
    preload?: boolean;
    sources: ImageSources;
    pictureProps?: JSX.HTMLAttributes<HTMLPictureElement>;
  } & JSX.HTMLAttributes<HTMLImageElement>
> = (p) => {
  let imageEl: HTMLImageElement = undefined!;
  const props = mergeProps(
    {
      class: "",
      sources: [],
    },
    p,
  );
  const [_, restProps] = splitProps(props, [
    "preload",
    "pictureProps",
    "lazy",
    "fadeDuration",
    "class",
    "sources",
    "alt",
  ]);
  const sources = createMemo(() => props.sources);
  const [showAltText, setShowAltText] = createSignal(false);
  const lazy = createMemo(function () {
    return props.lazy || props.fadeDuration != null;
  });

  const isVisible = useVisibilityObserver(() => lazy() && imageEl);
  const [loaded, setLoaded] = createSignal(false);
  const opened = createMemo(function (prev) {
    if (prev) {
      return prev;
    }

    return lazy() && loaded() && isVisible();
  });

  onMount(function () {
    if (!imageEl.complete) {
      return;
    }

    setLoaded(true);
  });

  const mainSource = createMemo(function () {
    return sources()[0];
  });

  const style = createMemo(function () {
    const fade = props.fadeDuration != null && props.fadeDuration > 0;

    return css({
      color: !showAltText() ? "transparent" : undefined,
      opacity: fade ? 0 : undefined,
      "&[data-loaded]": fade
        ? {
            animation: `${props.fadeDuration}ms ease-in forwards np-fade-in`,
          }
        : undefined,
    });
  });

  return (
    <Show when={mainSource()?.url?.length}>
      <picture {...props.pictureProps}>
        <For each={sources()}>
          {(source, idx) => {
            const maxWidth = createMemo(() => {
              const prevSource = sources()[idx() - 1];
              if (prevSource && prevSource.screen) {
                return ` and (max-width: ${prevSource.screen - 1}px)`;
              }

              return "";
            });
            return (
              <>
                <Show when={props.preload}>
                  <Link
                    rel="preload"
                    href={source.url}
                    as="image"
                    media={`(min-width: ${source.screen || 0}px)${maxWidth()}`}
                  />
                </Show>
                <source
                  media={`(min-width: ${source.screen || 0}px)`}
                  width={source.width}
                  height={source.height}
                  srcset={
                    source.url2x
                      ? `${source.url} 1x, ${source.url2x} 2x`
                      : source.url
                  }
                />
              </>
            );
          }}
        </For>
        <img
          class={clsx(style(), props.class)}
          width={mainSource().width as any}
          height={mainSource().height as any}
          alt={props.alt || undefined}
          data-loaded={opened() ? "" : undefined}
          loading={lazy() ? "lazy" : undefined}
          src={mainSource().url}
          ref={imageEl}
          onLoad={
            lazy()
              ? (evt: any) => {
                  const el = evt.target;
                  if (el.decode) {
                    el.decode()
                      .catch(function () {})
                      .finally(() => setLoaded(true));
                  } else {
                    setLoaded(true);
                  }
                }
              : undefined
          }
          onError={function () {
            setShowAltText(true);
          }}
          {...restProps}
        />
      </picture>
    </Show>
  );
};

export default Image;
