import type { NitroPageRevision, NitroProject } from "@prisma/client";
import { formatISO } from "date-fns";
import { Component, For, Show, Suspense } from "solid-js";
import { getHttpHost } from "../../server/util";
import { Dynamic, NoHydration, renderToStringAsync } from "solid-js/web";
import { Settings, getSettings, notFoundPage } from "../../server/settings";
import { prisma } from "../../prisma";
import { useMedia, useResize } from "../data/media";
import { StateContext } from "../context/page";
import server$ from "solid-start/server";
import { createResize } from "../../server/media";

declare module "solid-js" {
  namespace JSX {
    interface IntrinsicElements {
      feed: any;
      entry: any;
      icon: any;
      logo: any;
      id: any;
      updated: any;
      url: any;
      loc: any;
      lastmod: any;
      changefreq: any;
    }
  }
}

const Feed: Component<{
  project: NitroProject;
  url: URL;
  rootRevision?: NitroPageRevision | null;
  revisions: NitroPageRevision[];
  projectSettings: Settings;
  settings: Record<string, Settings>;
}> = function (props) {
  const host = getHttpHost(props.url, props.project.domain);
  const rootLink = host + "/";
  const updated = props.revisions[0]?.savedAt;

  const [faviconMedia] = useMedia(
    () => props.projectSettings.favicon?.value?.mediaIds,
    {
      noRouting: true,
    },
  );

  const [_, optimizeUrl, { Source }] = useResize(
    server$(
      createResize({
        type: "png",
        aspectRatios: ["1/1"],
        sizes: {
          fav32: { width: 32 },
        },
      }),
    ),
  );

  return (
    <NoHydration>
      <Show when={faviconMedia()}>
        <feed xmlns="http://www.w3.org/2005/Atom">
          <title>{props.rootRevision?.title || "Feed"}</title>
          <Show when={faviconMedia()?.byOrder[0]}>
            <icon>
              {host +
                optimizeUrl(faviconMedia()?.byOrder?.[0]?.rawUrl, "fav32", {
                  aspectRatio: "1/1",
                })}
            </icon>
          </Show>
          <Dynamic component="link" rel="alternate" href={rootLink} />
          <Dynamic component="link" rel="self" href={rootLink + "feed"} />
          <Show when={updated}>
            <updated>{formatISO(updated)}</updated>
          </Show>
          <id>{rootLink}</id>
          <For each={props.revisions}>
            {(rev) => {
              const description = props.settings[rev.id].metaDescription.value;
              const id = `urn:uuid:${rev.pageId}`;
              return (
                <entry>
                  <title>{rev.title}</title>
                  <Dynamic component="link" href={host + rev.urlPath} />
                  <id>{id}</id>
                  <updated>{formatISO(rev.savedAt)}</updated>
                  <Show when={description}>
                    <summary>{description}</summary>
                  </Show>
                </entry>
              );
            }}
          </For>
        </feed>
      </Show>
    </NoHydration>
  );
};

export const handleFeed = async function ({
  url,
  project,
}: {
  url: URL;
  project: NitroProject;
}) {
  const notFoundPage_ = await notFoundPage({ projectId: project.id });

  const rootRevision = await prisma.nitroPageRevision.findFirst({
    where: {
      page: {
        projectId: project.id,
      },
      urlPath: "/",
      publishedPage: { id: { not: undefined } },
    },
  });
  const revisions = await prisma.nitroPageRevision.findMany({
    where: {
      page: {
        id: notFoundPage_.value ? { not: notFoundPage_.value } : undefined,
        includeInFeed: true,
        projectId: project.id,
      },
      publishedPage: { id: { not: undefined } },
    },
    orderBy: {
      savedAt: "desc",
    },
  });

  const settings = await Promise.allSettled<Promise<Settings>>(
    revisions.map((rev) =>
      getSettings({ pageId: rev.pageId, projectId: project.id }),
    ),
  );

  const projectSettings = await getSettings({ projectId: project.id });

  const settingsMap = settings.reduce(
    (acc, item, idx) => {
      const revision = revisions[idx];

      if (item.status === "rejected") {
        return acc;
      }

      acc[revision.id] = item.value;
      return acc;
    },
    {} as Record<string, Settings>,
  );

  const feed = await renderToStringAsync(() => (
    <StateContext.Provider value={[{ admin: false, cacheKey: 0 }] as any}>
      <Suspense>
        <Feed
          settings={settingsMap}
          projectSettings={projectSettings}
          project={project}
          rootRevision={rootRevision}
          revisions={revisions}
          url={url}
        />
      </Suspense>
    </StateContext.Provider>
  ));

  const cleanedFeed = feed.replace(/<script>.*?<\/script>/, "");

  return new Response('<?xml version="1.0" encoding="UTF-8"?>' + cleanedFeed, {
    headers: {
      "Content-Type": "application/atom+xml",
    },
  });
};
