import { Component, For, createMemo, useContext } from "solid-js";
import { Link, Style } from "solid-start";
import { Font, FontFace, FontFaceFile } from "../../types";
import { StateContext } from "../context/page";

export const defaultFallbacks = [
  `ui-sans-serif`,
  `system-ui`,
  `-apple-system`,
  `BlinkMacSystemFont`,
  `'Segoe UI'`,
  `Roboto`,
  `'Helvetica Neue'`,
  `Arial`,
  `'Noto Sans'`,
  `sans-serif`,
  `'Apple Color Emoji'`,
  `'Segoe UI Emoji'`,
  `'Segoe UI Symbol'`,
  `'Noto Color Emoji'`,
];

export const Fonts: Component = function () {
  const [state] = useContext(StateContext)!;
  const fontData = createMemo(function () {
    const data = {
      familyCss: "",
      preloads: [] as (FontFaceFile & { publicId: string })[],
    };

    const fontFamily = function (font: Font) {
      const familyName = font.family || font.publicId;
      const face = function (face: FontFace) {
        let preloadRegistered = false;
        const src = face.files
          .reduce((acc, v) => {
            if (!preloadRegistered && (font.preload || face.preload)) {
              data.preloads.push({
                publicId: font.publicId,
                ...v,
              });
              preloadRegistered = true;
            }

            acc.push(
              `url(/fonts/${font.publicId}/${v.name}) format("${v.format}")`,
            );
            return acc;
          }, [] as string[])
          .join(",");

        return `
@font-face {
  font-family: '${familyName}';
  src: ${src};
  font-weight: ${face.weight || 400};
  font-style: ${face.style || "normal"};
  font-display: ${font.preload || face.preload ? "block" : "swap"};
}`;
      };

      return font.faces.map(face).join("\n") + "\n";
    };

    data["familyCss"] = state.fonts.map(fontFamily).join("\n");

    return data;
  });
  const fontCss = createMemo(function () {
    if (!state.fonts.length) {
      return;
    }

    return `

    ${fontData().familyCss}
    ${
      state.defaultFont
        ? `.app { font-family: var(--font-${state.defaultFont})}`
        : ""
    }

    :root {
      ${state.fonts
        .map((f, i) => {
          const familyName = f.family || f.publicId;
          const fallbacks = f.fallbacks[0] ? f.fallbacks : defaultFallbacks;
          let family = `'${familyName}'`;
          if (fallbacks.length) {
            family += "," + fallbacks.join(",");
          }

          return `--font-${f.publicId}: ${family};`;
        })
        .join("\n")}
    }
    
`;
  });

  return (
    <>
      <For each={fontData().preloads}>
        {(f) => {
          return (
            <Link
              rel="preload"
              href={`/fonts/${f.publicId}/${f.name}`}
              as="font"
              type={f.mime}
              crossorigin="anonymous"
            />
          );
        }}
      </For>
      <Style class={!state.admin ? "nitro-page-style" : undefined}>
        {fontCss()}
      </Style>
    </>
  );
};
