import {
  children,
  Component,
  createMemo,
  FlowComponent,
  For,
  lazy,
  ParentComponent,
  Resource,
  Show,
  splitProps,
  useContext,
  ValidComponent,
} from "solid-js";
import { Dynamic, isServer } from "solid-js/web";
import {
  Blueprint,
  CustomBlueprintComponent,
  State,
  Element,
  RteComponent,
  SlotComponent,
} from "../../types";
import { ConfigFn } from "../../config";
import { HttpStatusCode } from "solid-start/server";
import { useRteRouterLinks } from "../../hooks/rte";
import Styles from "./../page/styles";
import { hydrated } from "./root";
import { PageMeta } from "./meta";
import { ConfigContext, StateContext } from "../context/page";
import NoHydrate from "./nohydrate";

export const Richtext: FlowComponent<{
  component?: ValidComponent;
  class?: string;
}> = function (props) {
  const child = children(() => props.children);
  const html = createMemo(function () {
    const result = child();
    if (!isServer && !hydrated) {
      return;
    }

    return result || "";
  });
  return (
    <Show when={child()}>
      <Dynamic
        component={props.component || "span"}
        class={"rte " + (props.class || "")}
        innerHTML={html()}
      />
    </Show>
  );
};

const PageElement: ParentComponent<{
  element: Element;
  blueprint: Blueprint;
  state: State;
}> = function (p) {
  const element = p.element;
  const [config] = useContext(ConfigContext)!;
  const Rte: RteComponent = function (rteProps) {
    const content = function () {
      const value = element.ssrData;

      if (rteProps.path) {
        return rteProps.path((data) => value[data]);
      }

      if (rteProps.data) {
        return value[rteProps.data];
      }
    };
    return (
      <Richtext component={rteProps.component} class={rteProps.class}>
        {/*@once*/ content()}
      </Richtext>
    );
  };

  const Slot: SlotComponent = function (ps_) {
    const [localProps, restProps] = splitProps(ps_, [
      "direction",
      "name",
      "ref",
      "optional",
      "component",
    ]);
    const elements = createMemo(function () {
      if (element.blueprint === "root") {
        return p.state.root[localProps.name as "default"];
      } else {
        return element.slots[localProps.name]?.elements;
      }
    });

    return (
      <Show when={!localProps.optional || elements()?.length}>
        <Dynamic component={localProps.component || "div"} {...restProps}>
          <For each={elements()}>
            {(elementId, idx) => {
              return (
                <NoHydrate forceHydration={p.state.elements[elementId].dynamic}>
                  <PageElement
                    state={p.state}
                    blueprint={
                      config.blueprints[p.state.elements[elementId].blueprint]
                    }
                    element={p.state.elements[elementId]}
                  />
                </NoHydrate>
              );
            }}
          </For>
        </Dynamic>
      </Show>
    );
  };

  const slotElements = function (name: string) {
    return element.slots?.[name]?.elements || [];
  };

  return (
    <Dynamic
      blueprint={p.blueprint}
      element={{
        id: element.id,
      }}
      elementProps={{
        id: element.ssrData.id.length ? element.ssrData.id : undefined,
        "data-element": true,
      }}
      lax={true}
      component={p.blueprint?.component}
      slotElements={slotElements}
      Slot={Slot}
      Rte={Rte}
      data={element.ssrData}
    />
  );
};

const Root: CustomBlueprintComponent = function (p) {
  return (
    <>
      <p.Slot component="header" optional name="header" />
      <p.Slot component="main" optional name="default" />
      <p.Slot component="footer" optional name="footer" />
    </>
  );
};

export const Page: Component<{
  state: State;
}> = (p) => {
  const rootBlueprint: Blueprint = {
    id: "root",
    title: () => "Root",
    data: {},
    slots: {},
    component: Root as any,
  };

  useRteRouterLinks();

  // TODO: This logic is flawed with client side navigation:
  // If you navigate from a non-nitropage route (e.g. /abcdef) to a nitropage served one,
  // The styles will be missing!
  const ssrStyles = isServer && p.state.cacheKey === 0;
  if (isServer) {
    p.state.cacheKey++;
  }

  return (
    <StateContext.Provider value={[p.state, () => {}]}>
      <Show when={p.state.notFound}>
        <HttpStatusCode code={404} />
      </Show>

      <PageMeta state={p.state} />

      <Show when={ssrStyles || p.state.admin || import.meta.env.DEV}>
        <Styles />
      </Show>

      <PageElement
        state={p.state}
        blueprint={rootBlueprint}
        element={{
          blueprint: "root",
          id: "root",
          elementId: "root",
          parentSlot: "default",
          slots: {},
          data: {},
          ssrData: {},
        }}
      />
    </StateContext.Provider>
  );
};

const Loader: FlowComponent<{ load: () => Promise<any> }> = function (props) {
  const BlockRender = lazy(async () => {
    await props.load();

    return {
      default: function () {
        return props.children;
      },
    };
  });

  return <BlockRender />;
};

const Nitropage: Component<{
  page: Resource<
    | { blueprintFiles: Record<string, string>; state: State | undefined }
    | undefined
  >;
  config: ConfigFn;
}> = function (p) {
  const loadConfig = async () => {
    // On the server the blueprints are loaded during page$
    if (isServer) {
      return;
    }

    const page = p.page;
    await p.config({
      config,
      setConfig,
      admin: page ? !!page()?.state?.admin : false,
      page: page ? page() : undefined,
    });
  };

  const [config, setConfig] = useContext(ConfigContext)!;

  return (
    <Show when={!p.page.loading && p.page()?.state?.urlPath} keyed>
      {(_) => (
        <Loader load={loadConfig}>
          <Show when={p.page()}>
            <Page state={p.page()!.state!} />
          </Show>
        </Loader>
      )}
    </Show>
  );
};

export default Nitropage;
