import { Component, Show, createMemo, createResource } from "solid-js";
import { State } from "../../types";
import { Link, Meta, Title } from "solid-start";
import { useMedia, useResize } from "../data/media";
import server$ from "solid-start/server";
import { createResize } from "../../server/media";
import { getHttpHost } from "../../server/util";

export const PageMeta: Component<{
  state: State;
}> = function (p) {
  const [metaImageMedia] = useMedia(() =>
    p.state.metaImage ? [p.state.metaImage] : undefined,
  );
  const [faviconMedia] = useMedia(() =>
    p.state.favicon ? [p.state.favicon] : undefined,
  );

  const [optimizeMedia, optimizeUrl, { Source }] = useResize(
    server$(
      createResize({
        type: "png",
        aspectRatios: ["1/1"],
        sizes: {
          twitter: { width: 4096, height: 2048 },
          og: { width: 1200, height: 630 },
          fav192: { width: 192 },
          fav180: { width: 180 },
          fav167: { width: 167 },
          fav48: { width: 48 },
          fav32: { width: 32 },
          fav16: { width: 16 },
        },
      }),
    ),
  );

  const metaImage = createMemo(() => metaImageMedia().byOrder[0]?.rawUrl);
  const favicon = createMemo(() => faviconMedia().byOrder[0]?.rawUrl);

  const [host] = createResource(
    server$(function () {
      return getHttpHost(
        server$.locals.url as URL,
        (server$.locals.nitroProject as any).domain as string,
      );
    }),
  );

  const Favicon: Component<{ size: number; rel?: string; href: string }> =
    function (props) {
      return (
        <Link
          rel={props.rel || "icon"}
          type="image/png"
          sizes={`${props.size}x${props.size}`}
          href={props.href}
        />
      );
    };

  return (
    <>
      <Title>{p.state.title || " "}</Title>
      <Show when={p.state.metaDescription}>
        <Meta name="description" content={p.state.metaDescription} />
      </Show>
      <Show when={favicon() && host()}>
        <Favicon
          size={16}
          href={optimizeUrl(favicon(), "fav16", { aspectRatio: "1/1" })}
        />
        <Favicon
          size={32}
          href={optimizeUrl(favicon(), "fav32", { aspectRatio: "1/1" })}
        />
        <Favicon
          size={48}
          href={optimizeUrl(favicon(), "fav48", { aspectRatio: "1/1" })}
        />
        <Favicon
          size={192}
          href={optimizeUrl(favicon(), "fav192", { aspectRatio: "1/1" })}
        />
        <Favicon
          rel="apple-touch-icon"
          size={167}
          href={optimizeUrl(favicon(), "fav167", { aspectRatio: "1/1" })}
        />
        <Favicon
          rel="apple-touch-icon"
          size={180}
          href={optimizeUrl(favicon(), "fav180", { aspectRatio: "1/1" })}
        />
      </Show>
      <Meta property="og:type" content="website" />
      <Meta property="og:title" content={p.state.title || ""} />
      <Show when={p.state.metaDescription}>
        <Meta
          property="og:description"
          content={p.state.metaDescription || ""}
        />
      </Show>
      <Show when={metaImage() && host()}>
        <Meta
          property="og:image"
          content={host() + optimizeUrl(metaImage(), "og")}
        />
      </Show>
      <Meta name="twitter:card" content="summary_large_image" />
      <Meta name="twitter:title" content={p.state.title || ""} />
      <Show when={p.state.metaDescription}>
        <Meta
          name="twitter:description"
          content={p.state.metaDescription || ""}
        />
      </Show>
      <Show when={metaImage() && host()}>
        <Meta
          name="twitter:image"
          content={host() + optimizeUrl(metaImage(), "twitter")}
        />
      </Show>
      <Link rel="alternate" type="application/atom+xml" href={`/feed/`} />
    </>
  );
};
