import {
  Component,
  createEffect,
  createSignal,
  onCleanup,
  onMount,
  Show,
} from "solid-js";
/*
import "quill/dist/quill.bubble.css";
import "quill/dist/quill.snow.css";
*/
import "../css/quill/quill-core.css";
import "../css/quill/quill-bubble.css";
import "../css/quill/quill-snow.css";
import "../css/quill/overrides.css";
import "../css/quill/quill-list.css";
import type Quill from "@lufrai/quill";
import { css } from "../css/fela";
import { range } from "froebel";
import { ToolbarOption } from "../types";
import { colorPresets } from "./page/richtext";
import { onClientCleanup } from "../util";
import clsx from "clsx";

let quillGlobalsOptimized = false;

const toolbar = {
  bold: "bold",
  italic: "italic",
  underline: "underline",
  color: "color",
  background: "background",
  link: "link",
  clean: "clean",
};

const RteEditor: Component<{
  formats?: string[];
  toolbar?: ToolbarOption;
  inline?: boolean;
  disable?: boolean;
  value: any;
  onFocus?: () => void;
  setValue: (v: string) => void;
}> = (p) => {
  let editorEl: HTMLDivElement;
  const [isActive, setIsActive] = createSignal(false);
  const [editor, setEditor] = createSignal<Quill>();
  const [loaded, setLoaded] = createSignal(false);
  let lastValue: string | undefined = undefined;
  const emptyValueRegex = new RegExp(
    '^<span class="(?:(?!").)*">(?:<br>)?</span>$',
  );
  const linebreak = '<span class="block"><br></span>';

  createEffect(() => {
    const editor_ = editor();

    if (!editor_) {
      return;
    }

    if (lastValue != null && lastValue === p.value) {
      return;
    }

    lastValue = p.value || "";
    let newValue = lastValue;
    // Quill removes the last empty line
    // This adds an extra line as a workaround..
    // https://github.com/quilljs/quill/issues/2075
    if (newValue?.endsWith(linebreak)) {
      newValue += linebreak;
    }
    const delta = editor_.clipboard.convert({ html: newValue });
    editor_.setContents(delta, "silent");
  });

  createEffect(() => {
    const editor_ = editor();

    if (!editor_) {
      return;
    }

    if (p.disable) {
      editor_.disable();
    } else {
      editor_.enable();
    }
  });

  if (p.onFocus) {
    createEffect(function () {
      if (!isActive()) {
        return;
      }

      p.onFocus!();
    });
  }

  onMount(() => {
    // Clicking on the pageview iframe doesnt blur the editor by default
    const onBlur = function () {
      const editor_ = editor();
      if (!editor_) {
        return;
      }
      editorEl.querySelector(".ql-tooltip")?.classList.add("ql-hidden");
      editor_.blur();
      setIsActive(false);
    };
    window.addEventListener("blur", onBlur);
    onClientCleanup(function () {
      window.removeEventListener("blur", onBlur);
    });

    (async () => {
      const Quill = (await import("@lufrai/quill")).default;
      setLoaded(true);
      if (!quillGlobalsOptimized) {
        const Block: any = Quill.import("blots/block");
        class CustomBlock extends Block {}
        CustomBlock.tagName = "span";
        CustomBlock.className = "block";
        Quill.register(CustomBlock, true);

        const ColorPicker: any = Quill.import("ui/color-picker");
        const Picker = Object.getPrototypeOf(ColorPicker.prototype);
        ColorPicker.prototype.buildItem = function (option: any) {
          let item: HTMLSpanElement = Picker.buildItem.call(this, option);
          const value = option.getAttribute("value");
          if (!value) {
            item.style.backgroundSize = "5px 5px";
            item.style.backgroundImage =
              "repeating-linear-gradient(45deg, #ff0000 0, #ff0000 2px, #ffffff 0, #ffffff 50%)";
            return item;
          }
          item.classList.add(`ql-bg-${value}`);
          return item;
        };

        const ColorClass: any = Quill.import("attributors/class/color");
        Quill.register(ColorClass, true);
        const BgClass: any = Quill.import("attributors/class/background");
        Quill.register(BgClass, true);

        // Improve quill link handling
        // https://github.com/quilljs/quill/pull/2795
        // https://github.com/codebuddies/codebuddies/issues/429
        const Link: any = Quill.import("formats/link");
        const origCreate = Link.create;
        Link.create = function (value: string) {
          const node = origCreate.call(this, value) as Element;
          const url = Link.sanitize(value);
          if (url.startsWith("/")) {
            node.removeAttribute("rel");
            node.removeAttribute("target");
          }
          return node;
        };
        Quill.register(Link, true);
      }

      quillGlobalsOptimized = true;

      const colors: string[] = [""];
      const bgColors: string[] = [""];
      for (const color of range(1, colorPresets, 1)) {
        colors.push("g" + color);
        bgColors.push("bg-g" + color);
      }

      const defaultFormats = [
        "bold",
        "italic",
        "underline",
        "color",
        "background",
        "link",
        "indent",
        "align",
      ];

      const toolbarOpt: ToolbarOption = p.toolbar || {};
      let toolbarItems: any = [];

      const formatItems = [];
      if (toolbarOpt.bold) {
        formatItems.push(toolbar.bold);
      }
      if (toolbarOpt.italic) {
        formatItems.push(toolbar.italic);
      }
      if (toolbarOpt.underline) {
        formatItems.push(toolbar.underline);
      }
      if (formatItems.length) {
        toolbarItems.push(formatItems);
      }

      const colorsGroup = [];
      if (toolbarOpt.color) {
        colorsGroup.push({ color: colors });
      }
      if (toolbarOpt.background) {
        colorsGroup.push({ background: colors });
      }
      if (colorsGroup.length) {
        toolbarItems.push(colorsGroup);
      }

      if (toolbarOpt.script) {
        toolbarItems.push([{ script: "sub" }, { script: "super" }]);
      }

      const alignGroup = [];
      if (toolbarOpt.listOrdered) {
        alignGroup.push({ list: "ordered" });
      }
      if (toolbarOpt.listBullet) {
        alignGroup.push({ list: "bullet" });
      }
      if (toolbarOpt.listOrdered || toolbarOpt.listBullet) {
        defaultFormats.push("list");
      }
      if (toolbarOpt.indent) {
        alignGroup.push({ indent: "-1" });
        alignGroup.push({ indent: "+1" });
      }
      if (toolbarOpt.align) {
        alignGroup.push({ align: [] });
      }
      if (alignGroup.length) {
        toolbarItems.push(alignGroup);
      }

      const commonItems = [];
      if (toolbarOpt.link) {
        commonItems.push(toolbar.link);
      }
      if (toolbarOpt.clean) {
        commonItems.push(toolbar.clean);
      }
      if (commonItems.length) {
        toolbarItems.push(commonItems);
      }

      const editor_ = new Quill(editorEl, {
        // TODO: Enable quill formats option
        // formats: p.formats || defaultFormats,
        theme: p.inline ? "bubble" : "snow",
        modules: {
          keyboard: {
            bindings: {
              "list autofill": {
                // "1. " and "- " should only be converted to list elements, if ctrl is pressed
                shortKey: true,
              },
            },
          },
          toolbar: toolbarItems.length ? toolbarItems : false,
        },
      });

      editor_.root.classList.add("rte");
      const themeModule = (editor_ as any).theme;
      const tooltipEl = themeModule?.tooltip?.root;

      // Only blur the current editor if no other editors are focused
      // Quill's blur would otherwise blur all editors o.O
      const lazyBlur = function () {
        const activeEditors = document.querySelectorAll(
          ".ql-editor:focus-within",
        );
        if (
          activeEditors.length > 1 ||
          (activeEditors[0] != null && activeEditors[0] != editorEl)
        ) {
          return;
        }

        editor_.blur();
      };

      const tooltipOpened = () =>
        tooltipEl && !tooltipEl.classList.contains("ql-hidden");

      let toolTipInterval: NodeJS.Timeout | undefined = undefined;
      const activeHandler = (range: any = editor_.getSelection()!) => {
        if (p.disable) {
          setIsActive(false);
          return;
        }

        const hasSelection = !!range;
        const hasContentFocus = editor_.hasFocus();
        const hasFocusWithin = editorEl.parentElement!.contains(
          document.activeElement,
        );
        const hasFocus = hasContentFocus || hasFocusWithin;

        if (!hasSelection && !isActive() && !hasFocus) {
          return;
        }

        clearInterval(toolTipInterval);
        if (hasSelection && hasFocus) {
          setIsActive(true);
          return;
        }

        // If the Quill Tooltip is opened we shall not loose focus
        if (p.inline && tooltipOpened()) {
          setIsActive(true);

          toolTipInterval = setInterval(function () {
            if (tooltipOpened()) {
              return;
            }

            clearInterval(toolTipInterval);
            activeHandler();
          }, 100);
          return;
        }

        setIsActive(false);
      };

      editor_.root.addEventListener("focus", () => setIsActive(true));
      editor_.root.addEventListener("blur", function (evt) {
        // If relatedTarget has the class ql-clipboard,
        // this blur event was triggered by a "paste",
        // in such a case we leave the editor active
        if (
          evt.relatedTarget &&
          (evt.relatedTarget as HTMLElement).classList.contains("ql-clipboard")
        ) {
          return;
        }

        // After a toolbar click the editor blurs and for a short time looses focus
        // setTimeout gives the editor enough time to recover the focus in such a case
        setTimeout(activeHandler, 0);
      });

      editor_.on("selection-change", activeHandler);

      // https://github.com/quilljs/quill/issues/1290#issuecomment-803256886
      (editor_.getModule("toolbar") as any)?.container.addEventListener(
        "mousedown",
        (e: MouseEvent) => {
          e.preventDefault();
        },
      );

      editor_.on("text-change", () => {
        let newValue = editor_.getSemanticHTML();
        if (lastValue === newValue) {
          return;
        }

        newValue = newValue.replace(
          new RegExp('<span class="ql-cursor">\\s*</span>'),
          "",
        );

        if (newValue.match(emptyValueRegex)) {
          newValue = "";
        }

        if (lastValue === newValue) {
          return;
        }

        lastValue = newValue;
        p.setValue(newValue);
      });

      setEditor(editor_);
    })();
  });

  onCleanup(() => {
    const editor_ = editor();

    if (!editor_) {
      return;
    }

    editor_.destroy();
  });

  return (
    <div
      class={clsx(
        p.inline && !isActive() && !p.disable && "ql-editable",
        p.inline && isActive() && "ql-selected",
      )}
      onKeyDown={(e) => e.stopPropagation()}
    >
      <Show when={!loaded() && p.value}>
        <span class="ql-preview rte" innerHTML={p.value} />
      </Show>
      <Show when={p.inline && !p.value && !isActive() && !p.disable}>
        <div
          onClick={() => {
            const editor_ = editor()!;
            setIsActive(true);
            editor_.focus();
          }}
          class="np-admin-font ql-empty-type"
        >
          Type...
        </div>
      </Show>
      <div
        class={clsx(
          p.inline &&
            !p.value &&
            !isActive() &&
            css({
              position: "absolute",
              display: "none !important",
            }),
        )}
      >
        <div spellcheck={false} ref={editorEl!} />
      </div>
    </div>
  );
};

export default RteEditor;
