import {
  Accessor,
  createEffect,
  createRoot,
  createSignal,
  Setter,
  sharedConfig,
} from "solid-js";
import { reconcile, SetStoreFunction } from "solid-js/store";
import { isServer } from "solid-js/web";
import { ViteHotContext } from "vite/types/hot";
import { Config, Blueprint, State, BlueprintFn } from "./types";
import { capitalize } from "froebel";
import { textData } from "../data/text";

const effects: Record<
  string,
  {
    dispose: () => void;
    module: any;
  }
> = {};

type ConfigCtx = {
  config: Config;
  setConfig: SetStoreFunction<Config>;
  admin: boolean;
  page?: {
    blueprintFiles?: Record<string, string> | undefined;
    passiveBlueprints?: Record<string, boolean>;
    state?: State | undefined;
  };
};

export type ConfigFn = {
  (ctx: ConfigCtx): Promise<Config>;
  _getBlueprints: () => BlueprintModules;
};

type BlueprintModules = Record<string, () => Promise<any>>;

export const createConfig = function (fn: () => BlueprintModules) {
  const configFn = function (ctx: ConfigCtx) {
    return initConfig(fn(), ctx);
  };

  configFn._getBlueprints = fn;

  return configFn;
};

export const baseFilename = function (filePath: string) {
  const fileName = filePath.substring(filePath.lastIndexOf("/") + 1);
  return fileName.split(".")[0];
};

const getAllowedFiles = function (ctx: ConfigCtx, blueprintIds: string[]) {
  if (!ctx.page?.blueprintFiles) {
    return;
  }

  const files = ctx.page.blueprintFiles;
  const seenIds: Record<string, true> = {};
  const allowedFiles: string[] = [];
  for (const id of blueprintIds) {
    if (seenIds[id]) {
      continue;
    }

    seenIds[id] = true;
    const file = files[id];
    if (!file) {
      continue;
    }

    allowedFiles.push(file);
  }

  if (import.meta.env.DEV) {
    console.info("init config - allowed files", allowedFiles);
  }

  return allowedFiles;
};

export async function initConfig(modules: BlueprintModules, ctx: ConfigCtx) {
  const onlyDynamicBlueprints =
    !isServer && sharedConfig.context && !ctx.admin && !import.meta.env.DEV;
  let ids: string[] | undefined;
  if (ctx.page?.state?.blueprintIds) {
    if (onlyDynamicBlueprints) {
      ids = Object.entries(ctx.page.state.blueprintIds).reduce(
        (acc, [k, v]) => {
          if (!v) return acc;
          acc.push(k);
          return acc;
        },
        [] as string[],
      );
    } else {
      ids = Object.keys(ctx.page.state.blueprintIds);
    }
  }

  const allowedFiles = ids ? getAllowedFiles(ctx, ids) : undefined;

  const blueprintIds = ids || [];
  await Promise.all(
    Object.keys(modules).map((k) => {
      if (allowedFiles && !allowedFiles.includes(k)) {
        return;
      }

      return modules[k]().then((m: any) => {
        if (!m.default) {
          console.error("Invalid blueprint, doesn't export component", k);
          return;
        }

        if (!m.blueprint) {
          console.error("Invalid blueprint, doesn't export config", k);
          return;
        }

        if (effects[k]) {
          if (effects[k].module === m) {
            return;
          }

          effects[k].dispose();
        }

        const updateBlueprint = function () {
          const component = m.default;

          let blueprintFn: BlueprintFn = undefined!;
          let blueprint: Blueprint;

          try {
            blueprintFn = m.blueprint();
          } catch (err) {
            console.error(`Blueprint ${k} failed to run.`);
            console.error(err);
            return;
          }

          return (async function () {
            try {
              blueprint = (await blueprintFn({
                admin: ctx.admin,
              })) as Blueprint;
            } catch (err) {
              console.error(`Blueprint ${k} failed to run.`);
              console.error(err);
              return;
            }

            blueprint.component = component;
            const fileName = baseFilename(k);
            const capitalFileName = capitalize(fileName);
            blueprint.title = blueprint.title || (() => capitalFileName);

            // blueprintFn._id -> id after hot module replace
            // m.id -> first exported id
            blueprint.id = blueprintFn._id || m.id || fileName.toLowerCase();

            blueprint.slots = blueprint.slots || {};

            blueprint.data["id"] = textData({
              sanitizeSlug: true,
              title() {
                return "ID #";
              },
            });

            // Late skip adding this blueprint if its not needed
            // This is the case on ssr if ctx.page?.blueprintFiles isnt set
            if (
              ctx.page &&
              blueprintIds &&
              !blueprintIds.includes(blueprint.id)
            ) {
              return;
            }

            blueprint.version = Math.random();

            ctx.setConfig(
              "blueprints",
              blueprint.id,
              reconcile(blueprint, { key: null, merge: true }),
            );
          })();
        };

        if (isServer || import.meta.env.PROD) {
          return updateBlueprint();
        } else {
          createRoot(function (dispose) {
            createEffect(updateBlueprint);
            effects[k] = { dispose, module: m };
          });
        }
      });
    }),
  );

  if (import.meta.env.DEV) {
    console.info(
      "init config - loaded blueprints:",
      Object.keys(ctx.config.blueprints),
    );
  }

  ctx.setConfig("blueprints", "none", {
    id: "none",
    title: () => "None",
    data: {},
    component: () => [],
  });

  return ctx.config;
}

export const createBlueprint = function <T extends BlueprintFn>(
  fn: T,
  hot?: ViteHotContext,
): Accessor<T> {
  if (!hot) {
    return () => fn;
  }

  const signal: [state: Accessor<T>, setState: Setter<T>] =
    hot.data["npBlueprint"] || createSignal(fn);
  hot.data["npBlueprintFn"] = fn;
  hot.data["npBlueprint"] = signal;

  hot.accept(function (mod: any) {
    if (!mod.blueprint) {
      return;
    }

    hot.data["npBlueprintFn"]._id = mod.id;

    signal[1](() => hot.data["npBlueprintFn"]);
  });

  return signal[0];
};
