import { useMousePosition } from "@solid-primitives/mouse";
import { onCleanup } from "solid-js";
import { unwrap } from "solid-js/store";
import Sortable, { MoveEvent, SortableEvent } from "sortablejs";
// https://github.com/SortableJS/Sortable
// https://github.com/SortableJS/Sortable/wiki/Swap-Thresholds-and-Direction#dealing-with-swap-glitching

export const sortable = (el: HTMLElement, config: () => Sortable.Options) => {
  if (typeof window === "undefined") {
    return;
  }
  const config_ = config();
  const userOnEnd = config_?.onEnd;
  const userOnMove = config_?.onMove;
  const pos = useMousePosition();

  let oldY: false | number = false;
  let oldX: false | number = false;
  let lastElement: HTMLElement | undefined = undefined;

  config_.onEnd = (evt) => {
    lastElement = undefined;
    oldX = false;
    oldY = false;

    if (userOnEnd) {
      // Undoes the dom structure of the list after a drag event. Ultimately this is needed to gain compatibility between sortablejs and solidjs
      if (evt.oldIndex != null && evt.newIndex != null) {
        const parent = evt.from;
        const oldNextSibling =
          parent.children[evt.oldIndex + (evt.oldIndex > evt.newIndex ? 1 : 0)];
        if (oldNextSibling) {
          parent.insertBefore(evt.item, oldNextSibling);
        } else {
          parent.append(evt.item);
        }
      }
      userOnEnd(evt);
    }
  };
  config_.onMove = function (evt: MoveEvent, origEvent: MouseEvent) {
    const posX = pos.x;
    const posY = pos.y;
    const yChangeSmall = oldY !== false && Math.abs(posY - oldY) < 10;
    const xChangeSmall = oldX !== false && Math.abs(posX - oldX) < 10;
    if (
      lastElement &&
      evt.related != lastElement &&
      yChangeSmall &&
      xChangeSmall
    ) {
      return false;
    }

    const result = userOnMove ? userOnMove(evt, origEvent) : true;
    if (result !== false) {
      lastElement = evt.related;
      oldY = posY;
      oldX = posX;
    }

    return result;
  } as any;

  const sortable = Sortable.create(el, config_);
  onCleanup(function () {
    // We can only delete the current sortable of there is no current sorting happening
    if (!Sortable.ghost) {
      sortable.destroy();
    } else {
      if (!Sortable.active) {
        return;
      }

      const active = Sortable.active;

      const onEnd = () => {
        Sortable.utils.off(active.el, "end", onEnd);
        requestAnimationFrame(function () {
          // console.log("delete sortable late", el);
          // Delete our sortable after the active on is done
          sortable.destroy();
        });
      };

      // Listen to the end event on the currently running sortable
      Sortable.utils.on(active.el, "end", onEnd);
    }
  });

  return sortable;
};
export const createOnEnd = (
  setter: (
    itemsState: (draft: string[]) => string[],
    evt: SortableEvent,
  ) => any,
  callback?: (evt: SortableEvent) => any,
) => {
  return (evt: SortableEvent) => {
    const oldIndex = evt.oldIndex;
    const newIndex = evt.newIndex;
    if (oldIndex == null || newIndex == null) {
      return;
    }

    setter((positions: any[]) => {
      let item = positions[oldIndex];
      positions.splice(oldIndex, 1);
      positions.splice(newIndex, 0, unwrap(item));
      return positions;
    }, evt);
    callback && callback(evt);
  };
};
declare module "solid-js" {
  namespace JSX {
    interface Directives {
      sortable: Sortable.Options;
    }
  }
}
