import { ServerError, createServerData$, redirect } from "solid-start/server";
import type { AuthSession } from "../../hooks/auth";
import {
  clearIronSession,
  getIronSession,
  IronSessionConfig,
  renewIronSession,
  setIronSession,
} from "../session";
import { withServer, withServerOptions } from "../util";
import { prisma } from "../../prisma";
import { createId } from "@paralleldrive/cuid2";
import { hashPassword, validatePassword } from "./util";

const authSession = {
  name: "nitropage-auth",
  password: process.env["NP_AUTH_IRON_PASSWORD"]!,
  validate: (v) => v as AuthSession,
  cookieRenewAge: parseInt(process.env["NP_AUTH_COOKIE_RENEWAGE"] || "0"),
  cookie: {
    path: "/",
    maxAge: parseInt(process.env["NP_AUTH_COOKIE_MAXAGE"] || "0"),
  },
} satisfies IronSessionConfig;

export const signup = async function (
  response: Response,
  args: { username: string; password: string; email: string },
) {
  // TODO: allow signup of more users by logged in admins
  // TODO: add a check to avoid creating duplicate users

  const users = await prisma.nitroUser.count();
  if (users > 0) {
    return false;
  }

  const passwordHashed = await hashPassword(args.password);

  const user = await prisma.nitroUser.create({
    data: {
      admin: true,
      username: args.username,
      email: args.email,
      password: passwordHashed,
    },
  });

  return await signin(response, user.username, args.password);
};

export const signin = async function (
  response: Response,
  username: string,
  password: string,
) {
  const user = await prisma.nitroUser.findFirst({
    where: {
      [username.includes("@") ? "email" : "username"]: username,
    },
  });

  if (!user) {
    return false;
  }

  const passwordCorrect = await validatePassword(password, user.password);
  if (!passwordCorrect) {
    return false;
  }

  return await setAuth(response, {
    id: user.id,
    user: user.username,
    csrf: createId(),
  });
};

export const getAuth = async function (request: Request) {
  return await getIronSession(authSession, request);
};

export const setAuth = async function (response: Response, value: AuthSession) {
  return await setIronSession(authSession, response, value);
};

export const clearAuth = async function (response: Response) {
  return await clearIronSession(authSession, response);
};

export const renewAuth = async function (request: Request, response: Response) {
  return await renewIronSession(authSession, request, response);
};

type RouteDataEvent = Parameters<
  Parameters<typeof createServerData$>["0"]
>["1"];
type RouteDataOptions = Parameters<typeof createServerData$>["1"];

// https://twitter.com/mattpocockuk/status/1622730173446557697
// https://twitter.com/MichaelArnaldi/status/1622920514694946816
type Prettify<T> = {
  [K in keyof T]: T[K];
} extends infer X
  ? X
  : never;

type Middleware<R, O> = <
  K,
  E extends RouteDataEvent,
  F extends (key: K, event: E) => Promise<{}>,
>(
  fn: F,
  options: O,
) => (key: K, event: E) => Promise<Prettify<Awaited<ReturnType<F>> & R>>;

type AuthOptions = {
  error?: "throw" | "redirect";
  csrf?: string;
};

/*
export const withAuth: Middleware<
  { auth: AuthSession | undefined },
  AuthOptions
> = (fn, options = {}) => {
  return async (key, event) => {
    const data = await authData(event, options);

    const origResult = await fn(key, event);
    const result = {
      ...origResult,
      ...data,
    } as any;

    return result;
  };
};
*/

export const createAuthRpc = withServerOptions(function (
  _,
  server,
  options: AuthOptions = {},
) {
  // Normally server.request couldn't be accessed here in this separate file
  return authData(server.request, options);
});

export const auth2$ = withServer(function (_, server) {
  return authData(server.request, { error: "redirect" });
});

const handleAuthFailOptions = function (options: AuthOptions) {
  if (options.error === "redirect") {
    throw redirect("/");
  }
  if (options.error === "throw") {
    throw new ServerError("Unauthorized", { status: 401 });
  }
};

export const authData = async (request: Request, options: AuthOptions = {}) => {
  const auth = await getAuth(request);
  if (!auth) {
    handleAuthFailOptions(options);
  }

  if (options.error && options.csrf != null && options.csrf != auth?.csrf) {
    handleAuthFailOptions(options);
  }

  if (!auth) {
    return;
  }

  const user = await prisma.nitroUser.findFirst({
    select: {
      id: true,
      email: true,
      createdAt: true,
      updatedAt: true,
      username: true,
    },
    where: {
      id: auth.id,
    },
  });

  if (!user) {
    handleAuthFailOptions(options);
    return;
  }

  return { ...user, csrf: auth.csrf, ironDate: auth.ironDate };
};

/*

export const justAuth$2 = function(options: AuthOptions = {}) {
  return function() {
    const ctx$$ = this;
    console.log('this', this)

    return authData(ctx$$.request, options)
  }
}

export const withAuth2 = function <
  T,
  S,
  F extends (arg1: T, arg2: S) => Promise<{}>
>(fn: F, options: { loggedIn?: boolean } = {}) {
  console.log("options", options);
  const fetcher = async (args: T, event: S) => {
    const origResult = await fn(args, event);
    const auth = await getAuth((event as any).request);

    if (options.loggedIn && !auth) {
      throw redirect("/");
    }

    const result = {
      ...origResult,
      auth,
    };

    type Result = Awaited<ReturnType<F>> & { auth: typeof auth };

    return result as Prettify<Result>;
  };

  return fetcher;
};
*/
