import { ResizeOffer } from "../types";
import type { CreateServerFunction } from "solid-start/server/server-functions/types";
import { prisma } from "../prisma";
import path from "path";
import { globby } from "globby";
import { rm } from "fs/promises";
import { keyHash } from "./hash";

const offerIdKeyMap: Record<string, string> = {};
const offersCache: Record<string, ResizeOffer> = {};

export const getResizeOffer = function (key?: string) {
  if (!key) {
    return;
  }

  return offersCache[key];
};

export const createResize = function <const T extends ResizeOffer>(options: T) {
  const offerId = JSON.stringify(options);
  const key = offerIdKeyMap[offerId] || keyHash(offerId).slice(0, 16);
  if (!offersCache[key]) {
    offerIdKeyMap[offerId] = key;
    offersCache[key] = options;
  }

  const result = {
    key,
    Offer: {
      sizes: options.sizes,
    },
  } as {
    key: string;
    Offer: T;
  };

  return async function (this: CreateServerFunction) {
    return result;
  };
};

export const mediaPaths = function () {
  const publicPath = path.join(process.cwd(), "public");
  const mediaPath = path.join(publicPath, "media");
  const minMediaPath = path.join(mediaPath, "min");
  const cachePath = path.join(mediaPath, "cache");

  return {
    public: publicPath,
    raw: mediaPath,
    min: minMediaPath,
    cache: cachePath,
  };
};

const deleteMediaCache = async function (rawUrl: string) {
  const { cache: cachePath } = mediaPaths();
  const fileExt = path.extname(rawUrl);
  const fileName = path.basename(rawUrl, fileExt);
  const globPath = path.join(cachePath, fileName + ".*.(webp|png)");
  const cacheFiles = await globby([globPath]);
  await Promise.all(
    cacheFiles.map((f) => {
      return rm(f);
    }),
  );
};

export const deleteMedia = async function (id: string) {
  const media = await prisma.nitroMedia.findFirstOrThrow({
    where: {
      id,
    },
  });

  const { public: publicPath } = mediaPaths();

  if (media.url !== media.rawUrl) {
    const minFilePath = path.join(publicPath, media.url);
    try {
      await rm(minFilePath);
    } catch (err) {
      // TODO: add log file
    }
  }

  if (media.url2x) {
    const minFilePath = path.join(publicPath, media.url2x);
    try {
      await rm(minFilePath);
    } catch (err) {
      // TODO: add log file
    }
  }

  try {
    await deleteMediaCache(media.rawUrl);
    const rawFilePath = path.join(publicPath, media.rawUrl);
    await rm(rawFilePath);
  } catch (err) {
    // TODO: add log file
  }

  return await prisma.nitroMedia.delete({
    where: {
      id,
    },
  });
};
