import { useContext } from "solid-js";
import { createStore } from "solid-js/store";
import { baseFilename, ConfigFn } from "../config";
import { prisma } from "../prisma";
import { normalizeState } from "../util/normalization";
import { getElementData } from "../util/elementData";
import { getProjectFromHostname, withServerOptions } from "./util";
import { getSettings, notFoundPage } from "./settings";
import { NitroProject } from "@prisma/client";
import { ConfigContext } from "../components/context/page";

const getPublishedRevision = async function ({
  projectId,
  pageId,
  urlPath,
}: {
  projectId?: string;
  pageId?: string;
  urlPath?: string;
}) {
  return await prisma.nitroPageRevision.findFirst({
    where: {
      publishedPage: { id: { not: undefined } },
      urlPath: urlPath,
      page: pageId ? { id: pageId } : { projectId },
    },
    include: {
      publishedPage: true,
      elements: {
        include: {
          slots: {
            include: {
              elements: {
                orderBy: {
                  position: "asc",
                },
              },
            },
          },
          parentSlot: {
            include: {
              parentElement: true,
            },
          },
        },
      },
      rootSlots: {
        include: {
          elements: {
            orderBy: {
              position: "asc",
            },
          },
        },
      },
    },
  });
};

export const pageRoute = async function ({
  urlPath,
  project,
}: {
  urlPath: string;
  project: NitroProject;
}) {
  urlPath = "/" + urlPath;

  let pageRevision = await getPublishedRevision({
    urlPath,
    projectId: project.id,
  });

  if (!pageRevision) {
    const notFoundPage_ = await notFoundPage({
      projectId: project?.id,
    });
    if (notFoundPage_.value) {
      pageRevision = await getPublishedRevision({
        projectId: project?.id,
        pageId: notFoundPage_.value,
      });
    }
  }

  if (!pageRevision) {
    return;
  }

  const settings = await getSettings({
    projectId: project?.id,
    pageId: pageRevision.pageId,
  });

  const presets = await prisma.nitroPreset.findMany({
    where: {
      id: {
        in: pageRevision.elements.reduce((acc, el) => {
          if (!el.presetId) return acc;

          acc.push(el.presetId);

          return acc;
        }, [] as string[]),
      },
    },
  });

  const fonts = await prisma.nitroFont.findMany({
    where: {
      projectId: project.id,
    },
    include: {
      faces: true,
    },
  });

  return normalizeState({
    pageRevision,
    fonts,
    presets,
    settings,
    passiveBlueprints,
  });
};

let blueprintFiles: Record<string, string> = {};
let passiveBlueprints: Record<string, boolean> = {};
let firstLoad = true;

export const createPageRpc = withServerOptions(
  async (args: { urlPath: string }, server, options: { config: ConfigFn }) => {
    const configCtx = useContext(ConfigContext)!;

    const hostname = new URL(server.request.url).hostname;

    // TODO: Remove the fallback after further testing (nitroProject should always be set)
    const project: NitroProject | undefined =
      (server.locals.nitroProject as any) ||
      (await getProjectFromHostname(hostname));
    if (!project) {
      return;
    }

    const nitropageConfig = options.config;
    const updateBlueprintFiles = firstLoad || !import.meta.env.PROD;
    if (updateBlueprintFiles) {
      for (const [filePath, loader] of Object.entries(
        nitropageConfig._getBlueprints(),
      )) {
        const module = await loader();
        const id = module.id || baseFilename(filePath).toLowerCase();
        blueprintFiles[id] = filePath;
        if (module.passive) {
          passiveBlueprints[id] = true;
        }
      }

      firstLoad = false;
    }

    const state = await pageRoute({ urlPath: args.urlPath, project });
    if (!state) {
      return;
    }

    const [config, setConfig] = configCtx || createStore({ blueprints: {} });

    await nitropageConfig({
      setConfig,
      config,
      admin: !!state.admin,
      page: {
        blueprintFiles,
        passiveBlueprints,
        state,
      },
    });

    for (const element of Object.values(state.elements)) {
      const blueprint = config.blueprints[element.blueprint];
      if (!blueprint) {
        continue;
      }

      for (const dataKey of Object.keys(blueprint.data)) {
        element.ssrData[dataKey] = getElementData(
          element.id,
          state,
          config,
          dataKey,
        );
        delete element.data[dataKey];
      }
    }

    return { state, blueprintFiles };
  },
);
