import { mkdir, open, rm } from "fs/promises";
import { basename, extname, join } from "path";
import server$, { ServerError } from "solid-start/server";
import { Readable } from "stream";
import { prisma } from "../../prisma";
import { Font, FontFaceFile } from "../../types";
import { authData, pageEvent } from "nitropage/server";
import { deleteFont, getFontsPath } from "../font";
import { createId } from "@paralleldrive/cuid2";
import { demoEnabled } from "../../util";
import { fileTypeFromBlob } from "file-type";

export const getFonts = server$(
  async ({ projectId }: { projectId: string }) => {
    const event = pageEvent(this);
    await authData(event.request, { error: "throw" });

    return await prisma.nitroFont.findMany({
      where: {
        projectId,
      },
      orderBy: {
        createdAt: "desc",
      },
      include: {
        faces: {
          orderBy: [
            {
              weight: "asc",
            },
            {
              style: "asc",
            },
          ],
        },
      },
    });
  },
);

export const createFontRpc = server$(
  async ({ csrf, projectId }: { csrf: string; projectId: string }) => {
    await authData(server$.request, { csrf: csrf || "", error: "throw" });

    let publicId = createId();
    const existingFont = await prisma.nitroMedia.findUnique({
      where: {
        publicId,
      },
    });
    // If the publicId already exists, we try another one
    // TODO: Improve this (e.g. by trying multiple ones)
    if (existingFont) {
      publicId = createId();
    }

    const project = await prisma.nitroProject.findUniqueOrThrow({
      where: {
        id: projectId,
      },
    });

    return await prisma.nitroFont.create({
      data: {
        project: project
          ? {
              connect: {
                id: project.id,
              },
            }
          : undefined,
        publicId,
      },
    });
  },
);

export const deleteFontRpc = server$(
  async (args: { csrf: string; id: string }) => {
    await authData(server$.request, { csrf: args.csrf || "", error: "throw" });
    if (demoEnabled()) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    await deleteFont(args.id);
  },
);

export const createFontFace = server$(
  async (args: { csrf: string; id: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== args.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    return await prisma.nitroFontFace.create({
      data: {
        font: {
          connect: {
            id: args.id,
          },
        },
      },
    });
  },
);

export const deleteFontFace = server$(
  async (args: { csrf: string; id: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== args.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    return await prisma.nitroFontFace.delete({
      where: {
        id: args.id,
      },
    });
  },
);

export const updateFontsRpc = server$(
  async (args: { csrf: string; fonts: Font[] }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== args.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    for (const font of args.fonts) {
      const nitroFont = await prisma.nitroFont.findUniqueOrThrow({
        where: { id: font.id },
      });
      await prisma.nitroFont.update({
        data: {
          fallback: font.fallbacks.join(","),
          preload: font.preload,
          family: font.family,
        },
        where: {
          id: font.id,
        },
      });
      for (const face of font.faces) {
        await prisma.nitroFontFace.update({
          data: {
            files: JSON.stringify(face.files),
            preload: face.preload,
            style: face.style,
            weight: face.weight,
          },
          where: {
            id: face.id,
          },
        });
      }
    }
  },
);

const weightMappings: { weight: number; name: string }[] = [
  { weight: 100, name: "thin" },
  { weight: 100, name: "hairline" },
  { weight: 200, name: "extra light" },
  { weight: 300, name: "light" },
  { weight: 400, name: "normal" },
  { weight: 400, name: "regular" },
  { weight: 500, name: "medium" },
  { weight: 500, name: "dark" },
  { weight: 600, name: "semi bold" },
  { weight: 600, name: "demi bold" },
  { weight: 800, name: "extra bold" },
  { weight: 800, name: "ultra bold" },
  { weight: 700, name: "bold" },
  { weight: 900, name: "heavy" },
  { weight: 950, name: "extra black" },
  { weight: 950, name: "ultra black" },
  { weight: 900, name: "black" },
  { weight: 950, name: "super" },
];

const extractFontMeta = function (name: string) {
  const lowerName = name.toLowerCase();
  let weight = 400;
  let style = "";

  for (const mapping of weightMappings) {
    let mappingFound = lowerName.includes(mapping.weight + "");

    if (!mappingFound) {
      const variants = [
        mapping.name,
        mapping.name.replace(" ", "-"),
        mapping.name.replace(" ", "_"),
        mapping.name.replace(" ", ""),
      ];
      for (const variant of variants) {
        if (!lowerName.includes(variant)) {
          continue;
        }

        mappingFound = true;
        break;
      }
    }

    if (!mappingFound) {
      continue;
    }

    weight = mapping.weight;
    break;
  }

  if (
    lowerName.includes("slanted") ||
    lowerName.includes("italic") ||
    lowerName.includes("oblique")
  ) {
    style = "italic";
  }

  return { weight, style };
};

export const uploadFontsRpc = server$(async (form: FormData) => {
  const csrf = form.get("csrf");
  const user = await authData(server$.request);
  if (!user || user.csrf !== csrf) {
    throw new ServerError("Unauthorized", { status: 401 });
  }

  const fontId: string = form.get("id")?.toString() || "";
  const font = await prisma.nitroFont.findUniqueOrThrow({
    include: {
      faces: true,
    },
    where: {
      id: fontId,
    },
  });

  const faces: Record<
    string,
    {
      id?: string;
      weight: any;
      style: any;
      files: FontFaceFile[];
    }
  > = {};

  const faceKey = function (weight = 400, style = "") {
    return `w${weight}--s${style}`;
  };

  for (const face of font.faces) {
    const key = faceKey(face.weight!, face.style);
    faces[key] = {
      id: face.id,
      weight: face.weight,
      style: face.style,
      files: JSON.parse(face.files),
    };
  }

  const fontsPath = join(getFontsPath(), font.publicId);
  await mkdir(fontsPath, { recursive: true });

  const promises: Promise<any>[] = [];
  for (const file of form.getAll("files")) {
    const blob = file as unknown as File;
    const fileType = await fileTypeFromBlob(blob);

    const fileExt = extname(blob.name);
    const format = {
      "font/woff2": "woff2",
      "font/woff": "woff",
      "font/otf": "opentype",
      "font/ttf": "truetype",
    }[fileType?.mime as string];
    if (!fileType || !format) {
      console.warn(`Unsupported extension ${fileExt}`);
      continue;
    }

    const name = basename(blob.name, fileExt);
    const meta = extractFontMeta(name);
    const key = faceKey(meta.weight, meta.style);

    const face = (faces[key] = faces[key] || {
      weight: meta.weight,
      style: meta.style,
      files: [],
    });
    const fileName = `${name}${fileExt}`;
    const existingFile = face.files.find(function (f) {
      return f.name === fileName;
    });

    if (!existingFile) {
      face.files.push({ format, name: fileName, mime: fileType.mime });
    }

    const filePath = join(fontsPath, `${name}${fileExt}`);
    const fd = await open(filePath, "w");
    const fw = fd.createWriteStream({ autoClose: true });
    const fr = Readable.fromWeb(blob.stream() as any);
    const uploadStream = fr.pipe(fw);
    promises.push(
      (async () => {
        await new Promise((res) =>
          uploadStream.on("close", () => {
            res(true);
          }),
        );
      })(),
    );
  }

  for (const face of Object.values(faces)) {
    if (!face.id) {
      await prisma.nitroFontFace.create({
        data: {
          style: face.style,
          weight: face.weight,
          files: JSON.stringify(face.files),
          font: {
            connect: {
              id: fontId,
            },
          },
        },
      });
    } else {
      await prisma.nitroFontFace.update({
        data: {
          files: JSON.stringify(face.files),
        },
        where: {
          id: face.id,
        },
      });
    }
  }

  await Promise.all(promises);
});
