import { ElementDataRecord, State } from "../../types";
import server$, { ServerError } from "solid-start/server";
import { prisma } from "../../prisma";
import { authData, pageEvent } from "nitropage/server";
import { NitroElement, NitroElementSlot } from "@prisma/client";
import {
  Settings,
  getSettings as getSettings__,
  notFoundPage,
  setBlueprintDefaults,
  setColors,
  setDefaultFont,
  setFavicon,
  setMetaDescription,
  setMetaImage,
  setNotFoundPage,
  setPageBackgroundColor,
  setRteTextColors,
  setSettings,
} from "../settings";
import { demoEnabled } from "../../util";
import { deletePage, deletePageRevision } from "../page";

export { uploadFilesRpc, optimizeImage, getPublicMedia } from "./media";

export * from "./fonts";

export const getSettings = server$(async function (
  this: any,
  params: { projectId?: string; pageId?: string; inherit?: boolean } = {
    inherit: true,
  },
) {
  const event = pageEvent(this);
  await authData(server$.request || event.request, { error: "throw" });

  const page = params.pageId
    ? await prisma.nitroPage.findUniqueOrThrow({
        where: {
          id: params.pageId,
        },
      })
    : undefined;

  const projectId =
    params.projectId || (params.inherit ? page?.projectId : undefined);

  return await getSettings__({
    inherit: params.inherit,
    projectId,
    pageId: params.pageId,
  });
});

export const getRevision = server$(async (revisionId: string) => {
  if (!revisionId) {
    return;
  }

  return await prisma.nitroPageRevision.findUnique({
    where: {
      id: revisionId,
    },
    include: {
      elements: {
        include: {
          slots: {
            include: {
              elements: {
                orderBy: {
                  position: "asc",
                },
              },
            },
          },
          parentSlot: {
            include: {
              parentElement: true,
            },
          },
        },
      },
      rootSlots: {
        include: {
          elements: {
            orderBy: {
              position: "asc",
            },
          },
        },
      },
    },
  });
});

const serializeElementData = function (elementData: ElementDataRecord) {
  const filteredData: ElementDataRecord = {};
  for (const [key, data] of Object.entries(elementData)) {
    if (!data.override) {
      continue;
    }

    filteredData[key] = data;
  }

  return JSON.stringify(filteredData);
};

const savePresets = async ({
  projectId,
  presets,
}: {
  projectId: string;
  presets: State["presets"];
}) => {
  const updatedIds: string[] = [];
  const idMap: Record<string, string> = {};
  const deletedIds: string[] = [];

  for (const preset of Object.values(presets)) {
    if (preset.new) {
      const dbPreset = await prisma.nitroPreset.create({
        data: {
          name: preset.name,
          data: serializeElementData(preset.data),
          blueprint: preset.blueprint,
          project: {
            connect: {
              id: projectId,
            },
          },
        },
      });
      idMap[preset.id] = dbPreset.id;
      updatedIds.push(dbPreset.id);
    } else {
      const { count } = await prisma.nitroPreset.updateMany({
        where: {
          id: preset.id,
        },
        data: {
          name: preset.name,
          data: serializeElementData(preset.data),
        },
      });
      (count === 0 ? deletedIds : updatedIds).push(preset.id);
    }
  }

  return {
    deletedIds,
    updatedIds,
    idMap,
  };
};

export const savePage = server$(
  async (params: {
    csrf: string;
    publish?: boolean;
    pageRevisionId?: string;
    state: State;
    settings: {
      pages: Partial<Settings>;
      projects: Partial<Settings>;
    };
    pageId: string;
  }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== params.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    const page = await prisma.nitroPage.findFirst({
      where: {
        id: params.pageId,
      },
    });
    if (!page) {
      console.error("Page / PageRevision not found");
      return;
    }

    await prisma.nitroPage.update({
      where: {
        id: params.pageId,
      },
      data: {
        includeInFeed: params.state.includeInFeed,
      },
    });

    if (params.state.defaultFont) {
      const font = await prisma.nitroFont.findUnique({
        where: {
          publicId: params.state.defaultFont,
        },
      });

      await setDefaultFont(font?.id, {
        projectId: page.projectId,
      });
    } else {
      await setDefaultFont(undefined, {
        projectId: page.projectId,
      });
    }

    await setMetaDescription(
      params.state.metaDescription !== ""
        ? params.state.metaDescription
        : undefined,
      {
        pageId: page.id,
      },
    );

    await setMetaImage(params.settings.pages.metaImage?.value, {
      pageId: page.id,
    });
    await setMetaImage(params.settings.projects.metaImage?.value, {
      projectId: page.projectId,
    });

    await setFavicon(params.settings.projects.favicon?.value, {
      projectId: page.projectId,
    });

    await setColors(params.state.colors, {
      projectId: page.projectId,
    });

    await setPageBackgroundColor(params.state.backgroundColor, {
      projectId: page.projectId,
    });

    await setRteTextColors(
      Object.keys(params.state.richtext.textColors).length
        ? params.state.richtext.textColors
        : undefined,
      {
        projectId: page.projectId,
      },
    );

    let pageRevision = (await prisma.nitroPageRevision.findFirst({
      where: {
        page: { id: params.pageId },
      },
      orderBy: {
        savedAt: "desc",
      },
    }))!;

    if (!pageRevision) {
      return;
    }

    let createNewRevision =
      page.publishedRevisionId === pageRevision.id ||
      pageRevision.id !== params.pageRevisionId;

    if (params.pageRevisionId) {
      pageRevision = (await prisma.nitroPageRevision.findFirst({
        where: {
          id: params.pageRevisionId,
          page: { id: params.pageId },
        },
      }))!;
    }

    if (createNewRevision) {
      pageRevision = await prisma.nitroPageRevision.create({
        data: {
          title: pageRevision.title,
          urlPath: pageRevision.urlPath,
          page: {
            connect: {
              id: params.pageId,
            },
          },
        },
      });
    }

    /*
    console.log("revision", {
      pageRevision,
      createNewRevision,
      publishedId: page.publishedRevisionId,
    });
    */

    const { idMap: presetIdMap, deletedIds: deletedPresets } =
      await savePresets({
        projectId: page.projectId,
        presets: params.state.presets,
      });

    const elementIdMap: Record<string, string> = {};
    const updatedIds: string[] = [];

    const recursiveElementInsert = async function (
      elementIds: string[],
      slotId?: string,
    ) {
      let position = 0;
      for (const idx of elementIds) {
        const element = params.state.elements[idx];
        const isNewElement = element.elementId === "";

        if (element.presetId && presetIdMap[element.presetId]) {
          element.presetId = presetIdMap[element.presetId];
        }

        if (element.presetId && deletedPresets.includes(element.presetId)) {
          element.presetId = undefined;
        }

        if (isNewElement || createNewRevision) {
          const oldElement = !isNewElement
            ? await prisma.nitroElement.findFirst({
                where: {
                  id: element.elementId,
                },
              })
            : null;
          const newElement = await prisma.nitroElement.create({
            data: {
              position: 0,
              data: "{}",
              blueprint: element.blueprint,
              history: {
                create: !oldElement?.historyId
                  ? {
                      page: { connect: { id: params.pageId } },
                    }
                  : undefined,
                connect: oldElement?.historyId
                  ? { id: oldElement?.historyId }
                  : undefined,
              },
            },
          });

          elementIdMap[element.id] = newElement.historyId;
          element.elementId = newElement.id;
        }

        const elementId = element.elementId;

        const dbSlots = await prisma.nitroElementSlot.findMany({
          where: {
            parentElementId: elementId,
          },
        });
        const slots = dbSlots.reduce(
          (acc, slot) => {
            acc[slot.name] = slot;
            return acc;
          },
          {} as Record<string, NitroElementSlot>,
        );

        for (const [name, slot] of Object.entries(element.slots)) {
          if (!slot.elements.length) {
            continue;
          }
          let dbSlot = slots[name];
          if (!dbSlot) {
            dbSlot = await prisma.nitroElementSlot.create({
              data: {
                name,
                parentElement: {
                  connect: {
                    id: elementId,
                  },
                },
              },
            });
          }

          await recursiveElementInsert(slot.elements, dbSlot.id);
        }

        const dbElement = await prisma.nitroElement.update({
          where: {
            id: elementId,
          },
          data: {
            position,
            parentSlot: {
              connect: slotId
                ? {
                    id: slotId,
                  }
                : undefined,
              disconnect: !slotId ? true : undefined,
            },
            preset: {
              connect: element.presetId
                ? {
                    id: element.presetId,
                  }
                : undefined,
              disconnect: !element.presetId ? true : undefined,
            },
            page: {
              connect: {
                id: pageRevision.id,
              },
            },
            data: serializeElementData(element.data),
          },
        });

        updatedIds.push(dbElement.id);

        position++;
      }
    };

    if (params.state.pendingDeletes) {
      await prisma.nitroPreset.deleteMany({
        where: {
          id: { in: params.state.pendingDeletes.presets },
        },
      });
    }

    for (const slotName of ["header", "default", "footer"]) {
      let slot = await prisma.nitroElementSlot.findFirst({
        where: {
          parentPageRevision: {
            id: pageRevision.id,
          },
          name: slotName,
        },
      });
      if (!slot) {
        slot = await prisma.nitroElementSlot.create({
          data: {
            name: slotName,
            parentPageRevision: {
              connect: {
                id: pageRevision.id,
              },
            },
          },
        });
      }
      await recursiveElementInsert(
        params.state.root[slotName as "default"],
        slot.id,
      );
    }

    const obsoleteElements = await prisma.nitroElement.findMany({
      select: {
        id: true,
      },
      where: {
        id: { notIn: updatedIds },
        page: { id: pageRevision.id },
      },
    });

    const deleteIds = obsoleteElements.map((el) => el.id);

    await prisma.nitroElementSlot.deleteMany({
      where: {
        parentElement: {
          id: { in: deleteIds },
        },
      },
    });

    await prisma.nitroElement.deleteMany({
      where: {
        id: { in: deleteIds },
      },
    });

    await setBlueprintDefaults(params.state.blueprintDefaults.project, {
      projectId: page.projectId,
    });

    await prisma.nitroPageRevision.update({
      where: {
        id: pageRevision.id,
      },
      data: {
        title: params.state.title,
        urlPath: params.state.urlPath,
        savedAt: new Date(),
        // TODO: Remove page blueprint defaults
        // blueprintDefaults: JSON.stringify(params.state.blueprintDefaults.page),
        publishedPage: params.publish
          ? {
              connect: {
                id: params.pageId,
              },
            }
          : undefined,
      },
    });

    const notFoundPage_ = await notFoundPage({
      projectId: page.projectId,
    });
    if (params.state.notFound === true && notFoundPage_.value !== page.id) {
      await setNotFoundPage(page.id, {
        projectId: page.projectId,
      });
    }
    if (params.state.notFound === false && notFoundPage_.value === page.id) {
      await setNotFoundPage(undefined, {
        projectId: page.projectId,
      });
    }

    return {
      elementIdMap,
      newRevisionId: createNewRevision && pageRevision.id,
    };
  },
);

export const duplicatePageRevision = async function ({
  id,
  targetPageId,
}: {
  id: string;
  targetPageId: string;
}) {
  const sourceRevision = await prisma.nitroPageRevision.findUniqueOrThrow({
    include: {
      elements: {
        include: {
          history: true,
          slots: true,
        },
      },
      rootSlots: {
        include: {
          elements: true,
        },
      },
    },
    where: {
      id,
    },
  });

  const targetRevision = await prisma.nitroPageRevision.create({
    data: {
      blueprintDefaults: sourceRevision.blueprintDefaults,
      page: {
        connect: {
          id: targetPageId,
        },
      },
      title: sourceRevision.title,
      urlPath: sourceRevision.urlPath,
      savedAt: sourceRevision.savedAt,
    },
  });

  let newElementHistories: Record<string, string> = {};

  const duplicateElementRecursive = async function (
    sourceElement: NitroElement,
    parentSlot: NitroElementSlot,
  ) {
    let targetHistoryId = newElementHistories[sourceElement.historyId];
    if (!targetHistoryId) {
      const newHistory = await prisma.nitroElementHistory.create({
        data: {
          page: {
            connect: {
              id: targetPageId,
            },
          },
        },
      });
      newElementHistories[sourceElement.historyId] = newHistory.id;
      targetHistoryId = newHistory.id;
    }

    const targetElement = await prisma.nitroElement.create({
      data: {
        blueprint: sourceElement.blueprint,
        data: sourceElement.data,
        preset:
          sourceElement.presetId != null
            ? {
                connect: {
                  id: sourceElement.presetId,
                },
              }
            : undefined,
        position: sourceElement.position,
        parentSlot: {
          connect: {
            id: parentSlot.id,
          },
        },
        page:
          sourceElement.pageId != null
            ? {
                connect: {
                  id: targetRevision.id,
                },
              }
            : undefined,

        history: {
          connect: {
            id: targetHistoryId,
          },
        },
      },
    });

    const slots = await prisma.nitroElementSlot.findMany({
      include: {
        elements: true,
      },
      where: {
        parentElement: {
          id: sourceElement.id,
        },
      },
    });

    for (const sourceSlot of slots) {
      await duplicateSlotRecursive(sourceSlot, targetElement);
    }
  };

  const duplicateSlotRecursive = async function (
    sourceSlot: NitroElementSlot & {
      elements: NitroElement[];
    },
    parentElement?: NitroElement,
  ) {
    const targetSlot = await prisma.nitroElementSlot.create({
      data: {
        name: sourceSlot.name,
        parentElement: parentElement
          ? {
              connect: {
                id: parentElement.id,
              },
            }
          : undefined,
        parentPageRevision: !parentElement
          ? {
              connect: {
                id: targetRevision.id,
              },
            }
          : undefined,
      },
    });

    for (const element of sourceSlot.elements) {
      await duplicateElementRecursive(element, targetSlot);
    }
  };

  for (const sourceSlot of sourceRevision.rootSlots) {
    await duplicateSlotRecursive(sourceSlot);
  }
};

export const duplicatePage = server$(
  async (params: { csrf: string; id: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== params.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    await newPageAllowed();

    const sourcePage = await prisma.nitroPage.findUniqueOrThrow({
      include: {
        revisions: {
          take: 10,
          orderBy: {
            savedAt: "desc",
          },
        },
      },
      where: {
        id: params.id,
      },
    });

    const targetPage = await prisma.nitroPage.create({
      data: {
        includeInFeed: sourcePage.includeInFeed,
        project: {
          connect: {
            id: sourcePage.projectId,
          },
        },
      },
    });

    if (
      sourcePage.publishedRevisionId &&
      !sourcePage.revisions.find(
        (rev) => rev.id === sourcePage.publishedRevisionId,
      )
    ) {
      await duplicatePageRevision({
        id: sourcePage.publishedRevisionId,
        targetPageId: targetPage.id,
      });
    }

    for (const rev of sourcePage.revisions) {
      await duplicatePageRevision({ id: rev.id, targetPageId: targetPage.id });
    }

    const latestRevision = await prisma.nitroPageRevision.findFirst({
      where: {
        page: {
          id: targetPage.id,
        },
      },
      orderBy: {
        savedAt: "desc",
      },
    });
    if (latestRevision) {
      await prisma.nitroPageRevision.update({
        data: {
          title: latestRevision.title + " Copy",
        },
        where: {
          id: latestRevision.id,
        },
      });
    }

    const sourceSettings = await getSettings__({
      pageId: params.id,
      inherit: false,
    });

    type Vv = Parameters<typeof setSettings>[0];

    const newSettings = Object.entries(sourceSettings).reduce(
      (acc, [name, value]: any[]) => {
        acc[name as keyof Vv] =
          value.from === "pages" ? value.value : undefined;
        return acc;
      },
      {} as Vv,
    );

    await setSettings(newSettings, {
      pageId: targetPage.id,
    });
  },
);

export const deletePageRpc = server$(
  async (params: { csrf: string; id: string }) => {
    await authData(server$.request, {
      error: "throw",
      csrf: params.csrf || "",
    });
    await deletePage(params.id);
  },
);

export const deleteRevisionRpc = server$(
  async (params: { csrf: string; id: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== params.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    return deletePageRevision(params.id);
  },
);

export const unpublishPageRpc = server$(
  async (params: { csrf: string; id: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== params.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    await prisma.nitroPage.update({
      where: {
        id: params.id,
      },
      data: {
        publishedRevision: {
          disconnect: true,
        },
      },
    });
  },
);

export const newPageAllowed = async function () {
  if (!demoEnabled()) {
    return;
  }

  const pages = await prisma.nitroPage.count();
  if (pages <= 20) {
    return;
  }

  throw new ServerError("Unauthorized", { status: 401 });
};
