import server$, { ServerError } from "solid-start/server";
import {
  authData,
  prisma,
  validatePassword,
  hashPassword,
} from "nitropage/server";
import path from "path";
import { File, transcode } from "buffer";
import { access, FileHandle, mkdir, open } from "fs/promises";
import { Readable } from "stream";
import { createWriteStream } from "fs";
import sharp from "sharp";
import { decode, encode } from "blurhash";
import { z } from "zod";
import { NitroMedia } from "@prisma/client";
import { ImageSize, PositionSchema, SizeSchema } from "../../zodTypes";
import { klona } from "klona";
import { fileTypeFromStream } from "file-type";
import { createId } from "@paralleldrive/cuid2";
import { rgbVal } from "../../components/page/colors";
import { demoEnabled } from "../../util";
import { deleteMedia, getResizeOffer, mediaPaths } from "../media";
import { calculateAspectRatio, fileType } from "../../util/media";
import { fastHash } from "../hash";

const optimizeKeyContent = function (url: string, position: string) {
  return url + position;
};

export const setMediaResizePosition = server$(
  async (args: { id: string; resizePosition?: string; csrf: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== args.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    const media = await prisma.nitroMedia.findFirstOrThrow({
      where: { id: args.id },
    });

    const optimizeKey = args.resizePosition
      ? await hashPassword(
          optimizeKeyContent(media.rawUrl, args.resizePosition),
        )
      : null;
    const resizePosition = args.resizePosition ? args.resizePosition : null;

    await prisma.nitroMedia.update({
      where: { id: media.id },
      data: {
        resizePosition,
        optimizeKey,
      },
    });
  },
);

export const deleteMediaRpc = server$(
  async (args: { id: string; csrf: string }) => {
    const user = await authData(server$.request);
    if (!user || user.csrf !== args.csrf) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    if (demoEnabled()) {
      throw new ServerError("Unauthorized", { status: 401 });
    }

    return await deleteMedia(args.id);
  },
);

export const getAllMedia = server$(
  async ({ projectId }: { projectId: string }) => {
    await authData(server$.request, { error: "throw" });

    return await prisma.nitroMedia.findMany({
      where: {
        projectId,
      },
    });
  },
);

export const getPublicMedia = server$(
  async (args: {
    ids: string[];
    includeDataUrl?: boolean;
    admin?: boolean;
  }) => {
    if (args.admin) {
      await authData(server$.request, { error: "throw" });
    }
    const ids = args.ids.filter((id) => id != null);
    return prisma.nitroMedia.findMany({
      select: args.admin
        ? undefined
        : {
            id: true,
            publicId: true,
            url: true,
            url2x: true,
            dataUrl: args.includeDataUrl,
            rawUrl: true,
            mime: true,
            width: true,
            height: true,
            blurhash: true,
            color: true,
            description: true,
            resizePosition: true,
            optimizeKey: true,
          },
      where: {
        id: { in: ids },
      },
    });
  },
);

const minQuality = 90;

/**
 * Source: https://github.com/woltapp/blurhash/issues/43#issuecomment-759112713
 */
const generateBlurhashDataUrl = async (
  hash: string,
  width: number,
  height: number,
  options = {
    size: 10,
    quality: 25,
  },
) => {
  const hashWidth = options?.size;
  const hashHeight = Math.round(hashWidth * (height / width));

  const pixels = decode(hash, hashWidth, hashHeight);

  const resizedImageBuf = await sharp(Buffer.from(pixels), {
    raw: {
      channels: 4,
      width: hashWidth,
      height: hashHeight,
    },
  })
    .modulate({
      saturation: 1.2,
    })
    .jpeg({
      overshootDeringing: true,
      quality: options.quality,
    })
    .toBuffer();

  return `data:image/jpeg;base64,${resizedImageBuf.toString("base64")}`;
};

/**
 * Shrinks the crop target size, fitting it proportionally inside the original size
 *
 * E.g. if the original size is 300/300 and target is 400/300, this returns 300/225.
 * With target 300/400, this returns 225/300.
 *
 * In either case, the new target size is never larger than the original size
 *
 */
const getProportionalCropSize = function (
  originalWidth: number,
  originalHeight: number,
  targetWidth: number,
  targetHeight: number,
) {
  if (targetWidth > originalWidth) {
    targetHeight = Math.round((targetHeight / targetWidth) * originalWidth);
    targetWidth = originalWidth;
  }

  if (targetHeight > originalHeight) {
    targetWidth = Math.round((targetWidth / targetHeight) * originalHeight);
    targetHeight = originalHeight;
  }

  return {
    width: targetWidth,
    height: targetHeight,
  };
};

const toUtf8 = function (latinText: string) {
  return transcode(Buffer.from(latinText), "utf8", "latin1").toString("utf8");
};

const uploadFile = async function (blob: File) {
  let fd: FileHandle;

  const type = fileType(blob.type);
  const isConstant = blob.type === "image/svg+xml" || type !== "image";
  const fr = Readable.fromWeb(blob.stream());

  let publicId = createId();
  const existingFile = await prisma.nitroMedia.findUnique({
    where: {
      publicId,
    },
  });
  // If the publicId already exists, we try another one
  // TODO: Improve this (e.g. by trying multiple ones)
  if (existingFile) {
    publicId = createId();
  }

  const { min: minMediaPath, raw: mediaPath } = mediaPaths();

  const blobName = toUtf8(blob.name);
  const fileExt = path.extname(blobName);
  const originalName = path.basename(blobName, fileExt);
  const fileName = publicId;
  const filePath = path.join(mediaPath, fileName + fileExt);
  const minFilePath = path.join(minMediaPath, fileName + ".webp");
  const min2xFilePath = path.join(minMediaPath, fileName + ".2x.webp");

  const rawUrl = `/media/${fileName + fileExt}`;
  const minUrl = `/media/min/${fileName}.webp`;
  const min2xUrl = `/media/min/${fileName}.2x.webp`;

  fd = await open(filePath, "w");
  const fw = fd.createWriteStream({ autoClose: true });
  const uploadStream = fr.pipe(fw);

  const sharpPromise = (async () => {
    if (type !== "image") {
      return;
    }

    try {
      const content = new Uint8ClampedArray(await blob.arrayBuffer());
      const sharpStream = sharp(content).rotate().raw().ensureAlpha();
      const resized = await sharpStream
        .resize(32)
        .toBuffer({ resolveWithObject: true });

      const xComponent = 4;
      const yComponent = Math.round(
        (xComponent / resized.info.width) * resized.info.height,
      );

      const metadata = await sharpStream.metadata();
      const stats = await sharpStream.stats();
      const blurhash = encode(
        new Uint8ClampedArray(resized.data),
        resized.info.width!,
        resized.info.height!,
        xComponent,
        yComponent,
      );

      const base64 = await generateBlurhashDataUrl(
        blurhash,
        resized.info.width,
        resized.info.height,
      );

      return {
        blurhash,
        base64,
        width: metadata.width,
        height: metadata.height,
        color: stats.dominant,
      };
    } catch (err) {
      console.error(err);
      if (isConstant) {
        console.error("Media upload error: failed to parse svg data!");
      }
    }
  })();
  const uploadPromise = (async () => {
    await new Promise((res) =>
      uploadStream.on("close", () => {
        res(true);
      }),
    );

    if (isConstant) {
      return;
    }

    // rotate() makes sure that the original rotation stays as is
    let sharpBase = await sharp(filePath).rotate();

    await sharpBase
      .resize({
        width: 1920,
        withoutEnlargement: true,
      })
      .webp({ quality: minQuality })
      .toFile(minFilePath);

    await sharpBase
      .resize({
        width: 1920 * 2,
        withoutEnlargement: true,
      })
      .webp({ quality: minQuality })
      .toFile(min2xFilePath);
  })();
  const promises = [sharpPromise, uploadPromise] as [typeof sharpPromise, any];
  const results = await Promise.all(promises);
  const sharpResults = results[0];

  await fd.close();

  return {
    publicId: publicId,
    mime: blob.type,
    name: originalName,
    extension: fileExt.slice(1),
    file: fileName + fileExt,
    rawUrl: rawUrl,
    url: isConstant ? rawUrl : minUrl,
    url2x: isConstant ? undefined : min2xUrl,
    dataUrl: sharpResults?.base64,
    blurhash: sharpResults?.blurhash,
    color: sharpResults?.color ? rgbVal(sharpResults?.color) : undefined,
    width: sharpResults?.width,
    height: sharpResults?.height,
  } satisfies Partial<NitroMedia>;
};

export const uploadFilesRpc = server$(async (form: FormData) => {
  const csrf = form.get("csrf") as string | null;
  const projectId = form.get("projectId") as string | null;

  await authData(server$.request, { csrf: csrf || "", error: "throw" });

  if (demoEnabled()) {
    throw new ServerError("Unauthorized", { status: 401 });
  }

  const { min: minMediaPath, raw: mediaPath } = mediaPaths();

  await mkdir(minMediaPath, { recursive: true });

  const promises: Promise<Awaited<ReturnType<typeof uploadFile>>>[] = [];
  for (const file of form.getAll("files")) {
    promises.push(uploadFile(file as unknown as File));
  }

  const savedFiles = await Promise.all(promises);

  const project = await prisma.nitroProject.findUniqueOrThrow({
    where: {
      id: projectId as string,
    },
  });

  const media: NitroMedia[] = [];
  for (const file of savedFiles) {
    const nitroMedia = await prisma.nitroMedia.create({
      data: {
        project: project
          ? {
              connect: {
                id: project.id,
              },
            }
          : undefined,
        ...file,
      },
    });

    media.push(nitroMedia);
  }

  return media;
});

export const optimizeImage = server$(async (args) => {
  const params = new URL(server$.request.url).searchParams;

  const pixelRatio = Number.parseInt(params.get("pixelRatio") || "1");
  if (![1, 2].includes(pixelRatio)) {
    throw new ServerError("Pixel ratio is invalid");
  }

  const image = params.get("image");
  if (!image) {
    throw new ServerError("Image path missing");
  }

  const key = params.get("key") || undefined;
  const offer = getResizeOffer(key);
  const sizeId = params.get("size") || "";
  const size_ = offer?.sizes[sizeId];
  if (!offer || !size_) {
    throw new ServerError("Key/size invalid");
  }

  const aspectRatioParam = params.get("aspectRatio") || undefined;
  if (
    aspectRatioParam &&
    offer.aspectRatios &&
    !offer.aspectRatios.includes(aspectRatioParam)
  ) {
    throw new ServerError("AspectRatio invalid");
  }
  const aspectRatio = calculateAspectRatio(aspectRatioParam);
  const fileType = ["png", "webp"].includes(offer.type as any)
    ? offer.type
    : "webp";

  const size = { ...size_ };

  if (size.width) {
    if (aspectRatio && !size.height) {
      size.height = size.width * aspectRatio;
    }
    size.width *= pixelRatio;
  }
  if (size.height) {
    size.height *= pixelRatio;
    if (aspectRatio && !size.width) {
      size.width = size.height / aspectRatio;
    }
  }

  size.width = size.width ? Math.round(size.width) : undefined;
  size.height = size.height ? Math.round(size.height) : undefined;

  const optionsKey = params.get("optionsKey");
  const optionsSchema = z.object({
    position: PositionSchema,
  });
  const optionsParam = optionsSchema.parse(
    JSON.parse(params.get("options") || "{}"),
  );

  const settings: ImageSize = {
    fit: "cover",
    position: "attention",
    quality: minQuality,
    ...size,
    ...optionsParam,
  };

  const { cache: fileDistDir, public: publicPath } = mediaPaths();
  const fileSrcPath = path.join(publicPath, image);
  const fileExt = path.extname(image);
  const fileName = path.basename(image, fileExt);
  const hash = fastHash(
    JSON.stringify({
      i: image,
      f: settings.fit,
      w: settings.width,
      h: settings.height,
      q: settings.quality,
      p: settings.position,
    }),
  );
  const fileDistPath = path.join(
    fileDistDir,
    `${fileName}.${hash}${pixelRatio > 1 ? ".2x" : ""}.${fileType}`,
  );

  // Cache the generated image for one year
  const maxAge = 365 * 24 * 60 * 60;

  const responseOptions: any = {
    // TODO: This broke favicons in Chrome
    // Make a decision, either always return 200 or implement an option
    // status: 206,
    headers: {
      age: 0,
      "Cache-Control": `public,max-age=${maxAge}`,
    },
  };

  try {
    const fd = await open(fileDistPath);
    const stat = await fd.stat();

    const readStream = fd.createReadStream();
    const readableStream = Readable.toWeb(readStream);

    const type = await fileTypeFromStream(readStream);

    const responseOpts = klona(responseOptions);
    responseOpts.headers["Content-Length"] = stat.size + "";
    responseOpts.headers["Content-Type"] =
      type?.mime || "application/octet-stream";
    responseOpts.headers["Last-Modified"] = stat.mtime.toUTCString();

    const res = new Response(readableStream as any, responseOpts);
    return res;
  } catch (err) {
    // TODO: add log file
  }

  if (optionsKey && optionsParam.position) {
    const optionsKeyValid = await validatePassword(
      optimizeKeyContent(image, optionsParam.position),
      optionsKey,
    );
    if (!optionsKeyValid) {
      throw new ServerError("OptionsKey invalid");
    }
  }

  await mkdir(fileDistDir, { recursive: true });

  const exists = await access(fileSrcPath)
    .then(() => true)
    .catch(() => false);
  if (!exists) {
    return new Response("File not found", { status: 404 });
  }

  // rotate() makes sure that the original rotation stays as is
  let sharpStream = sharp(fileSrcPath).rotate();

  if (settings.width && settings.height) {
    const meta = await sharpStream.metadata();
    const cropSize = getProportionalCropSize(
      meta.width!,
      meta.height!,
      settings.width,
      settings.height,
    );
    settings.width = cropSize.width;
    settings.height = cropSize.height;
  }

  if (settings.width || settings.height) {
    sharpStream = sharpStream.resize(settings.width, settings.height, {
      fit: settings.fit,
      position: settings.position,
      withoutEnlargement: true,
    });
  }

  if (fileType === "webp") {
    sharpStream = sharpStream.webp({ quality: settings.quality });
  } else {
    sharpStream = sharpStream.png({ quality: settings.quality });
  }

  const responseOpts = klona(responseOptions);
  responseOpts.headers["Content-Type"] = `image/${fileType}`;
  responseOpts.headers["Last-Modified"] = new Date().toUTCString();

  sharpStream.pipe(createWriteStream(fileDistPath, { autoClose: true }));
  const res = new Response(sharpStream as any, responseOpts);
  return res;
});
