import {
  parseCookie,
  serializeCookie,
  CookieSerializeOptions,
} from "solid-start";
import Iron from "@hapi/iron";

export type IronSessionConfig<T extends {} = {}> = {
  name: string;
  password: string;
  validate?: (v: any) => T;
  cookieRenewAge?: number;
  cookie?: CookieSerializeOptions;
  seal?: Iron.SealOptions;
};

type AnyFunction = (...args: any) => any;
type IronSessionValue<C extends IronSessionConfig> =
  C["validate"] extends AnyFunction ? ReturnType<C["validate"]> : any;
export const getIronSession = async function <C extends IronSessionConfig>(
  config: C,
  request: Request,
): Promise<(IronSessionValue<C> & { ironDate: Date }) | undefined> {
  const requestCache = ((request as any)["_ironCache"] =
    (request as any)["_ironCache"] || {});
  if (requestCache[config.name]) {
    return requestCache[config.name];
  }

  const cookiesHeader = request.headers.get("Cookie") ?? "";
  if (!cookiesHeader.length) {
    return;
  }

  const cookies = await parseCookie(cookiesHeader);
  const cookieHeader = cookies[config.name] ?? "";
  if (!cookieHeader.length) {
    return;
  }

  let session: any;

  try {
    session = await Iron.unseal(
      cookies[config.name],
      config.password,
      config.seal || Iron.defaults,
    );
  } catch (err) {
    // TODO: Add error logging for failing iron unseals
    // console.info(err)
  }

  if (!session) {
    return;
  }

  session["ironDate"] = new Date(session["ironDate"]);

  const result = config.validate ? config.validate(session) : session;
  requestCache[config.name] = result;

  return result;
};

const createCookieOptions = function (
  config: IronSessionConfig,
): CookieSerializeOptions {
  return {
    sameSite: "lax",
    secure: config.cookie?.sameSite === "none",
    ...(config.cookie || {}),
  };
};

export const setIronSession = async function <C extends IronSessionConfig>(
  config: C,
  response: Response,
  value: IronSessionValue<C>,
) {
  const value_ = { ...value, ironDate: new Date() };
  const session = await Iron.seal(
    value_,
    config.password,
    config.seal || Iron.defaults,
  );
  const cookieOptions = createCookieOptions(config);

  const setCookie = serializeCookie(config.name, session, cookieOptions);
  response.headers.set("Set-Cookie", setCookie);
  return response;
};

export const renewIronSession = async function <C extends IronSessionConfig>(
  config: C,
  request: Request,
  response: Response,
) {
  const renewAge = config.cookieRenewAge;
  if (!renewAge) {
    return response;
  }

  const value = await getIronSession(config, request);

  if (!value) {
    return;
  }

  const timeNow = Math.round(new Date().getTime() / 1000);
  const timeSession = Math.round(value.ironDate.getTime() / 1000);
  const diff = timeNow - timeSession;

  if (diff < renewAge) {
    return response;
  }

  if (import.meta.env.DEV) {
    console.log("Renew session", {
      timeNow,
      timeSession,
      diff,
      renewAge,
    });
  }

  return await setIronSession(config, response, value);
};

export const clearIronSession = async function (
  config: IronSessionConfig,
  response: Response,
) {
  const setCookie = serializeCookie(config.name, "", {
    ...createCookieOptions(config),
    maxAge: 0,
  });
  response.headers.set("Set-Cookie", setCookie);
  return response;
};
