import { NitroSetting } from "@prisma/client";
import { prisma } from "../../prisma";

type CreateSettingArguments<T> = {};

type GetterOptions = {
  cache?: boolean;
  projectId?: string;
  pageId?: string;
  inherit?: boolean;
};
type GetterResult<T> = {
  value: T | undefined;
  updatedAt?: Date;
  from: Group | "default";
};
type Getter<T> = (args: GetterOptions) => Promise<GetterResult<T>>;

type SetterOptions = {
  projectId?: string;
  pageId?: string;
};
type Setter<T> = (v: T | undefined, args: SetterOptions) => Promise<void>;

type SettingCache = {
  readTimestamp?: number;
  value?: any;
  id?: string;
  updatedAt?: Date;
};

type Group = "projects" | "pages";

const settingsCache: Record<
  string,
  Record<Group, Record<string, SettingCache>>
> = {};

export const flushSettingsCache = function (group: Group, groupItem: string) {
  for (const cache of Object.values(settingsCache)) {
    delete cache[group][groupItem];
  }
};

const warmupSettingCache = async function (
  name: string,
  group: Group,
  groupItem: string,
  {
    setting,
  }: {
    setting?: Partial<NitroSetting>;
  } = {},
) {
  settingsCache[name][group][groupItem] =
    settingsCache[name][group][groupItem] || {};

  const dbSetting =
    setting ||
    (await prisma.nitroSetting.findFirst({
      where: {
        name,
        pageId: group === "pages" ? groupItem : undefined,
        projectId: group === "projects" ? groupItem : undefined,
      },
    }));

  settingsCache[name][group][groupItem] = {
    readTimestamp: Date.now(),
    id: dbSetting?.id,
    value: dbSetting?.value ? JSON.parse(dbSetting?.value) : undefined,
    updatedAt: dbSetting?.updatedAt,
  };

  return settingsCache[name][group][groupItem];
};

const updateSettingDb = async function (
  group: Group,
  groupItem: string,
  name: string,
  value: any,
) {
  if (value == null) {
    await prisma.nitroSetting.deleteMany({
      where: {
        name,
        pageId: group === "pages" ? groupItem : undefined,
        projectId: group === "projects" ? groupItem : undefined,
      },
    });

    return;
  }

  const setting = await prisma.nitroSetting.findFirst({
    where: {
      name,
      pageId: group === "pages" ? groupItem : undefined,
      projectId: group === "projects" ? groupItem : undefined,
    },
  });
  const encodedValue = JSON.stringify(value);
  if (setting) {
    return await prisma.nitroSetting.update({
      data: {
        value: encodedValue,
      },
      where: {
        id: setting.id,
      },
    });
  }

  return await prisma.nitroSetting.create({
    data: {
      name,
      page: group === "pages" ? { connect: { id: groupItem } } : undefined,
      project:
        group === "projects" ? { connect: { id: groupItem } } : undefined,
      value: encodedValue,
    },
  });
};

const cacheFlushAge =
  parseInt(process.env["NP_SETTINGS_CACHE_FLUSH_AGE"] || "0") * 1000;
const cacheMaxAge =
  parseInt(process.env["NP_SETTINGS_CACHE_MAXAGE"] || "0") * 1000;

export const createSetting = function <T>(
  name: string,
  defaultValue: () => Promise<T>,
  args: CreateSettingArguments<T>,
) {
  let cacheFlushTimestamp = Date.now();
  let nowTimestamp = cacheFlushTimestamp;

  settingsCache[name] = {
    projects: {},
    pages: {},
  };

  const flushCache = function () {
    if (nowTimestamp <= cacheFlushTimestamp + cacheFlushAge) {
      return;
    }

    cacheFlushTimestamp = nowTimestamp;
    setTimeout(function () {
      const flushedItems = [];

      for (const group of Object.values(settingsCache[name])) {
        for (const [itemKey, item] of Object.entries(group)) {
          if (
            item.readTimestamp &&
            nowTimestamp <= item.readTimestamp + cacheMaxAge
          ) {
            continue;
          }

          flushedItems.push(itemKey);
          delete group[itemKey];
        }
      }

      console.log(
        `Settings Cache "${name}" flushed (Items: ${flushedItems.length})`,
      );
    }, 0);
  };

  const getFromGroup = async function ({
    group,
    groupItem,
    cache,
  }: {
    group: Group;
    groupItem: string;
    cache?: boolean;
  }) {
    let setting = settingsCache[name][group][groupItem];
    if (!cache || !setting) {
      setting = await warmupSettingCache(name, group, groupItem);
    }

    if (!setting) {
      return;
    }

    setting.readTimestamp = nowTimestamp;

    if (!setting.id) {
      return;
    }

    return {
      value: setting.value,
      updatedAt: setting.updatedAt,
      from: group,
    } as GetterResult<any>;
  };

  const get: Getter<T> = async function ({
    inherit = true,
    projectId,
    pageId,
    cache = true,
  }) {
    nowTimestamp = Date.now();

    if (inherit && pageId && !projectId) {
      const page = await prisma.nitroPage.findUniqueOrThrow({
        where: {
          id: pageId,
        },
      });
      projectId = page.projectId;
    }

    let result: GetterResult<T> | undefined;

    if (pageId) {
      result = await getFromGroup({ group: "pages", groupItem: pageId, cache });
    }

    if (!result && projectId) {
      result = await getFromGroup({
        group: "projects",
        groupItem: projectId,
        cache,
      });
    }

    flushCache();

    if (result) {
      return result;
    }

    const defaultV = await defaultValue();

    return { value: defaultV, from: "default" };
  };

  const set: Setter<T> = async function (value, { pageId, projectId }) {
    if (projectId) {
      const setting = await updateSettingDb("projects", projectId, name, value);
      warmupSettingCache(name, "projects", projectId, {
        setting: setting || {},
      });
    }

    if (pageId) {
      const setting = await updateSettingDb("pages", pageId, name, value);
      warmupSettingCache(name, "pages", pageId, {
        setting: setting || {},
      });
    }
  };

  return [get, set] as [typeof get, typeof set];
};

export const combineSettings = function <
  T extends Record<string, [Getter<any>, Setter<any>]>,
>(settings: T) {
  type V = { [K in keyof T]: Awaited<ReturnType<T[K][0]>> };
  const get = async function (options: GetterOptions): Promise<V> {
    let result = {} as V;
    for (const [k, v] of Object.entries(settings)) {
      const key: keyof typeof result = k;
      result[key] = (await v[0](options)) as any;
    }

    return result;
  };

  type VV = { [K in keyof T]: Awaited<ReturnType<T[K][0]>>["value"] };
  const set = async function (value: VV, options: SetterOptions) {
    for (const [k, v] of Object.entries(settings)) {
      (await v[1](value[k], options)) as any;
    }
  };

  return [get, set] as [typeof get, typeof set];
};
