import type {
  Component,
  ComponentProps,
  Context,
  FlowComponent,
  JSX,
  ValidComponent,
} from "solid-js";
import type { ImageSize } from "./zodTypes";

export type BlueprintDataInputProps<V = any, C = any> = {
  element?: Element;
  setValue: (produce: (state: V) => V) => void;
  setOverride: (produce: (override: boolean) => boolean) => void;
  value?: V;
  data: BlueprintData<V, C>;
  override: boolean;
  key: string;
  context: "element" | "preset" | "project";
};

export type BlueprintDataInput<V = any, C = any> = Component<
  BlueprintDataInputProps<V, C>
>;

export type ResizeOffer = {
  sizes: Record<string, ImageSize>;
  aspectRatios?: readonly string[];
  type?: "png" | "webp";
};

export type BlueprintDataWrapper = FlowComponent;

export type ToolbarOption = {
  bold?: boolean;
  italic?: boolean;
  underline?: boolean;
  color?: boolean;
  background?: boolean;
  script?: boolean;
  align?: boolean;
  indent?: boolean;
  listOrdered?: boolean;
  listBullet?: boolean;
  link?: boolean;
  clean?: boolean;
};

export type BlueprintData<V = any, C = any> = {
  id?: string;
  title?: () => string;
  rte?: {
    // disabled this option for now is I need to figure out a good api first
    // formats?: string[];
    toolbar?: ToolbarOption;
  };
  defaultValue: () => V;
  custom?: C;
  getter?: (value: V | undefined, state: State) => unknown;
  wrapper?: BlueprintDataWrapper;
  loadWrappers?: { id: string; wrapper: BlueprintDataWrapper }[];
  context?: Context<any>;
  required?: boolean;
  input: BlueprintDataInput<V>;
  type?: String;
};

export type FontFaceFile = {
  format: string;
  mime: string;
  name: string;
};

export type FontFace = {
  id: string;
  files: FontFaceFile[];
  weight?: number;
  style?: string;
  preload?: boolean;
};
export type Font = {
  id?: string;
  publicId: string;
  family: string;
  external?: boolean;
  externalItalic?: boolean;
  externalWeights?: string[];
  fallbacks: string[];
  preload?: boolean;
  faces: FontFace[];
};

export type InputFromBlueprintData<T extends BlueprintData> = T["input"];

export type BlueprintDataRecord = Record<string, BlueprintData>;

type CustomProps<Props> = Omit<
  Props,
  "children" | "editor" | "dragging" | "ref" | "Slot" | "dragSlot"
>;

type ComponentData<
  T extends CustomBlueprintComponent = CustomBlueprintComponent,
  Props = CustomProps<ComponentProps<T>["data"]>,
> = {
  [P in keyof Props]: Props[P];
};

export type BlueprintDataPropDefault<T extends BlueprintDataRecord> = {
  [P in keyof T]: ReturnType<T[P]["defaultValue"]>;
};
export type BlueprintDataPropGetter<T extends BlueprintData> =
  T["getter"] extends Function
    ? ReturnType<T["getter"]>
    : ReturnType<T["defaultValue"]>;
export type BlueprintDataRecordPropGetter<T extends BlueprintDataRecord> = {
  [P in keyof T]: BlueprintDataPropGetter<T[P]>;
};

export type BlueprintDataPropsDefault<
  T extends BlueprintFn,
  R extends BaseBlueprint = Awaited<ReturnType<T>>,
> = BlueprintDataPropDefault<
  R["data"] extends BlueprintDataRecord ? R["data"] : BlueprintDataRecord
>;
export type BlueprintDataPropsGetter<
  T extends BlueprintFn,
  R extends BaseBlueprint = Awaited<ReturnType<T>>,
> = BlueprintDataRecordPropGetter<
  R["data"] extends BlueprintDataRecord ? R["data"] : BlueprintDataRecord
>;

export type Slots = Record<
  string,
  {
    max?: number;
    blueprint?: string;
    acceptBlueprint?: (blueprint: string) => boolean | void | undefined;
  }
>;

export type BaseBlueprint<
  SD extends BlueprintDataRecord = BlueprintDataRecord,
> = {
  component?: CustomBlueprintComponent;
  title?: () => string;
  acceptParentType?: "root" | "element";
  acceptParent?: (blueprint: string) => boolean | void | undefined;
  slots?: Slots;
  data: SD;
  previewData?: () => Partial<BlueprintDataPropDefault<SD>>;
};

export interface Blueprint extends BaseBlueprint {
  filename?: string;
  slots: Slots;
  title: () => string;
  id: string;
  version?: number;
}

export type BlueprintFn = {
  (opts: { admin: boolean }): BaseBlueprint | Promise<BaseBlueprint>;
  _id?: string;
};

export type ElementData = { value: any; override?: boolean };
export type ElementDataRecord = Record<string, ElementData>;

export type Element = {
  // corresponds to NitroElement.historyId
  id: string;
  elementId: string;
  blueprint: string;
  dynamic?: boolean;
  data: ElementDataRecord;
  ssrData: Record<string, any>;
  presetId?: string;
  parentId?: string;
  parentSlot: string;
  slots: Record<
    string,
    {
      elements: string[];
    }
  >;
};

export type Preset = {
  id: string;
  new?: boolean;
  blueprint: string;
  name: string;
  data: ElementDataRecord;
};

export type SlotComponent<
  ElementProps = JSX.HTMLAttributes<HTMLElement>,
  S = string,
> = Component<
  {
    optional?: boolean;
    component?: ValidComponent;
    name: S;
    direction?: "vertical" | "horizontal";
    editorPadding?: boolean;
  } & ElementProps
>;
export type DataLense<
  B extends BaseBlueprint = BaseBlueprint,
  D extends BlueprintDataKeys<B> = BlueprintDataKeys<B>,
> = <T extends D>(data: T) => BlueprintDataPropDefault<B["data"]>[T];

export type RteComponent<
  B extends BaseBlueprint = BaseBlueprint,
  D extends BlueprintDataKeys<B> = BlueprintDataKeys<B>,
> = Component<{
  component?: ValidComponent;
  class?: string;
  data?: D;
  path?: (lense: DataLense<B, D>) => any;
}>;
type DragSlotConfig = {
  slot?: string;
  direction?: "vertical" | "horizontal";
};
export type DragSlotFunction = (
  el: HTMLElement,
  config: () => DragSlotConfig,
) => any;
type BlueprintComponentProps<B extends BaseBlueprint, S = BlueprintSlots<B>> = {
  admin?: boolean;
  preview?: boolean;
  handles?: JSX.Element;
  blueprint: B;
  element: {
    id: string;
  };
  // TODO: Maybe rename to rootProps
  elementProps: { "data-element": string | boolean; id?: string };
  slotElements: (name: S) => string[];
  Slot: SlotComponent<JSX.HTMLAttributes<HTMLElement>, S>;
  Rte: RteComponent<B>;
  lax?: boolean;
};

export type BlueprintSlots<T extends BaseBlueprint> =
  T["slots"] extends Record<any, any> ? keyof T["slots"] : string;
export type BlueprintDataKeys<T extends BaseBlueprint> =
  T["data"] extends Record<any, any> ? keyof T["data"] : string;

export type CustomBlueprintComponent<
  T = {},
  B extends BaseBlueprint = BaseBlueprint,
> = Component<{ data: T } & BlueprintComponentProps<B>>;

export type SidebarElementsTab = "element" | "preset" | "presets" | "project";

export type EditorUiState = {
  pageId: string;
  expandElementsTree: boolean;
  selectedRevisionId: string;
  sidebarStatus: "idle" | "chooseType" | "createBlueprint";
  sidebarIdleTab: "elements" | "page" | "project";
  sidebarElementsTab: SidebarElementsTab;
  toolbarStatus:
    | "idle"
    | "write"
    | "selecting"
    | "creating"
    | "locked"
    | "moving"
    | "deleting";
  movingElementId: false | string;
  createMoveTarget: {
    elementId?: string;
    slot?: string;
  };
  darkMode?: boolean;
  hoveredElementId: string | undefined;
  creatingElementId: string | undefined;
  deletingElementId: string | undefined;
  selectedElementId: string | undefined;
  selectedBlueprint: string | undefined;
};

export interface RgbColor {
  r: number;
  g: number;
  b: number;
}

export type Colors = Record<string, { light: RgbColor; dark?: RgbColor }>;
export type Color = {
  css?: string;
  rgb?: RgbColor;
  id?: string;
  invert?: boolean;
  alpha?: number;
};

export type RteTextColors = Record<string, Color>;

export type BlueprintDefaults = Record<string, Record<string, ElementData>>;

export type State = {
  admin?: boolean;
  title: string;
  cacheKey: number;
  urlPath: string;
  defaultFont?: string;
  notFound?: boolean;
  includeInFeed?: boolean;
  metaDescription?: string;
  metaImage?: string;
  favicon?: string;
  fonts: Font[];
  colors: Colors;
  backgroundColor?: Color;
  richtext: {
    textColors: RteTextColors;
  };
  root: {
    header: string[];
    default: string[];
    footer: string[];
  };
  // true -> blueprint needs to always be loaded
  // false -> blueprint is static, can be skipped during hydration
  blueprintIds?: Record<string, boolean>;
  blueprintDefaults: {
    project: BlueprintDefaults;
    // TODO: Remove page blueprint defaults
    // page: BlueprintDefaults;
  };
  elements: Record<string, Element>;
  presets: Record<string, Preset>;
  pendingDeletes?: {
    presets: string[];
  };
};

export type Config = {
  blueprints: Record<string, Blueprint>;
};

export type BlueprintComponent<TS extends () => BlueprintFn> =
  CustomBlueprintComponent<
    BlueprintDataPropsGetter<ReturnType<TS>>,
    Awaited<ReturnType<ReturnType<TS>>>
  >;
