import {
  NitroElement,
  NitroElementSlot,
  NitroFont,
  NitroFontFace,
  NitroPage,
  NitroPageRevision,
  NitroPreset,
} from "@prisma/client";
import { Settings } from "../server/settings";
import { Element, FontFaceFile, State } from "../types";

export const normalizeFonts = function ({
  settings,
  fonts,
}: {
  settings?: Settings;
  fonts: (NitroFont & { faces: NitroFontFace[] })[];
}) {
  let defaultId: string | undefined;

  const normalizedFonts = fonts.map(function (font) {
    const normalizedFaces = font.faces.map(function (face) {
      const normalizedFiles = (JSON.parse(face.files) as FontFaceFile[]).sort(
        function (fileA, fileB) {
          const sortA = fontTypesSorting[fileA.format] || 1;
          const sortB = fontTypesSorting[fileB.format] || 1;

          return sortB - sortA;
        },
      );

      return {
        id: face.id,
        preload: face.preload,
        weight: face.weight!,
        style: face.style,
        files: normalizedFiles,
      };
    });

    if (settings?.defaultFont.value === font.id) {
      defaultId = font.publicId;
    }

    return {
      id: font.id,
      publicId: font.publicId,
      family: font.family,
      preload: font.preload,
      fallbacks: font.fallback.split(","),
      faces: normalizedFaces,
    };
  });

  return {
    defaultFont: defaultId,
    fonts: normalizedFonts,
  } satisfies Partial<State>;
};
export const normalizePresets = function ({
  presets,
}: {
  presets: NitroPreset[];
}) {
  return presets.reduce(
    (acc, preset) => {
      acc[preset.id] = {
        id: preset.id,
        name: preset.name,
        blueprint: preset.blueprint,
        data: JSON.parse(preset.data),
      };
      return acc;
    },
    {} as State["presets"],
  );
};
export const normalizeSettings = function ({
  settings,
  pageRevision,
}: {
  pageRevision: PageRevision;
  settings: Settings;
}) {
  return {
    notFound: settings.notFoundPage.value === pageRevision.pageId,
    metaDescription: settings.metaDescription.value,
    metaImage: settings.metaImage.value?.mediaIds[0],
    favicon: settings.favicon.value?.mediaIds[0],
    colors: settings.colors.value || {},
    backgroundColor: settings.pageBackgroundColor.value,
    richtext: {
      textColors: settings.rteTextColors.value || {},
    },
    blueprintDefaults: {
      project: settings.blueprintDefaults.value || {},
      // TODO: Remove page blueprint defaults
      //page: JSON.parse(pageRevision.blueprintDefaults),
    },
  } satisfies Partial<State>;
};
export const normalizePage = function ({
  pageRevision,
  page,
  passiveBlueprints = {},
}: {
  pageRevision: PageRevision;
  page?: NitroPage;
  passiveBlueprints?: Record<string, boolean>;
}) {
  const blueprintIds: Record<string, boolean> = {};
  const dynamicElements: string[] = [];
  const elements = pageRevision.elements.reduce(
    (acc, el) => {
      blueprintIds[el.blueprint] = false;
      const dynamic = !passiveBlueprints[el.blueprint];
      if (dynamic) {
        dynamicElements.push(el.historyId);
      }

      acc[el.historyId] = {
        id: el.historyId,
        elementId: el.id,
        presetId: el.presetId || undefined,
        parentId: el.parentSlot?.parentElement?.historyId,
        dynamic,
        blueprint: el.blueprint,
        data: JSON.parse(el.data),
        ssrData: {},
        slots: el.slots.reduce(
          (acc, slot) => {
            acc[slot.name] = {
              elements: slot.elements.map((el) => el.historyId),
            };
            return acc;
          },
          {} as Element["slots"],
        ),
        parentSlot: el.parentSlot?.name || "default",
      };

      return acc;
    },
    {} as Record<string, Element>,
  );

  const setIds: Record<string, boolean> = {};
  const setDynamic = function (id: string) {
    if (setIds[id]) {
      return;
    }

    setIds[id] = true;

    const element = elements[id];
    elements[id].dynamic = true;
    blueprintIds[element.blueprint] = true;

    if (element.parentId == null) {
      return;
    }

    setDynamic(element.parentId);
  };

  for (const id of dynamicElements) {
    setDynamic(id);
  }

  const root = pageRevision.rootSlots.reduce(
    function (acc, next) {
      const slotName = next.name as unknown as keyof typeof acc;
      if (!acc[slotName]) {
        return acc;
      }

      for (const el of next.elements) {
        acc[slotName].push(el.historyId);
      }

      return acc;
    },
    { default: [], footer: [], header: [] } as State["root"],
  );

  return {
    title: pageRevision.title,
    urlPath: pageRevision.urlPath,
    includeInFeed: page ? page.includeInFeed : undefined,
    root,
    elements,
    blueprintIds,
  };
};
export const normalizeState = function ({
  pageRevision,
  page,
  fonts,
  presets,
  settings,
  passiveBlueprints = {},
}: {
  pageRevision: PageRevision;
  page?: NitroPage;
  settings: Settings;
  presets: NitroPreset[];
  fonts: (NitroFont & { faces: NitroFontFace[] })[];
  passiveBlueprints?: Record<string, boolean>;
}): State {
  const settingsData = normalizeSettings({
    pageRevision,
    settings,
  });
  const pageData = normalizePage({ page, pageRevision, passiveBlueprints });
  const presetData = normalizePresets({ presets });
  const fontData = normalizeFonts({ fonts, settings });

  const result = {
    cacheKey: 0,
    presets: presetData,
    ...pageData,
    ...fontData,
    ...settingsData,
  } satisfies State;

  return result;
};
export type PageRevision = NitroPageRevision & {
  rootSlots: (NitroElementSlot & { elements: NitroElement[] })[];
  elements: (NitroElement & {
    parentSlot:
      | (NitroElementSlot & { parentElement: NitroElement | null })
      | null;
    slots: (NitroElementSlot & { elements: NitroElement[] })[];
  })[];
};
export const fontTypesSorting: Record<string, number> = {
  woff2: 4,
  woff: 3,
  opentype: 2,
  truetype: 1,
};
