import clsx from "clsx";
import {
  A,
  createBlueprint,
  BlueprintComponent,
  DataLense,
  compose,
} from "nitropage";
import { css, colorData, rgb } from "nitropage/data/color";
import { fontCss, fontData } from "nitropage/data/font";
import { listData, sortItems } from "nitropage/data/list";
import {
  Image,
  ImageContainer,
  MediaContainer,
  mediaData,
  PlaceholderImage,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { rteData, toolbarDefaults } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import { createResize } from "nitropage/server";
import { Component, For, Show, createMemo } from "solid-js";
import server$ from "solid-start/server";

/*
 * The design of this blueprint is inspired by Meraki UI, check it out!
 * https://github.com/merakiui/merakiui
 */

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const sortedItems = createMemo(() =>
    [...sortItems(props.data.items)].reverse(),
  );
  const recentItems = createMemo(() => sortedItems().slice(1));
  const mainPost_ = createMemo(() => sortedItems()[0]);
  const mainPost = createMemo(() => mainPost_()?.[0]);
  const [images] = useMedia(() => mainPost()?.image);
  const [authorImages] = useMedia(() => props.data.authorImage);
  const [optimizeMediaScreen] = useResize(
    server$(
      createResize({
        aspectRatios: ["16/9"],
        sizes: {
          xs: { width: 512 },
          lg: { width: 720 },
          lgWide: { width: 1080, height: 405 },
          author: { width: 100, height: 100 },
        },
      }),
    ),
  );

  type Lense = DataLense<typeof props.blueprint>;

  const TeaserImage = function () {
    return (
      <A href={mainPost()?.url || ""} class="group/teaser block lg:h-96">
        <For
          each={images().byOrder.length ? images().byOrder : [PlaceholderImage]}
        >
          {(image) => {
            return (
              <MediaContainer
                blurhash
                fadeDuration={2000}
                media={image}
                sources={optimizeMediaScreen(image, [
                  {
                    size: recentItems().length ? "lgWide" : "lg",
                    screen: 1024,
                  },
                  { size: "lg", aspectRatio: "16/9", screen: 500 },
                  { size: "xs", aspectRatio: "16/9" },
                ])}
                imageClass="group-hover/teaser:scale-105 transition-transform"
                class="overflow-hidden rounded-xl media-outline"
              />
            );
          }}
        </For>
      </A>
    );
  };

  const Author = function () {
    return (
      <div class="flex items-center mt-4 lg:mt-6">
        <For
          each={
            authorImages().byOrder.length
              ? authorImages().byOrder
              : [PlaceholderImage]
          }
        >
          {(image) => {
            return (
              <ImageContainer
                width={false}
                height={false}
                class="overflow-hidden w-10 h-10 rounded-full"
              >
                <Image
                  sources={optimizeMediaScreen(image, [{ size: "author" }])}
                  alt={image.description}
                />
              </ImageContainer>
            );
          }}
        </For>

        <div class="mx-4">
          <props.Rte
            data="author"
            component="span"
            class="text-sm text-slate-700 teaser-text--light"
          />
          <props.Rte
            data="authorSubText"
            component="p"
            class="text-sm text-slate-500 teaser-text--lighter"
          />
        </div>
      </div>
    );
  };

  const HeaderLayout: Component<{
    post: {
      url: string;
      category?: string;
      categoryUrl?: string;
      date?: string;
    };
    id: string;
  }> = function (p) {
    const itemLense = (d: Lense) => d("items").byId[p.id];

    return (
      <div class="w-full flex items-center gap-6">
        <Show when={props.admin || p.post.category}>
          <A
            href={p.post.categoryUrl || p.post.url}
            class="text-sm teaser-link rte-ellipsis min-w-0 text-blue-600 uppercase hover:underline"
          >
            <props.Rte path={compose(itemLense, (d) => d.category)} />
          </A>
        </Show>
        <Show when={p.post.date && p.post.category}>
          <div class="teaser-text--light">&bull;</div>
        </Show>
        <props.Rte
          class="text-sm teaser-text--light"
          path={compose(itemLense, (d) => d.date)}
        />
      </div>
    );
  };

  const MainPost = function () {
    const postLense = (d: Lense) => d("items").byId[mainPost_()[1]];

    return (
      <>
        <HeaderLayout post={mainPost()} id={mainPost_()[1]} />

        <A
          href={mainPost().url}
          class={
            css({
              ...fontCss(props.data.titleFont),
            }) + " block mt-4 text-2xl font-semibold hover:underline"
          }
        >
          <props.Rte path={compose(postLense, (d) => d.title)} />
        </A>
        <props.Rte
          component="p"
          path={compose(postLense, (d) => d.teaser)}
          class="mt-3 text-sm teaser-text--lighter md:text-base"
        />
        <Show when={mainPost().teaser}>
          <A
            href={mainPost().url}
            class={clsx(
              "mt-2 teaser-link text-blue-600 underline hover:no-underline",
              recentItems().length ? "hidden lg:inline-block" : "inline-block",
            )}
          >
            <props.Rte data="urlText" />
          </A>
        </Show>
        <Author />
      </>
    );
  };

  const RecentPosts = function () {
    return (
      <div class="mt-8 lg:w-1/4 lg:mt-0 text-sm lg:text-base">
        <hr class="my-6 lg:hidden border-slate-200 teaser-border" />
        <For each={recentItems()}>
          {([item, id]) => {
            return (
              <div class="group/recent-item">
                <HeaderLayout post={item} id={id} />
                <A
                  href={item.url}
                  class="block mt-2 font-semibold text-slate-700 teaser-text--light hover:underline"
                >
                  <props.Rte path={(d) => d("items").byId[id].title} />
                </A>
                <hr class="my-6 border-slate-200 teaser-border group-last/recent-item:hidden" />
              </div>
            );
          }}
        </For>
      </div>
    );
  };

  return (
    <div
      class={
        "text-slate-900 " +
        css({
          color: rgb(props.data.textColor),
          "& .teaser-link": {
            color: rgb(props.data.linkColor),
          },
          "& .teaser-text--light": {
            color: rgb({ ...props.data.textColor, alpha: 0.85 }),
          },
          "& .teaser-text--lighter": {
            color: rgb({ ...props.data.textColor, alpha: 0.65 }),
          },
          "& .teaser-border": {
            borderColor: rgb(props.data.borderColor),
          },
        })
      }
      {...props.elementProps}
    >
      {props.handles}
      <div
        class={clsx(
          "lg:flex gap-12",
          !recentItems().length && "lg:items-center",
        )}
      >
        <div class="flex flex-col flex-1">
          <TeaserImage />
          <Show when={recentItems().length}>
            <div class="mt-6">
              <MainPost />
            </div>
          </Show>
        </div>

        <Show when={mainPost()}>
          <Show when={!recentItems().length} fallback={<RecentPosts />}>
            <div class="mt-6 flex-1 lg:mt-0">
              <MainPost />
            </div>
          </Show>
        </Show>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    const toolbar = {
      ...toolbarDefaults,
      underline: false,
      link: false,
    };

    return {
      title() {
        return "Blog Post Teaser";
      },
      acceptParentType: "element",
      slots: {},
      data: {
        items: listData({
          defaultItems: 1,
          itemTitle: (data) => data.title?.content,
          itemData: {
            image: mediaData({
              type: "image",
              max: 1,
            }),
            title: rteData({
              shortLorem: true,
              toolbar,
            }),
            teaser: rteData({
              lorem: true,
              toolbar,
            }),
            url: textData({
              wide: true,
              title: () => "Link",
            }),
            category: rteData({
              shortLorem: true,
              toolbar,
            }),
            categoryUrl: textData({
              wide: true,
            }),
            date: rteData({ toolbar }),
          },
        }),
        author: rteData({
          defaultValue() {
            return "Author Name";
          },
        }),
        authorSubText: rteData({
          defaultValue() {
            return "Author Subtext";
          },
        }),
        authorImage: mediaData({
          type: "image",
          max: 1,
        }),
        urlText: rteData({
          defaultValue() {
            return `READ MORE`;
          },
          title() {
            return "Link Text";
          },
          toolbar,
        }),
        titleFont: fontData(),
        textColor: colorData({
          alpha: false,
        }),
        linkColor: colorData(),
        borderColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "fh2n2ya9tstf76uhh1l1iqtg";
