import clsx from "clsx";
import {
  createBlueprint,
  BlueprintComponent,
  DataLense,
  A,
  compose,
} from "nitropage";
import { css, rgb } from "nitropage/css";
import { colorData } from "nitropage/data/color";
import { fontCss, fontData } from "nitropage/data/font";
import { listData, sortItems } from "nitropage/data/list";
import {
  MediaContainer,
  PlaceholderImage,
  mediaData,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { rteData, toolbarDefaults } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import { createResize } from "nitropage/server";
import { For, createMemo } from "solid-js";
import server$ from "solid-start/server";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const sortedItems = createMemo(() => sortItems(props.data.items));
  const [images] = useMedia(() => sortedItems().map((i) => i[0].image?.[0]!), {
    dataUrl: true,
  });
  const [resize] = useResize(
    server$(
      createResize({
        aspectRatios: ["1/1"],
        sizes: {
          lg: { width: 160 },
        },
      }),
    ),
  );

  type Lense = DataLense<typeof props.blueprint>;

  return (
    <div
      class={css({
        "& .item__title": {
          color: rgb(props.data.color),
          ...fontCss(props.data.titleFont),
        },
        "& .item__link": {
          color: rgb(props.data.linkColor || props.data.color),
        },
      })}
      {...props.elementProps}
    >
      {props.handles}
      <div
        class={clsx(
          "grid lg:grid-cols-2 gap-x-8 gap-y-10",
          sortedItems().length >= 3 && "2xl:grid-cols-3",
        )}
      >
        <For each={sortedItems()}>
          {([item, id]) => {
            const itemLense = (d: Lense) => d("items").byId[id];
            const image = () =>
              images().byId[item.image?.[0] || ""] || PlaceholderImage;

            return (
              <div>
                <A
                  href={item.link}
                  class="h-full flex items-center gap-4 group"
                >
                  <div class="w-24 xl:w-32 aspect-square flex-shrink-0">
                    <MediaContainer
                      media={image()}
                      lazy
                      fadeDuration={1000}
                      sources={resize(image(), [
                        { size: "lg", aspectRatio: "1/1" },
                      ])}
                      placeholder
                      class="rounded-lg media-outline overflow-hidden"
                      imageClass="transition-transform group-hover:scale-110"
                    />
                  </div>
                  <div class="flex-1 min-w-0 flex flex-col">
                    <props.Rte
                      class="item__title text-slate-900 rte-ellipsis text-lg lg:text-xl font-bold group-hover:underline mb-4 lg:mb-6"
                      path={compose(itemLense, (d) => d.title)}
                    />
                    <props.Rte
                      class="item__link text-slate-700 underline text-sm lg:text-base"
                      data="linkText"
                    />
                  </div>
                </A>
              </div>
            );
          }}
        </For>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    const toolbar = {
      ...toolbarDefaults,
      underline: false,
      link: false,
    };

    return {
      title: () => "Blog Post Teaser - Small",
      acceptParentType: "element",
      data: {
        items: listData({
          defaultItems: 3,
          itemTitle: (data) => data.title?.content,
          itemData: {
            image: mediaData({
              type: "image",
              max: 1,
            }),
            title: rteData({
              shortLorem: true,
              toolbar,
            }),
            link: textData({
              wide: true,
            }),
          },
        }),
        linkText: rteData({
          defaultValue() {
            return `READ MORE`;
          },
          toolbar,
        }),
        titleFont: fontData(),
        color: colorData(),
        linkColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "may03wu7iylv7y3v62hyz9gc";
