import { createBlueprint, BlueprintComponent } from "nitropage";
import { css, colorData, rgb } from "nitropage/data/color";
import PostLayout from "./blog/postLayout";
import { toggleData } from "nitropage/data/toggle";
import { Dynamic } from "solid-js/web";
import Fragment from "~/components/fragment";
import hljs from "highlight.js/lib/core";
import { textData } from "nitropage/data/text";
import { createResource } from "solid-js";
import { selectData } from "nitropage/data/select";

const languageMap = {
  bash: () => import("highlight.js/lib/languages/bash"),
  c: () => import("highlight.js/lib/languages/c"),
  csharp: () => import("highlight.js/lib/languages/csharp"),
  css: () => import("highlight.js/lib/languages/css"),
  go: () => import("highlight.js/lib/languages/go"),
  php: () => import("highlight.js/lib/languages/php"),
  python: () => import("highlight.js/lib/languages/python"),
  java: () => import("highlight.js/lib/languages/java"),
  json: () => import("highlight.js/lib/languages/json"),
  jsx: () => import("highlight.js/lib/languages/javascript"),
  ruby: () => import("highlight.js/lib/languages/ruby"),
  rust: () => import("highlight.js/lib/languages/rust"),
  sql: () => import("highlight.js/lib/languages/sql"),
  tsx: () => import("highlight.js/lib/languages/typescript"),
  xml: () => import("highlight.js/lib/languages/xml"),
  yaml: () => import("highlight.js/lib/languages/yaml"),
};

const loadLanguage = async function (language: keyof typeof languageMap) {
  const loader = languageMap[language];
  if (!loader) return;

  if (hljs.getLanguage(language)) {
    return;
  }

  if (["tsx", "jsx"].includes(language)) {
    await loadLanguage("xml");
  }

  const lang = await loader();
  hljs.registerLanguage(language, lang.default);
};

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [html] = createResource(
    () => [props.data.code, props.data.language || "jsx"],
    async function ([code, language]) {
      await loadLanguage(language as any);
      return hljs.highlight(code, { language }).value;
    },
  );
  return (
    <div
      class={
        css({
          "& code": {
            borderColor: rgb(props.data.borderColor),
            backgroundColor: rgb(props.data.bgColor),
          },
        }) + " md:text-lg"
      }
      {...props.elementProps}
    >
      {props.handles}
      <Dynamic component={props.data.fullWidth ? Fragment : PostLayout}>
        <pre>
          <code
            class="text-[0.9em] hljs !p-2 md:!p-3 rounded-lg border border-transparent"
            innerHTML={html()}
          />
        </pre>
      </Dynamic>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      acceptParentType: "element",
      data: {
        code: textData({ multiline: true }),
        language: selectData({
          options: [
            { value: "bash" },
            { value: "c" },
            { value: "csharp", title: () => "C#" },
            { value: "css" },
            { value: "go" },
            { value: "php" },
            { value: "python" },
            { value: "java" },
            { value: "json" },
            { title: () => "Jsx" },
            { value: "ruby" },
            { value: "rust" },
            { value: "sql" },
            { value: "tsx" },
            { value: "xml" },
            { value: "yaml" },
          ],
        }),
        fullWidth: toggleData({ title: () => "Full Width" }),
        bgColor: colorData(),
        borderColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "ussx0679sbtu0kqcez8wyu70";
