import { Button } from "@kobalte/core";
import clsx from "clsx";
import { createBlueprint, BlueprintComponent, A, Richtext } from "nitropage";
import { colorData, css, rgb } from "nitropage/data/color";
import { listData, sortItems } from "nitropage/data/list";
import { rteData } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import {
  Component,
  FlowComponent,
  For,
  Show,
  createMemo,
  createSignal,
  onMount,
} from "solid-js";
import { isServer } from "solid-js/web";
import { useLocation } from "solid-start";
import { Container } from "~/components/container";
import MenuIcon from "~icons/tabler/selector";

/*
 * The design of this blueprint is adapted from the Kobalte Site components
 * https://kobalte.dev
 * https://github.com/kobaltedev/kobalte/tree/e8ca243e95a8c5513c6648e44be98ac3660b5f54/apps/docs/src/components
 */

type TocItem = { text?: string; id: string; level: number };

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  let articleEl: HTMLElement = undefined!;
  const path = useLocation();
  const sections = createMemo(() => sortItems(props.data.sections));
  const [showToc, setShowToc] = createSignal(false);

  const [toc, setToc] = createSignal<TocItem[]>();
  const pagination = createMemo(function () {
    let prev: { url: string; title: string } | undefined;
    let found: { url: string; title: string } | undefined;

    for (const [section] of sections()) {
      for (const linkId of section.links?.ids || []) {
        const link = section.links?.byId[linkId]!;

        if (found) {
          return {
            prev,
            current: found,
            next: link,
          };
        }

        if (link.url === path.pathname || props.admin) {
          found = link;
          continue;
        }

        prev = link;
      }
    }

    return {};
  });

  onMount(function () {
    if (props.admin) {
      return;
    }

    const toc = [...articleEl.querySelectorAll("h2, h3")].reduce((acc, el) => {
      let id = el.id;
      if (!id) {
        const parentWithId = el.closest("[id]");
        if (!parentWithId) {
          return acc;
        }

        id = parentWithId.id;
      }

      acc.push({
        text: el.textContent || undefined,
        id,
        level: Number.parseInt(el.tagName.slice(1)),
      });

      return acc;
    }, [] as TocItem[]);

    setToc(toc);
  });

  const PaginationLabel: FlowComponent = function (p) {
    return (
      <dt class="text-sm font-bold docs-text text-slate-900 ">{p.children}</dt>
    );
  };

  const PaginationLink: FlowComponent<{ href: string }> = function (p) {
    return (
      <dd class="mt-1">
        <A
          href={p.href}
          class="flex gap-1 text-base font-semibold docs-text-muted text-slate-500 hover:text-slate-600"
        >
          {p.children}
        </A>
      </dd>
    );
  };

  const Pagination: Component = function () {
    return (
      <div class="docs-border mt-12 flex justify-between gap-8 border-t border-slate-200 pt-6 ">
        <dl>
          <Show when={pagination().prev}>
            {(link) => (
              <>
                <PaginationLabel>
                  <props.Rte data="previousLabel" />
                </PaginationLabel>
                <PaginationLink href={link().url}>
                  <span aria-hidden="true" class="mr-1">
                    &larr;
                  </span>
                  <Richtext>{link().title}</Richtext>
                </PaginationLink>
              </>
            )}
          </Show>
        </dl>
        <Show when={pagination().next}>
          {(link) => (
            <dl class="text-right">
              <dt class="text-sm font-bold docs-text text-slate-900 ">
                <props.Rte data="nextLabel" />
              </dt>
              <PaginationLink href={link().url}>
                <Richtext>{link().title}</Richtext>
                <span aria-hidden="true" class="ml-1">
                  &rarr;
                </span>
              </PaginationLink>
            </dl>
          )}
        </Show>
      </div>
    );
  };

  const Navigation: Component = function () {
    return (
      <nav class={clsx("text-base lg:text-sm lg:w-52")}>
        <ul class="space-y-9">
          <For each={sections()}>
            {([section, id]) => (
              <li>
                <h2 class="font-bold docs-text text-slate-900 ">
                  <props.Rte path={(d) => d("sections").byId[id].title} />
                </h2>
                <ul class="mt-2 lg:mt-3 text-sm">
                  <For each={sortItems(section.links)}>
                    {([link, linkId]) => (
                      <li class="relative group">
                        <A
                          href={link.url}
                          activeClass="text-blue-700 docs-link--active"
                          inactiveClass="text-slate-600 docs-text-muted"
                          matchHash={false}
                          end
                          class={clsx(
                            "docs-link flex items-center justify-between w-full font-sans transition font-normal rounded px-3 py-2 hover:bg-blue-50",
                          )}
                        >
                          <props.Rte
                            path={(d) =>
                              d("sections").byId[id].links?.byId[linkId].title
                            }
                          />
                          <Show when={link.status}>
                            <span class="docs-status inline-flex items-center font-medium leading-none uppercase px-1.5 py-1 text-xs rounded text-white bg-blue-600">
                              {link.status}
                            </span>
                          </Show>
                        </A>
                      </li>
                    )}
                  </For>
                </ul>
              </li>
            )}
          </For>
        </ul>
      </nav>
    );
  };

  const TableOfContents: Component = function () {
    const [id, setId] = createSignal(
      (!isServer && location.hash.slice(1)) || undefined,
    );
    return (
      <div class="hidden xl:sticky xl:top-[4.5rem] xl:block xl:h-[calc(100vh-4.5rem)] xl:flex-none xl:overflow-y-auto xl:py-4 xl:pr-6">
        <nav aria-labelledby="on-this-page-title" class="w-56">
          <Show when={toc()}>
            <h2
              id="on-this-page-title"
              class="text-sm font-bold docs-text text-slate-900 "
            >
              <props.Rte data="tocLabel" />
            </h2>
          </Show>
          <ol class="mt-3 text-sm">
            <For each={toc()}>
              {(item) => (
                <li>
                  <h3>
                    <A
                      href={`${path.pathname}#${item.id}`}
                      onClick={() => setId(item.id)}
                      activeClass="text-blue-700 docs-link--active"
                      inactiveClass="text-slate-600 docs-text-muted"
                      class={clsx(
                        "docs-link block w-full font-sans transition font-normal rounded px-3 py-2 hover:bg-blue-50",
                        item.level === 3 && "pl-6",
                      )}
                    >
                      {item.text}
                    </A>
                  </h3>
                </li>
              )}
            </For>
          </ol>
        </nav>
      </div>
    );
  };

  return (
    <div
      class={css({
        "& .docs-text": {
          color: rgb(props.data.textColor),
        },
        "& .docs-text-muted": {
          color: rgb({ ...props.data.textColor, alpha: 0.6 }),
          "&:hover": {
            color: rgb({ ...props.data.textColor, alpha: 0.75 }),
          },
        },
        "& .docs-border": {
          borderColor: rgb(props.data.borderColor),
        },
        "& .docs-link": {
          "&:hover": {
            backgroundColor: rgb(props.data.activeButtonColor),
          },
          "&.docs-link--active": {
            color: rgb(props.data.activeTextColor),
          },
        },
        "& .docs-nav-toggle": {
          borderColor: rgb({ ...props.data.borderColor, alpha: 0.6 }),
          "&:hover,&.docs-nav-toggle--active": {
            borderColor: rgb(props.data.borderColor),
            backgroundColor: rgb(props.data.activeButtonColor),
          },
        },
        "& .docs-status": {
          color: rgb(props.data.statusColor),
          backgroundColor: rgb(props.data.statusBgColor),
        },
      })}
      {...props.elementProps}
    >
      {props.handles}
      <div class="relative flex flex-col lg:flex-row justify-center">
        <div class="hidden lg:relative lg:block lg:flex-none">
          <div class="overflow-x-hidden lg:sticky top-[61px] h-[calc(100vh-61px)] overflow-y-auto py-6 pl-6 pr-2">
            <Navigation />
          </div>
        </div>
        <Container class="min-w-0 mx-auto flex-auto px-5 xl:px-16 py-16 lg:max-w-4xl xl:max-w-5xl">
          <div class="lg:hidden flex justify-end mb-3">
            <Button.Root
              class={clsx(
                "docs-nav-toggle text-sm max-w-full font-bold flex items-center gap-3 border shadow-sm border-slate-200 px-3 py-2 rounded-lg hover:bg-slate-100 hover:text-slate-700",
                showToc() && "bg-slate-100 docs-text docs-nav-toggle--active",
                !showToc() && "docs-text-muted text-slate-500",
              )}
              aria-label="Show navigation"
              onClick={() => setShowToc(!showToc())}
            >
              <Richtext class="flex-1 min-w-0 truncate [&_*]:truncate">
                {pagination()?.current?.title || props.data.tocLabel}
              </Richtext>
              <MenuIcon width="1.7em" height="1.7em" />
            </Button.Root>
          </div>
          <Show when={showToc()}>
            <aside class="docs-border shadow lg:hidden border rounded-md border-slate-200 p-5 mb-7">
              <Navigation />
            </aside>
          </Show>
          <article ref={articleEl}>
            <props.Slot class="flex flex-col gap-10 lg:gap-12" name="default" />
          </article>
          <Pagination />
        </Container>
        <TableOfContents />
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      slots: { default: {} },
      data: {
        sections: listData({
          defaultItems: 2,
          itemTitle(item) {
            return item.title?.content;
          },
          itemData: {
            title: rteData({
              shortLorem: true,
              toolbar: {
                color: true,
                italic: true,
                underline: true,
                clean: true,
              },
            }),
            links: listData({
              defaultItems: 3,
              itemTitle(item) {
                return item.title?.content;
              },
              itemData: {
                title: rteData({
                  shortLorem: true,
                  toolbar: {
                    bold: true,
                    color: true,
                    italic: true,
                    underline: true,
                    clean: true,
                  },
                }),
                url: textData({ wide: true }),
                status: textData(),
              },
            }),
          },
        }),

        textColor: colorData({ alpha: false }),
        activeTextColor: colorData(),
        borderColor: colorData({ alpha: false }),
        activeButtonColor: colorData(),
        statusColor: colorData(),
        statusBgColor: colorData(),
        previousLabel: rteData({
          defaultValue: () => "Previous",
          toolbar: {},
        }),
        nextLabel: rteData({
          defaultValue: () => "Next",
          toolbar: {},
        }),
        tocLabel: rteData({
          title: () => "TableOfContentsLabel",
          defaultValue: () => "On this page",
          toolbar: {},
        }),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "n0u4th54fk63w7er8w22uyje";
