import { createBlueprint, BlueprintComponent } from "nitropage";
import { css, colorData, rgb } from "nitropage/data/color";
import { listData, sortItems } from "nitropage/data/list";
import { rteData, toolbarDefaults, toolbarFull } from "nitropage/data/rte";
import { createMemo, createSignal, For, onMount, Show } from "solid-js";
import { Accordion } from "@kobalte/core";
import PlusIcon from "~icons/tabler/plus";
import MinusIcon from "~icons/tabler/minus";
import { Style } from "solid-start";
import clsx from "clsx";

/*
 * The design of this blueprint is inspired by Meraki UI, check it out!
 * https://github.com/merakiui/merakiui
 */

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [sectionId, setSectionId] = createSignal<string | undefined>();
  const activeSectionId = createMemo(function () {
    if (sectionId() && props.data.sections?.byId[sectionId()!]) {
      return sectionId();
    }

    return props.data.sections?.ids[0];
  });
  const activeSection = createMemo(function () {
    return props.data.sections?.byId[activeSectionId()!];
  });

  const [mounted, setMounted] = createSignal(false);
  onMount(function () {
    setMounted(true);
  });

  return (
    <div
      class={clsx(
        css({
          "& .accordion__hr": {
            color: rgb(props.data.lineColor),
          },
          "& .accordion__text": {
            color: rgb(props.data.textColor),
          },
          "& .accordion__text--faded": {
            color: rgb({ ...props.data.textColor, alpha: 0.8 }),
          },
          "& .accordion__text--faded-xl": {
            color: rgb({ ...props.data.textColor, alpha: 0.7 }),
          },
          "& .accordion__highlight-text": {
            color: rgb(props.data.highlightColor),
          },
          // TODO: Move these to the css file when https://github.com/solidjs/solid-start/issues/306 is fixed
          "& .accordion__item[data-closed] .accordion__content-opened": {
            display: "none",
          },
          "& .accordion__item[data-expanded] .accordion__content-closed": {
            display: "none",
          },
        }),
        mounted() && "accordion--mounted",
      )}
      {...props.elementProps}
    >
      {props.handles}

      <Style>{
        /* css */ `
.accordion__content {
  animation-name: slideUp;
}
.accordion__content[data-expanded] {
  animation-name: slideDown;
}
.accordion--mounted .accordion__content {
  animation-duration: 300ms;
}
@keyframes slideDown {
  from {
    height: 0;
  }
  to {
    height: var(--kb-accordion-content-height);
  }
}
@keyframes slideUp {
  from {
    height: var(--kb-accordion-content-height);
  }
  to {
    height: 0;
  }
}
`
      }</Style>

      <div class="lg:flex gap-24 justify-between">
        <Show
          when={
            props.data.sections?.ids.length &&
            props.data.sections?.ids.length > 1
          }
        >
          <div class="">
            <props.Rte
              data="title"
              class="text-xl font-semibold accordion__text text-slate-800 block mb-4 lg:mb-8"
            />

            <div class="space-y-4">
              <For each={sortItems(props.data.sections)}>
                {([section, id]) => (
                  <button
                    onClick={() => setSectionId(id)}
                    class={clsx(
                      "block hover:underline",
                      activeSectionId() === id
                        ? "text-blue-700 accordion__highlight-text"
                        : "text-slate-600 accordion__text--faded-xl",
                    )}
                  >
                    <props.Rte path={(p) => p("sections").byId[id].title} />
                  </button>
                )}
              </For>
            </div>
          </div>
        </Show>
        <Show when={activeSection()} keyed>
          <Accordion.Root
            class="flex-1 mt-8 lg:mt-0"
            defaultValue={[activeSection()!.items!.ids[0]]}
            multiple={false}
          >
            <For each={sortItems(activeSection()!.items)}>
              {([item, id], idx) => (
                <Accordion.Item class="accordion__item" value={id}>
                  <Show when={idx() != 0}>
                    <hr class="accordion__hr my-8 border-current text-slate-200 " />
                  </Show>

                  <Accordion.Header>
                    <Accordion.Trigger
                      class={clsx("flex items-center", props.admin && "w-full")}
                    >
                      <MinusIcon class="accordion__content-opened flex-shrink-0 w-6 h-6 accordion__highlight-text text-blue-600" />
                      <PlusIcon class="accordion__content-closed flex-shrink-0 w-6 h-6 accordion__highlight-text text-blue-600" />

                      <props.Rte
                        component={"div"}
                        path={(p) =>
                          p("sections").byId[activeSectionId()!].items?.byId[id]
                            .question
                        }
                        class="flex-1 text-left mx-4 text-xl text-slate-700 accordion__text--faded "
                      />
                    </Accordion.Trigger>
                  </Accordion.Header>

                  <Accordion.Content
                    class={
                      "overflow-hidden focus-within:overflow-visible ease-in-out accordion__content"
                    }
                  >
                    <div class="flex mt-8 md:mx-10">
                      <span class="border border-current accordion__highlight-text text-blue-600"></span>

                      <props.Rte
                        component={"div"}
                        path={(p) =>
                          p("sections").byId[activeSectionId()!].items?.byId[id]
                            .answer
                        }
                        class="w-full max-w-3xl px-4 accordion__text--faded-xl text-slate-600 "
                      />
                    </div>
                  </Accordion.Content>
                </Accordion.Item>
              )}
            </For>
          </Accordion.Root>
        </Show>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      title: () => "FAQ",
      acceptParentType: "element",
      data: {
        title: rteData({
          defaultValue() {
            return "Table of Content";
          },
        }),
        sections: listData({
          defaultItems: 3,
          itemTitle(item) {
            return item.title?.content;
          },
          itemData: {
            title: rteData({
              defaultValue() {
                return "Lorem Ipsum";
              },
              toolbar: {
                ...toolbarDefaults,
                underline: false,
                link: false,
              },
            }),
            items: listData({
              defaultItems: 3,
              itemTitle(item) {
                return item.question?.content;
              },
              itemData: {
                question: rteData({
                  defaultValue() {
                    return "A question?";
                  },
                  toolbar: {
                    ...toolbarDefaults,
                    underline: false,
                    link: false,
                  },
                }),
                answer: rteData({
                  defaultValue() {
                    return "An Answer.";
                  },
                  toolbar: toolbarFull,
                }),
              },
            }),
          },
        }),
        textColor: colorData({ alpha: false }),
        highlightColor: colorData(),
        lineColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "fil2tlugy1pltj9006rv354c";
