import { createBlueprint, BlueprintComponent } from "nitropage";
import { css, colorData, rgb } from "nitropage/data/color";
import { listData, sortItems } from "nitropage/data/list";
import {
  MediaContainer,
  mediaData,
  PlaceholderImage,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { rteData, toolbarDefaults } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import { toggleData } from "nitropage/data/toggle";
import { createResize } from "nitropage/server";
import { createSignal, Show } from "solid-js";
import { createMemo, For } from "solid-js";
import server$ from "solid-start/server";
import Button from "~/components/button";
import CatIcon from "~icons/tabler/cat";
import { Transition, TransitionGroup } from "solid-transition-group";
import clsx from "clsx";
import classes from "./style.module.css";
import { buttonColorData, prefixedKeys } from "~/data/button";

/*
 * The design of this blueprint is inspired by Meraki UI, check it out!
 * https://github.com/merakiui/merakiui
 */

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [optimizeMedia] = useResize(
    server$(
      createResize({
        sizes: {
          xs: { width: 500, height: 500 },
          xl: { width: 680, height: 680 },
        },
      }),
    ),
  );

  const groups = createMemo(() => sortItems(props.data.groups));
  const [groupId, setGroupId] = createSignal<string | undefined>();
  const activeGroupId = createMemo(() => {
    let id = groupId();

    if (id && !props.data.groups?.byId[id]) {
      id = undefined;
    }

    return id || props.data.groups?.ids[0];
  });
  const group = createMemo(() => {
    if (!activeGroupId()) {
      return;
    }
    return props.data.groups?.byId[activeGroupId()!];
  });
  const [images] = useMedia(() => groups().flatMap((g) => g[0].image || []));
  const image = createMemo(function () {
    const id = group()?.image?.[0];
    return (id && images().byId[id]) || PlaceholderImage;
  });

  return (
    <div {...props.elementProps}>
      {props.handles}
      <Show when={groups().length > 1}>
        <div class="absolute z-10 flex items-center mb-16 gap-4 xl:gap-6 flex-wrap">
          <For each={groups()}>
            {([group, id]) => (
              <Button
                clickScale={!props.admin}
                textColor={[
                  rgb(props.data.buttonTextColor),
                  rgb(props.data.buttonTextColorHover),
                ]}
                bgColor={[
                  rgb(props.data.buttonBgColor),
                  rgb(props.data.buttonBgColorHover),
                ]}
                active={activeGroupId() === id}
                onClick={() => setGroupId(id)}
              >
                <props.Rte path={(p) => p("groups").byId[id].title} />
              </Button>
            )}
          </For>
        </div>
      </Show>
      <div class={clsx("lg:flex", groups().length === 1 && "lg:items-center")}>
        <Show when={group()}>
          <div
            class={clsx(
              "relative w-full lg:w-1/2 flex flex-col",
              groups().length > 1 && "py-24",
            )}
          >
            <div class="relative">
              <TransitionGroup
                name={classes["group"]}
                enterClass={classes["group-enter"]}
                exitActiveClass={classes["group-exit-active"]}
                exitToClass={classes["group-exit-to"]}
              >
                <Show when={group()} keyed>
                  {function (group) {
                    return (
                      <div
                        class={clsx(
                          css({
                            "& svg": {
                              color: rgb(props.data.iconColor),
                              width: "1.5rem",
                              height: "1.5rem",
                            },
                            "& .feature__title": {
                              color: rgb(props.data.textColor),
                            },
                            "& .feature__description": {
                              color: rgb({
                                ...props.data.textColor,
                                alpha: 0.8,
                              }),
                            },
                            "& .feature__icon": {
                              backgroundColor: rgb(
                                props.data.iconBackgroundColor,
                              ),
                            },
                          }),
                          classes["group"],
                          "w-full grid grid-cols-1 gap-8 xl:gap-16 md:grid-cols-2",
                        )}
                      >
                        <For each={sortItems(group.features)}>
                          {([feature, id]) => (
                            <div
                              class={clsx(
                                classes["group__feature"],
                                "space-y-3",
                              )}
                            >
                              <span class="feature__icon inline-block p-3 text-blue-600 bg-blue-100 rounded-xl">
                                <Show
                                  when={feature.iconSvg}
                                  fallback={<CatIcon />}
                                >
                                  <span innerHTML={feature.iconSvg} />
                                </Show>
                              </span>

                              <props.Rte
                                path={(p) =>
                                  p("groups").byId[activeGroupId()!].features
                                    ?.byId[id].title
                                }
                                component="h3"
                                class="feature__title text-xl font-semibold text-slate-700 capitalize"
                              />

                              <props.Rte
                                component="p"
                                path={(p) =>
                                  p("groups").byId[activeGroupId()!].features
                                    ?.byId[id].description
                                }
                                class="feature__description text-slate-500"
                              />
                            </div>
                          )}
                        </For>
                      </div>
                    );
                  }}
                </Show>
              </TransitionGroup>
            </div>
          </div>

          <div
            class="hidden relative lg:block lg:w-1/2 lg:pl-24 xl:pl-12"
            data-rounded={props.data.imageMask}
          >
            <div
              class={clsx(
                "relative aspect-square overflow-hidden",
                props.data.imageMask && "rounded-full drop-shadow-2xl",
              )}
            >
              <Transition
                name={classes["image"]}
                enterClass={classes["image-enter"]}
                enterActiveClass={classes["image-enter-active"]}
                exitActiveClass={classes["image-exit-active"]}
                exitToClass={classes["image-exit-to"]}
              >
                <Show when={image()} keyed>
                  <MediaContainer
                    media={image()}
                    placeholder={props.data.imageMask}
                    blurhash={groups().length === 1 && props.data.imageMask}
                    fadeDuration={1500}
                    sources={optimizeMedia(image(), [
                      { size: "xl", screen: 1280 },
                      { size: "xs" },
                    ])}
                    alt="An image about this feature"
                    class={clsx(
                      classes["image"],
                      "overflow-hidden",
                      props.data.imageMask && "rounded-full",
                    )}
                  />
                </Show>
              </Transition>
            </div>
          </div>
        </Show>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      acceptParentType: "element",
      title() {
        return "Feature Grid";
      },
      data: {
        groups: listData({
          defaultItems: 1,
          itemTitle(item) {
            return item.title?.content;
          },
          itemData: {
            title: rteData({
              shortLorem: true,
              toolbar: {
                ...toolbarDefaults,
                link: false,
              },
            }),
            image: mediaData({
              max: 1,
              type: "image",
            }),
            features: listData({
              defaultItems: 4,
              itemTitle(item) {
                return item.title?.content;
              },
              itemData: {
                title: rteData({
                  defaultValue() {
                    return "A feature";
                  },
                }),
                description: rteData({
                  lorem: true,
                }),
                iconSvg: textData({
                  wide: true,
                }),
              },
            }),
          },
        }),
        imageMask: toggleData({
          defaultValue: true,
        }),
        textColor: colorData({ alpha: false }),
        iconColor: colorData(),
        iconBackgroundColor: colorData(),
        ...buttonColorData({
          title: "Button",
          keys: prefixedKeys,
        }),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "vt5a4mjfbbbzn5d6y08pho35";
