import clsx from "clsx";
import { A, createBlueprint, BlueprintComponent } from "nitropage";
import { css, rgb } from "nitropage/css";
import { colorData } from "nitropage/data/color";
import { listData, sortItems } from "nitropage/data/list";
import {
  Image,
  mediaData,
  PlaceholderImage,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { textData } from "nitropage/data/text";
import { createResize } from "nitropage/server";
import { Component, createMemo, For, Show } from "solid-js";
import server$ from "solid-start/server";

/*
 * The design of this blueprint is inspired by Meraki UI, check it out!
 * https://github.com/merakiui/merakiui
 */

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [optimizeMedia] = useResize(
    server$(
      createResize({
        sizes: {
          md: { height: 70 },
        },
      }),
    ),
  );
  const items = createMemo(() => sortItems(props.data.items));

  const Item: Component<ReturnType<typeof items>[0][0]> = function (props) {
    const [image] = useMedia(() => props.image);
    const [imageDark] = useMedia(() => props.imageDark);
    const type = createMemo(() => {
      if (image().byOrder[0]) {
        return "image";
      }

      if (props.imageSvg) {
        return "svg";
      }

      if (props.title) {
        return "text";
      }

      return "image";
    });

    return (
      <A
        href={props.url}
        title={props.title}
        target="_blank"
        class={clsx(
          "flex-grow-0 flex-shrink-0 flex items-center justify-center hover:scale-110 transform-gpu transition-all ease-out duration-500",
          type() !== "text" && "max-w-[12rem]",
        )}
      >
        <span class="h-8 lg:h-10 xl:h-12 object-contain">
          <Show when={type() === "image"}>
            <Image
              class={clsx(
                "h-full w-auto object-contain",
                imageDark().byOrder[0] && "dark:hidden",
              )}
              alt={props.title || "Third-party logo"}
              sources={optimizeMedia(image().byOrder[0] || PlaceholderImage, [
                { size: "md" },
              ])}
            />
          </Show>
          <Show when={type() === "image" && imageDark().byOrder[0]}>
            <Image
              class={clsx("h-full w-auto object-contain hidden dark:block")}
              alt={props.title || "Third-party logo"}
              sources={optimizeMedia(
                imageDark().byOrder[0] || PlaceholderImage,
                [{ size: "md" }],
              )}
            />
          </Show>
          <Show when={type() === "svg"}>
            <span
              class="h-8 lg:h-10 xl:h-12 block"
              innerHTML={props.imageSvg}
            />
          </Show>
          <Show when={type() === "text"}>
            <div class="h-full flex items-center justify-center">
              <div class="lg:text-lg tracking-tight select-none font-semibold whitespace-nowrap">
                {props.title}
              </div>
            </div>
          </Show>
        </span>
      </A>
    );
  };

  return (
    <div
      class={css({
        color: rgb(props.data.color),
        "& svg": {
          width: "100%",
          height: "100%",
        },
      })}
      {...props.elementProps}
    >
      {props.handles}
      <div class="py-2 flex flex-col sm:flex-row flex-wrap items-center justify-evenly gap-x-16 lg:gap-x-28 gap-y-9 lg:gap-y-11">
        <For each={items()}>{([item]) => <Item {...item} />}</For>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      title() {
        return "Logo Grid";
      },
      acceptParentType: "element",
      data: {
        items: listData({
          defaultItems: 5,
          itemTitle(item) {
            return item.title?.content;
          },
          itemData: {
            title: textData(),
            url: textData(),
            image: mediaData({
              max: 1,
              type: "image",
            }),
            imageDark: mediaData({
              max: 1,
              type: "image",
            }),
            imageSvg: textData({
              multiline: true,
            }),
          },
        }),
        color: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "zuvec1ijvwpdq9pu19jpqylg";
