import { createBlueprint, BlueprintComponent } from "nitropage";
import {
  mediaData,
  useMedia,
  PlaceholderImage,
  useResize,
  MediaContainer,
} from "nitropage/data/media";
import { rteData } from "nitropage/data/rte";
import { Accessor, createMemo } from "solid-js";
import PostLayout from "../blog/postLayout";
import ImageMagnifier from "~/components/imageMagnifier";
import server$ from "solid-start/server";
import { createResize } from "nitropage/server";
import { selectData } from "nitropage/data/select";
import clsx from "clsx";
import { toggleData } from "nitropage/data/toggle";
import { Dynamic } from "solid-js/web";
import Fragment from "~/components/fragment";
import { colorClass, colorData } from "nitropage/data/color";

export const useImageResize = function () {
  const [optimizeMedia, _, { Source }] = useResize(
    server$(
      createResize({
        aspectRatios: ["16/9", "4/3", "1/1"],
        sizes: {
          xl: { width: 1520 },
          lg: { width: 800 },
          sm: { width: 400 },
        },
      }),
    ),
  );

  const createSources = (
    props: Accessor<{
      aspectRatio?: "" | (typeof Source)["aspectRatio"];
      fullWidth?: boolean;
    }>,
  ) => {
    return createMemo<(typeof Source)[]>(function () {
      const aspect = props().aspectRatio;
      const aspectRatio = aspect !== "" ? aspect : undefined;
      const common: (typeof Source)[] = [
        { size: "lg", screen: 600, aspectRatio },
        { size: "sm", aspectRatio },
      ];

      if (!props().fullWidth) {
        return common;
      }

      return [{ size: "xl", screen: 1024, aspectRatio }, ...common];
    });
  };

  return [optimizeMedia, createSources] as [
    typeof optimizeMedia,
    typeof createSources,
  ];
};

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [images] = useMedia(() => props.data.image, { dataUrl: true });
  const image = createMemo(function () {
    const maybeImage = images().byOrder[0];
    if (maybeImage) {
      return maybeImage;
    }

    return PlaceholderImage;
  });

  const [optimizeMedia, createSources] = useImageResize();
  const sources = createSources(() => ({
    aspectRatio: props.data.aspectRatio,
    fullWidth: props.data.fullWidth,
  }));

  return (
    <div {...props.elementProps}>
      {props.handles}

      <Dynamic component={props.data.fullWidth ? Fragment : PostLayout}>
        <ImageMagnifier
          url={image().url2x || image().url}
          width={image().width || undefined}
          height={image().height || undefined}
        >
          <MediaContainer
            media={image()}
            placeholder={props.data.placeholder}
            fadeDuration={2000}
            sources={optimizeMedia(image(), sources())}
            class={clsx(
              "media-outline",
              props.data.aspectRatio === "1/1" && "aspect-square",
              props.data.aspectRatio === "4/3" && "aspect-[4/3]",
              props.data.aspectRatio === "16/9" && "aspect-video",
            )}
          />
        </ImageMagnifier>
        <props.Rte
          class={clsx(
            "w-full inline-block text-slate-600 text-sm md:text-base text-center mt-1 md:mt-2",
            colorClass(props.data.textColor),
          )}
          data="description"
        />
      </Dynamic>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      acceptParentType: "element",
      data: {
        image: mediaData({
          type: "image",
          max: 1,
        }),
        description: rteData({
          title() {
            return "Description";
          },
          defaultValue() {
            return "Image Description";
          },
        }),
        textColor: colorData(),
        placeholder: toggleData({ defaultValue: true }),
        fullWidth: toggleData({ title: () => "Full Width" }),
        aspectRatio: selectData<"" | "1/1" | "4/3" | "16/9">({
          title: () => "Aspect Ratio",
          options: [
            { title: () => "None" },
            { value: "1/1" },
            { value: "4/3" },
            { value: "16/9" },
          ],
        }),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "ezmzzfo6uu2aeuqgbmez4pfi";
