import {
  createBlueprint,
  BlueprintComponent,
  useVisibilityObserver,
} from "nitropage";
import { MediaContainer, mediaData, useMedia } from "nitropage/data/media";
import { toggleData } from "nitropage/data/toggle";
import {
  createEffect,
  createMemo,
  createSignal,
  Show,
  untrack,
} from "solid-js";
import { Dynamic } from "solid-js/web";
import Fragment from "~/components/fragment";
import LoaderIcon from "~icons/tabler/loader-2";
import PlayIcon from "~icons/tabler/player-play-filled";
import PostLayout from "../blog/postLayout";
import clsx from "clsx";
import { selectData } from "nitropage/data/select";
import { useImageResize } from "./image.np";
import { colorData, css, rgb } from "nitropage/data/color";

const [loaded, setLoaded] = createSignal(false);
const [loading, setLoading] = createSignal(false);

const load = async function () {
  if (loaded() || loading()) {
    return;
  }

  setLoading(true);
  if (!import.meta.env.SSR) {
    await import("vidstack/player/styles/default/theme.css");
    await import("vidstack/player/styles/default/layouts/video.css");
    await import("vidstack/player");
    await import("vidstack/player/layouts");
    await import("vidstack/player/ui");
  }
  setLoading(false);
  setLoaded(true);
};

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [videos] = useMedia(() => props.data.video);
  const [thumbnails] = useMedia(() => props.data.thumbnail, { dataUrl: true });
  const [showPlayer, setShowPlayer] = createSignal(false);
  const [optimizeMedia, createSources] = useImageResize();
  const sources = createSources(() => ({
    aspectRatio: props.data.aspectRatio,
    fullWidth: props.data.fullWidth,
  }));
  let wrapperElement: HTMLDivElement = undefined!;
  const isVisible = useVisibilityObserver(
    () => props.data.autoload && wrapperElement,
  );
  const thumbnail = createMemo(() => thumbnails().byOrder?.[0]);

  createEffect(function () {
    if (loaded() || !props.data.autoload || !isVisible()) {
      return;
    }

    untrack(load);
  });

  return (
    <div
      class={clsx(
        props.admin && "min-h-[20px]",
        props.admin && !videos().byOrder[0] && "bg-slate-300",
      )}
      {...props.elementProps}
    >
      {props.handles}

      <Show when={videos().byOrder[0]}>
        <Dynamic component={props.data.fullWidth ? Fragment : PostLayout}>
          <div
            ref={wrapperElement}
            class={clsx(
              css({
                backgroundColor: rgb(props.data.backgroundColor),
                "--video-border-radius": "0.7rem",
                "--video-bg": "transparent",
                "--video-border": "1px solid transparent",
              }),
              "relative z-0 flex rounded-xl overflow-hidden media-outline",
              props.data.aspectRatio === "1/1" && "aspect-square",
              props.data.aspectRatio === "4/3" && "aspect-[4/3]",
              props.data.aspectRatio === "" && "aspect-video",
            )}
          >
            <Show
              when={loaded() && (showPlayer() || props.data.autoload)}
              fallback={
                <>
                  <Show when={thumbnail()}>
                    <MediaContainer
                      class="absolute inset-0 !-z-10"
                      fadeDuration={2000}
                      media={thumbnail()}
                      sources={optimizeMedia(thumbnail(), sources())}
                    />
                  </Show>
                  <button
                    aria-label="Play video"
                    class="group absolute w-full text-4xl hover:text-[2.5rem] duration-75 ease-out transition-all cursor-pointer inset-0 flex items-center justify-center"
                    onClick={() => {
                      setShowPlayer(true);
                      load();
                    }}
                  >
                    <div class="rounded-full bg-black/60 backdrop-blur p-4 text-white shadow-2xl">
                      <Show
                        when={showPlayer() && loading()}
                        fallback={<PlayIcon />}
                      >
                        <LoaderIcon class="animate-spin" />
                      </Show>
                    </div>
                  </button>
                </>
              }
            >
              <media-player
                ref={(el) => {
                  el.addEventListener("started", () => el.controls.hide(0));
                }}
                autoPlay={!props.data.autoload ? true : undefined}
                src={videos().byOrder[0].rawUrl}
              >
                <media-provider>
                  <Show when={props.data.autoload && thumbnail()}>
                    <media-poster
                      class={clsx(
                        css({ "& img": { objectFit: "cover" } }),
                        "vds-poster",
                      )}
                      src={optimizeMedia(thumbnail(), [sources()[0]])[0].url}
                    />
                  </Show>
                </media-provider>
                <media-video-layout></media-video-layout>
              </media-player>
            </Show>
          </div>
        </Dynamic>
      </Show>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      acceptParentType: "element",
      data: {
        video: mediaData({
          type: "video",
          max: 1,
        }),
        thumbnail: mediaData({
          type: "image",
          max: 1,
        }),
        fullWidth: toggleData({ title: () => "Full Width" }),
        aspectRatio: selectData<"" | "1/1" | "4/3">({
          title: () => "Aspect Ratio",
          options: [
            { value: "1/1" },
            { value: "4/3" },
            { title: () => "16/9" },
          ],
        }),
        autoload: toggleData(),
        backgroundColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "uhkonhxg2jcray627kltt8dk";
