import clsx from "clsx";
import { A, createBlueprint, BlueprintComponent, Richtext } from "nitropage";
import { colorData, rgb, css } from "nitropage/data/color";
import { fontCss, fontData } from "nitropage/data/font";
import { listData, sortItems } from "nitropage/data/list";
import { rteData } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import { children, createMemo, FlowComponent, For, Show } from "solid-js";
import { Container } from "~/components/container";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const columns = createMemo(function () {
    return (
      (props.data.columns?.ids.length || 0) +
      (!props.lax || props.slotElements("customColumn").length ? 1 : 0)
    );
  });

  const Column: FlowComponent<{
    id: string;
    links?: [{ text: string; url: string }, string][];
  }> = function (columnProps) {
    const child = children(() => columnProps.children);
    return (
      <nav>
        <ul class="flex flex-col gap-3">
          <Show when={child()}>
            <li class="font-bold mb-2">{child()}</li>
          </Show>
          <For each={columnProps.links}>
            {([item, id]) => {
              return (
                <li>
                  <A
                    class="column__link text-slate-700 hover:underline"
                    href={item.url}
                  >
                    <props.Rte
                      path={(d) =>
                        d("columns").byId[columnProps.id].links?.byId[id].text
                      }
                    />
                  </A>
                </li>
              );
            }}
          </For>
        </ul>
      </nav>
    );
  };

  return (
    <div
      class={
        "print:hidden " +
        css({
          backgroundColor: rgb(props.data.bgColor),
          color: rgb(props.data.titleColor),
          "& .footer__hr": {
            backgroundColor: rgb(props.data.lineColor),
          },
          "& .column__link": {
            color: rgb({ ...props.data.titleColor, alpha: 0.8 }),
            ":hover": {
              color: rgb(props.data.titleColor),
            },
          },
        })
      }
      {...props.elementProps}
    >
      {props.handles}
      <Container>
        <div class="py-32 text-base flex flex-col">
          <props.Slot name="header" optional />
          <Show when={props.slotElements("header").length}>
            <hr class="footer__hr h-px bg-slate-200 border-none my-11" />
          </Show>
          <div class="flex justify-between gap-8 flex-wrap">
            <div
              class={
                "w-full lg:w-auto lg:min-w-[30%] text-lg font-bold text-slate-800 " +
                css({
                  color: rgb(props.data.titleColor),
                  ...fontCss(props.data.titleFont),
                })
              }
            >
              <props.Rte data="title" />
            </div>
            <div
              class={clsx(
                "flex-1 grid gap-7",
                columns() === 2 && "lg:grid-cols-2",
                columns() > 2 && "lg:grid-cols-3",
              )}
            >
              <For each={sortItems(props.data.columns)}>
                {([column, id]) => (
                  <Column id={id} links={sortItems(column.links)}>
                    <props.Rte path={(p) => p("columns").byId[id].title} />
                  </Column>
                )}
              </For>
              <props.Slot
                class="flex flex-col gap-3 min-w-[20px]"
                name="customColumn"
                optional
              />
            </div>
          </div>
          <div class="mt-8 flex flex-col md:flex-row gap-6 justify-between flex-wrap">
            <div class="text-xs flex gap-1 items-center">
              <props.Rte data="copyrightPrefix" />
              <span>&copy {new Date().getFullYear()}</span>
              <props.Rte data="copyrightSuffix" />
            </div>
            <props.Slot
              class="flex-1 md:flex-auto flex flex-col gap-3 min-w-[20px]"
              name="social"
              optional
            />
          </div>
        </div>
      </Container>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      title() {
        return "Page Footer";
      },
      acceptParentType: "root",
      slots: { header: {}, customColumn: {}, social: {} },
      data: {
        title: rteData({
          defaultValue() {
            return "Your Logo";
          },
        }),
        titleColor: colorData({ alpha: false }),
        titleFont: fontData({
          weight: true,
          letterSpacing: true,
          italic: true,
        }),
        columns: listData({
          itemTitle(item) {
            return item.title?.content;
          },
          itemData: {
            title: rteData({
              shortLorem: true,
            }),
            links: listData({
              itemTitle(item) {
                return item.text?.content;
              },
              itemData: {
                url: textData({
                  title() {
                    return "Url";
                  },
                  wide: true,
                }),
                text: rteData({
                  title() {
                    return "Text";
                  },
                  defaultValue() {
                    return "A link";
                  },
                  toolbar: {
                    bold: true,
                    italic: true,
                    clean: true,
                    color: true,
                    underline: true,
                  },
                }),
              },
            }),
          },
        }),
        copyrightPrefix: rteData({
          defaultValue() {
            return "Your Logo";
          },
        }),
        copyrightSuffix: rteData({
          defaultValue() {
            return "- Made with Nitropage";
          },
        }),
        bgColor: colorData({
          title() {
            return "BackgroundColor";
          },
        }),
        lineColor: colorData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "b0ailno63q17afjl6wr83xz1";
