import { A, createBlueprint, BlueprintComponent, Richtext } from "nitropage";
import { css, colorData, rgb, CssColor } from "nitropage/data/color";
import { fontData, fontCss } from "nitropage/data/font";
import { listData, sortItems } from "nitropage/data/list";
import { rteData } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import {
  Component,
  createEffect,
  createMemo,
  createSignal,
  For,
  onCleanup,
  Show,
  useContext,
} from "solid-js";
import { Container } from "~/components/container";
import { spring, timeline } from "motion";
import {
  motion,
  Presence,
  createMotion,
  PresenceContext,
} from "solid-motionone";
import { Dialog } from "@kobalte/core";
import { Style } from "solid-start";
import { selectData } from "nitropage/data/select";
import { Image, mediaData, useMedia } from "nitropage/data/media";
import { createScrollPosition } from "@solid-primitives/scroll";
import { toggleData } from "nitropage/data/toggle";
import clsx from "clsx";

const MenuIcon: Component<{ open: boolean }> = function (props) {
  let menuLineTop: SVGLineElement = undefined!;
  let menuLineMiddle: SVGLineElement = undefined!;
  let menuLineBottom: SVGLineElement = undefined!;

  let initial = true;
  createEffect(function () {
    if (props.open) {
      timeline(
        [
          [menuLineTop, { y: [0, 25] }, { easing: "ease-out" }],
          [menuLineBottom, { y: -25 }, { easing: "ease-out", at: 0 }],
          [menuLineMiddle, { opacity: [1, 0] }, { at: 0 }],
          "rotate",
          [
            menuLineTop,
            { rotate: [0, 45] },
            { easing: spring({ stiffness: 300 }) },
          ],
          [
            menuLineBottom,
            { rotate: [0, -45] },
            { easing: spring({ stiffness: 300 }), at: "rotate" },
          ],
        ],
        { duration: 0.7 },
      );

      return;
    }

    if (initial) {
      initial = false;
      return;
    }

    timeline(
      [
        [menuLineTop, { rotate: 0 }, { easing: "ease-out" }],
        [menuLineBottom, { rotate: 0 }, { easing: "ease-out", at: 0 }],
        "move",
        [menuLineMiddle, { opacity: [0.1, 1] }, { duration: 0.4 }],
        [
          menuLineTop,
          { y: [25, 0] },
          { easing: spring({ stiffness: 300 }), at: "move" },
        ],
        [
          menuLineBottom,
          { y: [-25, 0] },
          { easing: spring({ stiffness: 300 }), at: "move" },
        ],
      ],
      { duration: 1 },
    );
  });

  return (
    <svg class="w-10" stroke="currentColor" viewBox="0 0 100 100">
      <line
        class={css({
          transformBox: "fill-box",
          transformOrigin: "center",
        })}
        ref={menuLineTop}
        x1={20}
        x2={80}
        y1={25}
        y2={25}
        stroke-width={10}
        stroke-linecap="round"
      />
      <line
        ref={menuLineMiddle}
        x1={20}
        x2={80}
        y1={50}
        y2={50}
        stroke-width={10}
        stroke-linecap="round"
      />
      <line
        class={css({
          transformBox: "fill-box",
          transformOrigin: "center",
        })}
        ref={menuLineBottom}
        x1={20}
        x2={80}
        y1={75}
        y2={75}
        stroke-width={10}
        stroke-linecap="round"
      />
    </svg>
  );
};

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [mobileNavState, setMobileNavState] = createSignal<"close" | "open">(
    "close",
  );
  let contentEl: HTMLDivElement = undefined!;
  const mobileNavMotion = (el: HTMLElement) => {
    createMotion(
      el,
      {
        initial: {
          opacity: 0,
          width: "97%",
          clipPath: "circle(1% at 100% 0)",
        },
        animate: {
          opacity: 1,
          width: "100%",
          clipPath: ["circle(10% at 100% 0)", "circle(150% at 100% 0%)"],
          transition: {
            duration: 0.7,
            easing: "ease-out",
          },
        },
        exit: {
          opacity: 0,
          width: "97%",
          clipPath: "circle(1% at 100% 0%)",
          transition: {
            duration: 1.4,
            easing: "ease-in-out",
          },
        },
        transition: { duration: 5, easing: "ease-out" },
      },
      useContext(PresenceContext),
    );
  };
  const [mobileNavTrigger, setMobileNavTrigger] = createSignal<HTMLElement>();
  let element: HTMLDivElement = undefined!;

  const scrolling = createMemo(
    () => !["", "relative"].includes(props.data.position),
  );
  const [sticky, setSticky] = createSignal(false);
  const windowScroll = createScrollPosition();
  let observer: IntersectionObserver | undefined;
  createEffect(function () {
    if (!scrolling()) {
      if (observer) {
        observer.disconnect();
        observer = undefined;
      }

      return;
    }

    if (observer) {
      return;
    }

    observer = new IntersectionObserver(
      ([e]) => setSticky(e.intersectionRatio < 1),
      { threshold: [1] },
    );
    observer.observe(element);
  });
  onCleanup(function () {
    observer?.disconnect();
  });
  const shrink = createMemo(function () {
    return sticky() && windowScroll.y > 0;
  });

  const [logos] = useMedia(() => props.data.logo);
  const logo = createMemo(function () {
    return logos().byOrder[0];
  });

  const stickyColor = function (color?: CssColor) {
    if (!color?.value || (!shrink() && props.lax)) {
      return;
    }

    return color;
  };

  const overlayColor = createMemo(() => ({
    value: "255, 255, 255",
    alpha: props.data.glassyOverlay ? 0.7 : undefined,
    ...props.data.overlayColor,
  }));

  const style = createMemo(function () {
    return css({
      backgroundColor:
        mobileNavState() !== "open"
          ? rgb(
              stickyColor(props.data.backgroundColorSticky) ||
                props.data.backgroundColor,
            )
          : rgb({ ...overlayColor(), alpha: 0.3 }),
    });
  });

  const logoClass = createMemo(() =>
    clsx(
      "flex h-9 lg:h-12 w-auto !transition-all !duration-700",
      shrink() && "!h-9",
    ),
  );

  const linkColor = createMemo(
    () => stickyColor(props.data.linkColorSticky) || props.data.linkColor,
  );

  return (
    <div
      ref={element}
      class={clsx(
        "text-slate-900 h-14 lg:h-20 box-content print:hidden pointer-events-none",

        props.lax && "left-0 right-0",
        props.data.position === "" && props.lax && "absolute",
        props.data.position === "relative" && "relative",
        scrolling() && "-top-[1px] pt-px",
        props.data.position === "fixed" && props.lax && "fixed",
        // the -1px shift applies for sticky only if its sticked, only then we need to apply the 1px top padding
        props.data.position === "sticky" && props.lax && !shrink() && "!pt-0",
        props.data.position === "sticky" && props.lax && "sticky",
        props.lax ? "z-40" : "bg-black/10",
      )}
      {...props.elementProps}
    >
      <Show when={props.admin}>
        <Style>{"body { padding-right: 0px !important; }"}</Style>
      </Show>
      <div
        class={clsx(
          "h-14 lg:h-20",
          shrink() && "shadow !h-14",
          props.lax && "pointer-events-auto",
          mobileNavState() === "open" && "shadow-none",
          style(),
          props.lax &&
            (scrolling() || !!props.data.backgroundColor) &&
            "backdrop-blur-lg duration-1000 transition-all",
          mobileNavState() === "open" && "!backdrop-blur",
          mobileNavState() === "open" &&
            props.data.position === "sticky" &&
            "top-0 left-0 right-0 !fixed",
        )}
      >
        {props.handles}
        <Container class="h-full">
          <div class="h-full gap-10 md:gap-12 lg:gap-16 flex justify-between items-center">
            <A
              href="/"
              aria-label="Home logo"
              class={
                css({
                  color: rgb(
                    stickyColor(props.data.titleColorSticky) ||
                      props.data.titleColor,
                  ),
                  opacity: mobileNavState() === "open" ? 0 : 1,
                  translate: mobileNavState() === "open" ? "-20px" : 0,
                  pointerEvents:
                    mobileNavState() === "open" ? "none" : undefined,
                  ...fontCss(props.data.titleFont),
                  "& svg": {
                    width: "100%",
                    height: "100%",
                  },
                }) + " font-bold transition-all duration-300"
              }
            >
              <Show when={logo()} keyed>
                {(logo) => {
                  return (
                    <Image
                      class={logoClass()}
                      preload
                      alt={logo.description || "Logo"}
                      sources={[
                        {
                          url: logo.url,
                          width: logo.width,
                          height: logo.height,
                        },
                      ]}
                    />
                  );
                }}
              </Show>
              <Show when={!logo() && props.data.logoSvg}>
                <span class={logoClass()} innerHTML={props.data.logoSvg} />
              </Show>
              <Show when={!logo() && !props.data.logoSvg}>
                <props.Rte data="title" />
              </Show>
            </A>

            <nav
              class={
                css({
                  display: mobileNavState() === "open" ? "none" : undefined,
                  color: rgb(linkColor()),
                  "& > :hover": props.data.linkColorHover
                    ? { color: rgb(props.data.linkColorHover) }
                    : undefined,
                }) + " hidden lg:flex gap-10 xl:gap-12 items-center justify-end"
              }
            >
              <For each={sortItems(props.data.links)}>
                {([link, id]) => {
                  return (
                    <A
                      end={true}
                      class="transition-colors ease-out hover:underline"
                      activeClass="underline"
                      href={link.url!}
                    >
                      <props.Rte path={(p) => p("links").byId[id].text} />
                    </A>
                  );
                }}
              </For>
              <props.Slot
                class="gap-12 items-center justify-end"
                name="cta"
                optional
              />
            </nav>
            <nav class="flex shrink-0 lg:hidden gap-4 md:gap-6 lg:gap-8 items-center justify-end">
              <props.Slot
                class={clsx(
                  "gap-12 items-center justify-end opacity-100 transition-opacity",
                  mobileNavState() === "open" &&
                    "!opacity-0 pointer-events-none",
                )}
                name="cta"
                optional
              />
              <Show when={props.data.links?.ids.length}>
                <Dialog.Root
                  modal={false}
                  preventScroll={true}
                  forceMount={true}
                  open={mobileNavState() == "open"}
                  onOpenChange={(isOpen) => {
                    setMobileNavState(isOpen ? "open" : "close");
                  }}
                >
                  <Dialog.Trigger
                    ref={setMobileNavTrigger}
                    tabIndex={0}
                    onFocusOut={(evt) => {
                      if (mobileNavState() != "open") {
                        return;
                      }

                      evt.preventDefault();
                      if (contentEl.firstElementChild) {
                        (
                          contentEl.firstElementChild as HTMLAnchorElement
                        ).focus();
                      } else {
                        contentEl.focus();
                      }
                    }}
                    aria-label="menu"
                    class={clsx(
                      "transition-colors duration-500 relative z-50 w-10",
                      mobileNavState() == "open" && "block",
                      css({
                        color: rgb(
                          mobileNavState() == "open"
                            ? props.data.overlayLinkColor || linkColor()
                            : linkColor(),
                        ),
                      }),
                    )}
                  >
                    <MenuIcon open={mobileNavState() == "open"} />
                  </Dialog.Trigger>
                  <Dialog.Portal>
                    <Presence exitBeforeEnter>
                      <Show when={mobileNavState() == "open"}>
                        <Dialog.Content
                          as="nav"
                          id="mobile-nav"
                          ref={mobileNavMotion}
                          class={clsx(
                            css({
                              color: rgb(
                                props.data.overlayLinkColor ||
                                  props.data.linkColor,
                              ),
                              backgroundColor: rgb(overlayColor()),
                              "& *": {
                                overflow: "hidden",
                                textOverflow: "ellipsis",
                              },
                            }),
                            "fixed z-30 w-full h-full top-0 right-0 bottom-0 overflow-y-auto overflow-x-hidden ",
                            props.data.glassyOverlay && "backdrop-blur-3xl",
                          )}
                        >
                          <div
                            ref={contentEl}
                            class="w-screen min-h-full flex flex-col gap-6 justify-center items-center text-3xl p-9 pt-20 text-ellipsis"
                          >
                            <For each={sortItems(props.data.links)}>
                              {([link]) => {
                                return (
                                  <A
                                    class="max-w-full text-center hover:underline"
                                    activeClass="underline"
                                    onClick={() => {
                                      requestAnimationFrame(function () {
                                        setMobileNavState("close");
                                      });
                                    }}
                                    onFocusOut={(evt) => {
                                      const nextElement = evt.relatedTarget as
                                        | HTMLElement
                                        | undefined;
                                      if (nextElement === mobileNavTrigger()) {
                                        return;
                                      }
                                      if (
                                        nextElement &&
                                        nextElement.dataset.leave !== "" &&
                                        contentEl.contains(nextElement)
                                      ) {
                                        return;
                                      }

                                      mobileNavTrigger()?.focus();
                                    }}
                                    end={true}
                                    href={link.url!}
                                  >
                                    <Richtext>{link.text}</Richtext>
                                  </A>
                                );
                              }}
                            </For>
                            <button
                              class="absolute"
                              data-leave
                              aria-hidden={true}
                            />
                          </div>
                        </Dialog.Content>
                      </Show>
                    </Presence>
                  </Dialog.Portal>
                </Dialog.Root>
              </Show>
            </nav>
          </div>
        </Container>
      </div>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    motion;
    const commonColor = colorData();
    return {
      title() {
        return "Page Header";
      },
      acceptParentType: "root",
      slots: {
        cta: {},
      },
      data: {
        titleFont: fontData({
          weight: true,
          size: true,
          letterSpacing: true,
          italic: true,
        }),
        logo: mediaData({
          max: 1,
          type: "image",
        }),
        logoSvg: textData({
          multiline: true,
        }),
        title: rteData({
          defaultValue() {
            return "Your Logo";
          },
        }),
        links: listData({
          title() {
            return "Links";
          },
          defaultItems: 3,
          itemData: {
            url: textData({
              title() {
                return "Url";
              },
              wide: true,
            }),
            text: rteData({
              title() {
                return "Text";
              },
              shortLorem: true,
              toolbar: {
                bold: true,
                italic: true,
                clean: true,
                color: true,
                underline: true,
              },
            }),
          },
          itemTitle(i) {
            return i.text?.content;
          },
        }),
        position: selectData({
          title() {
            return "Position";
          },
          options: [
            { title: () => "Absolute" },
            { value: "relative" },
            { value: "fixed" },
            { value: "sticky" },
          ],
        }),
        titleColor: commonColor,
        titleColorSticky: commonColor,
        linkColor: commonColor,
        linkColorHover: commonColor,
        linkColorSticky: commonColor,
        backgroundColor: commonColor,
        backgroundColorSticky: commonColor,
        overlayColor: colorData({
          alpha: false,
        }),
        overlayLinkColor: commonColor,
        glassyOverlay: toggleData({
          defaultValue: true,
        }),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const id = "pynj0gs5qlwztg60j596bkgm";
