import clsx from "clsx";
import { createBlueprint, BlueprintComponent } from "nitropage";
import { colorData, rgb, css } from "nitropage/data/color";
import { fontCss, fontData } from "nitropage/data/font";
import {
  mediaData,
  MediaImage,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { numberData } from "nitropage/data/number";
import { rteData, toolbarDefaults } from "nitropage/data/rte";
import { selectData } from "nitropage/data/select";
import { toggleData } from "nitropage/data/toggle";
import { createResize } from "nitropage/server";
import { For, Show } from "solid-js";
import server$ from "solid-start/server";
import { Container } from "~/components/container";
import Heading from "~/components/heading";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [optimizeMedia] = useResize(
    server$(
      createResize({
        sizes: {
          xs: { width: 500 },
          sm: { width: 600 },
          md: { width: 800 },
          lg: { width: 1100 },
          xl: { width: 1200 },
          xxl: { width: 1450 },
        },
      }),
    ),
  );

  const [sideImages] = useMedia(() => props.data.sideImage);
  return (
    <div
      class={
        "min-h-screen flex flex-col justify-center pt-28 pb-24 overflow-x-clip"
      }
      {...props.elementProps}
    >
      {props.handles}
      <Container>
        <div
          class={clsx(
            "h-full flex flex-wrap items-center",
            props.data.sideImagePosition == "" && "gap-x-5 xl:gap-x-10",
          )}
        >
          <div
            class={clsx(
              "flex flex-col w-full",
              props.lax && "relative z-10",
              sideImages().byOrder.length ? "lg:w-7/12" : "lg:w-8/12",
            )}
          >
            <Heading
              level={1}
              size="xl"
              fullWidth
              class={css({
                color: rgb(props.data.titleColor),
                ...fontCss(props.data.titleFont),
              })}
            >
              <props.Rte data="title" />
            </Heading>
            <Show when={props.admin || props.data.description}>
              <div
                class={
                  "mt-12 text-slate-700 text-lg lg:text-xl " +
                  css({
                    color: rgb(props.data.descriptionColor),
                    ...fontCss(props.data.descriptionFont),
                  })
                }
              >
                <props.Rte class="inline-block md:w-5/6" data="description" />
              </div>
            </Show>
            <props.Slot
              name="default"
              editorPadding={true}
              class="flex flex-col gap-4 mt-8 md:w-5/6"
            />
          </div>
          <div class="relative w-full lg:flex-1 drop-shadow-2xl">
            <For each={sideImages().byOrder}>
              {(media) => {
                return (
                  <MediaImage
                    media={media}
                    preload={true}
                    alt="Side image"
                    sources={optimizeMedia(media, [
                      { size: "xxl", screen: 1536 },
                      {
                        size:
                          props.data.sideImagePosition === "" ? "xl" : "xxl",
                        screen: 1280,
                      },
                      {
                        size: props.data.sideImagePosition === "" ? "lg" : "xl",
                        screen: 1024,
                      },
                      { size: "md", screen: 768 },
                      { size: "sm", screen: 500 },
                      { size: "xs" },
                    ])}
                    class={clsx(
                      css({
                        "@media (min-width: 1024px)": {
                          opacity: props.data.sideImageOpacity / 100,
                          maskImage: props.data.sideImageFade
                            ? "linear-gradient(to right, transparent 0%, #000D 30%, black 45%)"
                            : undefined,
                        },
                      }),
                      "w-full max-w-none mt-28 lg:mt-0 pointer-events-none origin-left",
                      props.data.sideImagePosition === "" &&
                        props.lax &&
                        "max-h-[70vh] lg:scale-125 !origin-center",
                      props.data.sideImagePosition === "expand" &&
                        props.lax &&
                        "lg:!absolute lg:w-auto lg:-translate-y-1/2 lg:-translate-x-52 xl:-translate-x-56 2xl:-translate-x-64 lg:h-[50vh] xl:h-[65vh]",
                      props.data.sideImagePosition === "expandXl" &&
                        props.lax &&
                        "lg:!absolute lg:w-auto lg:-translate-y-1/2 lg:-translate-x-52 xl:-translate-x-56 2xl:-translate-x-64 lg:h-[70vh] xl:h-[90vh]",
                    )}
                  />
                );
              }}
            </For>
          </div>
        </div>
      </Container>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      title: () => "Page Intro",
      acceptParentType: "element",
      slots: {
        default: {},
      },
      data: {
        title: rteData({
          shortLorem: true,
          toolbar: {
            ...toolbarDefaults,
            link: false,
            background: true,
          },
        }),
        titleColor: colorData(),
        titleFont: fontData({
          weight: true,
          letterSpacing: true,
          lineHeight: true,
          italic: true,
          size: true,
        }),
        description: rteData({
          lorem: true,
        }),
        descriptionColor: colorData(),
        descriptionFont: fontData({
          weight: true,
          letterSpacing: true,
          lineHeight: true,
          italic: true,
          size: true,
        }),
        sideImage: mediaData({
          max: 1,
          type: "image",
        }),
        sideImageFade: toggleData(),
        sideImageOpacity: numberData({ defaultValue: 100, min: 0, max: 100 }),
        sideImagePosition: selectData({
          options: [
            {
              title() {
                return "Shrink";
              },
            },
            {
              title() {
                return "Expand";
              },
              value: "expand",
            },
            {
              title() {
                return "Expand XL";
              },
              value: "expandXl",
            },
          ],
        }),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "ae8f3r696oky0tpbdlu1accp";
