import clsx from "clsx";
import { BaseBlueprint, createBlueprint, BlueprintComponent } from "nitropage";
import { colorData, rgb, css } from "nitropage/data/color";
import {
  ImageSources,
  MediaContainer,
  mediaData,
  useMedia,
  useResize,
} from "nitropage/data/media";
import { numberData } from "nitropage/data/number";
import { selectData } from "nitropage/data/select";
import { toggleData } from "nitropage/data/toggle";
import { createResize } from "nitropage/server";
import { createMemo, For, Show } from "solid-js";
import server$ from "solid-start/server";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [backgroundImages] = useMedia(() => props.data.backgroundImage, {
    dataUrl: true,
  });

  const [optimizeMediaScreen] = useResize(
    server$(
      createResize({
        sizes: {
          xs: { width: 500, height: 800 },
          md: { width: 1500, height: 800 },
          lg: { width: 1920, height: 1080 },
        },
      }),
    ),
  );

  const [optimizeMediaBanner] = useResize(
    server$(
      createResize({
        sizes: {
          xs: { width: 700, height: 300 },
          md: { width: 1500, height: 600 },
          lg: { width: 1920, height: 720 },
        },
      }),
    ),
  );

  const clipPath = function () {
    if (props.data.topMask !== "diag" && props.data.bottomMask !== "diag") {
      return;
    }

    if (!props.lax) {
      return;
    }

    const topY = props.data.topMask === "diag" ? "var(--m-h)" : 0;
    const bottomY =
      props.data.bottomMask === "diag" ? "calc(100% - var(--m-h))" : "100%";
    return `polygon(0 ${topY},100% 0,100% ${bottomY},0 100%)`;
  };

  // https://css-generators.com/section-divider/
  // https://css-generators.com/custom-borders/
  // https://css-tricks.com/how-to-create-wavy-shapes-patterns-in-css/
  const mask = function () {
    let masks: string[] = [];
    let topMask: string | undefined = undefined;
    let bottomMask: string | undefined = undefined;

    if (!props.lax) {
      return;
    }

    if (props.data.topMask === "wavy") {
      topMask = `radial-gradient(var(--m-h),#0000 98%,#000) 0 var(--m-h-inv)/var(--m-h-r) var(--m-h-twice) round no-repeat`;
      masks.push(topMask);
    }
    if (props.data.bottomMask === "wavy") {
      bottomMask = `radial-gradient(var(--m-h),#0000 98%,#000) 0 calc(100% + var(--m-h))/var(--m-h-r) var(--m-h-twice)  round no-repeat`;
      masks.push(bottomMask);
    }

    if (props.data.topMask === "wavy2") {
      topMask = `radial-gradient(var(--m-h),#000 98%,#0000) top/var(--m-h-r) var(--m-h-twice) repeat-x`;
      masks.push(topMask);
    }
    if (props.data.bottomMask === "wavy2") {
      bottomMask = `radial-gradient(var(--m-h),#000 98%,#0000) bottom/var(--m-h-r) var(--m-h-twice) repeat-x`;
      masks.push(bottomMask);
    }

    if (props.data.topMask === "wavy3") {
      topMask = `radial-gradient(var(--m-h-half) at 25% 0,#0000 98%,#000) 50% var(--m-h-half)/var(--m-h-twice) 51% repeat-x,
      radial-gradient(var(--m-h-half) at 75% 50% ,#000 99%,#0000 101%) top/var(--m-h-twice) var(--m-h) repeat-x`;
      masks.push(topMask);
    }
    if (props.data.bottomMask === "wavy3") {
      bottomMask = `radial-gradient(var(--m-h-half) at 75% 100%,#0000 98%,#000) 50% calc(100% - var(--m-h-half))/var(--m-h-twice) 51% repeat-x,
      radial-gradient(var(--m-h-half) at 25% 50% ,#000 99%,#0000 101%) bottom/var(--m-h-twice) var(--m-h) repeat-x`;
      masks.push(bottomMask);
    }

    if (masks.length === 0) {
      return undefined;
    }

    const fillerPosition =
      masks.length === 2 ? "50%" : topMask ? "bottom" : "top";
    const fillerHeight =
      masks.length === 2
        ? "calc(100% - var(--m-h-twice))"
        : "calc(100% - var(--m-h))";
    const filler = `linear-gradient(#000 0 0) ${fillerPosition} / calc(100% - 0px) ${fillerHeight} no-repeat`;

    masks.push(filler);

    return masks.join(",");
  };

  const sectionStyle = createMemo(function () {
    return css({
      zIndex: props.data.zIndex + "",
      "--m-h": "clamp(2rem, 5vw, 4rem)",
      "--m-h-r": "calc(var(--m-h) * 1.85)",
      "--m-h-inv": "calc(var(--m-h) * -1)",
      "--m-h-half": "calc(var(--m-h) / 2)",
      "--m-h-twice": "calc(var(--m-h) * 2)",
      marginTop: props.data.topMask && props.lax ? "var(--m-h-inv)" : undefined,
      marginBottom:
        props.data.bottomMask && props.lax ? "var(--m-h-inv)" : undefined,
    });
  });

  const opacity = () => props.data.backgroundImageOpacity / 100;

  const maskElementStyle = createMemo(function () {
    return css({
      backgroundColor: rgb(props.data.backgroundColor),
      clipPath: clipPath(),
      mask: mask(),
      // TODO: Remove these or add as setting
      // paddingTop: props.lax && props.data.topMask ? "var(--m-h)" : undefined,
      // paddingBottom: props.lax && props.data.bottomMask ? "var(--m-h)" : undefined,
    });
  });

  return (
    <section
      class={clsx(
        sectionStyle(),
        props.data.shadow === "sm" && "drop-shadow-md",
        props.data.shadow === "md" && "drop-shadow-xl",
        props.data.shadow === "lg" && "drop-shadow-2xl",
        !props.lax && "py-4 !drop-shadow-none",
      )}
      {...props.elementProps}
    >
      <div
        class={clsx(
          maskElementStyle(),
          "z-0",
          !props.lax && "min-h-[20px]",
          props.preview && "bg-blue-200",
        )}
      >
        <Show
          when={!props.data.backgroundImageRepeat}
          fallback={
            <div
              class={
                "absolute inset-0 " +
                css({
                  backgroundImage:
                    props.data.backgroundImageRepeat &&
                    backgroundImages().byOrder[0]
                      ? `url("${backgroundImages().byOrder[0].url}")`
                      : undefined,
                  backgroundRepeat: props.data.backgroundImageRepeat
                    ? "repeat"
                    : undefined,
                  backgroundSize:
                    props.data.repeatSize != null
                      ? props.data.repeatSize + "rem"
                      : undefined,
                  opacity: opacity(),
                })
              }
            />
          }
        >
          <For each={backgroundImages().byOrder}>
            {(media) => {
              const sources = createMemo<ImageSources | undefined>(function () {
                if (props.data.backgroundImageRatio === "screen") {
                  return optimizeMediaScreen(media, [
                    { size: "lg", screen: 1200 },
                    { size: "md", screen: 900 },
                    { size: "xs" },
                  ]);
                }

                if (props.data.backgroundImageRatio === "banner") {
                  return optimizeMediaBanner(media, [
                    { size: "lg", screen: 1200 },
                    { size: "md", screen: 900 },
                    { size: "xs" },
                  ]);
                }

                return;
              });

              return (
                <MediaContainer
                  media={
                    props.data.backgroundColor
                      ? { ...media, color: null }
                      : media
                  }
                  height={false}
                  width={false}
                  fit="cover"
                  placeholder
                  class={css({ opacity: opacity() }) + " !absolute inset-0"}
                  preload={props.data.backgroundImagePreload}
                  sources={sources()}
                />
              );
            }}
          </For>
        </Show>
        <Show when={props.data.overlayColor}>
          <div
            class={clsx(
              css({
                backgroundColor: rgb(props.data.overlayColor),
                mixBlendMode:
                  props.data.overlayBlendMode !== ""
                    ? props.data.overlayBlendMode
                    : undefined,
              }),
              "absolute inset-0",
            )}
          />
        </Show>
        {props.handles}
        <props.Slot
          optional
          class={clsx("flex flex-col justify-center", props.lax && "relative")}
          name="default"
        />
      </div>
    </section>
  );
};

export const blueprint = createBlueprint(
  () => {
    const blendModes = [
      "multiply",
      "screen",
      "overlay",
      "darken",
      "lighten",
      "color-dodge",
      "color-burn",
      "hard-light",
      "soft-light",
      "difference",
      "exclusion",
      "hue",
      "saturation",
      "color",
      "luminosity",
      // TODO: upstream css types miss plus-darker at the moment
      "plus-darker" as never,
      "plus-lighter",
    ] as const;

    return {
      title: () => "Page Section",
      acceptParentType: "root",
      data: {
        backgroundColor: colorData({
          title: () => "Background Color",
        }),
        backgroundImage: mediaData({
          max: 1,
          type: "image",
          title: () => "Background Image",
        }),
        backgroundImagePreload: toggleData({
          title: () => "Preload Background Image",
        }),
        backgroundImageOpacity: numberData({
          defaultValue: 100,
          title: () => "Background Image Opacity",
        }),
        backgroundImageRatio: selectData<"" | "screen" | "banner" | "3">({
          title() {
            return "Background Image Aspect Ratio";
          },
          options: [
            { title: () => "None" },
            { value: "screen", title: () => "Fullscreen" },
            { value: "banner", title: () => "Banner (3/1)" },
          ],
        }),
        backgroundImageRepeat: toggleData(),
        repeatSize: numberData({
          min: 0,
          max: 48,
          step: 0.125,
          optional: true,
        }),
        overlayColor: colorData(),
        overlayBlendMode: selectData<"" | (typeof blendModes)[any]>({
          options: [
            { title: () => "Normal" },
            ...blendModes.map((m) => ({ value: m })),
          ],
        }),
        zIndex: numberData({
          max: 100,
          title() {
            return "Layer";
          },
        }),
        shadow: selectData<"" | "sm" | "md" | "lg">({
          title() {
            return "Shadow";
          },
          options: [
            {
              title() {
                return "None";
              },
            },
            {
              title() {
                return "Small";
              },
              value: "sm",
            },
            {
              title() {
                return "Middle";
              },
              value: "md",
            },
            {
              title() {
                return "Large";
              },
              value: "lg",
            },
          ],
        }),
        topMask: selectData<"" | "diag" | "wavy" | "wavy2" | "wavy3">({
          title() {
            return "Top Mask";
          },
          options: [
            { title: () => "None" },
            { value: "diag", title: () => "Diagonal" },
            { value: "wavy", title: () => "Wavy" },
            { value: "wavy2", title: () => "Wavy 2" },
            { value: "wavy3", title: () => "Wavy 3" },
          ],
        }),
        bottomMask: selectData<"" | "diag" | "wavy" | "wavy2" | "wavy3">({
          title() {
            return "Bottom Mask";
          },
          options: [
            { title: () => "None" },
            { value: "diag", title: () => "Diagonal" },
            { value: "wavy", title: () => "Wavy" },
            { value: "wavy2", title: () => "Wavy 2" },
            { value: "wavy3", title: () => "Wavy 3" },
          ],
        }),
      },
      slots: {
        default: {},
      },
    } satisfies BaseBlueprint;
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "usedfsqfwa4g4aedh2wjkqkp";
