import clsx from "clsx";
import { A, createBlueprint, BlueprintComponent } from "nitropage";
import { css, colorData, rgb } from "nitropage/data/color";
import { fontCss, fontData } from "nitropage/data/font";
import { listData, sortItems } from "nitropage/data/list";
import {
  mediaData,
  Image,
  useMedia,
  PlaceholderImage,
  useResize,
} from "nitropage/data/media";
import { rteData } from "nitropage/data/rte";
import { textData } from "nitropage/data/text";
import { createResize } from "nitropage/server";
import { createMemo, For, Show } from "solid-js";
import server$ from "solid-start/server";
import { Container } from "~/components/container";
import Heading from "~/components/heading";
import ArrowRight from "~icons/tabler/arrowNarrowRight";

/*
 * The design of this blueprint is inspired by Meraki UI, check it out!
 * https://github.com/merakiui/merakiui
 */

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const [images] = useMedia(() => props.data.image);
  const image = createMemo(() => images().byOrder[0] || PlaceholderImage);
  const [optimizeMedia] = useResize(
    server$(
      createResize({
        sizes: {
          xs: { width: 700, height: 400 },
          lg: { width: 800, height: 500 },
        },
      }),
    ),
  );

  return (
    <div
      class={css({
        "& .notfound__highlight": {
          color: rgb(props.data.highlightColor),
        },
        "& .notfound__text": {
          color: rgb(props.data.textColor),
        },
        "& .notfound__text--faded": {
          color: rgb({ ...props.data.textColor, alpha: 0.8 }),
        },
      })}
      {...props.elementProps}
    >
      {props.handles}
      <Container>
        <div class="min-h-screen py-24 pb-12 lg:pb-24 lg:flex lg:items-center lg:gap-12">
          <div class="lg:w-1/2">
            <props.Rte
              data="error"
              component="p"
              class={"notfound__highlight text-sm font-medium text-blue-600"}
            />
            <Heading
              level={1}
              size="sm"
              class={clsx(
                "mt-3 notfound__text",
                css({
                  ...fontCss(props.data.titleFont),
                }),
              )}
            >
              <props.Rte data="title" />
            </Heading>
            <props.Rte
              data="description"
              component="p"
              class={"mt-4 notfound__text--faded text-slate-500"}
            />

            <props.Slot name="default" class="flex items-center mt-6 gap-x-3" />

            <div class="mt-10 space-y-6">
              <For each={sortItems(props.data.links)}>
                {([item, id]) => (
                  <div>
                    <A
                      href={item.link}
                      class={
                        "inline-flex items-center text-sm notfound__highlight text-blue-600 gap-x-2 hover:underline"
                      }
                    >
                      <props.Rte path={(d) => d("links").byId[id].title} />
                      <ArrowRight class="w-5 h-5" />
                    </A>

                    <props.Rte
                      path={(d) => d("links").byId[id].description}
                      class="block mt-2 text-sm notfound__text--faded text-slate-500"
                    />
                  </div>
                )}
              </For>
            </div>
          </div>

          <Show when={image()}>
            <div class="relative w-full mt-8 lg:w-1/2 lg:mt-0">
              <Image
                preload={true}
                sources={optimizeMedia(image(), [
                  { size: "lg", screen: 1000 },
                  { size: "xs" },
                ])}
                class=" w-full lg:h-[32rem] h-80 md:h-96 rounded-lg object-cover "
                alt={image().description}
              />
            </div>
          </Show>
        </div>
      </Container>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    return {
      title() {
        return "Page not found";
      },
      slots: { default: {} },
      data: {
        title: rteData({
          defaultValue() {
            return "Page not found";
          },
        }),
        image: mediaData({
          type: "image",
          max: 1,
        }),
        error: rteData({
          defaultValue() {
            return "404 error";
          },
        }),
        description: rteData({
          defaultValue() {
            return "Sorry, the page you are looking for doesn't exist.";
          },
        }),
        textColor: colorData({ alpha: false }),
        highlightColor: colorData(),
        links: listData({
          defaultItems: 3,
          itemData: {
            link: textData({
              wide: true,
            }),
            title: rteData({
              shortLorem: true,
            }),
            description: rteData({
              defaultValue() {
                return "An awesome description.";
              },
            }),
          },
          itemTitle(item) {
            return item.title?.content;
          },
        }),
        titleFont: fontData(),
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "hfhknwqzq0f2o9qzrf2taxaj";
