import { createBlueprint, BlueprintComponent } from "nitropage";
import { css, rgb } from "nitropage/css";
import { Dynamic } from "solid-js/web";
import Fragment from "~/components/fragment";
import PostLayout from "./blog/postLayout";
import { rteData, toolbarFull } from "nitropage/data/rte";
import { toggleData } from "nitropage/data/toggle";
import { colorData } from "nitropage/data/color";
import { listData, sortItems } from "nitropage/data/list";
import { For, Show, createMemo } from "solid-js";
import { numberData } from "nitropage/data/number";

const MyComponent: BlueprintComponent<typeof blueprint> = function (props) {
  const columns = createMemo(function () {
    for (let idx = 9; idx >= 0; idx--) {
      const column = "column" + idx;
      if (props.data[column as "column1"]) {
        return idx;
      }
    }
  });

  const columnsList = createMemo(function () {
    return Array(columns()).fill(0);
  });

  const spreadColumn = createMemo(function () {
    const column = props.data.spreadColumn;
    if (column == null || column == 0) {
      return;
    }

    return column + (props.data.showNumeration ? 1 : 0);
  });

  return (
    <div
      class={
        css({
          color: rgb(props.data.textColor),
          "& .table-color": {
            backgroundColor: rgb(props.data.backgroundColor),
          },
          "& th, & td": {
            borderColor: rgb(props.data.borderColor),
          },
          "& thead": {
            color: rgb(props.data.headTextColor),
            backgroundColor: rgb(props.data.headBackgroundColor),
          },
          "& tbody > tr > *": props.data.alignTop
            ? {
                verticalAlign: "top",
              }
            : undefined,
          "& tbody > tr:nth-of-type(2n+1)": {
            backgroundColor: rgb(props.data.stripBackgroundColor),
          },
          [`& tbody > tr > *:nth-child(${spreadColumn()})`]:
            spreadColumn() != null
              ? {
                  width: "100%",
                }
              : undefined,
          [`& tbody > tr > *:not(:nth-child(${spreadColumn()})) *`]:
            spreadColumn() != null
              ? {
                  whiteSpace: "pre",
                }
              : undefined,
        }) + " text-slate-900 md:text-lg"
      }
      {...props.elementProps}
    >
      {props.handles}
      <Dynamic component={props.data.fullWidth ? Fragment : PostLayout}>
        <div class="table-color overflow-x-auto">
          <table class="w-full">
            <thead class="text-left">
              <tr>
                <Show when={props.data.showNumeration}>
                  <th class="py-1 px-3 border border-b-2 w-0" scope="col">
                    <props.Rte data="numerationLabel" />
                  </th>
                </Show>
                <For each={columnsList()}>
                  {(_, column) => {
                    return (
                      <th class="py-1 px-3 border border-b-2" scope="col">
                        <props.Rte
                          data={("column" + (column() + 1)) as "column1"}
                        />
                      </th>
                    );
                  }}
                </For>
              </tr>
            </thead>
            <tbody>
              <For each={sortItems(props.data.rows)}>
                {([row, id], index) => {
                  return (
                    <tr>
                      <Show when={props.data.showNumeration}>
                        <th class="text-left py-1 px-3 border" scope="row">
                          {index() + 1}
                        </th>
                      </Show>
                      <For each={columnsList()}>
                        {(_, column) => {
                          return (
                            <td class="py-1 px-3 border">
                              <props.Rte
                                data="rows"
                                path={(d) =>
                                  d("rows").byId[id][
                                    ("column" + (column() + 1)) as "column1"
                                  ]
                                }
                              />
                            </td>
                          );
                        }}
                      </For>
                    </tr>
                  );
                }}
              </For>
            </tbody>
          </table>
        </div>
      </Dynamic>
    </div>
  );
};

export const blueprint = createBlueprint(
  () => {
    const columnTitleOptions: Parameters<typeof rteData>[0] = {
      toolbar: {
        background: true,
        color: true,
        italic: true,
        underline: true,
        link: true,
        align: true,
        clean: true,
      },
    };
    const columnRowOptions: Parameters<typeof rteData>[0] = {
      toolbar: toolbarFull,
    };
    const genericRteData = rteData(columnRowOptions);
    const titleRteData = rteData(columnTitleOptions);
    const genericColorData = colorData();

    return {
      slots: { default: {} },
      data: {
        rows: listData({
          itemData: {
            column1: genericRteData,
            column2: genericRteData,
            column3: genericRteData,
            column4: genericRteData,
            column5: genericRteData,
            column6: genericRteData,
            column7: genericRteData,
            column8: genericRteData,
            column9: genericRteData,
          },
        }),
        column1: rteData({ ...columnTitleOptions, shortLorem: true }),
        column2: rteData({ ...columnTitleOptions, shortLorem: true }),
        column3: rteData({ ...columnTitleOptions, shortLorem: true }),
        column4: rteData({ ...columnTitleOptions, shortLorem: true }),
        column5: titleRteData,
        column6: titleRteData,
        column7: titleRteData,
        column8: titleRteData,
        column9: titleRteData,
        fullWidth: toggleData(),
        spreadColumn: numberData({
          optional: true,
          min: 1,
          max: 9,
        }),
        alignTop: toggleData(),
        showNumeration: toggleData(),
        numerationLabel: rteData({
          defaultValue: () => "#",
          ...columnTitleOptions,
        }),
        textColor: genericColorData,
        headTextColor: genericColorData,
        borderColor: genericColorData,
        backgroundColor: genericColorData,
        headBackgroundColor: genericColorData,
        stripBackgroundColor: genericColorData,
      },
    };
  },
  import.meta.hot,
);

export default MyComponent;
export const passive = true;
export const id = "f9zu52lepi4wlteaz0k8w79n";
